(*
 * Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
 * SPDX-License-Identifier: Apache-2.0 OR ISC OR MIT-0
 *)

(* The number of instructions of functions. *)

let count_insts (execth:thm*thm option array) =
  let length_th = fst execth in
  dest_small_numeral (snd (dest_eq (concl length_th))) / 4;;

let len_montsqr_p384_neon, len_montmul_p384_neon, len_sub_p384 =
  count_insts BIGNUM_MONTSQR_P384_NEON_EXEC,
  count_insts BIGNUM_MONTMUL_P384_NEON_EXEC,
  count_insts BIGNUM_SUB_P384_EXEC;;

(* ------------------------------------------------------------------------- *)
(* Building actions for proving correctness of optimization step 1 inlining. *)
(* The inlined code is simply the result of turning 'bl' to macro invocation *)
(* and removing the 'ret' instruction at the end of subroutine.              *)
(* The functions were inlined by hand, not using a script.                   *)
(* ------------------------------------------------------------------------- *)

let code_blocks = [
  (* (action("equal"|"call <subroutine>"), # instructions if "equal") *)
  ("equal", Some 3);
  ("call montsqr_p384", None);
  ("call montsqr_p384", None);
  ("call montmul_p384", None);
  ("call montmul_p384", None);
  ("call montmul_p384", None);
  ("call montmul_p384", None);
  ("call montmul_p384", None);
  ("call montmul_p384", None);
  ("call sub_p384", None);
  ("call sub_p384", None);
  ("call montsqr_p384", None);
  ("call montsqr_p384", None);
  ("call montmul_p384", None);
  ("call montmul_p384", None);
  ("call sub_p384", None);
  ("call sub_p384", None);
  ("call montmul_p384", None);
  ("call sub_p384", None);
  ("call sub_p384", None);
  ("call montmul_p384", None);
  ("call montmul_p384", None);
  ("call montmul_p384", None);
  ("call sub_p384", None);
  ("equal", Some ((0xe94 - 0xd38) / 4))
];;

(* Build the actions. *)

let actions1 = ref [];;
let last_n1 = ref 0 and last_n2 = ref 0;;
List.iter (fun code_block ->
  let n1,n2 = !last_n1,!last_n2 in
  match code_block with
  | "equal", Some nsteps ->
    actions1 := !actions1 @ [("equal",n1,n1+nsteps,n2,n2+nsteps)];
    last_n1 := nsteps + n1;
    last_n2 := nsteps + n2
  | callfn, None ->
    (* nsteps_prologue = # fn arguments
       nsteps_fnbody = # insts of fn except last ret *)
    let nsteps_prologue, nsteps_fnbody = match callfn with
      | "call montsqr_p384" -> 2, len_montsqr_p384_neon - 1
      | "call montmul_p384" ->
        (* +2 for sp updates *)
        3, len_montmul_p384_neon - 1 + 2
      | "call sub_p384" -> 3, len_sub_p384 - 1
      | _ -> failwith callfn in begin
    actions1 := !actions1 @
      [(* For argument assignments, use 'replace' not 'equiv' because
          the symbolic expression of memory addresses shouldn't be abbreviated.
          Otherwise symbolic simulator will be confused. *)
       ("replace",n1,n1+nsteps_prologue,n2,n2+nsteps_prologue)];
    let n1 = n1+nsteps_prologue and n2 = n2+nsteps_prologue in begin
    actions1 := !actions1 @
      [(* bl *)
       ("delete", n1, n1+1, n2, n2);
       (* fn body *)
       ("equal",  n1+1, n1+1+nsteps_fnbody,
                  n2, n2+nsteps_fnbody);
       (* ret *)
       ("delete", n1+1+nsteps_fnbody, n1+1+nsteps_fnbody+1,
                  n2+nsteps_fnbody, n2+nsteps_fnbody)];
    last_n1 := n1+1+nsteps_fnbody+1;
    last_n2 := n2+nsteps_fnbody
    end end)
  code_blocks;;


(* ------------------------------------------------------------------------- *)
(* Actions for proving correctness of step 2 memory optimizations.           *)
(* To get this, trim the callee-save register spills & ret from the assembly *)
(* source files after the first stage (inlining) and second stage (mem opt), *)
(* and run                                                                   *)
(* 'tools/get-actions.py <trimmed first asm> <trimmed second asm>'.          *)
(* ------------------------------------------------------------------------- *)

let actions2 = [
  ("equal", 0, 4, 0, 4);
  ("replace", 4, 9, 4, 8);
  ("equal", 9, 14, 8, 13);
  ("replace", 14, 15, 13, 14);
  ("equal", 15, 16, 14, 15);
  ("replace", 16, 17, 15, 16);
  ("equal", 17, 18, 16, 17);
  ("replace", 18, 19, 17, 18);
  ("equal", 19, 311, 18, 310);
  ("replace", 311, 317, 310, 314);
  ("equal", 317, 322, 314, 319);
  ("replace", 322, 323, 319, 320);
  ("equal", 323, 324, 320, 321);
  ("replace", 324, 325, 321, 322);
  ("equal", 325, 326, 322, 323);
  ("replace", 326, 327, 323, 324);
  ("equal", 327, 435, 324, 432);
  ("replace", 435, 436, 432, 433);
  ("equal", 436, 439, 433, 436);
  ("replace", 439, 440, 436, 437);
  ("equal", 440, 442, 437, 439);
  ("replace", 442, 443, 439, 440);
  ("equal", 443, 489, 440, 486);
  ("replace", 489, 490, 486, 487);
  ("equal", 490, 495, 487, 492);
  ("replace", 495, 496, 492, 493);
  ("equal", 496, 498, 493, 495);
  ("replace", 498, 499, 495, 496);
  ("equal", 499, 613, 496, 610);
  ("replace", 613, 614, 610, 611);
  ("equal", 614, 616, 611, 613);
  ("replace", 616, 617, 613, 614);
  ("equal", 617, 618, 614, 615);
  ("replace", 618, 630, 615, 621);
  ("equal", 630, 635, 621, 626);
  ("replace", 635, 637, 626, 628);
  ("equal", 637, 638, 628, 629);
  ("replace", 638, 639, 629, 630);
  ("equal", 639, 649, 630, 640);
  ("replace", 649, 650, 640, 641);
  ("equal", 650, 716, 641, 707);
  ("replace", 716, 717, 707, 708);
  ("equal", 717, 719, 708, 710);
  ("replace", 719, 720, 710, 711);
  ("equal", 720, 768, 711, 759);
  ("replace", 768, 769, 759, 760);
  ("equal", 769, 771, 760, 762);
  ("replace", 771, 772, 762, 763);
  ("equal", 772, 782, 763, 773);
  ("replace", 782, 783, 773, 774);
  ("equal", 783, 851, 774, 842);
  ("replace", 851, 852, 842, 843);
  ("equal", 852, 855, 843, 846);
  ("replace", 855, 856, 846, 847);
  ("equal", 856, 859, 847, 850);
  ("replace", 859, 860, 850, 851);
  ("equal", 860, 865, 851, 856);
  ("replace", 865, 866, 856, 857);
  ("equal", 866, 869, 857, 860);
  ("replace", 869, 870, 860, 861);
  ("equal", 870, 874, 861, 865);
  ("replace", 874, 875, 865, 866);
  ("equal", 875, 931, 866, 922);
  ("replace", 931, 933, 922, 924);
  ("equal", 933, 939, 924, 930);
  ("replace", 939, 940, 930, 931);
  ("equal", 940, 1026, 931, 1017);
  ("replace", 1026, 1027, 1017, 1018);
  ("equal", 1027, 1029, 1018, 1020);
  ("replace", 1029, 1030, 1020, 1021);
  ("equal", 1030, 1031, 1021, 1022);
  ("replace", 1031, 1047, 1022, 1027);
  ("equal", 1047, 1052, 1027, 1032);
  ("replace", 1052, 1054, 1032, 1034);
  ("equal", 1054, 1055, 1034, 1035);
  ("replace", 1055, 1056, 1035, 1036);
  ("equal", 1056, 1066, 1036, 1046);
  ("replace", 1066, 1067, 1046, 1047);
  ("equal", 1067, 1133, 1047, 1113);
  ("replace", 1133, 1134, 1113, 1114);
  ("equal", 1134, 1136, 1114, 1116);
  ("replace", 1136, 1137, 1116, 1117);
  ("equal", 1137, 1185, 1117, 1165);
  ("replace", 1185, 1186, 1165, 1166);
  ("equal", 1186, 1188, 1166, 1168);
  ("replace", 1188, 1189, 1168, 1169);
  ("equal", 1189, 1199, 1169, 1179);
  ("replace", 1199, 1200, 1179, 1180);
  ("equal", 1200, 1268, 1180, 1248);
  ("replace", 1268, 1269, 1248, 1249);
  ("equal", 1269, 1272, 1249, 1252);
  ("replace", 1272, 1273, 1252, 1253);
  ("equal", 1273, 1276, 1253, 1256);
  ("replace", 1276, 1277, 1256, 1257);
  ("equal", 1277, 1282, 1257, 1262);
  ("replace", 1282, 1283, 1262, 1263);
  ("equal", 1283, 1286, 1263, 1266);
  ("replace", 1286, 1287, 1266, 1267);
  ("equal", 1287, 1291, 1267, 1271);
  ("replace", 1291, 1292, 1271, 1272);
  ("equal", 1292, 1348, 1272, 1328);
  ("replace", 1348, 1350, 1328, 1330);
  ("equal", 1350, 1356, 1330, 1336);
  ("replace", 1356, 1357, 1336, 1337);
  ("equal", 1357, 1443, 1337, 1423);
  ("replace", 1443, 1444, 1423, 1424);
  ("equal", 1444, 1446, 1424, 1426);
  ("replace", 1446, 1447, 1426, 1427);
  ("equal", 1447, 1448, 1427, 1428);
  ("replace", 1448, 1454, 1428, 1429);
  ("equal", 1454, 1455, 1429, 1430);
  ("delete", 1455, 1460, 1430, 1430);
  ("equal", 1460, 1461, 1430, 1431);
  ("replace", 1461, 1463, 1431, 1433);
  ("equal", 1463, 1469, 1433, 1439);
  ("replace", 1469, 1470, 1439, 1440);
  ("equal", 1470, 1483, 1440, 1453);
  ("replace", 1483, 1484, 1453, 1454);
  ("equal", 1484, 1550, 1454, 1520);
  ("replace", 1550, 1551, 1520, 1521);
  ("equal", 1551, 1602, 1521, 1572);
  ("replace", 1602, 1603, 1572, 1573);
  ("equal", 1603, 1605, 1573, 1575);
  ("replace", 1605, 1606, 1575, 1576);
  ("equal", 1606, 1616, 1576, 1586);
  ("replace", 1616, 1617, 1586, 1587);
  ("equal", 1617, 1685, 1587, 1655);
  ("replace", 1685, 1686, 1655, 1656);
  ("equal", 1686, 1689, 1656, 1659);
  ("replace", 1689, 1690, 1659, 1660);
  ("equal", 1690, 1693, 1660, 1663);
  ("replace", 1693, 1694, 1663, 1664);
  ("equal", 1694, 1699, 1664, 1669);
  ("replace", 1699, 1700, 1669, 1670);
  ("equal", 1700, 1703, 1670, 1673);
  ("replace", 1703, 1704, 1673, 1674);
  ("equal", 1704, 1708, 1674, 1678);
  ("replace", 1708, 1709, 1678, 1679);
  ("equal", 1709, 1765, 1679, 1735);
  ("replace", 1765, 1767, 1735, 1737);
  ("equal", 1767, 1773, 1737, 1743);
  ("replace", 1773, 1774, 1743, 1744);
  ("equal", 1774, 1860, 1744, 1830);
  ("replace", 1860, 1861, 1830, 1831);
  ("equal", 1861, 1863, 1831, 1833);
  ("replace", 1863, 1864, 1833, 1834);
  ("equal", 1864, 1865, 1834, 1835);
  ("replace", 1865, 1881, 1835, 1840);
  ("equal", 1881, 1886, 1840, 1845);
  ("replace", 1886, 1888, 1845, 1847);
  ("equal", 1888, 1889, 1847, 1848);
  ("replace", 1889, 1890, 1848, 1849);
  ("equal", 1890, 1900, 1849, 1859);
  ("replace", 1900, 1901, 1859, 1860);
  ("equal", 1901, 1967, 1860, 1926);
  ("replace", 1967, 1968, 1926, 1927);
  ("equal", 1968, 1970, 1927, 1929);
  ("replace", 1970, 1971, 1929, 1930);
  ("equal", 1971, 2019, 1930, 1978);
  ("replace", 2019, 2020, 1978, 1979);
  ("equal", 2020, 2022, 1979, 1981);
  ("replace", 2022, 2023, 1981, 1982);
  ("equal", 2023, 2033, 1982, 1992);
  ("replace", 2033, 2034, 1992, 1993);
  ("equal", 2034, 2102, 1993, 2061);
  ("replace", 2102, 2103, 2061, 2062);
  ("equal", 2103, 2106, 2062, 2065);
  ("replace", 2106, 2107, 2065, 2066);
  ("equal", 2107, 2110, 2066, 2069);
  ("replace", 2110, 2111, 2069, 2070);
  ("equal", 2111, 2116, 2070, 2075);
  ("replace", 2116, 2117, 2075, 2076);
  ("equal", 2117, 2120, 2076, 2079);
  ("replace", 2120, 2121, 2079, 2080);
  ("equal", 2121, 2125, 2080, 2084);
  ("replace", 2125, 2126, 2084, 2085);
  ("equal", 2126, 2182, 2085, 2141);
  ("replace", 2182, 2184, 2141, 2143);
  ("equal", 2184, 2190, 2143, 2149);
  ("replace", 2190, 2191, 2149, 2150);
  ("equal", 2191, 2277, 2150, 2236);
  ("replace", 2277, 2278, 2236, 2237);
  ("equal", 2278, 2280, 2237, 2239);
  ("replace", 2280, 2281, 2239, 2240);
  ("equal", 2281, 2282, 2240, 2241);
  ("replace", 2282, 2288, 2241, 2242);
  ("equal", 2288, 2289, 2242, 2243);
  ("delete", 2289, 2294, 2243, 2243);
  ("equal", 2294, 2295, 2243, 2244);
  ("replace", 2295, 2297, 2244, 2246);
  ("equal", 2297, 2303, 2246, 2252);
  ("replace", 2303, 2304, 2252, 2253);
  ("equal", 2304, 2317, 2253, 2266);
  ("replace", 2317, 2318, 2266, 2267);
  ("equal", 2318, 2384, 2267, 2333);
  ("replace", 2384, 2385, 2333, 2334);
  ("equal", 2385, 2436, 2334, 2385);
  ("replace", 2436, 2437, 2385, 2386);
  ("equal", 2437, 2439, 2386, 2388);
  ("replace", 2439, 2440, 2388, 2389);
  ("equal", 2440, 2450, 2389, 2399);
  ("replace", 2450, 2451, 2399, 2400);
  ("equal", 2451, 2519, 2400, 2468);
  ("replace", 2519, 2520, 2468, 2469);
  ("equal", 2520, 2523, 2469, 2472);
  ("replace", 2523, 2524, 2472, 2473);
  ("equal", 2524, 2527, 2473, 2476);
  ("replace", 2527, 2528, 2476, 2477);
  ("equal", 2528, 2533, 2477, 2482);
  ("replace", 2533, 2534, 2482, 2483);
  ("equal", 2534, 2537, 2483, 2486);
  ("replace", 2537, 2538, 2486, 2487);
  ("equal", 2538, 2542, 2487, 2491);
  ("replace", 2542, 2543, 2491, 2492);
  ("equal", 2543, 2599, 2492, 2548);
  ("replace", 2599, 2601, 2548, 2550);
  ("equal", 2601, 2607, 2550, 2556);
  ("replace", 2607, 2608, 2556, 2557);
  ("equal", 2608, 2694, 2557, 2643);
  ("replace", 2694, 2695, 2643, 2644);
  ("equal", 2695, 2697, 2644, 2646);
  ("replace", 2697, 2698, 2646, 2647);
  ("equal", 2698, 2699, 2647, 2648);
  ("replace", 2699, 2715, 2648, 2653);
  ("equal", 2715, 2720, 2653, 2658);
  ("replace", 2720, 2722, 2658, 2660);
  ("equal", 2722, 2723, 2660, 2661);
  ("replace", 2723, 2724, 2661, 2662);
  ("equal", 2724, 2734, 2662, 2672);
  ("replace", 2734, 2735, 2672, 2673);
  ("equal", 2735, 2801, 2673, 2739);
  ("replace", 2801, 2802, 2739, 2740);
  ("equal", 2802, 2804, 2740, 2742);
  ("replace", 2804, 2805, 2742, 2743);
  ("equal", 2805, 2853, 2743, 2791);
  ("replace", 2853, 2854, 2791, 2792);
  ("equal", 2854, 2856, 2792, 2794);
  ("replace", 2856, 2857, 2794, 2795);
  ("equal", 2857, 2867, 2795, 2805);
  ("replace", 2867, 2868, 2805, 2806);
  ("equal", 2868, 2936, 2806, 2874);
  ("replace", 2936, 2937, 2874, 2875);
  ("equal", 2937, 2940, 2875, 2878);
  ("replace", 2940, 2941, 2878, 2879);
  ("equal", 2941, 2944, 2879, 2882);
  ("replace", 2944, 2945, 2882, 2883);
  ("equal", 2945, 2950, 2883, 2888);
  ("replace", 2950, 2951, 2888, 2889);
  ("equal", 2951, 2954, 2889, 2892);
  ("replace", 2954, 2955, 2892, 2893);
  ("equal", 2955, 2959, 2893, 2897);
  ("replace", 2959, 2960, 2897, 2898);
  ("equal", 2960, 3016, 2898, 2954);
  ("replace", 3016, 3018, 2954, 2956);
  ("equal", 3018, 3024, 2956, 2962);
  ("replace", 3024, 3025, 2962, 2963);
  ("equal", 3025, 3110, 2963, 3048);
  ("replace", 3110, 3126, 3048, 3057);
  ("equal", 3126, 3128, 3057, 3059);
  ("replace", 3128, 3130, 3059, 3061);
  ("equal", 3130, 3132, 3061, 3063);
  ("replace", 3132, 3134, 3063, 3065);
  ("equal", 3134, 3148, 3065, 3079);
  ("replace", 3148, 3156, 3079, 3084);
  ("equal", 3156, 3158, 3084, 3086);
  ("replace", 3158, 3166, 3086, 3092);
  ("equal", 3166, 3178, 3092, 3104);
  ("replace", 3178, 3187, 3104, 3111);
  ("equal", 3187, 3192, 3111, 3116);
  ("replace", 3192, 3193, 3116, 3117);
  ("equal", 3193, 3194, 3117, 3118);
  ("replace", 3194, 3195, 3118, 3119);
  ("equal", 3195, 3196, 3119, 3120);
  ("replace", 3196, 3197, 3120, 3121);
  ("equal", 3197, 3305, 3121, 3229);
  ("replace", 3305, 3306, 3229, 3230);
  ("equal", 3306, 3309, 3230, 3233);
  ("replace", 3309, 3310, 3233, 3234);
  ("equal", 3310, 3312, 3234, 3236);
  ("replace", 3312, 3313, 3236, 3237);
  ("equal", 3313, 3359, 3237, 3283);
  ("replace", 3359, 3360, 3283, 3284);
  ("equal", 3360, 3365, 3284, 3289);
  ("replace", 3365, 3366, 3289, 3290);
  ("equal", 3366, 3368, 3290, 3292);
  ("replace", 3368, 3369, 3292, 3293);
  ("equal", 3369, 3483, 3293, 3407);
  ("replace", 3483, 3484, 3407, 3408);
  ("equal", 3484, 3486, 3408, 3410);
  ("replace", 3486, 3487, 3410, 3411);
  ("equal", 3487, 3488, 3411, 3412);
  ("replace", 3488, 3489, 3412, 3413);
  ("equal", 3489, 3490, 3413, 3414);
  ("replace", 3490, 3495, 3414, 3418);
  ("equal", 3495, 3500, 3418, 3423);
  ("replace", 3500, 3501, 3423, 3424);
  ("equal", 3501, 3502, 3424, 3425);
  ("replace", 3502, 3503, 3425, 3426);
  ("equal", 3503, 3504, 3426, 3427);
  ("replace", 3504, 3505, 3427, 3428);
  ("equal", 3505, 3797, 3428, 3720);
  ("replace", 3797, 3808, 3720, 3724);
  ("equal", 3808, 3813, 3724, 3729);
  ("replace", 3813, 3815, 3729, 3731);
  ("equal", 3815, 3816, 3731, 3732);
  ("replace", 3816, 3817, 3732, 3733);
  ("equal", 3817, 3827, 3733, 3743);
  ("replace", 3827, 3828, 3743, 3744);
  ("equal", 3828, 3894, 3744, 3810);
  ("replace", 3894, 3895, 3810, 3811);
  ("equal", 3895, 3897, 3811, 3813);
  ("replace", 3897, 3898, 3813, 3814);
  ("equal", 3898, 3946, 3814, 3862);
  ("replace", 3946, 3947, 3862, 3863);
  ("equal", 3947, 3949, 3863, 3865);
  ("replace", 3949, 3950, 3865, 3866);
  ("equal", 3950, 3960, 3866, 3876);
  ("replace", 3960, 3961, 3876, 3877);
  ("equal", 3961, 4029, 3877, 3945);
  ("replace", 4029, 4030, 3945, 3946);
  ("equal", 4030, 4033, 3946, 3949);
  ("replace", 4033, 4034, 3949, 3950);
  ("equal", 4034, 4037, 3950, 3953);
  ("replace", 4037, 4038, 3953, 3954);
  ("equal", 4038, 4043, 3954, 3959);
  ("replace", 4043, 4044, 3959, 3960);
  ("equal", 4044, 4047, 3960, 3963);
  ("replace", 4047, 4048, 3963, 3964);
  ("equal", 4048, 4052, 3964, 3968);
  ("replace", 4052, 4053, 3968, 3969);
  ("equal", 4053, 4109, 3969, 4025);
  ("replace", 4109, 4111, 4025, 4027);
  ("equal", 4111, 4117, 4027, 4033);
  ("replace", 4117, 4118, 4033, 4034);
  ("equal", 4118, 4204, 4034, 4120);
  ("replace", 4204, 4205, 4120, 4121);
  ("equal", 4205, 4207, 4121, 4123);
  ("replace", 4207, 4208, 4123, 4124);
  ("equal", 4208, 4209, 4124, 4125);
  ("replace", 4209, 4225, 4125, 4130);
  ("equal", 4225, 4230, 4130, 4135);
  ("replace", 4230, 4232, 4135, 4137);
  ("equal", 4232, 4233, 4137, 4138);
  ("replace", 4233, 4234, 4138, 4139);
  ("equal", 4234, 4244, 4139, 4149);
  ("replace", 4244, 4245, 4149, 4150);
  ("equal", 4245, 4311, 4150, 4216);
  ("replace", 4311, 4312, 4216, 4217);
  ("equal", 4312, 4314, 4217, 4219);
  ("replace", 4314, 4315, 4219, 4220);
  ("equal", 4315, 4363, 4220, 4268);
  ("replace", 4363, 4364, 4268, 4269);
  ("equal", 4364, 4366, 4269, 4271);
  ("replace", 4366, 4367, 4271, 4272);
  ("equal", 4367, 4377, 4272, 4282);
  ("replace", 4377, 4378, 4282, 4283);
  ("equal", 4378, 4446, 4283, 4351);
  ("replace", 4446, 4447, 4351, 4352);
  ("equal", 4447, 4450, 4352, 4355);
  ("replace", 4450, 4451, 4355, 4356);
  ("equal", 4451, 4454, 4356, 4359);
  ("replace", 4454, 4455, 4359, 4360);
  ("equal", 4455, 4460, 4360, 4365);
  ("replace", 4460, 4461, 4365, 4366);
  ("equal", 4461, 4464, 4366, 4369);
  ("replace", 4464, 4465, 4369, 4370);
  ("equal", 4465, 4469, 4370, 4374);
  ("replace", 4469, 4470, 4374, 4375);
  ("equal", 4470, 4526, 4375, 4431);
  ("replace", 4526, 4528, 4431, 4433);
  ("equal", 4528, 4534, 4433, 4439);
  ("replace", 4534, 4535, 4439, 4440);
  ("equal", 4535, 4620, 4440, 4525);
  ("replace", 4620, 4631, 4525, 4532);
  ("equal", 4631, 4633, 4532, 4534);
  ("delete", 4633, 4634, 4534, 4534);
  ("equal", 4634, 4635, 4534, 4535);
  ("replace", 4635, 4636, 4535, 4536);
  ("equal", 4636, 4639, 4536, 4539);
  ("replace", 4639, 4640, 4539, 4540);
  ("equal", 4640, 4643, 4540, 4543);
  ("replace", 4643, 4644, 4543, 4544);
  ("equal", 4644, 4661, 4544, 4561);
  ("replace", 4661, 4666, 4561, 4563);
  ("equal", 4666, 4668, 4563, 4565);
  ("replace", 4668, 4676, 4565, 4571);
  ("equal", 4676, 4688, 4571, 4583);
  ("replace", 4688, 4702, 4583, 4590);
  ("equal", 4702, 4707, 4590, 4595);
  ("replace", 4707, 4709, 4595, 4597);
  ("equal", 4709, 4710, 4597, 4598);
  ("replace", 4710, 4711, 4598, 4599);
  ("equal", 4711, 4721, 4599, 4609);
  ("replace", 4721, 4722, 4609, 4610);
  ("equal", 4722, 4788, 4610, 4676);
  ("replace", 4788, 4789, 4676, 4677);
  ("equal", 4789, 4791, 4677, 4679);
  ("replace", 4791, 4792, 4679, 4680);
  ("equal", 4792, 4840, 4680, 4728);
  ("replace", 4840, 4841, 4728, 4729);
  ("equal", 4841, 4843, 4729, 4731);
  ("replace", 4843, 4844, 4731, 4732);
  ("equal", 4844, 4854, 4732, 4742);
  ("replace", 4854, 4855, 4742, 4743);
  ("equal", 4855, 4923, 4743, 4811);
  ("replace", 4923, 4924, 4811, 4812);
  ("equal", 4924, 4927, 4812, 4815);
  ("replace", 4927, 4928, 4815, 4816);
  ("equal", 4928, 4931, 4816, 4819);
  ("replace", 4931, 4932, 4819, 4820);
  ("equal", 4932, 4937, 4820, 4825);
  ("replace", 4937, 4938, 4825, 4826);
  ("equal", 4938, 4941, 4826, 4829);
  ("replace", 4941, 4942, 4829, 4830);
  ("equal", 4942, 4946, 4830, 4834);
  ("replace", 4946, 4947, 4834, 4835);
  ("equal", 4947, 5003, 4835, 4891);
  ("replace", 5003, 5005, 4891, 4893);
  ("equal", 5005, 5011, 4893, 4899);
  ("replace", 5011, 5012, 4899, 4900);
  ("equal", 5012, 5098, 4900, 4986);
  ("replace", 5098, 5099, 4986, 4987);
  ("equal", 5099, 5101, 4987, 4989);
  ("replace", 5101, 5102, 4989, 4990);
  ("equal", 5102, 5103, 4990, 4991);
  ("replace", 5103, 5108, 4991, 4992);
  ("equal", 5108, 5110, 4992, 4994);
  ("delete", 5110, 5111, 4994, 4994);
  ("equal", 5111, 5112, 4994, 4995);
  ("replace", 5112, 5113, 4995, 4996);
  ("equal", 5113, 5116, 4996, 4999);
  ("replace", 5116, 5117, 4999, 5000);
  ("equal", 5117, 5120, 5000, 5003);
  ("replace", 5120, 5121, 5003, 5004);
  ("equal", 5121, 5126, 5004, 5009);
  ("replace", 5126, 5127, 5009, 5010);
  ("equal", 5127, 5128, 5010, 5011);
  ("replace", 5128, 5129, 5011, 5012);
  ("equal", 5129, 5131, 5012, 5014);
  ("replace", 5131, 5147, 5014, 5025);
  ("equal", 5147, 5148, 5025, 5026);
  ("replace", 5148, 5152, 5026, 5029);
  ("equal", 5152, 5165, 5029, 5042);
  ("replace", 5165, 5179, 5042, 5049);
  ("equal", 5179, 5184, 5049, 5054);
  ("replace", 5184, 5186, 5054, 5056);
  ("equal", 5186, 5187, 5056, 5057);
  ("replace", 5187, 5188, 5057, 5058);
  ("equal", 5188, 5198, 5058, 5068);
  ("replace", 5198, 5199, 5068, 5069);
  ("equal", 5199, 5265, 5069, 5135);
  ("replace", 5265, 5266, 5135, 5136);
  ("equal", 5266, 5268, 5136, 5138);
  ("replace", 5268, 5269, 5138, 5139);
  ("equal", 5269, 5317, 5139, 5187);
  ("replace", 5317, 5318, 5187, 5188);
  ("equal", 5318, 5320, 5188, 5190);
  ("replace", 5320, 5321, 5190, 5191);
  ("equal", 5321, 5331, 5191, 5201);
  ("replace", 5331, 5332, 5201, 5202);
  ("equal", 5332, 5400, 5202, 5270);
  ("replace", 5400, 5401, 5270, 5271);
  ("equal", 5401, 5404, 5271, 5274);
  ("replace", 5404, 5405, 5274, 5275);
  ("equal", 5405, 5408, 5275, 5278);
  ("replace", 5408, 5409, 5278, 5279);
  ("equal", 5409, 5414, 5279, 5284);
  ("replace", 5414, 5415, 5284, 5285);
  ("equal", 5415, 5418, 5285, 5288);
  ("replace", 5418, 5419, 5288, 5289);
  ("equal", 5419, 5423, 5289, 5293);
  ("replace", 5423, 5424, 5293, 5294);
  ("equal", 5424, 5480, 5294, 5350);
  ("replace", 5480, 5482, 5350, 5352);
  ("equal", 5482, 5488, 5352, 5358);
  ("replace", 5488, 5489, 5358, 5359);
  ("equal", 5489, 5575, 5359, 5445);
  ("replace", 5575, 5576, 5445, 5446);
  ("equal", 5576, 5578, 5446, 5448);
  ("replace", 5578, 5579, 5448, 5449);
  ("equal", 5579, 5580, 5449, 5450);
  ("replace", 5580, 5596, 5450, 5455);
  ("equal", 5596, 5601, 5455, 5460);
  ("replace", 5601, 5603, 5460, 5462);
  ("equal", 5603, 5604, 5462, 5463);
  ("replace", 5604, 5605, 5463, 5464);
  ("equal", 5605, 5615, 5464, 5474);
  ("replace", 5615, 5616, 5474, 5475);
  ("equal", 5616, 5682, 5475, 5541);
  ("replace", 5682, 5683, 5541, 5542);
  ("equal", 5683, 5685, 5542, 5544);
  ("replace", 5685, 5686, 5544, 5545);
  ("equal", 5686, 5734, 5545, 5593);
  ("replace", 5734, 5735, 5593, 5594);
  ("equal", 5735, 5737, 5594, 5596);
  ("replace", 5737, 5738, 5596, 5597);
  ("equal", 5738, 5748, 5597, 5607);
  ("replace", 5748, 5749, 5607, 5608);
  ("equal", 5749, 5817, 5608, 5676);
  ("replace", 5817, 5818, 5676, 5677);
  ("equal", 5818, 5821, 5677, 5680);
  ("replace", 5821, 5822, 5680, 5681);
  ("equal", 5822, 5825, 5681, 5684);
  ("replace", 5825, 5826, 5684, 5685);
  ("equal", 5826, 5831, 5685, 5690);
  ("replace", 5831, 5832, 5690, 5691);
  ("equal", 5832, 5835, 5691, 5694);
  ("replace", 5835, 5836, 5694, 5695);
  ("equal", 5836, 5840, 5695, 5699);
  ("replace", 5840, 5841, 5699, 5700);
  ("equal", 5841, 5897, 5700, 5756);
  ("replace", 5897, 5899, 5756, 5758);
  ("equal", 5899, 5905, 5758, 5764);
  ("replace", 5905, 5906, 5764, 5765);
  ("equal", 5906, 5992, 5765, 5851);
  ("replace", 5992, 5993, 5851, 5852);
  ("equal", 5993, 5995, 5852, 5854);
  ("replace", 5995, 5996, 5854, 5855);
  ("equal", 5996, 5997, 5855, 5856);
  ("replace", 5997, 6013, 5856, 5862);
  ("equal", 6013, 6018, 5862, 5867);
  ("replace", 6018, 6020, 5867, 5869);
  ("equal", 6020, 6021, 5869, 5870);
  ("replace", 6021, 6022, 5870, 5871);
  ("equal", 6022, 6032, 5871, 5881);
  ("replace", 6032, 6033, 5881, 5882);
  ("equal", 6033, 6099, 5882, 5948);
  ("replace", 6099, 6100, 5948, 5949);
  ("equal", 6100, 6102, 5949, 5951);
  ("replace", 6102, 6103, 5951, 5952);
  ("equal", 6103, 6151, 5952, 6000);
  ("replace", 6151, 6152, 6000, 6001);
  ("equal", 6152, 6154, 6001, 6003);
  ("replace", 6154, 6155, 6003, 6004);
  ("equal", 6155, 6165, 6004, 6014);
  ("replace", 6165, 6166, 6014, 6015);
  ("equal", 6166, 6234, 6015, 6083);
  ("replace", 6234, 6235, 6083, 6084);
  ("equal", 6235, 6238, 6084, 6087);
  ("replace", 6238, 6239, 6087, 6088);
  ("equal", 6239, 6242, 6088, 6091);
  ("replace", 6242, 6243, 6091, 6092);
  ("equal", 6243, 6248, 6092, 6097);
  ("replace", 6248, 6249, 6097, 6098);
  ("equal", 6249, 6252, 6098, 6101);
  ("replace", 6252, 6253, 6101, 6102);
  ("equal", 6253, 6257, 6102, 6106);
  ("replace", 6257, 6258, 6106, 6107);
  ("equal", 6258, 6314, 6107, 6163);
  ("replace", 6314, 6316, 6163, 6165);
  ("equal", 6316, 6322, 6165, 6171);
  ("replace", 6322, 6323, 6171, 6172);
  ("equal", 6323, 6405, 6172, 6254);
  ("replace", 6405, 6406, 6254, 6255);
  ("equal", 6406, 6407, 6255, 6256);
  ("replace", 6407, 6425, 6256, 6263);
  ("equal", 6425, 6426, 6263, 6264);
  ("replace", 6426, 6428, 6264, 6265);
  ("equal", 6428, 6429, 6265, 6266);
  ("replace", 6429, 6434, 6266, 6270);
  ("equal", 6434, 6437, 6270, 6273);
  ("replace", 6437, 6438, 6273, 6274);
  ("equal", 6438, 6439, 6274, 6275);
  ("replace", 6439, 6440, 6275, 6276);
  ("equal", 6440, 6446, 6276, 6282);
  ("replace", 6446, 6449, 6282, 6284);
  ("equal", 6449, 6486, 6284, 6321);
  ("replace", 6486, 6487, 6321, 6322);
  ("equal", 6487, 6507, 6322, 6342);
  ("replace", 6507, 6510, 6342, 6344);
  ("equal", 6510, 6527, 6344, 6361);
  ("replace", 6527, 6536, 6361, 6370);
];;


(* ------------------------------------------------------------------------- *)
(* Full actions.                                                             *)
(* ------------------------------------------------------------------------- *)

let actions_merged = merge_actions (!actions1,actions2);;


(* ------------------------------------------------------------------------- *)
(* Define arrays describing which registers are containing dead values at    *)
(* each program location.                                                    *)
(*                                                                           *)
(* First, let's build dead value info of the original program. These don't   *)
(* consider callee-save register spills and ret.                             *)
(* ------------------------------------------------------------------------- *)

let local_montmul_p384_dead_value_info = [|
  []; (* 1: ldr q3, [x1] *)
  []; (* 2: ldr q25, [x2] *)
  []; (* 3: ldp x13, x23, [x2] *)
  []; (* 4: ldp x3, x21, [x1] *)
  []; (* 5: rev64 v23.4S, v25.4S *)
  [`Q25`]; (* 6: uzp1 v17.4S, v25.4S, v3.4S *)
  []; (* 7: umulh x15, x3, x13 *)
  [`Q23`]; (* 8: mul v6.4S, v23.4S, v3.4S *)
  []; (* 9: uzp1 v3.4S, v3.4S, v3.4S *)
  []; (* 10: ldr q27, [x2, #32] *)
  []; (* 11: ldp x8, x24, [x1, #16] *)
  []; (* 12: subs x6, x3, x21 *)
  []; (* 13: ldr q0, [x1, #32] *)
  []; (* 14: movi v23.2D, #0x00000000ffffffff *)
  []; (* 15: csetm x10, cc *)
  []; (* 16: umulh x19, x21, x23 *)
  []; (* 17: rev64 v4.4S, v27.4S *)
  []; (* 18: uzp2 v25.4S, v27.4S, v27.4S *)
  [`CF`; `NF`; `ZF`; `VF`; `X6`]; (* 19: cneg x4, x6, cc *)
  []; (* 20: subs x7, x23, x13 *)
  []; (* 21: xtn v22.2S, v0.2D *)
  [`Q27`]; (* 22: xtn v24.2S, v27.2D *)
  [`X7`]; (* 23: cneg x20, x7, cc *)
  []; (* 24: ldp x6, x14, [x2, #16] *)
  [`Q4`]; (* 25: mul v27.4S, v4.4S, v0.4S *)
  [`Q6`]; (* 26: uaddlp v20.2D, v6.4S *)
  [`CF`; `NF`; `ZF`; `VF`; `X10`]; (* 27: cinv x5, x10, cc *)
  []; (* 28: mul x16, x4, x20 *)
  [`Q0`]; (* 29: uzp2 v6.4S, v0.4S, v0.4S *)
  []; (* 30: umull v21.2D, v22.2S, v25.2S *)
  [`Q20`]; (* 31: shl v0.2D, v20.2D, #32 *)
  [`Q17`; `Q3`]; (* 32: umlal v0.2D, v3.2S, v17.2S *)
  []; (* 33: mul x22, x8, x6 *)
  [`Q25`]; (* 34: umull v1.2D, v6.2S, v25.2S *)
  []; (* 35: subs x12, x3, x8 *)
  []; (* 36: umull v20.2D, v22.2S, v24.2S *)
  [`X12`]; (* 37: cneg x17, x12, cc *)
  []; (* 38: umulh x9, x8, x6 *)
  []; (* 39: mov x12, v0.d[1] *)
  [`X16`]; (* 40: eor x11, x16, x5 *)
  [`Q0`]; (* 41: mov x7, v0.d[0] *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 42: csetm x10, cc *)
  [`Q20`]; (* 43: usra v21.2D, v20.2D, #32 *)
  [`X12`]; (* 44: adds x15, x15, x12 *)
  [`X19`; `X22`]; (* 45: adcs x12, x19, x22 *)
  [`X4`]; (* 46: umulh x20, x4, x20 *)
  [`CF`; `NF`; `ZF`; `VF`; `X9`]; (* 47: adc x19, x9, xzr *)
  []; (* 48: usra v1.2D, v21.2D, #32 *)
  []; (* 49: adds x22, x15, x7 *)
  [`Q21`; `Q23`]; (* 50: and v26.16B, v21.16B, v23.16B *)
  []; (* 51: adcs x16, x12, x15 *)
  [`Q27`]; (* 52: uaddlp v25.2D, v27.4S *)
  []; (* 53: adcs x9, x19, x12 *)
  [`Q6`]; (* 54: umlal v26.2D, v6.2S, v24.2S *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 55: adc x4, x19, xzr *)
  []; (* 56: adds x16, x16, x7 *)
  [`Q25`]; (* 57: shl v27.2D, v25.2D, #32 *)
  [`X15`]; (* 58: adcs x9, x9, x15 *)
  [`X12`]; (* 59: adcs x4, x4, x12 *)
  [`X20`]; (* 60: eor x12, x20, x5 *)
  [`CF`; `NF`; `ZF`; `VF`; `X19`]; (* 61: adc x15, x19, xzr *)
  []; (* 62: subs x20, x6, x13 *)
  []; (* 63: cneg x20, x20, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 64: cinv x10, x10, cc *)
  []; (* 65: cmn x5, #0x1 *)
  []; (* 66: mul x19, x17, x20 *)
  [`X22`]; (* 67: adcs x11, x22, x11 *)
  [`X16`]; (* 68: adcs x12, x16, x12 *)
  []; (* 69: adcs x9, x9, x5 *)
  [`X20`]; (* 70: umulh x17, x17, x20 *)
  [`X4`]; (* 71: adcs x22, x4, x5 *)
  [`CF`; `NF`; `ZF`; `VF`; `X15`]; (* 72: adc x5, x15, x5 *)
  []; (* 73: subs x16, x21, x8 *)
  [`X16`]; (* 74: cneg x20, x16, cc *)
  []; (* 75: eor x19, x19, x10 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 76: csetm x4, cc *)
  []; (* 77: subs x16, x6, x23 *)
  []; (* 78: cneg x16, x16, cc *)
  [`Q22`; `Q24`]; (* 79: umlal v27.2D, v22.2S, v24.2S *)
  []; (* 80: mul x15, x20, x16 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 81: cinv x4, x4, cc *)
  []; (* 82: cmn x10, #0x1 *)
  [`Q26`]; (* 83: usra v1.2D, v26.2D, #32 *)
  [`X12`]; (* 84: adcs x19, x12, x19 *)
  []; (* 85: eor x17, x17, x10 *)
  [`X17`]; (* 86: adcs x9, x9, x17 *)
  []; (* 87: adcs x22, x22, x10 *)
  []; (* 88: lsl x12, x7, #32 *)
  [`X16`]; (* 89: umulh x20, x20, x16 *)
  [`X15`]; (* 90: eor x16, x15, x4 *)
  []; (* 91: ldp x15, x17, [x2, #32] *)
  [`X12`; `X7`]; (* 92: add x2, x12, x7 *)
  [`CF`; `NF`; `ZF`; `VF`; `X10`; `X5`]; (* 93: adc x7, x5, x10 *)
  []; (* 94: ldp x5, x10, [x1, #32] *)
  []; (* 95: lsr x1, x2, #32 *)
  [`X20`]; (* 96: eor x12, x20, x4 *)
  []; (* 97: subs x1, x1, x2 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 98: sbc x20, x2, xzr *)
  []; (* 99: cmn x4, #0x1 *)
  [`X16`]; (* 100: adcs x9, x9, x16 *)
  []; (* 101: extr x1, x20, x1, #32 *)
  []; (* 102: lsr x20, x20, #32 *)
  [`X12`]; (* 103: adcs x22, x22, x12 *)
  [`CF`; `NF`; `ZF`; `VF`; `X4`; `X7`]; (* 104: adc x16, x7, x4 *)
  [`X20`]; (* 105: adds x12, x20, x2 *)
  []; (* 106: umulh x7, x24, x14 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 107: adc x4, xzr, xzr *)
  [`X11`]; (* 108: subs x1, x11, x1 *)
  [`X12`; `X19`]; (* 109: sbcs x20, x19, x12 *)
  [`X4`; `X9`]; (* 110: sbcs x12, x9, x4 *)
  []; (* 111: lsl x9, x1, #32 *)
  [`X9`]; (* 112: add x1, x9, x1 *)
  [`X22`]; (* 113: sbcs x9, x22, xzr *)
  []; (* 114: mul x22, x24, x14 *)
  []; (* 115: sbcs x16, x16, xzr *)
  []; (* 116: lsr x4, x1, #32 *)
  [`CF`; `NF`; `ZF`; `VF`; `X2`]; (* 117: sbc x19, x2, xzr *)
  []; (* 118: subs x4, x4, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 119: sbc x11, x1, xzr *)
  [`X4`]; (* 120: extr x2, x11, x4, #32 *)
  [`X11`]; (* 121: lsr x4, x11, #32 *)
  []; (* 122: adds x4, x4, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 123: adc x11, xzr, xzr *)
  [`X20`]; (* 124: subs x2, x20, x2 *)
  [`X12`]; (* 125: sbcs x4, x12, x4 *)
  [`X11`; `X9`]; (* 126: sbcs x20, x9, x11 *)
  []; (* 127: lsl x12, x2, #32 *)
  [`X12`]; (* 128: add x2, x12, x2 *)
  [`X16`]; (* 129: sbcs x9, x16, xzr *)
  []; (* 130: lsr x11, x2, #32 *)
  []; (* 131: sbcs x19, x19, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 132: sbc x1, x1, xzr *)
  [`X11`]; (* 133: subs x16, x11, x2 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 134: sbc x12, x2, xzr *)
  []; (* 135: extr x16, x12, x16, #32 *)
  []; (* 136: lsr x12, x12, #32 *)
  [`X12`]; (* 137: adds x11, x12, x2 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 138: adc x12, xzr, xzr *)
  [`X4`]; (* 139: subs x16, x4, x16 *)
  []; (* 140: mov x4, v27.d[0] *)
  [`X20`]; (* 141: sbcs x11, x20, x11 *)
  [`X12`; `X9`]; (* 142: sbcs x20, x9, x12 *)
  [`X11`; `X16`]; (* 143: stp x16, x11, [x0] *)
  [`X19`]; (* 144: sbcs x11, x19, xzr *)
  [`X1`]; (* 145: sbcs x9, x1, xzr *)
  [`X11`; `X20`]; (* 146: stp x20, x11, [x0, #16] *)
  []; (* 147: mov x1, v1.d[0] *)
  [`CF`; `NF`; `ZF`; `VF`; `X2`]; (* 148: sbc x20, x2, xzr *)
  []; (* 149: subs x12, x24, x5 *)
  [`Q27`]; (* 150: mov x11, v27.d[1] *)
  [`X12`]; (* 151: cneg x16, x12, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 152: csetm x2, cc *)
  []; (* 153: subs x19, x15, x14 *)
  [`Q1`]; (* 154: mov x12, v1.d[1] *)
  []; (* 155: cinv x2, x2, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 156: cneg x19, x19, cc *)
  [`X20`; `X9`]; (* 157: stp x9, x20, [x0, #32] *)
  []; (* 158: mul x9, x16, x19 *)
  [`X7`]; (* 159: adds x4, x7, x4 *)
  [`X1`]; (* 160: adcs x11, x1, x11 *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 161: adc x1, x12, xzr *)
  []; (* 162: adds x20, x4, x22 *)
  [`X16`]; (* 163: umulh x19, x16, x19 *)
  []; (* 164: adcs x7, x11, x4 *)
  [`X9`]; (* 165: eor x16, x9, x2 *)
  []; (* 166: adcs x9, x1, x11 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 167: adc x12, x1, xzr *)
  []; (* 168: adds x7, x7, x22 *)
  [`X9`]; (* 169: adcs x4, x9, x4 *)
  [`X11`; `X12`]; (* 170: adcs x9, x12, x11 *)
  [`CF`; `NF`; `ZF`; `VF`; `X1`]; (* 171: adc x12, x1, xzr *)
  []; (* 172: cmn x2, #0x1 *)
  [`X19`]; (* 173: eor x1, x19, x2 *)
  [`X16`; `X20`]; (* 174: adcs x11, x20, x16 *)
  [`X1`; `X7`]; (* 175: adcs x19, x7, x1 *)
  [`X4`]; (* 176: adcs x1, x4, x2 *)
  [`X9`]; (* 177: adcs x20, x9, x2 *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 178: adc x2, x12, x2 *)
  []; (* 179: subs x12, x24, x10 *)
  [`X12`]; (* 180: cneg x16, x12, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 181: csetm x12, cc *)
  []; (* 182: subs x9, x17, x14 *)
  []; (* 183: cinv x12, x12, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 184: cneg x9, x9, cc *)
  [`X24`]; (* 185: subs x3, x24, x3 *)
  []; (* 186: sbcs x21, x5, x21 *)
  []; (* 187: mul x24, x16, x9 *)
  [`X8`]; (* 188: sbcs x4, x10, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 189: ngc x8, xzr *)
  [`X5`]; (* 190: subs x10, x5, x10 *)
  [`X24`]; (* 191: eor x5, x24, x12 *)
  []; (* 192: csetm x7, cc *)
  [`CF`; `NF`; `ZF`; `VF`; `X10`]; (* 193: cneg x24, x10, cc *)
  []; (* 194: subs x10, x17, x15 *)
  []; (* 195: cinv x7, x7, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 196: cneg x10, x10, cc *)
  [`X13`]; (* 197: subs x14, x13, x14 *)
  [`X23`]; (* 198: sbcs x15, x23, x15 *)
  [`X21`]; (* 199: eor x13, x21, x8 *)
  []; (* 200: mul x23, x24, x10 *)
  [`X6`]; (* 201: sbcs x17, x6, x17 *)
  [`X3`]; (* 202: eor x6, x3, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 203: ngc x21, xzr *)
  [`X16`]; (* 204: umulh x9, x16, x9 *)
  []; (* 205: cmn x8, #0x1 *)
  [`X23`]; (* 206: eor x3, x23, x7 *)
  [`X6`]; (* 207: adcs x23, x6, xzr *)
  []; (* 208: adcs x13, x13, xzr *)
  [`X4`]; (* 209: eor x16, x4, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 210: adc x16, x16, xzr *)
  [`X17`]; (* 211: eor x4, x17, x21 *)
  [`X10`; `X24`]; (* 212: umulh x17, x24, x10 *)
  []; (* 213: cmn x21, #0x1 *)
  [`X14`]; (* 214: eor x24, x14, x21 *)
  [`X15`]; (* 215: eor x6, x15, x21 *)
  [`X24`]; (* 216: adcs x15, x24, xzr *)
  [`X6`]; (* 217: adcs x14, x6, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X4`]; (* 218: adc x6, x4, xzr *)
  []; (* 219: cmn x12, #0x1 *)
  [`X9`]; (* 220: eor x4, x9, x12 *)
  [`X5`]; (* 221: adcs x19, x19, x5 *)
  []; (* 222: umulh x5, x23, x15 *)
  [`X4`]; (* 223: adcs x1, x1, x4 *)
  [`X20`]; (* 224: adcs x10, x20, x12 *)
  [`X17`]; (* 225: eor x4, x17, x7 *)
  []; (* 226: ldp x20, x9, [x0] *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 227: adc x2, x2, x12 *)
  []; (* 228: cmn x7, #0x1 *)
  [`X1`; `X3`]; (* 229: adcs x12, x1, x3 *)
  []; (* 230: ldp x17, x24, [x0, #16] *)
  []; (* 231: mul x1, x16, x6 *)
  [`X10`; `X4`]; (* 232: adcs x3, x10, x4 *)
  [`CF`; `NF`; `ZF`; `VF`; `X7`]; (* 233: adc x2, x2, x7 *)
  []; (* 234: ldp x7, x4, [x0, #32] *)
  [`X22`]; (* 235: adds x20, x22, x20 *)
  []; (* 236: mul x10, x13, x14 *)
  [`X9`]; (* 237: adcs x11, x11, x9 *)
  [`X21`; `X8`]; (* 238: eor x9, x8, x21 *)
  [`X17`; `X19`]; (* 239: adcs x21, x19, x17 *)
  [`X11`; `X20`]; (* 240: stp x20, x11, [x0] *)
  [`X24`]; (* 241: adcs x12, x12, x24 *)
  []; (* 242: mul x8, x23, x15 *)
  [`X7`]; (* 243: adcs x3, x3, x7 *)
  [`X12`; `X21`]; (* 244: stp x21, x12, [x0, #16] *)
  [`X2`; `X4`]; (* 245: adcs x12, x2, x4 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 246: adc x19, xzr, xzr *)
  []; (* 247: subs x21, x23, x16 *)
  []; (* 248: umulh x2, x16, x6 *)
  [`X12`; `X3`]; (* 249: stp x3, x12, [x0, #32] *)
  [`X21`]; (* 250: cneg x3, x21, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 251: csetm x24, cc *)
  []; (* 252: umulh x11, x13, x14 *)
  [`X16`]; (* 253: subs x21, x13, x16 *)
  []; (* 254: eor x7, x8, x9 *)
  [`X21`]; (* 255: cneg x17, x21, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 256: csetm x16, cc *)
  []; (* 257: subs x21, x6, x15 *)
  [`X21`]; (* 258: cneg x22, x21, cc *)
  [`CF`; `NF`; `ZF`; `VF`; `X24`]; (* 259: cinv x21, x24, cc *)
  [`X13`; `X23`]; (* 260: subs x20, x23, x13 *)
  []; (* 261: umulh x12, x3, x22 *)
  [`X20`]; (* 262: cneg x23, x20, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 263: csetm x24, cc *)
  [`X15`]; (* 264: subs x20, x14, x15 *)
  []; (* 265: cinv x24, x24, cc *)
  [`X3`]; (* 266: mul x22, x3, x22 *)
  [`CF`; `NF`; `ZF`; `VF`; `X20`]; (* 267: cneg x3, x20, cc *)
  [`X14`; `X6`]; (* 268: subs x13, x6, x14 *)
  [`X13`]; (* 269: cneg x20, x13, cc *)
  [`CF`; `NF`; `ZF`; `VF`; `X16`]; (* 270: cinv x15, x16, cc *)
  [`X10`; `X5`]; (* 271: adds x13, x5, x10 *)
  []; (* 272: mul x4, x23, x3 *)
  [`X1`]; (* 273: adcs x11, x11, x1 *)
  [`CF`; `NF`; `ZF`; `VF`; `X2`]; (* 274: adc x14, x2, xzr *)
  []; (* 275: adds x5, x13, x8 *)
  []; (* 276: adcs x16, x11, x13 *)
  [`X3`]; (* 277: umulh x23, x23, x3 *)
  []; (* 278: adcs x3, x14, x11 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 279: adc x1, x14, xzr *)
  [`X16`; `X8`]; (* 280: adds x10, x16, x8 *)
  [`X13`; `X3`]; (* 281: adcs x6, x3, x13 *)
  [`X1`; `X11`]; (* 282: adcs x8, x1, x11 *)
  []; (* 283: umulh x13, x17, x20 *)
  [`X4`]; (* 284: eor x1, x4, x24 *)
  [`CF`; `NF`; `ZF`; `VF`; `X14`]; (* 285: adc x4, x14, xzr *)
  []; (* 286: cmn x24, #0x1 *)
  [`X5`]; (* 287: adcs x1, x5, x1 *)
  [`X23`]; (* 288: eor x16, x23, x24 *)
  [`X1`]; (* 289: eor x11, x1, x9 *)
  [`X10`; `X16`]; (* 290: adcs x23, x10, x16 *)
  [`X22`]; (* 291: eor x2, x22, x21 *)
  [`X6`]; (* 292: adcs x3, x6, x24 *)
  [`X17`; `X20`]; (* 293: mul x14, x17, x20 *)
  [`X13`]; (* 294: eor x17, x13, x15 *)
  [`X8`]; (* 295: adcs x13, x8, x24 *)
  [`CF`; `NF`; `ZF`; `VF`; `X24`; `X4`]; (* 296: adc x8, x4, x24 *)
  []; (* 297: cmn x21, #0x1 *)
  [`X2`; `X23`]; (* 298: adcs x6, x23, x2 *)
  []; (* 299: mov x16, #0xfffffffffffffffe *)
  [`X12`]; (* 300: eor x20, x12, x21 *)
  [`X3`]; (* 301: adcs x20, x3, x20 *)
  [`X14`]; (* 302: eor x23, x14, x15 *)
  [`X13`]; (* 303: adcs x2, x13, x21 *)
  [`CF`; `NF`; `ZF`; `VF`; `X21`]; (* 304: adc x8, x8, x21 *)
  []; (* 305: cmn x15, #0x1 *)
  []; (* 306: ldp x5, x4, [x0] *)
  []; (* 307: ldp x21, x12, [x0, #16] *)
  [`X20`; `X23`]; (* 308: adcs x22, x20, x23 *)
  [`X22`]; (* 309: eor x23, x22, x9 *)
  [`X2`]; (* 310: adcs x17, x2, x17 *)
  [`CF`; `NF`; `ZF`; `VF`; `X15`; `X8`]; (* 311: adc x22, x8, x15 *)
  []; (* 312: cmn x9, #0x1 *)
  [`X7`]; (* 313: adcs x15, x7, x5 *)
  []; (* 314: ldp x10, x14, [x0, #32] *)
  [`X6`]; (* 315: eor x1, x6, x9 *)
  []; (* 316: lsl x2, x15, #32 *)
  [`X11`]; (* 317: adcs x8, x11, x4 *)
  [`X1`]; (* 318: adcs x13, x1, x21 *)
  [`X22`]; (* 319: eor x1, x22, x9 *)
  [`X23`]; (* 320: adcs x24, x23, x12 *)
  [`X17`]; (* 321: eor x11, x17, x9 *)
  [`X11`]; (* 322: adcs x23, x11, x10 *)
  [`X1`]; (* 323: adcs x7, x1, x14 *)
  []; (* 324: adcs x17, x9, x19 *)
  []; (* 325: adcs x20, x9, xzr *)
  [`X15`; `X2`]; (* 326: add x1, x2, x15 *)
  []; (* 327: lsr x3, x1, #32 *)
  []; (* 328: adcs x11, x9, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 329: adc x9, x9, xzr *)
  []; (* 330: subs x3, x3, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 331: sbc x6, x1, xzr *)
  [`X5`]; (* 332: adds x24, x24, x5 *)
  [`X23`]; (* 333: adcs x4, x23, x4 *)
  []; (* 334: extr x3, x6, x3, #32 *)
  []; (* 335: lsr x6, x6, #32 *)
  [`X7`]; (* 336: adcs x21, x7, x21 *)
  [`X12`; `X17`]; (* 337: adcs x15, x17, x12 *)
  [`X10`; `X20`]; (* 338: adcs x7, x20, x10 *)
  [`X11`; `X14`]; (* 339: adcs x20, x11, x14 *)
  []; (* 340: mov x14, #0xffffffff *)
  [`CF`; `NF`; `ZF`; `VF`; `X19`; `X9`]; (* 341: adc x22, x9, x19 *)
  [`X6`]; (* 342: adds x12, x6, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 343: adc x10, xzr, xzr *)
  [`X8`]; (* 344: subs x3, x8, x3 *)
  [`X13`]; (* 345: sbcs x12, x13, x12 *)
  []; (* 346: lsl x9, x3, #32 *)
  [`X9`]; (* 347: add x3, x9, x3 *)
  [`X24`]; (* 348: sbcs x10, x24, x10 *)
  [`X4`]; (* 349: sbcs x24, x4, xzr *)
  []; (* 350: lsr x9, x3, #32 *)
  []; (* 351: sbcs x21, x21, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 352: sbc x1, x1, xzr *)
  []; (* 353: subs x9, x9, x3 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 354: sbc x13, x3, xzr *)
  []; (* 355: extr x9, x13, x9, #32 *)
  []; (* 356: lsr x13, x13, #32 *)
  []; (* 357: adds x13, x13, x3 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 358: adc x6, xzr, xzr *)
  [`X9`]; (* 359: subs x12, x12, x9 *)
  [`X10`; `X13`]; (* 360: sbcs x17, x10, x13 *)
  []; (* 361: lsl x2, x12, #32 *)
  [`X24`; `X6`]; (* 362: sbcs x10, x24, x6 *)
  [`X12`; `X2`]; (* 363: add x9, x2, x12 *)
  [`X21`]; (* 364: sbcs x6, x21, xzr *)
  []; (* 365: lsr x5, x9, #32 *)
  [`X1`]; (* 366: sbcs x21, x1, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X3`]; (* 367: sbc x13, x3, xzr *)
  [`X5`]; (* 368: subs x8, x5, x9 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 369: sbc x19, x9, xzr *)
  []; (* 370: lsr x12, x19, #32 *)
  [`X19`; `X8`]; (* 371: extr x3, x19, x8, #32 *)
  [`X12`]; (* 372: adds x8, x12, x9 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 373: adc x1, xzr, xzr *)
  [`X17`; `X3`]; (* 374: subs x2, x17, x3 *)
  [`X10`; `X8`]; (* 375: sbcs x12, x10, x8 *)
  [`X1`; `X6`]; (* 376: sbcs x5, x6, x1 *)
  [`X21`]; (* 377: sbcs x3, x21, xzr *)
  [`X13`]; (* 378: sbcs x19, x13, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X9`]; (* 379: sbc x24, x9, xzr *)
  [`X15`; `X3`]; (* 380: adds x23, x15, x3 *)
  [`X19`; `X7`]; (* 381: adcs x8, x7, x19 *)
  [`X20`; `X24`]; (* 382: adcs x11, x20, x24 *)
  [`CF`; `NF`; `ZF`; `VF`; `X22`]; (* 383: adc x9, x22, xzr *)
  [`X9`]; (* 384: add x24, x9, #0x1 *)
  []; (* 385: lsl x7, x24, #32 *)
  []; (* 386: subs x21, x24, x7 *)
  [`CF`; `NF`; `ZF`; `VF`; `X7`]; (* 387: sbc x10, x7, xzr *)
  [`X2`; `X21`]; (* 388: adds x6, x2, x21 *)
  [`X10`; `X12`]; (* 389: adcs x7, x12, x10 *)
  [`X5`]; (* 390: adcs x24, x5, x24 *)
  [`X23`]; (* 391: adcs x13, x23, xzr *)
  []; (* 392: adcs x8, x8, xzr *)
  [`X11`]; (* 393: adcs x15, x11, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 394: csetm x23, cc *)
  [`X16`]; (* 395: and x11, x16, x23 *)
  [`X14`]; (* 396: and x20, x14, x23 *)
  [`X6`]; (* 397: adds x22, x6, x20 *)
  [`X20`]; (* 398: eor x3, x20, x23 *)
  [`X3`; `X7`]; (* 399: adcs x5, x7, x3 *)
  [`X11`; `X24`]; (* 400: adcs x14, x24, x11 *)
  [`X22`; `X5`]; (* 401: stp x22, x5, [x0] *)
  [`X13`]; (* 402: adcs x5, x13, x23 *)
  [`X8`]; (* 403: adcs x21, x8, x23 *)
  [`X14`; `X5`]; (* 404: stp x14, x5, [x0, #16] *)
  [`CF`; `NF`; `ZF`; `VF`; `X15`; `X23`]; (* 405: adc x12, x15, x23 *)
  [`X12`; `X21`]; (* 406: stp x21, x12, [x0, #32] *)
|];;

let local_montsqr_p384_dead_value_info = [|
  []; (* 1: ldr q1, [x1] *)
  []; (* 2: ldp x9, x2, [x1] *)
  []; (* 3: ldr q0, [x1] *)
  []; (* 4: ldp x4, x6, [x1, #16] *)
  []; (* 5: rev64 v21.4S, v1.4S *)
  []; (* 6: uzp2 v28.4S, v1.4S, v1.4S *)
  []; (* 7: umulh x7, x9, x2 *)
  [`Q1`]; (* 8: xtn v17.2S, v1.2D *)
  [`Q21`]; (* 9: mul v27.4S, v21.4S, v0.4S *)
  []; (* 10: ldr q20, [x1, #32] *)
  []; (* 11: xtn v30.2S, v0.2D *)
  []; (* 12: ldr q1, [x1, #32] *)
  [`Q0`]; (* 13: uzp2 v31.4S, v0.4S, v0.4S *)
  []; (* 14: ldp x5, x10, [x1, #32] *)
  []; (* 15: umulh x8, x9, x4 *)
  [`Q27`]; (* 16: uaddlp v3.2D, v27.4S *)
  []; (* 17: umull v16.2D, v30.2S, v17.2S *)
  []; (* 18: mul x16, x9, x4 *)
  []; (* 19: umull v27.2D, v30.2S, v28.2S *)
  []; (* 20: shrn v0.2S, v20.2D, #32 *)
  [`Q20`]; (* 21: xtn v7.2S, v20.2D *)
  [`Q3`]; (* 22: shl v20.2D, v3.2D, #32 *)
  [`Q28`]; (* 23: umull v3.2D, v31.2S, v28.2S *)
  []; (* 24: mul x3, x2, x4 *)
  [`Q30`]; (* 25: umlal v20.2D, v30.2S, v17.2S *)
  [`Q0`]; (* 26: umull v22.2D, v7.2S, v0.2S *)
  [`Q16`]; (* 27: usra v27.2D, v16.2D, #32 *)
  []; (* 28: umulh x11, x2, x4 *)
  []; (* 29: movi v21.2D, #0x00000000ffffffff *)
  []; (* 30: uzp2 v28.4S, v1.4S, v1.4S *)
  [`X16`; `X7`]; (* 31: adds x15, x16, x7 *)
  [`Q21`]; (* 32: and v5.16B, v27.16B, v21.16B *)
  [`X8`]; (* 33: adcs x3, x3, x8 *)
  [`Q27`]; (* 34: usra v3.2D, v27.2D, #32 *)
  []; (* 35: dup v29.2D, x6 *)
  [`CF`; `NF`; `ZF`; `VF`; `X11`]; (* 36: adcs x16, x11, xzr *)
  []; (* 37: mov x14, v20.d[0] *)
  [`Q17`; `Q31`]; (* 38: umlal v5.2D, v31.2S, v17.2S *)
  []; (* 39: mul x8, x9, x2 *)
  [`Q20`]; (* 40: mov x7, v20.d[1] *)
  [`Q22`]; (* 41: shl v19.2D, v22.2D, #33 *)
  []; (* 42: xtn v25.2S, v29.2D *)
  []; (* 43: rev64 v31.4S, v1.4S *)
  []; (* 44: lsl x13, x14, #32 *)
  []; (* 45: uzp2 v6.4S, v29.4S, v29.4S *)
  [`Q7`]; (* 46: umlal v19.2D, v7.2S, v7.2S *)
  [`Q5`]; (* 47: usra v3.2D, v5.2D, #32 *)
  [`X8`]; (* 48: adds x1, x8, x8 *)
  []; (* 49: umulh x8, x4, x4 *)
  [`X13`; `X14`]; (* 50: add x12, x13, x14 *)
  [`Q29`; `Q31`]; (* 51: mul v17.4S, v31.4S, v29.4S *)
  [`Q1`]; (* 52: xtn v4.2S, v1.2D *)
  [`X15`]; (* 53: adcs x14, x15, x15 *)
  []; (* 54: lsr x13, x12, #32 *)
  [`X3`]; (* 55: adcs x15, x3, x3 *)
  []; (* 56: umull v31.2D, v25.2S, v28.2S *)
  [`X16`]; (* 57: adcs x11, x16, x16 *)
  []; (* 58: umull v21.2D, v25.2S, v4.2S *)
  []; (* 59: mov x17, v3.d[0] *)
  [`Q28`]; (* 60: umull v18.2D, v6.2S, v28.2S *)
  [`CF`; `NF`; `ZF`; `VF`; `X8`]; (* 61: adc x16, x8, xzr *)
  [`Q17`]; (* 62: uaddlp v16.2D, v17.4S *)
  []; (* 63: movi v1.2D, #0x00000000ffffffff *)
  []; (* 64: subs x13, x13, x12 *)
  [`Q21`]; (* 65: usra v31.2D, v21.2D, #32 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 66: sbc x8, x12, xzr *)
  [`X1`]; (* 67: adds x17, x17, x1 *)
  []; (* 68: mul x1, x4, x4 *)
  [`Q16`]; (* 69: shl v28.2D, v16.2D, #32 *)
  [`Q3`]; (* 70: mov x3, v3.d[1] *)
  [`X7`]; (* 71: adcs x14, x7, x14 *)
  [`X13`]; (* 72: extr x7, x8, x13, #32 *)
  [`X15`; `X3`]; (* 73: adcs x13, x3, x15 *)
  [`Q1`]; (* 74: and v3.16B, v31.16B, v1.16B *)
  [`X1`]; (* 75: adcs x11, x1, x11 *)
  [`X8`]; (* 76: lsr x1, x8, #32 *)
  [`Q6`]; (* 77: umlal v3.2D, v6.2S, v4.2S *)
  [`Q31`]; (* 78: usra v18.2D, v31.2D, #32 *)
  [`CF`; `NF`; `ZF`; `VF`; `X16`]; (* 79: adc x3, x16, xzr *)
  []; (* 80: adds x1, x1, x12 *)
  [`Q25`; `Q4`]; (* 81: umlal v28.2D, v25.2S, v4.2S *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 82: adc x16, xzr, xzr *)
  [`X17`; `X7`]; (* 83: subs x15, x17, x7 *)
  [`X1`; `X14`]; (* 84: sbcs x7, x14, x1 *)
  []; (* 85: lsl x1, x15, #32 *)
  [`X13`]; (* 86: sbcs x16, x13, x16 *)
  [`X1`; `X15`]; (* 87: add x8, x1, x15 *)
  [`Q3`]; (* 88: usra v18.2D, v3.2D, #32 *)
  [`X11`]; (* 89: sbcs x14, x11, xzr *)
  []; (* 90: lsr x1, x8, #32 *)
  [`X3`]; (* 91: sbcs x17, x3, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 92: sbc x11, x12, xzr *)
  [`X1`]; (* 93: subs x13, x1, x8 *)
  []; (* 94: umulh x12, x4, x10 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 95: sbc x1, x8, xzr *)
  []; (* 96: extr x13, x1, x13, #32 *)
  []; (* 97: lsr x1, x1, #32 *)
  [`X1`]; (* 98: adds x15, x1, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 99: adc x1, xzr, xzr *)
  [`X13`]; (* 100: subs x7, x7, x13 *)
  [`X15`; `X16`]; (* 101: sbcs x13, x16, x15 *)
  []; (* 102: lsl x3, x7, #32 *)
  []; (* 103: umulh x16, x2, x5 *)
  [`X1`; `X14`]; (* 104: sbcs x15, x14, x1 *)
  [`X3`]; (* 105: add x7, x3, x7 *)
  [`X17`]; (* 106: sbcs x3, x17, xzr *)
  []; (* 107: lsr x1, x7, #32 *)
  [`X11`]; (* 108: sbcs x14, x11, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X8`]; (* 109: sbc x11, x8, xzr *)
  [`X1`]; (* 110: subs x8, x1, x7 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 111: sbc x1, x7, xzr *)
  []; (* 112: extr x8, x1, x8, #32 *)
  []; (* 113: lsr x1, x1, #32 *)
  []; (* 114: adds x1, x1, x7 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 115: adc x17, xzr, xzr *)
  [`X8`]; (* 116: subs x13, x13, x8 *)
  []; (* 117: umulh x8, x9, x6 *)
  [`X15`]; (* 118: sbcs x1, x15, x1 *)
  [`X17`; `X3`]; (* 119: sbcs x15, x3, x17 *)
  [`X14`]; (* 120: sbcs x3, x14, xzr *)
  []; (* 121: mul x17, x2, x5 *)
  []; (* 122: sbcs x11, x11, xzr *)
  [`X1`; `X13`]; (* 123: stp x13, x1, [x0] *)
  [`CF`; `NF`; `ZF`; `VF`; `X7`]; (* 124: sbc x14, x7, xzr *)
  []; (* 125: mul x7, x4, x10 *)
  []; (* 126: subs x1, x9, x2 *)
  [`X15`; `X3`]; (* 127: stp x15, x3, [x0, #16] *)
  []; (* 128: csetm x15, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 129: cneg x1, x1, cc *)
  [`X11`; `X14`]; (* 130: stp x11, x14, [x0, #32] *)
  []; (* 131: mul x14, x9, x6 *)
  [`X8`]; (* 132: adds x17, x8, x17 *)
  [`X16`]; (* 133: adcs x7, x16, x7 *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 134: adc x13, x12, xzr *)
  []; (* 135: subs x12, x5, x6 *)
  [`X12`]; (* 136: cneg x3, x12, cc *)
  [`CF`; `NF`; `ZF`; `VF`; `X15`]; (* 137: cinv x16, x15, cc *)
  []; (* 138: mul x8, x1, x3 *)
  [`X3`]; (* 139: umulh x1, x1, x3 *)
  [`X8`]; (* 140: eor x12, x8, x16 *)
  []; (* 141: adds x11, x17, x14 *)
  []; (* 142: adcs x3, x7, x17 *)
  []; (* 143: adcs x15, x13, x7 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 144: adc x8, x13, xzr *)
  []; (* 145: adds x3, x3, x14 *)
  [`X17`]; (* 146: adcs x15, x15, x17 *)
  [`X7`; `X8`]; (* 147: adcs x17, x8, x7 *)
  []; (* 148: eor x1, x1, x16 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 149: adc x13, x13, xzr *)
  []; (* 150: subs x9, x9, x4 *)
  []; (* 151: csetm x8, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 152: cneg x9, x9, cc *)
  [`X2`]; (* 153: subs x4, x2, x4 *)
  []; (* 154: cneg x4, x4, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 155: csetm x7, cc *)
  []; (* 156: subs x2, x10, x6 *)
  []; (* 157: cinv x8, x8, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 158: cneg x2, x2, cc *)
  []; (* 159: cmn x16, #0x1 *)
  [`X12`]; (* 160: adcs x11, x11, x12 *)
  []; (* 161: mul x12, x9, x2 *)
  [`X1`]; (* 162: adcs x3, x3, x1 *)
  []; (* 163: adcs x15, x15, x16 *)
  [`X2`]; (* 164: umulh x9, x9, x2 *)
  []; (* 165: adcs x17, x17, x16 *)
  [`CF`; `NF`; `ZF`; `VF`; `X16`]; (* 166: adc x13, x13, x16 *)
  []; (* 167: subs x1, x10, x5 *)
  [`X7`]; (* 168: cinv x2, x7, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 169: cneg x1, x1, cc *)
  []; (* 170: eor x9, x9, x8 *)
  []; (* 171: cmn x8, #0x1 *)
  [`X12`]; (* 172: eor x7, x12, x8 *)
  []; (* 173: mul x12, x4, x1 *)
  [`X7`]; (* 174: adcs x3, x3, x7 *)
  [`X15`; `X9`]; (* 175: adcs x7, x15, x9 *)
  [`X17`]; (* 176: adcs x15, x17, x8 *)
  []; (* 177: ldp x9, x17, [x0, #16] *)
  [`X1`]; (* 178: umulh x4, x4, x1 *)
  [`CF`; `NF`; `ZF`; `VF`; `X13`]; (* 179: adc x8, x13, x8 *)
  []; (* 180: cmn x2, #0x1 *)
  [`X12`]; (* 181: eor x1, x12, x2 *)
  [`X7`]; (* 182: adcs x1, x7, x1 *)
  []; (* 183: ldp x7, x16, [x0] *)
  [`X4`]; (* 184: eor x12, x4, x2 *)
  [`X12`; `X15`]; (* 185: adcs x4, x15, x12 *)
  []; (* 186: ldp x15, x12, [x0, #32] *)
  [`CF`; `NF`; `ZF`; `VF`; `X2`]; (* 187: adc x8, x8, x2 *)
  [`X14`]; (* 188: adds x13, x14, x14 *)
  []; (* 189: umulh x14, x5, x10 *)
  [`X11`]; (* 190: adcs x2, x11, x11 *)
  []; (* 191: adcs x3, x3, x3 *)
  []; (* 192: adcs x1, x1, x1 *)
  []; (* 193: adcs x4, x4, x4 *)
  [`X8`]; (* 194: adcs x11, x8, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 195: adc x8, xzr, xzr *)
  [`X7`]; (* 196: adds x13, x13, x7 *)
  [`X16`]; (* 197: adcs x2, x2, x16 *)
  []; (* 198: mul x16, x5, x10 *)
  [`X9`]; (* 199: adcs x3, x3, x9 *)
  [`X17`]; (* 200: adcs x1, x1, x17 *)
  []; (* 201: umulh x5, x5, x5 *)
  []; (* 202: lsl x9, x13, #32 *)
  [`X13`]; (* 203: add x9, x9, x13 *)
  [`X15`]; (* 204: adcs x4, x4, x15 *)
  []; (* 205: mov x13, v28.d[1] *)
  [`X11`; `X12`]; (* 206: adcs x15, x11, x12 *)
  []; (* 207: lsr x7, x9, #32 *)
  [`CF`; `NF`; `ZF`; `VF`; `X8`]; (* 208: adc x11, x8, xzr *)
  []; (* 209: subs x7, x7, x9 *)
  []; (* 210: umulh x10, x10, x10 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 211: sbc x17, x9, xzr *)
  []; (* 212: extr x7, x17, x7, #32 *)
  []; (* 213: lsr x17, x17, #32 *)
  []; (* 214: adds x17, x17, x9 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 215: adc x12, xzr, xzr *)
  [`X2`; `X7`]; (* 216: subs x8, x2, x7 *)
  [`X3`]; (* 217: sbcs x17, x3, x17 *)
  []; (* 218: lsl x7, x8, #32 *)
  [`X1`; `X12`]; (* 219: sbcs x2, x1, x12 *)
  [`X7`; `X8`]; (* 220: add x3, x7, x8 *)
  [`X4`]; (* 221: sbcs x12, x4, xzr *)
  []; (* 222: lsr x1, x3, #32 *)
  [`X15`]; (* 223: sbcs x7, x15, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X9`]; (* 224: sbc x15, x9, xzr *)
  []; (* 225: subs x1, x1, x3 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 226: sbc x4, x3, xzr *)
  []; (* 227: lsr x9, x4, #32 *)
  [`X1`; `X4`]; (* 228: extr x8, x4, x1, #32 *)
  []; (* 229: adds x9, x9, x3 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 230: adc x4, xzr, xzr *)
  [`X17`; `X8`]; (* 231: subs x1, x17, x8 *)
  []; (* 232: lsl x17, x1, #32 *)
  [`X2`; `X9`]; (* 233: sbcs x8, x2, x9 *)
  [`X12`; `X4`]; (* 234: sbcs x9, x12, x4 *)
  [`X1`]; (* 235: add x17, x17, x1 *)
  []; (* 236: mov x1, v18.d[1] *)
  []; (* 237: lsr x2, x17, #32 *)
  []; (* 238: sbcs x7, x7, xzr *)
  [`Q18`]; (* 239: mov x12, v18.d[0] *)
  []; (* 240: sbcs x15, x15, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 241: sbc x3, x3, xzr *)
  [`X2`]; (* 242: subs x4, x2, x17 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 243: sbc x2, x17, xzr *)
  [`X13`]; (* 244: adds x12, x13, x12 *)
  [`X1`]; (* 245: adcs x16, x16, x1 *)
  []; (* 246: lsr x13, x2, #32 *)
  [`X2`; `X4`]; (* 247: extr x1, x2, x4, #32 *)
  [`CF`; `NF`; `ZF`; `VF`; `X14`]; (* 248: adc x2, x14, xzr *)
  [`X13`]; (* 249: adds x4, x13, x17 *)
  []; (* 250: mul x13, x6, x6 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 251: adc x14, xzr, xzr *)
  [`X8`]; (* 252: subs x1, x8, x1 *)
  [`X9`]; (* 253: sbcs x4, x9, x4 *)
  [`Q28`]; (* 254: mov x9, v28.d[0] *)
  [`X14`]; (* 255: sbcs x7, x7, x14 *)
  [`X15`]; (* 256: sbcs x8, x15, xzr *)
  []; (* 257: sbcs x3, x3, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X17`]; (* 258: sbc x14, x17, xzr *)
  [`X9`]; (* 259: adds x17, x9, x9 *)
  []; (* 260: adcs x12, x12, x12 *)
  []; (* 261: mov x15, v19.d[0] *)
  [`X16`]; (* 262: adcs x9, x16, x16 *)
  []; (* 263: umulh x6, x6, x6 *)
  [`X2`]; (* 264: adcs x16, x2, x2 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 265: adc x2, xzr, xzr *)
  [`X8`]; (* 266: adds x11, x11, x8 *)
  []; (* 267: adcs x3, x3, xzr *)
  []; (* 268: adcs x14, x14, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 269: adcs x8, xzr, xzr *)
  [`X1`]; (* 270: adds x13, x1, x13 *)
  [`Q19`]; (* 271: mov x1, v19.d[1] *)
  [`X4`]; (* 272: adcs x6, x4, x6 *)
  []; (* 273: mov x4, #0xffffffff *)
  [`X7`]; (* 274: adcs x15, x7, x15 *)
  [`X11`; `X5`]; (* 275: adcs x7, x11, x5 *)
  [`X3`]; (* 276: adcs x1, x3, x1 *)
  [`X10`]; (* 277: adcs x14, x14, x10 *)
  [`CF`; `NF`; `ZF`; `VF`; `X8`]; (* 278: adc x11, x8, xzr *)
  [`X17`]; (* 279: adds x6, x6, x17 *)
  [`X12`; `X15`]; (* 280: adcs x8, x15, x12 *)
  [`X7`; `X9`]; (* 281: adcs x3, x7, x9 *)
  [`X1`; `X16`]; (* 282: adcs x15, x1, x16 *)
  []; (* 283: mov x16, #0xffffffff00000001 *)
  [`X2`]; (* 284: adcs x14, x14, x2 *)
  []; (* 285: mov x2, #0x1 *)
  [`CF`; `NF`; `ZF`; `VF`; `X11`]; (* 286: adc x17, x11, xzr *)
  []; (* 287: cmn x13, x16 *)
  []; (* 288: adcs xzr, x6, x4 *)
  []; (* 289: adcs xzr, x8, x2 *)
  []; (* 290: adcs xzr, x3, xzr *)
  []; (* 291: adcs xzr, x15, xzr *)
  []; (* 292: adcs xzr, x14, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X17`]; (* 293: adc x1, x17, xzr *)
  [`X1`]; (* 294: neg x9, x1 *)
  [`X16`]; (* 295: and x1, x16, x9 *)
  [`X1`; `X13`]; (* 296: adds x11, x13, x1 *)
  [`X4`]; (* 297: and x13, x4, x9 *)
  [`X13`; `X6`]; (* 298: adcs x5, x6, x13 *)
  [`X2`; `X9`]; (* 299: and x1, x2, x9 *)
  [`X1`; `X8`]; (* 300: adcs x7, x8, x1 *)
  [`X11`; `X5`]; (* 301: stp x11, x5, [x0] *)
  [`X3`]; (* 302: adcs x11, x3, xzr *)
  [`X15`]; (* 303: adcs x2, x15, xzr *)
  [`X11`; `X7`]; (* 304: stp x7, x11, [x0, #16] *)
  [`CF`; `NF`; `ZF`; `VF`; `X14`]; (* 305: adc x17, x14, xzr *)
  [`X17`; `X2`]; (* 306: stp x2, x17, [x0, #32] *)
|];;

let local_sub_p384_dead_value_info = [|
  []; (* 1: ldp x5, x6, [x1] *)
  []; (* 2: ldp x4, x3, [x2] *)
  [`X4`]; (* 3: subs x5, x5, x4 *)
  [`X3`]; (* 4: sbcs x6, x6, x3 *)
  []; (* 5: ldp x7, x8, [x1, #16] *)
  []; (* 6: ldp x4, x3, [x2, #16] *)
  [`X4`]; (* 7: sbcs x7, x7, x4 *)
  [`X3`]; (* 8: sbcs x8, x8, x3 *)
  []; (* 9: ldp x9, x10, [x1, #32] *)
  []; (* 10: ldp x4, x3, [x2, #32] *)
  [`X4`]; (* 11: sbcs x9, x9, x4 *)
  [`X3`]; (* 12: sbcs x10, x10, x3 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 13: csetm x3, cc *)
  []; (* 14: mov x4, #0xffffffff *)
  []; (* 15: and x4, x4, x3 *)
  []; (* 16: adds x5, x5, x4 *)
  []; (* 17: eor x4, x4, x3 *)
  [`X4`]; (* 18: adcs x6, x6, x4 *)
  []; (* 19: mov x4, #0xfffffffffffffffe *)
  []; (* 20: and x4, x4, x3 *)
  [`X4`]; (* 21: adcs x7, x7, x4 *)
  []; (* 22: adcs x8, x8, x3 *)
  []; (* 23: adcs x9, x9, x3 *)
  [`CF`; `NF`; `ZF`; `VF`; `X3`]; (* 24: adc x10, x10, x3 *)
  [`X5`; `X6`]; (* 25: stp x5, x6, [x0] *)
  [`X7`; `X8`]; (* 26: stp x7, x8, [x0, #16] *)
  [`X10`; `X9`]; (* 27: stp x9, x10, [x0, #32] *)
|];;

let p384_montjadd_unopt_dead_value_info =
  let _,_,n,_,_ = last !actions1 in Array.init n (fun _ -> []);;

let i = ref 0;;
List.iter (fun code_block ->
  match code_block with
  | "equal", Some instcnt ->
    (* simply, don't add any data. *)
    i := instcnt + !i
  | callstr, None -> begin
    match callstr with
    | "call montmul_p384" ->
      i := 4 + !i; (* arguments + bl *)
      i := 4 + !i; (* callee-save reg stores *)
      for j = 0 to Array.length local_montmul_p384_dead_value_info - 1 do
        p384_montjadd_unopt_dead_value_info.(!i) <-
          local_montmul_p384_dead_value_info.(j);
        i := 1 + !i;
      done;
      i := 5 + !i (* callee-save restores and ret *)
    | "call montsqr_p384" ->
      i := 3 + !i; (* arguments + bl *)
      for j = 0 to Array.length local_montsqr_p384_dead_value_info - 1 do
        p384_montjadd_unopt_dead_value_info.(!i) <-
          local_montsqr_p384_dead_value_info.(j);
        i := 1 + !i;
      done;
      i := 1 + !i (* ret *)
    | "call sub_p384" ->
      i := 4 + !i; (* arguments + bl *)
      for j = 0 to Array.length local_sub_p384_dead_value_info - 1 do
        p384_montjadd_unopt_dead_value_info.(!i) <-
          local_sub_p384_dead_value_info.(j);
        i := 1 + !i;
      done;
      i := 1 + !i (* ret *)
  end)
  code_blocks;;

assert (!i = Array.length p384_montjadd_unopt_dead_value_info);;

(* ------------------------------------------------------------------------- *)
(* Second, build dead value info of the optimized program.                   *)
(* ------------------------------------------------------------------------- *)

let p384_montjadd_dead_value_info = [|
  []; (* 1: mov x24, x0 *)
  []; (* 2: mov x25, x1 *)
  []; (* 3: mov x26, x2 *)
  []; (* 4: mov x0, sp *)
  []; (* 5: ldr q1, [x25, #96] *)
  []; (* 6: ldp x9, x2, [x25, #96] *)
  []; (* 7: ldr q0, [x25, #96] *)
  []; (* 8: ldp x4, x6, [x25, #112] *)
  []; (* 9: rev64 v21.4s, v1.4s *)
  []; (* 10: uzp2 v28.4s, v1.4s, v1.4s *)
  []; (* 11: umulh x7, x9, x2 *)
  [`Q1`]; (* 12: xtn v17.2s, v1.2d *)
  [`Q21`]; (* 13: mul v27.4s, v21.4s, v0.4s *)
  []; (* 14: ldr q20, [x25, #128] *)
  []; (* 15: xtn v30.2s, v0.2d *)
  []; (* 16: ldr q1, [x25, #128] *)
  [`Q0`]; (* 17: uzp2 v31.4s, v0.4s, v0.4s *)
  []; (* 18: ldp x5, x10, [x25, #128] *)
  []; (* 19: umulh x8, x9, x4 *)
  [`Q27`]; (* 20: uaddlp v3.2d, v27.4s *)
  []; (* 21: umull v16.2d, v30.2s, v17.2s *)
  []; (* 22: mul x16, x9, x4 *)
  []; (* 23: umull v27.2d, v30.2s, v28.2s *)
  []; (* 24: shrn v0.2s, v20.2d, #32 *)
  [`Q20`]; (* 25: xtn v7.2s, v20.2d *)
  [`Q3`]; (* 26: shl v20.2d, v3.2d, #32 *)
  [`Q28`]; (* 27: umull v3.2d, v31.2s, v28.2s *)
  []; (* 28: mul x3, x2, x4 *)
  [`Q30`]; (* 29: umlal v20.2d, v30.2s, v17.2s *)
  [`Q0`]; (* 30: umull v22.2d, v7.2s, v0.2s *)
  [`Q16`]; (* 31: usra v27.2d, v16.2d, #32 *)
  []; (* 32: umulh x11, x2, x4 *)
  []; (* 33: movi v21.2d, #0xffffffff *)
  []; (* 34: uzp2 v28.4s, v1.4s, v1.4s *)
  [`X16`; `X7`]; (* 35: adds x15, x16, x7 *)
  [`Q21`]; (* 36: and v5.16b, v27.16b, v21.16b *)
  [`X8`]; (* 37: adcs x3, x3, x8 *)
  [`Q27`]; (* 38: usra v3.2d, v27.2d, #32 *)
  []; (* 39: dup v29.2d, x6 *)
  [`CF`; `NF`; `ZF`; `VF`; `X11`]; (* 40: adcs x16, x11, xzr *)
  []; (* 41: mov x14, v20.d[0] *)
  [`Q17`; `Q31`]; (* 42: umlal v5.2d, v31.2s, v17.2s *)
  []; (* 43: mul x8, x9, x2 *)
  [`Q20`]; (* 44: mov x7, v20.d[1] *)
  [`Q22`]; (* 45: shl v19.2d, v22.2d, #33 *)
  []; (* 46: xtn v25.2s, v29.2d *)
  []; (* 47: rev64 v31.4s, v1.4s *)
  []; (* 48: lsl x13, x14, #32 *)
  []; (* 49: uzp2 v6.4s, v29.4s, v29.4s *)
  [`Q7`]; (* 50: umlal v19.2d, v7.2s, v7.2s *)
  [`Q5`]; (* 51: usra v3.2d, v5.2d, #32 *)
  [`X8`]; (* 52: adds x1, x8, x8 *)
  []; (* 53: umulh x8, x4, x4 *)
  [`X13`; `X14`]; (* 54: add x12, x13, x14 *)
  [`Q29`; `Q31`]; (* 55: mul v17.4s, v31.4s, v29.4s *)
  [`Q1`]; (* 56: xtn v4.2s, v1.2d *)
  [`X15`]; (* 57: adcs x14, x15, x15 *)
  []; (* 58: lsr x13, x12, #32 *)
  [`X3`]; (* 59: adcs x15, x3, x3 *)
  []; (* 60: umull v31.2d, v25.2s, v28.2s *)
  [`X16`]; (* 61: adcs x11, x16, x16 *)
  []; (* 62: umull v21.2d, v25.2s, v4.2s *)
  []; (* 63: mov x17, v3.d[0] *)
  [`Q28`]; (* 64: umull v18.2d, v6.2s, v28.2s *)
  [`CF`; `NF`; `ZF`; `VF`; `X8`]; (* 65: adc x16, x8, xzr *)
  [`Q17`]; (* 66: uaddlp v16.2d, v17.4s *)
  []; (* 67: movi v1.2d, #0xffffffff *)
  []; (* 68: subs x13, x13, x12 *)
  [`Q21`]; (* 69: usra v31.2d, v21.2d, #32 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 70: sbc x8, x12, xzr *)
  [`X1`]; (* 71: adds x17, x17, x1 *)
  []; (* 72: mul x1, x4, x4 *)
  [`Q16`]; (* 73: shl v28.2d, v16.2d, #32 *)
  [`Q3`]; (* 74: mov x3, v3.d[1] *)
  [`X7`]; (* 75: adcs x14, x7, x14 *)
  [`X13`]; (* 76: extr x7, x8, x13, #32 *)
  [`X15`; `X3`]; (* 77: adcs x13, x3, x15 *)
  [`Q1`]; (* 78: and v3.16b, v31.16b, v1.16b *)
  [`X1`]; (* 79: adcs x11, x1, x11 *)
  [`X8`]; (* 80: lsr x1, x8, #32 *)
  [`Q6`]; (* 81: umlal v3.2d, v6.2s, v4.2s *)
  [`Q31`]; (* 82: usra v18.2d, v31.2d, #32 *)
  [`CF`; `NF`; `ZF`; `VF`; `X16`]; (* 83: adc x3, x16, xzr *)
  []; (* 84: adds x1, x1, x12 *)
  [`Q25`; `Q4`]; (* 85: umlal v28.2d, v25.2s, v4.2s *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 86: adc x16, xzr, xzr *)
  [`X17`; `X7`]; (* 87: subs x15, x17, x7 *)
  [`X1`; `X14`]; (* 88: sbcs x7, x14, x1 *)
  []; (* 89: lsl x1, x15, #32 *)
  [`X13`]; (* 90: sbcs x16, x13, x16 *)
  [`X1`; `X15`]; (* 91: add x8, x1, x15 *)
  [`Q3`]; (* 92: usra v18.2d, v3.2d, #32 *)
  [`X11`]; (* 93: sbcs x14, x11, xzr *)
  []; (* 94: lsr x1, x8, #32 *)
  [`X3`]; (* 95: sbcs x17, x3, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 96: sbc x11, x12, xzr *)
  [`X1`]; (* 97: subs x13, x1, x8 *)
  []; (* 98: umulh x12, x4, x10 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 99: sbc x1, x8, xzr *)
  []; (* 100: extr x13, x1, x13, #32 *)
  []; (* 101: lsr x1, x1, #32 *)
  [`X1`]; (* 102: adds x15, x1, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 103: adc x1, xzr, xzr *)
  [`X13`]; (* 104: subs x7, x7, x13 *)
  [`X15`; `X16`]; (* 105: sbcs x13, x16, x15 *)
  []; (* 106: lsl x3, x7, #32 *)
  []; (* 107: umulh x16, x2, x5 *)
  [`X1`; `X14`]; (* 108: sbcs x15, x14, x1 *)
  [`X3`]; (* 109: add x7, x3, x7 *)
  [`X17`]; (* 110: sbcs x3, x17, xzr *)
  []; (* 111: lsr x1, x7, #32 *)
  [`X11`]; (* 112: sbcs x14, x11, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X8`]; (* 113: sbc x11, x8, xzr *)
  [`X1`]; (* 114: subs x8, x1, x7 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 115: sbc x1, x7, xzr *)
  []; (* 116: extr x8, x1, x8, #32 *)
  []; (* 117: lsr x1, x1, #32 *)
  []; (* 118: adds x1, x1, x7 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 119: adc x17, xzr, xzr *)
  [`X8`]; (* 120: subs x13, x13, x8 *)
  []; (* 121: umulh x8, x9, x6 *)
  [`X15`]; (* 122: sbcs x1, x15, x1 *)
  [`X17`; `X3`]; (* 123: sbcs x15, x3, x17 *)
  [`X14`]; (* 124: sbcs x3, x14, xzr *)
  []; (* 125: mul x17, x2, x5 *)
  []; (* 126: sbcs x11, x11, xzr *)
  [`X1`; `X13`]; (* 127: stp x13, x1, [x0] *)
  [`CF`; `NF`; `ZF`; `VF`; `X7`]; (* 128: sbc x14, x7, xzr *)
  []; (* 129: mul x7, x4, x10 *)
  []; (* 130: subs x1, x9, x2 *)
  [`X15`; `X3`]; (* 131: stp x15, x3, [x0, #16] *)
  []; (* 132: csetm x15, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 133: cneg x1, x1, cc *)
  [`X11`; `X14`]; (* 134: stp x11, x14, [x0, #32] *)
  []; (* 135: mul x14, x9, x6 *)
  [`X8`]; (* 136: adds x17, x8, x17 *)
  [`X16`]; (* 137: adcs x7, x16, x7 *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 138: adc x13, x12, xzr *)
  []; (* 139: subs x12, x5, x6 *)
  [`X12`]; (* 140: cneg x3, x12, cc *)
  [`CF`; `NF`; `ZF`; `VF`; `X15`]; (* 141: cinv x16, x15, cc *)
  []; (* 142: mul x8, x1, x3 *)
  [`X3`]; (* 143: umulh x1, x1, x3 *)
  [`X8`]; (* 144: eor x12, x8, x16 *)
  []; (* 145: adds x11, x17, x14 *)
  []; (* 146: adcs x3, x7, x17 *)
  []; (* 147: adcs x15, x13, x7 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 148: adc x8, x13, xzr *)
  []; (* 149: adds x3, x3, x14 *)
  [`X17`]; (* 150: adcs x15, x15, x17 *)
  [`X7`; `X8`]; (* 151: adcs x17, x8, x7 *)
  []; (* 152: eor x1, x1, x16 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 153: adc x13, x13, xzr *)
  []; (* 154: subs x9, x9, x4 *)
  []; (* 155: csetm x8, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 156: cneg x9, x9, cc *)
  [`X2`]; (* 157: subs x4, x2, x4 *)
  []; (* 158: cneg x4, x4, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 159: csetm x7, cc *)
  []; (* 160: subs x2, x10, x6 *)
  []; (* 161: cinv x8, x8, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 162: cneg x2, x2, cc *)
  []; (* 163: cmn x16, #0x1 *)
  [`X12`]; (* 164: adcs x11, x11, x12 *)
  []; (* 165: mul x12, x9, x2 *)
  [`X1`]; (* 166: adcs x3, x3, x1 *)
  []; (* 167: adcs x15, x15, x16 *)
  [`X2`]; (* 168: umulh x9, x9, x2 *)
  []; (* 169: adcs x17, x17, x16 *)
  [`CF`; `NF`; `ZF`; `VF`; `X16`]; (* 170: adc x13, x13, x16 *)
  []; (* 171: subs x1, x10, x5 *)
  [`X7`]; (* 172: cinv x2, x7, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 173: cneg x1, x1, cc *)
  []; (* 174: eor x9, x9, x8 *)
  []; (* 175: cmn x8, #0x1 *)
  [`X12`]; (* 176: eor x7, x12, x8 *)
  []; (* 177: mul x12, x4, x1 *)
  [`X7`]; (* 178: adcs x3, x3, x7 *)
  [`X15`; `X9`]; (* 179: adcs x7, x15, x9 *)
  [`X17`]; (* 180: adcs x15, x17, x8 *)
  []; (* 181: ldp x9, x17, [x0, #16] *)
  [`X1`]; (* 182: umulh x4, x4, x1 *)
  [`CF`; `NF`; `ZF`; `VF`; `X13`]; (* 183: adc x8, x13, x8 *)
  []; (* 184: cmn x2, #0x1 *)
  [`X12`]; (* 185: eor x1, x12, x2 *)
  [`X7`]; (* 186: adcs x1, x7, x1 *)
  []; (* 187: ldp x7, x16, [x0] *)
  [`X4`]; (* 188: eor x12, x4, x2 *)
  [`X12`; `X15`]; (* 189: adcs x4, x15, x12 *)
  []; (* 190: ldp x15, x12, [x0, #32] *)
  [`CF`; `NF`; `ZF`; `VF`; `X2`]; (* 191: adc x8, x8, x2 *)
  [`X14`]; (* 192: adds x13, x14, x14 *)
  []; (* 193: umulh x14, x5, x10 *)
  [`X11`]; (* 194: adcs x2, x11, x11 *)
  []; (* 195: adcs x3, x3, x3 *)
  []; (* 196: adcs x1, x1, x1 *)
  []; (* 197: adcs x4, x4, x4 *)
  [`X8`]; (* 198: adcs x11, x8, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 199: adc x8, xzr, xzr *)
  [`X7`]; (* 200: adds x13, x13, x7 *)
  [`X16`]; (* 201: adcs x2, x2, x16 *)
  []; (* 202: mul x16, x5, x10 *)
  [`X9`]; (* 203: adcs x3, x3, x9 *)
  [`X17`]; (* 204: adcs x1, x1, x17 *)
  []; (* 205: umulh x5, x5, x5 *)
  []; (* 206: lsl x9, x13, #32 *)
  [`X13`]; (* 207: add x9, x9, x13 *)
  [`X15`]; (* 208: adcs x4, x4, x15 *)
  []; (* 209: mov x13, v28.d[1] *)
  [`X11`; `X12`]; (* 210: adcs x15, x11, x12 *)
  []; (* 211: lsr x7, x9, #32 *)
  [`CF`; `NF`; `ZF`; `VF`; `X8`]; (* 212: adc x11, x8, xzr *)
  []; (* 213: subs x7, x7, x9 *)
  []; (* 214: umulh x10, x10, x10 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 215: sbc x17, x9, xzr *)
  []; (* 216: extr x7, x17, x7, #32 *)
  []; (* 217: lsr x17, x17, #32 *)
  []; (* 218: adds x17, x17, x9 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 219: adc x12, xzr, xzr *)
  [`X2`; `X7`]; (* 220: subs x8, x2, x7 *)
  [`X3`]; (* 221: sbcs x17, x3, x17 *)
  []; (* 222: lsl x7, x8, #32 *)
  [`X1`; `X12`]; (* 223: sbcs x2, x1, x12 *)
  [`X7`; `X8`]; (* 224: add x3, x7, x8 *)
  [`X4`]; (* 225: sbcs x12, x4, xzr *)
  []; (* 226: lsr x1, x3, #32 *)
  [`X15`]; (* 227: sbcs x7, x15, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X9`]; (* 228: sbc x15, x9, xzr *)
  []; (* 229: subs x1, x1, x3 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 230: sbc x4, x3, xzr *)
  []; (* 231: lsr x9, x4, #32 *)
  [`X1`; `X4`]; (* 232: extr x8, x4, x1, #32 *)
  []; (* 233: adds x9, x9, x3 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 234: adc x4, xzr, xzr *)
  [`X17`; `X8`]; (* 235: subs x1, x17, x8 *)
  []; (* 236: lsl x17, x1, #32 *)
  [`X2`; `X9`]; (* 237: sbcs x8, x2, x9 *)
  [`X12`; `X4`]; (* 238: sbcs x9, x12, x4 *)
  [`X1`]; (* 239: add x17, x17, x1 *)
  []; (* 240: mov x1, v18.d[1] *)
  []; (* 241: lsr x2, x17, #32 *)
  []; (* 242: sbcs x7, x7, xzr *)
  [`Q18`]; (* 243: mov x12, v18.d[0] *)
  []; (* 244: sbcs x15, x15, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 245: sbc x3, x3, xzr *)
  [`X2`]; (* 246: subs x4, x2, x17 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 247: sbc x2, x17, xzr *)
  [`X13`]; (* 248: adds x12, x13, x12 *)
  [`X1`]; (* 249: adcs x16, x16, x1 *)
  []; (* 250: lsr x13, x2, #32 *)
  [`X2`; `X4`]; (* 251: extr x1, x2, x4, #32 *)
  [`CF`; `NF`; `ZF`; `VF`; `X14`]; (* 252: adc x2, x14, xzr *)
  [`X13`]; (* 253: adds x4, x13, x17 *)
  []; (* 254: mul x13, x6, x6 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 255: adc x14, xzr, xzr *)
  [`X8`]; (* 256: subs x1, x8, x1 *)
  [`X9`]; (* 257: sbcs x4, x9, x4 *)
  [`Q28`]; (* 258: mov x9, v28.d[0] *)
  [`X14`]; (* 259: sbcs x7, x7, x14 *)
  [`X15`]; (* 260: sbcs x8, x15, xzr *)
  []; (* 261: sbcs x3, x3, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X17`]; (* 262: sbc x14, x17, xzr *)
  [`X9`]; (* 263: adds x17, x9, x9 *)
  []; (* 264: adcs x12, x12, x12 *)
  []; (* 265: mov x15, v19.d[0] *)
  [`X16`]; (* 266: adcs x9, x16, x16 *)
  []; (* 267: umulh x6, x6, x6 *)
  [`X2`]; (* 268: adcs x16, x2, x2 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 269: adc x2, xzr, xzr *)
  [`X8`]; (* 270: adds x11, x11, x8 *)
  []; (* 271: adcs x3, x3, xzr *)
  []; (* 272: adcs x14, x14, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 273: adcs x8, xzr, xzr *)
  [`X1`]; (* 274: adds x13, x1, x13 *)
  [`Q19`]; (* 275: mov x1, v19.d[1] *)
  [`X4`]; (* 276: adcs x6, x4, x6 *)
  []; (* 277: mov x4, #0xffffffff *)
  [`X7`]; (* 278: adcs x15, x7, x15 *)
  [`X11`; `X5`]; (* 279: adcs x7, x11, x5 *)
  [`X3`]; (* 280: adcs x1, x3, x1 *)
  [`X10`]; (* 281: adcs x14, x14, x10 *)
  [`CF`; `NF`; `ZF`; `VF`; `X8`]; (* 282: adc x11, x8, xzr *)
  [`X17`]; (* 283: adds x6, x6, x17 *)
  [`X12`; `X15`]; (* 284: adcs x8, x15, x12 *)
  [`X7`; `X9`]; (* 285: adcs x3, x7, x9 *)
  [`X1`; `X16`]; (* 286: adcs x15, x1, x16 *)
  []; (* 287: mov x16, #0xffffffff00000001 *)
  [`X2`]; (* 288: adcs x14, x14, x2 *)
  []; (* 289: mov x2, #0x1 *)
  [`CF`; `NF`; `ZF`; `VF`; `X11`]; (* 290: adc x17, x11, xzr *)
  []; (* 291: cmn x13, x16 *)
  []; (* 292: adcs xzr, x6, x4 *)
  []; (* 293: adcs xzr, x8, x2 *)
  []; (* 294: adcs xzr, x3, xzr *)
  []; (* 295: adcs xzr, x15, xzr *)
  []; (* 296: adcs xzr, x14, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X17`]; (* 297: adc x1, x17, xzr *)
  [`X1`]; (* 298: neg x9, x1 *)
  [`X16`]; (* 299: and x1, x16, x9 *)
  [`X1`; `X13`]; (* 300: adds x11, x13, x1 *)
  [`X4`]; (* 301: and x13, x4, x9 *)
  [`X13`; `X6`]; (* 302: adcs x5, x6, x13 *)
  [`X2`; `X9`]; (* 303: and x1, x2, x9 *)
  [`X1`; `X8`]; (* 304: adcs x7, x8, x1 *)
  [`X11`; `X5`]; (* 305: stp x11, x5, [x0] *)
  [`X3`]; (* 306: adcs x11, x3, xzr *)
  [`X15`]; (* 307: adcs x2, x15, xzr *)
  [`X11`; `X7`]; (* 308: stp x7, x11, [x0, #16] *)
  [`CF`; `NF`; `ZF`; `VF`; `X14`]; (* 309: adc x17, x14, xzr *)
  [`X0`; `X17`; `X2`]; (* 310: stp x2, x17, [x0, #32] *)
  []; (* 311: ldr q1, [x26, #96] *)
  []; (* 312: ldp x9, x2, [x26, #96] *)
  []; (* 313: ldr q0, [x26, #96] *)
  []; (* 314: ldp x4, x6, [x26, #112] *)
  []; (* 315: rev64 v21.4s, v1.4s *)
  []; (* 316: uzp2 v28.4s, v1.4s, v1.4s *)
  []; (* 317: umulh x7, x9, x2 *)
  [`Q1`]; (* 318: xtn v17.2s, v1.2d *)
  [`Q21`]; (* 319: mul v27.4s, v21.4s, v0.4s *)
  []; (* 320: ldr q20, [x26, #128] *)
  []; (* 321: xtn v30.2s, v0.2d *)
  []; (* 322: ldr q1, [x26, #128] *)
  [`Q0`]; (* 323: uzp2 v31.4s, v0.4s, v0.4s *)
  []; (* 324: ldp x5, x10, [x26, #128] *)
  []; (* 325: umulh x8, x9, x4 *)
  [`Q27`]; (* 326: uaddlp v3.2d, v27.4s *)
  []; (* 327: umull v16.2d, v30.2s, v17.2s *)
  []; (* 328: mul x16, x9, x4 *)
  []; (* 329: umull v27.2d, v30.2s, v28.2s *)
  []; (* 330: shrn v0.2s, v20.2d, #32 *)
  [`Q20`]; (* 331: xtn v7.2s, v20.2d *)
  [`Q3`]; (* 332: shl v20.2d, v3.2d, #32 *)
  [`Q28`]; (* 333: umull v3.2d, v31.2s, v28.2s *)
  []; (* 334: mul x3, x2, x4 *)
  [`Q30`]; (* 335: umlal v20.2d, v30.2s, v17.2s *)
  [`Q0`]; (* 336: umull v22.2d, v7.2s, v0.2s *)
  [`Q16`]; (* 337: usra v27.2d, v16.2d, #32 *)
  []; (* 338: umulh x11, x2, x4 *)
  []; (* 339: movi v21.2d, #0xffffffff *)
  []; (* 340: uzp2 v28.4s, v1.4s, v1.4s *)
  [`X16`; `X7`]; (* 341: adds x15, x16, x7 *)
  [`Q21`]; (* 342: and v5.16b, v27.16b, v21.16b *)
  [`X8`]; (* 343: adcs x3, x3, x8 *)
  [`Q27`]; (* 344: usra v3.2d, v27.2d, #32 *)
  []; (* 345: dup v29.2d, x6 *)
  [`CF`; `NF`; `ZF`; `VF`; `X11`]; (* 346: adcs x16, x11, xzr *)
  []; (* 347: mov x14, v20.d[0] *)
  [`Q17`; `Q31`]; (* 348: umlal v5.2d, v31.2s, v17.2s *)
  []; (* 349: mul x8, x9, x2 *)
  [`Q20`]; (* 350: mov x7, v20.d[1] *)
  [`Q22`]; (* 351: shl v19.2d, v22.2d, #33 *)
  []; (* 352: xtn v25.2s, v29.2d *)
  []; (* 353: rev64 v31.4s, v1.4s *)
  []; (* 354: lsl x13, x14, #32 *)
  []; (* 355: uzp2 v6.4s, v29.4s, v29.4s *)
  [`Q7`]; (* 356: umlal v19.2d, v7.2s, v7.2s *)
  [`Q5`]; (* 357: usra v3.2d, v5.2d, #32 *)
  [`X8`]; (* 358: adds x1, x8, x8 *)
  []; (* 359: umulh x8, x4, x4 *)
  [`X13`; `X14`]; (* 360: add x12, x13, x14 *)
  [`Q29`; `Q31`]; (* 361: mul v17.4s, v31.4s, v29.4s *)
  [`Q1`]; (* 362: xtn v4.2s, v1.2d *)
  [`X15`]; (* 363: adcs x14, x15, x15 *)
  []; (* 364: lsr x13, x12, #32 *)
  [`X3`]; (* 365: adcs x15, x3, x3 *)
  []; (* 366: umull v31.2d, v25.2s, v28.2s *)
  [`X16`]; (* 367: adcs x11, x16, x16 *)
  []; (* 368: umull v21.2d, v25.2s, v4.2s *)
  []; (* 369: mov x17, v3.d[0] *)
  [`Q28`]; (* 370: umull v18.2d, v6.2s, v28.2s *)
  [`CF`; `NF`; `ZF`; `VF`; `X8`]; (* 371: adc x16, x8, xzr *)
  [`Q17`]; (* 372: uaddlp v16.2d, v17.4s *)
  []; (* 373: movi v1.2d, #0xffffffff *)
  []; (* 374: subs x13, x13, x12 *)
  [`Q21`]; (* 375: usra v31.2d, v21.2d, #32 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 376: sbc x8, x12, xzr *)
  [`X1`]; (* 377: adds x17, x17, x1 *)
  []; (* 378: mul x1, x4, x4 *)
  [`Q16`]; (* 379: shl v28.2d, v16.2d, #32 *)
  [`Q3`]; (* 380: mov x3, v3.d[1] *)
  [`X7`]; (* 381: adcs x14, x7, x14 *)
  [`X13`]; (* 382: extr x7, x8, x13, #32 *)
  [`X15`; `X3`]; (* 383: adcs x13, x3, x15 *)
  [`Q1`]; (* 384: and v3.16b, v31.16b, v1.16b *)
  [`X1`]; (* 385: adcs x11, x1, x11 *)
  [`X8`]; (* 386: lsr x1, x8, #32 *)
  [`Q6`]; (* 387: umlal v3.2d, v6.2s, v4.2s *)
  [`Q31`]; (* 388: usra v18.2d, v31.2d, #32 *)
  [`CF`; `NF`; `ZF`; `VF`; `X16`]; (* 389: adc x3, x16, xzr *)
  []; (* 390: adds x1, x1, x12 *)
  [`Q25`; `Q4`]; (* 391: umlal v28.2d, v25.2s, v4.2s *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 392: adc x16, xzr, xzr *)
  [`X17`; `X7`]; (* 393: subs x15, x17, x7 *)
  [`X1`; `X14`]; (* 394: sbcs x7, x14, x1 *)
  []; (* 395: lsl x1, x15, #32 *)
  [`X13`]; (* 396: sbcs x16, x13, x16 *)
  [`X1`; `X15`]; (* 397: add x8, x1, x15 *)
  [`Q3`]; (* 398: usra v18.2d, v3.2d, #32 *)
  [`X11`]; (* 399: sbcs x14, x11, xzr *)
  []; (* 400: lsr x1, x8, #32 *)
  [`X3`]; (* 401: sbcs x17, x3, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 402: sbc x11, x12, xzr *)
  [`X1`]; (* 403: subs x13, x1, x8 *)
  []; (* 404: umulh x12, x4, x10 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 405: sbc x1, x8, xzr *)
  []; (* 406: extr x13, x1, x13, #32 *)
  []; (* 407: lsr x1, x1, #32 *)
  [`X1`]; (* 408: adds x15, x1, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 409: adc x1, xzr, xzr *)
  [`X13`]; (* 410: subs x7, x7, x13 *)
  [`X15`; `X16`]; (* 411: sbcs x13, x16, x15 *)
  []; (* 412: lsl x3, x7, #32 *)
  []; (* 413: umulh x16, x2, x5 *)
  [`X1`; `X14`]; (* 414: sbcs x15, x14, x1 *)
  [`X3`]; (* 415: add x7, x3, x7 *)
  [`X17`]; (* 416: sbcs x3, x17, xzr *)
  []; (* 417: lsr x1, x7, #32 *)
  [`X11`]; (* 418: sbcs x14, x11, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X8`]; (* 419: sbc x11, x8, xzr *)
  [`X1`]; (* 420: subs x8, x1, x7 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 421: sbc x1, x7, xzr *)
  []; (* 422: extr x8, x1, x8, #32 *)
  []; (* 423: lsr x1, x1, #32 *)
  []; (* 424: adds x1, x1, x7 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 425: adc x17, xzr, xzr *)
  [`X8`]; (* 426: subs x13, x13, x8 *)
  []; (* 427: umulh x8, x9, x6 *)
  [`X15`]; (* 428: sbcs x1, x15, x1 *)
  [`X17`; `X3`]; (* 429: sbcs x15, x3, x17 *)
  [`X14`]; (* 430: sbcs x3, x14, xzr *)
  []; (* 431: mul x17, x2, x5 *)
  []; (* 432: sbcs x11, x11, xzr *)
  [`X1`; `X13`]; (* 433: stp x13, x1, [sp, #240] *)
  [`CF`; `NF`; `ZF`; `VF`; `X7`]; (* 434: sbc x14, x7, xzr *)
  []; (* 435: mul x7, x4, x10 *)
  []; (* 436: subs x1, x9, x2 *)
  [`X15`; `X3`]; (* 437: stp x15, x3, [sp, #256] *)
  []; (* 438: csetm x15, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 439: cneg x1, x1, cc *)
  [`X11`; `X14`]; (* 440: stp x11, x14, [sp, #272] *)
  []; (* 441: mul x14, x9, x6 *)
  [`X8`]; (* 442: adds x17, x8, x17 *)
  [`X16`]; (* 443: adcs x7, x16, x7 *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 444: adc x13, x12, xzr *)
  []; (* 445: subs x12, x5, x6 *)
  [`X12`]; (* 446: cneg x3, x12, cc *)
  [`CF`; `NF`; `ZF`; `VF`; `X15`]; (* 447: cinv x16, x15, cc *)
  []; (* 448: mul x8, x1, x3 *)
  [`X3`]; (* 449: umulh x1, x1, x3 *)
  [`X8`]; (* 450: eor x12, x8, x16 *)
  []; (* 451: adds x11, x17, x14 *)
  []; (* 452: adcs x3, x7, x17 *)
  []; (* 453: adcs x15, x13, x7 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 454: adc x8, x13, xzr *)
  []; (* 455: adds x3, x3, x14 *)
  [`X17`]; (* 456: adcs x15, x15, x17 *)
  [`X7`; `X8`]; (* 457: adcs x17, x8, x7 *)
  []; (* 458: eor x1, x1, x16 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 459: adc x13, x13, xzr *)
  []; (* 460: subs x9, x9, x4 *)
  []; (* 461: csetm x8, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 462: cneg x9, x9, cc *)
  [`X2`]; (* 463: subs x4, x2, x4 *)
  []; (* 464: cneg x4, x4, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 465: csetm x7, cc *)
  []; (* 466: subs x2, x10, x6 *)
  []; (* 467: cinv x8, x8, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 468: cneg x2, x2, cc *)
  []; (* 469: cmn x16, #0x1 *)
  [`X12`]; (* 470: adcs x11, x11, x12 *)
  []; (* 471: mul x12, x9, x2 *)
  [`X1`]; (* 472: adcs x3, x3, x1 *)
  []; (* 473: adcs x15, x15, x16 *)
  [`X2`]; (* 474: umulh x9, x9, x2 *)
  []; (* 475: adcs x17, x17, x16 *)
  [`CF`; `NF`; `ZF`; `VF`; `X16`]; (* 476: adc x13, x13, x16 *)
  []; (* 477: subs x1, x10, x5 *)
  [`X7`]; (* 478: cinv x2, x7, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 479: cneg x1, x1, cc *)
  []; (* 480: eor x9, x9, x8 *)
  []; (* 481: cmn x8, #0x1 *)
  [`X12`]; (* 482: eor x7, x12, x8 *)
  []; (* 483: mul x12, x4, x1 *)
  [`X7`]; (* 484: adcs x3, x3, x7 *)
  [`X15`; `X9`]; (* 485: adcs x7, x15, x9 *)
  [`X17`]; (* 486: adcs x15, x17, x8 *)
  []; (* 487: ldp x9, x17, [sp, #256] *)
  [`X1`]; (* 488: umulh x4, x4, x1 *)
  [`CF`; `NF`; `ZF`; `VF`; `X13`]; (* 489: adc x8, x13, x8 *)
  []; (* 490: cmn x2, #0x1 *)
  [`X12`]; (* 491: eor x1, x12, x2 *)
  [`X7`]; (* 492: adcs x1, x7, x1 *)
  []; (* 493: ldp x7, x16, [sp, #240] *)
  [`X4`]; (* 494: eor x12, x4, x2 *)
  [`X12`; `X15`]; (* 495: adcs x4, x15, x12 *)
  []; (* 496: ldp x15, x12, [sp, #272] *)
  [`CF`; `NF`; `ZF`; `VF`; `X2`]; (* 497: adc x8, x8, x2 *)
  [`X14`]; (* 498: adds x13, x14, x14 *)
  []; (* 499: umulh x14, x5, x10 *)
  [`X11`]; (* 500: adcs x2, x11, x11 *)
  []; (* 501: adcs x3, x3, x3 *)
  []; (* 502: adcs x1, x1, x1 *)
  []; (* 503: adcs x4, x4, x4 *)
  [`X8`]; (* 504: adcs x11, x8, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 505: adc x8, xzr, xzr *)
  [`X7`]; (* 506: adds x13, x13, x7 *)
  [`X16`]; (* 507: adcs x2, x2, x16 *)
  []; (* 508: mul x16, x5, x10 *)
  [`X9`]; (* 509: adcs x3, x3, x9 *)
  [`X17`]; (* 510: adcs x1, x1, x17 *)
  []; (* 511: umulh x5, x5, x5 *)
  []; (* 512: lsl x9, x13, #32 *)
  [`X13`]; (* 513: add x9, x9, x13 *)
  [`X15`]; (* 514: adcs x4, x4, x15 *)
  []; (* 515: mov x13, v28.d[1] *)
  [`X11`; `X12`]; (* 516: adcs x15, x11, x12 *)
  []; (* 517: lsr x7, x9, #32 *)
  [`CF`; `NF`; `ZF`; `VF`; `X8`]; (* 518: adc x11, x8, xzr *)
  []; (* 519: subs x7, x7, x9 *)
  []; (* 520: umulh x10, x10, x10 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 521: sbc x17, x9, xzr *)
  []; (* 522: extr x7, x17, x7, #32 *)
  []; (* 523: lsr x17, x17, #32 *)
  []; (* 524: adds x17, x17, x9 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 525: adc x12, xzr, xzr *)
  [`X2`; `X7`]; (* 526: subs x8, x2, x7 *)
  [`X3`]; (* 527: sbcs x17, x3, x17 *)
  []; (* 528: lsl x7, x8, #32 *)
  [`X1`; `X12`]; (* 529: sbcs x2, x1, x12 *)
  [`X7`; `X8`]; (* 530: add x3, x7, x8 *)
  [`X4`]; (* 531: sbcs x12, x4, xzr *)
  []; (* 532: lsr x1, x3, #32 *)
  [`X15`]; (* 533: sbcs x7, x15, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X9`]; (* 534: sbc x15, x9, xzr *)
  []; (* 535: subs x1, x1, x3 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 536: sbc x4, x3, xzr *)
  []; (* 537: lsr x9, x4, #32 *)
  [`X1`; `X4`]; (* 538: extr x8, x4, x1, #32 *)
  []; (* 539: adds x9, x9, x3 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 540: adc x4, xzr, xzr *)
  [`X17`; `X8`]; (* 541: subs x1, x17, x8 *)
  []; (* 542: lsl x17, x1, #32 *)
  [`X2`; `X9`]; (* 543: sbcs x8, x2, x9 *)
  [`X12`; `X4`]; (* 544: sbcs x9, x12, x4 *)
  [`X1`]; (* 545: add x17, x17, x1 *)
  []; (* 546: mov x1, v18.d[1] *)
  []; (* 547: lsr x2, x17, #32 *)
  []; (* 548: sbcs x7, x7, xzr *)
  [`Q18`]; (* 549: mov x12, v18.d[0] *)
  []; (* 550: sbcs x15, x15, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 551: sbc x3, x3, xzr *)
  [`X2`]; (* 552: subs x4, x2, x17 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 553: sbc x2, x17, xzr *)
  [`X13`]; (* 554: adds x12, x13, x12 *)
  [`X1`]; (* 555: adcs x16, x16, x1 *)
  []; (* 556: lsr x13, x2, #32 *)
  [`X2`; `X4`]; (* 557: extr x1, x2, x4, #32 *)
  [`CF`; `NF`; `ZF`; `VF`; `X14`]; (* 558: adc x2, x14, xzr *)
  [`X13`]; (* 559: adds x4, x13, x17 *)
  []; (* 560: mul x13, x6, x6 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 561: adc x14, xzr, xzr *)
  [`X8`]; (* 562: subs x1, x8, x1 *)
  [`X9`]; (* 563: sbcs x4, x9, x4 *)
  [`Q28`]; (* 564: mov x9, v28.d[0] *)
  [`X14`]; (* 565: sbcs x7, x7, x14 *)
  [`X15`]; (* 566: sbcs x8, x15, xzr *)
  []; (* 567: sbcs x3, x3, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X17`]; (* 568: sbc x14, x17, xzr *)
  [`X9`]; (* 569: adds x17, x9, x9 *)
  []; (* 570: adcs x12, x12, x12 *)
  []; (* 571: mov x15, v19.d[0] *)
  [`X16`]; (* 572: adcs x9, x16, x16 *)
  []; (* 573: umulh x6, x6, x6 *)
  [`X2`]; (* 574: adcs x16, x2, x2 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 575: adc x2, xzr, xzr *)
  [`X8`]; (* 576: adds x11, x11, x8 *)
  []; (* 577: adcs x3, x3, xzr *)
  []; (* 578: adcs x14, x14, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 579: adcs x8, xzr, xzr *)
  [`X1`]; (* 580: adds x13, x1, x13 *)
  [`Q19`]; (* 581: mov x1, v19.d[1] *)
  [`X4`]; (* 582: adcs x6, x4, x6 *)
  []; (* 583: mov x4, #0xffffffff *)
  [`X7`]; (* 584: adcs x15, x7, x15 *)
  [`X11`; `X5`]; (* 585: adcs x7, x11, x5 *)
  [`X3`]; (* 586: adcs x1, x3, x1 *)
  [`X10`]; (* 587: adcs x14, x14, x10 *)
  [`CF`; `NF`; `ZF`; `VF`; `X8`]; (* 588: adc x11, x8, xzr *)
  [`X17`]; (* 589: adds x6, x6, x17 *)
  [`X12`; `X15`]; (* 590: adcs x8, x15, x12 *)
  [`X7`; `X9`]; (* 591: adcs x3, x7, x9 *)
  [`X1`; `X16`]; (* 592: adcs x15, x1, x16 *)
  []; (* 593: mov x16, #0xffffffff00000001 *)
  [`X2`]; (* 594: adcs x14, x14, x2 *)
  []; (* 595: mov x2, #0x1 *)
  [`CF`; `NF`; `ZF`; `VF`; `X11`]; (* 596: adc x17, x11, xzr *)
  []; (* 597: cmn x13, x16 *)
  []; (* 598: adcs xzr, x6, x4 *)
  []; (* 599: adcs xzr, x8, x2 *)
  []; (* 600: adcs xzr, x3, xzr *)
  []; (* 601: adcs xzr, x15, xzr *)
  []; (* 602: adcs xzr, x14, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X17`]; (* 603: adc x1, x17, xzr *)
  [`X1`]; (* 604: neg x9, x1 *)
  [`X16`]; (* 605: and x1, x16, x9 *)
  [`X1`; `X13`]; (* 606: adds x11, x13, x1 *)
  [`X4`]; (* 607: and x13, x4, x9 *)
  [`X13`; `X6`]; (* 608: adcs x5, x6, x13 *)
  [`X2`; `X9`]; (* 609: and x1, x2, x9 *)
  [`X1`; `X8`]; (* 610: adcs x7, x8, x1 *)
  [`X11`; `X5`]; (* 611: stp x11, x5, [sp, #240] *)
  [`X3`]; (* 612: adcs x11, x3, xzr *)
  [`X15`]; (* 613: adcs x2, x15, xzr *)
  [`X11`; `X7`]; (* 614: stp x7, x11, [sp, #256] *)
  [`CF`; `NF`; `ZF`; `VF`; `X14`]; (* 615: adc x17, x14, xzr *)
  [`X17`; `X2`]; (* 616: stp x2, x17, [sp, #272] *)
  [`X24`]; (* 617: stp x23, x24, [sp, #-48] *)
  []; (* 618: ldr q3, [x26, #96] *)
  []; (* 619: ldr q25, [x25, #48] *)
  []; (* 620: ldp x13, x23, [x25, #48] *)
  []; (* 621: ldp x3, x21, [x26, #96] *)
  []; (* 622: rev64 v23.4s, v25.4s *)
  [`Q25`]; (* 623: uzp1 v17.4s, v25.4s, v3.4s *)
  []; (* 624: umulh x15, x3, x13 *)
  [`Q23`]; (* 625: mul v6.4s, v23.4s, v3.4s *)
  []; (* 626: uzp1 v3.4s, v3.4s, v3.4s *)
  []; (* 627: ldr q27, [x25, #80] *)
  []; (* 628: ldp x8, x24, [x26, #112] *)
  []; (* 629: subs x6, x3, x21 *)
  []; (* 630: ldr q0, [x26, #128] *)
  []; (* 631: movi v23.2d, #0xffffffff *)
  []; (* 632: csetm x10, cc *)
  []; (* 633: umulh x19, x21, x23 *)
  []; (* 634: rev64 v4.4s, v27.4s *)
  []; (* 635: uzp2 v25.4s, v27.4s, v27.4s *)
  [`CF`; `NF`; `ZF`; `VF`; `X6`]; (* 636: cneg x4, x6, cc *)
  []; (* 637: subs x7, x23, x13 *)
  []; (* 638: xtn v22.2s, v0.2d *)
  [`Q27`]; (* 639: xtn v24.2s, v27.2d *)
  [`X7`]; (* 640: cneg x20, x7, cc *)
  []; (* 641: ldp x6, x14, [x25, #64] *)
  [`Q4`]; (* 642: mul v27.4s, v4.4s, v0.4s *)
  [`Q6`]; (* 643: uaddlp v20.2d, v6.4s *)
  [`CF`; `NF`; `ZF`; `VF`; `X10`]; (* 644: cinv x5, x10, cc *)
  []; (* 645: mul x16, x4, x20 *)
  [`Q0`]; (* 646: uzp2 v6.4s, v0.4s, v0.4s *)
  []; (* 647: umull v21.2d, v22.2s, v25.2s *)
  [`Q20`]; (* 648: shl v0.2d, v20.2d, #32 *)
  [`Q17`; `Q3`]; (* 649: umlal v0.2d, v3.2s, v17.2s *)
  []; (* 650: mul x22, x8, x6 *)
  [`Q25`]; (* 651: umull v1.2d, v6.2s, v25.2s *)
  []; (* 652: subs x12, x3, x8 *)
  []; (* 653: umull v20.2d, v22.2s, v24.2s *)
  [`X12`]; (* 654: cneg x17, x12, cc *)
  []; (* 655: umulh x9, x8, x6 *)
  []; (* 656: mov x12, v0.d[1] *)
  [`X16`]; (* 657: eor x11, x16, x5 *)
  [`Q0`]; (* 658: mov x7, v0.d[0] *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 659: csetm x10, cc *)
  [`Q20`]; (* 660: usra v21.2d, v20.2d, #32 *)
  [`X12`]; (* 661: adds x15, x15, x12 *)
  [`X19`; `X22`]; (* 662: adcs x12, x19, x22 *)
  [`X4`]; (* 663: umulh x20, x4, x20 *)
  [`CF`; `NF`; `ZF`; `VF`; `X9`]; (* 664: adc x19, x9, xzr *)
  []; (* 665: usra v1.2d, v21.2d, #32 *)
  []; (* 666: adds x22, x15, x7 *)
  [`Q21`; `Q23`]; (* 667: and v26.16b, v21.16b, v23.16b *)
  []; (* 668: adcs x16, x12, x15 *)
  [`Q27`]; (* 669: uaddlp v25.2d, v27.4s *)
  []; (* 670: adcs x9, x19, x12 *)
  [`Q6`]; (* 671: umlal v26.2d, v6.2s, v24.2s *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 672: adc x4, x19, xzr *)
  []; (* 673: adds x16, x16, x7 *)
  [`Q25`]; (* 674: shl v27.2d, v25.2d, #32 *)
  [`X15`]; (* 675: adcs x9, x9, x15 *)
  [`X12`]; (* 676: adcs x4, x4, x12 *)
  [`X20`]; (* 677: eor x12, x20, x5 *)
  [`CF`; `NF`; `ZF`; `VF`; `X19`]; (* 678: adc x15, x19, xzr *)
  []; (* 679: subs x20, x6, x13 *)
  []; (* 680: cneg x20, x20, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 681: cinv x10, x10, cc *)
  []; (* 682: cmn x5, #0x1 *)
  []; (* 683: mul x19, x17, x20 *)
  [`X22`]; (* 684: adcs x11, x22, x11 *)
  [`X16`]; (* 685: adcs x12, x16, x12 *)
  []; (* 686: adcs x9, x9, x5 *)
  [`X20`]; (* 687: umulh x17, x17, x20 *)
  [`X4`]; (* 688: adcs x22, x4, x5 *)
  [`CF`; `NF`; `ZF`; `VF`; `X15`]; (* 689: adc x5, x15, x5 *)
  []; (* 690: subs x16, x21, x8 *)
  [`X16`]; (* 691: cneg x20, x16, cc *)
  []; (* 692: eor x19, x19, x10 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 693: csetm x4, cc *)
  []; (* 694: subs x16, x6, x23 *)
  []; (* 695: cneg x16, x16, cc *)
  [`Q22`; `Q24`]; (* 696: umlal v27.2d, v22.2s, v24.2s *)
  []; (* 697: mul x15, x20, x16 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 698: cinv x4, x4, cc *)
  []; (* 699: cmn x10, #0x1 *)
  [`Q26`]; (* 700: usra v1.2d, v26.2d, #32 *)
  [`X12`]; (* 701: adcs x19, x12, x19 *)
  []; (* 702: eor x17, x17, x10 *)
  [`X17`]; (* 703: adcs x9, x9, x17 *)
  []; (* 704: adcs x22, x22, x10 *)
  []; (* 705: lsl x12, x7, #32 *)
  [`X16`]; (* 706: umulh x20, x20, x16 *)
  [`X15`]; (* 707: eor x16, x15, x4 *)
  []; (* 708: ldp x15, x17, [x25, #80] *)
  [`X12`; `X7`]; (* 709: add x2, x12, x7 *)
  [`CF`; `NF`; `ZF`; `VF`; `X10`; `X5`]; (* 710: adc x7, x5, x10 *)
  []; (* 711: ldp x5, x10, [x26, #128] *)
  []; (* 712: lsr x1, x2, #32 *)
  [`X20`]; (* 713: eor x12, x20, x4 *)
  []; (* 714: subs x1, x1, x2 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 715: sbc x20, x2, xzr *)
  []; (* 716: cmn x4, #0x1 *)
  [`X16`]; (* 717: adcs x9, x9, x16 *)
  []; (* 718: extr x1, x20, x1, #32 *)
  []; (* 719: lsr x20, x20, #32 *)
  [`X12`]; (* 720: adcs x22, x22, x12 *)
  [`CF`; `NF`; `ZF`; `VF`; `X4`; `X7`]; (* 721: adc x16, x7, x4 *)
  [`X20`]; (* 722: adds x12, x20, x2 *)
  []; (* 723: umulh x7, x24, x14 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 724: adc x4, xzr, xzr *)
  [`X11`]; (* 725: subs x1, x11, x1 *)
  [`X12`; `X19`]; (* 726: sbcs x20, x19, x12 *)
  [`X4`; `X9`]; (* 727: sbcs x12, x9, x4 *)
  []; (* 728: lsl x9, x1, #32 *)
  [`X9`]; (* 729: add x1, x9, x1 *)
  [`X22`]; (* 730: sbcs x9, x22, xzr *)
  []; (* 731: mul x22, x24, x14 *)
  []; (* 732: sbcs x16, x16, xzr *)
  []; (* 733: lsr x4, x1, #32 *)
  [`CF`; `NF`; `ZF`; `VF`; `X2`]; (* 734: sbc x19, x2, xzr *)
  []; (* 735: subs x4, x4, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 736: sbc x11, x1, xzr *)
  [`X4`]; (* 737: extr x2, x11, x4, #32 *)
  [`X11`]; (* 738: lsr x4, x11, #32 *)
  []; (* 739: adds x4, x4, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 740: adc x11, xzr, xzr *)
  [`X20`]; (* 741: subs x2, x20, x2 *)
  [`X12`]; (* 742: sbcs x4, x12, x4 *)
  [`X11`; `X9`]; (* 743: sbcs x20, x9, x11 *)
  []; (* 744: lsl x12, x2, #32 *)
  [`X12`]; (* 745: add x2, x12, x2 *)
  [`X16`]; (* 746: sbcs x9, x16, xzr *)
  []; (* 747: lsr x11, x2, #32 *)
  []; (* 748: sbcs x19, x19, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 749: sbc x1, x1, xzr *)
  [`X11`]; (* 750: subs x16, x11, x2 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 751: sbc x12, x2, xzr *)
  []; (* 752: extr x16, x12, x16, #32 *)
  []; (* 753: lsr x12, x12, #32 *)
  [`X12`]; (* 754: adds x11, x12, x2 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 755: adc x12, xzr, xzr *)
  [`X4`]; (* 756: subs x16, x4, x16 *)
  []; (* 757: mov x4, v27.d[0] *)
  [`X20`]; (* 758: sbcs x11, x20, x11 *)
  [`X12`; `X9`]; (* 759: sbcs x20, x9, x12 *)
  [`X11`; `X16`]; (* 760: stp x16, x11, [sp, #288] *)
  [`X19`]; (* 761: sbcs x11, x19, xzr *)
  [`X1`]; (* 762: sbcs x9, x1, xzr *)
  [`X11`; `X20`]; (* 763: stp x20, x11, [sp, #304] *)
  []; (* 764: mov x1, v1.d[0] *)
  [`CF`; `NF`; `ZF`; `VF`; `X2`]; (* 765: sbc x20, x2, xzr *)
  []; (* 766: subs x12, x24, x5 *)
  [`Q27`]; (* 767: mov x11, v27.d[1] *)
  [`X12`]; (* 768: cneg x16, x12, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 769: csetm x2, cc *)
  []; (* 770: subs x19, x15, x14 *)
  [`Q1`]; (* 771: mov x12, v1.d[1] *)
  []; (* 772: cinv x2, x2, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 773: cneg x19, x19, cc *)
  [`X20`; `X9`]; (* 774: stp x9, x20, [sp, #320] *)
  []; (* 775: mul x9, x16, x19 *)
  [`X7`]; (* 776: adds x4, x7, x4 *)
  [`X1`]; (* 777: adcs x11, x1, x11 *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 778: adc x1, x12, xzr *)
  []; (* 779: adds x20, x4, x22 *)
  [`X16`]; (* 780: umulh x19, x16, x19 *)
  []; (* 781: adcs x7, x11, x4 *)
  [`X9`]; (* 782: eor x16, x9, x2 *)
  []; (* 783: adcs x9, x1, x11 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 784: adc x12, x1, xzr *)
  []; (* 785: adds x7, x7, x22 *)
  [`X9`]; (* 786: adcs x4, x9, x4 *)
  [`X11`; `X12`]; (* 787: adcs x9, x12, x11 *)
  [`CF`; `NF`; `ZF`; `VF`; `X1`]; (* 788: adc x12, x1, xzr *)
  []; (* 789: cmn x2, #0x1 *)
  [`X19`]; (* 790: eor x1, x19, x2 *)
  [`X16`; `X20`]; (* 791: adcs x11, x20, x16 *)
  [`X1`; `X7`]; (* 792: adcs x19, x7, x1 *)
  [`X4`]; (* 793: adcs x1, x4, x2 *)
  [`X9`]; (* 794: adcs x20, x9, x2 *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 795: adc x2, x12, x2 *)
  []; (* 796: subs x12, x24, x10 *)
  [`X12`]; (* 797: cneg x16, x12, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 798: csetm x12, cc *)
  []; (* 799: subs x9, x17, x14 *)
  []; (* 800: cinv x12, x12, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 801: cneg x9, x9, cc *)
  [`X24`]; (* 802: subs x3, x24, x3 *)
  []; (* 803: sbcs x21, x5, x21 *)
  []; (* 804: mul x24, x16, x9 *)
  [`X8`]; (* 805: sbcs x4, x10, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 806: ngc x8, xzr *)
  [`X5`]; (* 807: subs x10, x5, x10 *)
  [`X24`]; (* 808: eor x5, x24, x12 *)
  []; (* 809: csetm x7, cc *)
  [`CF`; `NF`; `ZF`; `VF`; `X10`]; (* 810: cneg x24, x10, cc *)
  []; (* 811: subs x10, x17, x15 *)
  []; (* 812: cinv x7, x7, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 813: cneg x10, x10, cc *)
  [`X13`]; (* 814: subs x14, x13, x14 *)
  [`X23`]; (* 815: sbcs x15, x23, x15 *)
  [`X21`]; (* 816: eor x13, x21, x8 *)
  []; (* 817: mul x23, x24, x10 *)
  [`X6`]; (* 818: sbcs x17, x6, x17 *)
  [`X3`]; (* 819: eor x6, x3, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 820: ngc x21, xzr *)
  [`X16`]; (* 821: umulh x9, x16, x9 *)
  []; (* 822: cmn x8, #0x1 *)
  [`X23`]; (* 823: eor x3, x23, x7 *)
  [`X6`]; (* 824: adcs x23, x6, xzr *)
  []; (* 825: adcs x13, x13, xzr *)
  [`X4`]; (* 826: eor x16, x4, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 827: adc x16, x16, xzr *)
  [`X17`]; (* 828: eor x4, x17, x21 *)
  [`X10`; `X24`]; (* 829: umulh x17, x24, x10 *)
  []; (* 830: cmn x21, #0x1 *)
  [`X14`]; (* 831: eor x24, x14, x21 *)
  [`X15`]; (* 832: eor x6, x15, x21 *)
  [`X24`]; (* 833: adcs x15, x24, xzr *)
  [`X6`]; (* 834: adcs x14, x6, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X4`]; (* 835: adc x6, x4, xzr *)
  []; (* 836: cmn x12, #0x1 *)
  [`X9`]; (* 837: eor x4, x9, x12 *)
  [`X5`]; (* 838: adcs x19, x19, x5 *)
  []; (* 839: umulh x5, x23, x15 *)
  [`X4`]; (* 840: adcs x1, x1, x4 *)
  [`X20`]; (* 841: adcs x10, x20, x12 *)
  [`X17`]; (* 842: eor x4, x17, x7 *)
  []; (* 843: ldp x20, x9, [sp, #288] *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 844: adc x2, x2, x12 *)
  []; (* 845: cmn x7, #0x1 *)
  [`X1`; `X3`]; (* 846: adcs x12, x1, x3 *)
  []; (* 847: ldp x17, x24, [sp, #304] *)
  []; (* 848: mul x1, x16, x6 *)
  [`X10`; `X4`]; (* 849: adcs x3, x10, x4 *)
  [`CF`; `NF`; `ZF`; `VF`; `X7`]; (* 850: adc x2, x2, x7 *)
  []; (* 851: ldp x7, x4, [sp, #320] *)
  [`X22`]; (* 852: adds x20, x22, x20 *)
  []; (* 853: mul x10, x13, x14 *)
  [`X9`]; (* 854: adcs x11, x11, x9 *)
  [`X21`; `X8`]; (* 855: eor x9, x8, x21 *)
  [`X17`; `X19`]; (* 856: adcs x21, x19, x17 *)
  [`X11`; `X20`]; (* 857: stp x20, x11, [sp, #288] *)
  [`X24`]; (* 858: adcs x12, x12, x24 *)
  []; (* 859: mul x8, x23, x15 *)
  [`X7`]; (* 860: adcs x3, x3, x7 *)
  [`X12`; `X21`]; (* 861: stp x21, x12, [sp, #304] *)
  [`X2`; `X4`]; (* 862: adcs x12, x2, x4 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 863: adc x19, xzr, xzr *)
  []; (* 864: subs x21, x23, x16 *)
  []; (* 865: umulh x2, x16, x6 *)
  [`X12`; `X3`]; (* 866: stp x3, x12, [sp, #320] *)
  [`X21`]; (* 867: cneg x3, x21, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 868: csetm x24, cc *)
  []; (* 869: umulh x11, x13, x14 *)
  [`X16`]; (* 870: subs x21, x13, x16 *)
  []; (* 871: eor x7, x8, x9 *)
  [`X21`]; (* 872: cneg x17, x21, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 873: csetm x16, cc *)
  []; (* 874: subs x21, x6, x15 *)
  [`X21`]; (* 875: cneg x22, x21, cc *)
  [`CF`; `NF`; `ZF`; `VF`; `X24`]; (* 876: cinv x21, x24, cc *)
  [`X13`; `X23`]; (* 877: subs x20, x23, x13 *)
  []; (* 878: umulh x12, x3, x22 *)
  [`X20`]; (* 879: cneg x23, x20, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 880: csetm x24, cc *)
  [`X15`]; (* 881: subs x20, x14, x15 *)
  []; (* 882: cinv x24, x24, cc *)
  [`X3`]; (* 883: mul x22, x3, x22 *)
  [`CF`; `NF`; `ZF`; `VF`; `X20`]; (* 884: cneg x3, x20, cc *)
  [`X14`; `X6`]; (* 885: subs x13, x6, x14 *)
  [`X13`]; (* 886: cneg x20, x13, cc *)
  [`CF`; `NF`; `ZF`; `VF`; `X16`]; (* 887: cinv x15, x16, cc *)
  [`X10`; `X5`]; (* 888: adds x13, x5, x10 *)
  []; (* 889: mul x4, x23, x3 *)
  [`X1`]; (* 890: adcs x11, x11, x1 *)
  [`CF`; `NF`; `ZF`; `VF`; `X2`]; (* 891: adc x14, x2, xzr *)
  []; (* 892: adds x5, x13, x8 *)
  []; (* 893: adcs x16, x11, x13 *)
  [`X3`]; (* 894: umulh x23, x23, x3 *)
  []; (* 895: adcs x3, x14, x11 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 896: adc x1, x14, xzr *)
  [`X16`; `X8`]; (* 897: adds x10, x16, x8 *)
  [`X13`; `X3`]; (* 898: adcs x6, x3, x13 *)
  [`X1`; `X11`]; (* 899: adcs x8, x1, x11 *)
  []; (* 900: umulh x13, x17, x20 *)
  [`X4`]; (* 901: eor x1, x4, x24 *)
  [`CF`; `NF`; `ZF`; `VF`; `X14`]; (* 902: adc x4, x14, xzr *)
  []; (* 903: cmn x24, #0x1 *)
  [`X5`]; (* 904: adcs x1, x5, x1 *)
  [`X23`]; (* 905: eor x16, x23, x24 *)
  [`X1`]; (* 906: eor x11, x1, x9 *)
  [`X10`; `X16`]; (* 907: adcs x23, x10, x16 *)
  [`X22`]; (* 908: eor x2, x22, x21 *)
  [`X6`]; (* 909: adcs x3, x6, x24 *)
  [`X17`; `X20`]; (* 910: mul x14, x17, x20 *)
  [`X13`]; (* 911: eor x17, x13, x15 *)
  [`X8`]; (* 912: adcs x13, x8, x24 *)
  [`CF`; `NF`; `ZF`; `VF`; `X24`; `X4`]; (* 913: adc x8, x4, x24 *)
  []; (* 914: cmn x21, #0x1 *)
  [`X2`; `X23`]; (* 915: adcs x6, x23, x2 *)
  []; (* 916: mov x16, #0xfffffffffffffffe *)
  [`X12`]; (* 917: eor x20, x12, x21 *)
  [`X3`]; (* 918: adcs x20, x3, x20 *)
  [`X14`]; (* 919: eor x23, x14, x15 *)
  [`X13`]; (* 920: adcs x2, x13, x21 *)
  [`CF`; `NF`; `ZF`; `VF`; `X21`]; (* 921: adc x8, x8, x21 *)
  []; (* 922: cmn x15, #0x1 *)
  []; (* 923: ldp x5, x4, [sp, #288] *)
  []; (* 924: ldp x21, x12, [sp, #304] *)
  [`X20`; `X23`]; (* 925: adcs x22, x20, x23 *)
  [`X22`]; (* 926: eor x23, x22, x9 *)
  [`X2`]; (* 927: adcs x17, x2, x17 *)
  [`CF`; `NF`; `ZF`; `VF`; `X15`; `X8`]; (* 928: adc x22, x8, x15 *)
  []; (* 929: cmn x9, #0x1 *)
  [`X7`]; (* 930: adcs x15, x7, x5 *)
  []; (* 931: ldp x10, x14, [sp, #320] *)
  [`X6`]; (* 932: eor x1, x6, x9 *)
  []; (* 933: lsl x2, x15, #32 *)
  [`X11`]; (* 934: adcs x8, x11, x4 *)
  [`X1`]; (* 935: adcs x13, x1, x21 *)
  [`X22`]; (* 936: eor x1, x22, x9 *)
  [`X23`]; (* 937: adcs x24, x23, x12 *)
  [`X17`]; (* 938: eor x11, x17, x9 *)
  [`X11`]; (* 939: adcs x23, x11, x10 *)
  [`X1`]; (* 940: adcs x7, x1, x14 *)
  []; (* 941: adcs x17, x9, x19 *)
  []; (* 942: adcs x20, x9, xzr *)
  [`X15`; `X2`]; (* 943: add x1, x2, x15 *)
  []; (* 944: lsr x3, x1, #32 *)
  []; (* 945: adcs x11, x9, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 946: adc x9, x9, xzr *)
  []; (* 947: subs x3, x3, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 948: sbc x6, x1, xzr *)
  [`X5`]; (* 949: adds x24, x24, x5 *)
  [`X23`]; (* 950: adcs x4, x23, x4 *)
  []; (* 951: extr x3, x6, x3, #32 *)
  []; (* 952: lsr x6, x6, #32 *)
  [`X7`]; (* 953: adcs x21, x7, x21 *)
  [`X12`; `X17`]; (* 954: adcs x15, x17, x12 *)
  [`X10`; `X20`]; (* 955: adcs x7, x20, x10 *)
  [`X11`; `X14`]; (* 956: adcs x20, x11, x14 *)
  []; (* 957: mov x14, #0xffffffff *)
  [`CF`; `NF`; `ZF`; `VF`; `X19`; `X9`]; (* 958: adc x22, x9, x19 *)
  [`X6`]; (* 959: adds x12, x6, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 960: adc x10, xzr, xzr *)
  [`X8`]; (* 961: subs x3, x8, x3 *)
  [`X13`]; (* 962: sbcs x12, x13, x12 *)
  []; (* 963: lsl x9, x3, #32 *)
  [`X9`]; (* 964: add x3, x9, x3 *)
  [`X24`]; (* 965: sbcs x10, x24, x10 *)
  [`X4`]; (* 966: sbcs x24, x4, xzr *)
  []; (* 967: lsr x9, x3, #32 *)
  []; (* 968: sbcs x21, x21, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 969: sbc x1, x1, xzr *)
  []; (* 970: subs x9, x9, x3 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 971: sbc x13, x3, xzr *)
  []; (* 972: extr x9, x13, x9, #32 *)
  []; (* 973: lsr x13, x13, #32 *)
  []; (* 974: adds x13, x13, x3 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 975: adc x6, xzr, xzr *)
  [`X9`]; (* 976: subs x12, x12, x9 *)
  [`X10`; `X13`]; (* 977: sbcs x17, x10, x13 *)
  []; (* 978: lsl x2, x12, #32 *)
  [`X24`; `X6`]; (* 979: sbcs x10, x24, x6 *)
  [`X12`; `X2`]; (* 980: add x9, x2, x12 *)
  [`X21`]; (* 981: sbcs x6, x21, xzr *)
  []; (* 982: lsr x5, x9, #32 *)
  [`X1`]; (* 983: sbcs x21, x1, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X3`]; (* 984: sbc x13, x3, xzr *)
  [`X5`]; (* 985: subs x8, x5, x9 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 986: sbc x19, x9, xzr *)
  []; (* 987: lsr x12, x19, #32 *)
  [`X19`; `X8`]; (* 988: extr x3, x19, x8, #32 *)
  [`X12`]; (* 989: adds x8, x12, x9 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 990: adc x1, xzr, xzr *)
  [`X17`; `X3`]; (* 991: subs x2, x17, x3 *)
  [`X10`; `X8`]; (* 992: sbcs x12, x10, x8 *)
  [`X1`; `X6`]; (* 993: sbcs x5, x6, x1 *)
  [`X21`]; (* 994: sbcs x3, x21, xzr *)
  [`X13`]; (* 995: sbcs x19, x13, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X9`]; (* 996: sbc x24, x9, xzr *)
  [`X15`; `X3`]; (* 997: adds x23, x15, x3 *)
  [`X19`; `X7`]; (* 998: adcs x8, x7, x19 *)
  [`X20`; `X24`]; (* 999: adcs x11, x20, x24 *)
  [`CF`; `NF`; `ZF`; `VF`; `X22`]; (* 1000: adc x9, x22, xzr *)
  [`X9`]; (* 1001: add x24, x9, #0x1 *)
  []; (* 1002: lsl x7, x24, #32 *)
  []; (* 1003: subs x21, x24, x7 *)
  [`CF`; `NF`; `ZF`; `VF`; `X7`]; (* 1004: sbc x10, x7, xzr *)
  [`X2`; `X21`]; (* 1005: adds x6, x2, x21 *)
  [`X10`; `X12`]; (* 1006: adcs x7, x12, x10 *)
  [`X5`]; (* 1007: adcs x24, x5, x24 *)
  [`X23`]; (* 1008: adcs x13, x23, xzr *)
  []; (* 1009: adcs x8, x8, xzr *)
  [`X11`]; (* 1010: adcs x15, x11, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1011: csetm x23, cc *)
  [`X16`]; (* 1012: and x11, x16, x23 *)
  [`X14`]; (* 1013: and x20, x14, x23 *)
  [`X6`]; (* 1014: adds x22, x6, x20 *)
  [`X20`]; (* 1015: eor x3, x20, x23 *)
  [`X3`; `X7`]; (* 1016: adcs x5, x7, x3 *)
  [`X11`; `X24`]; (* 1017: adcs x14, x24, x11 *)
  [`X22`; `X5`]; (* 1018: stp x22, x5, [sp, #288] *)
  [`X13`]; (* 1019: adcs x5, x13, x23 *)
  [`X8`]; (* 1020: adcs x21, x8, x23 *)
  [`X14`; `X5`]; (* 1021: stp x14, x5, [sp, #304] *)
  [`CF`; `NF`; `ZF`; `VF`; `X15`; `X23`]; (* 1022: adc x12, x15, x23 *)
  [`X12`; `X21`]; (* 1023: stp x21, x12, [sp, #320] *)
  []; (* 1024: ldr q3, [x25, #96] *)
  []; (* 1025: ldr q25, [x26, #48] *)
  []; (* 1026: ldp x13, x23, [x26, #48] *)
  []; (* 1027: ldp x3, x21, [x25, #96] *)
  []; (* 1028: rev64 v23.4s, v25.4s *)
  [`Q25`]; (* 1029: uzp1 v17.4s, v25.4s, v3.4s *)
  []; (* 1030: umulh x15, x3, x13 *)
  [`Q23`]; (* 1031: mul v6.4s, v23.4s, v3.4s *)
  []; (* 1032: uzp1 v3.4s, v3.4s, v3.4s *)
  []; (* 1033: ldr q27, [x26, #80] *)
  []; (* 1034: ldp x8, x24, [x25, #112] *)
  []; (* 1035: subs x6, x3, x21 *)
  []; (* 1036: ldr q0, [x25, #128] *)
  []; (* 1037: movi v23.2d, #0xffffffff *)
  []; (* 1038: csetm x10, cc *)
  []; (* 1039: umulh x19, x21, x23 *)
  []; (* 1040: rev64 v4.4s, v27.4s *)
  []; (* 1041: uzp2 v25.4s, v27.4s, v27.4s *)
  [`CF`; `NF`; `ZF`; `VF`; `X6`]; (* 1042: cneg x4, x6, cc *)
  []; (* 1043: subs x7, x23, x13 *)
  []; (* 1044: xtn v22.2s, v0.2d *)
  [`Q27`]; (* 1045: xtn v24.2s, v27.2d *)
  [`X7`]; (* 1046: cneg x20, x7, cc *)
  []; (* 1047: ldp x6, x14, [x26, #64] *)
  [`Q4`]; (* 1048: mul v27.4s, v4.4s, v0.4s *)
  [`Q6`]; (* 1049: uaddlp v20.2d, v6.4s *)
  [`CF`; `NF`; `ZF`; `VF`; `X10`]; (* 1050: cinv x5, x10, cc *)
  []; (* 1051: mul x16, x4, x20 *)
  [`Q0`]; (* 1052: uzp2 v6.4s, v0.4s, v0.4s *)
  []; (* 1053: umull v21.2d, v22.2s, v25.2s *)
  [`Q20`]; (* 1054: shl v0.2d, v20.2d, #32 *)
  [`Q17`; `Q3`]; (* 1055: umlal v0.2d, v3.2s, v17.2s *)
  []; (* 1056: mul x22, x8, x6 *)
  [`Q25`]; (* 1057: umull v1.2d, v6.2s, v25.2s *)
  []; (* 1058: subs x12, x3, x8 *)
  []; (* 1059: umull v20.2d, v22.2s, v24.2s *)
  [`X12`]; (* 1060: cneg x17, x12, cc *)
  []; (* 1061: umulh x9, x8, x6 *)
  []; (* 1062: mov x12, v0.d[1] *)
  [`X16`]; (* 1063: eor x11, x16, x5 *)
  [`Q0`]; (* 1064: mov x7, v0.d[0] *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1065: csetm x10, cc *)
  [`Q20`]; (* 1066: usra v21.2d, v20.2d, #32 *)
  [`X12`]; (* 1067: adds x15, x15, x12 *)
  [`X19`; `X22`]; (* 1068: adcs x12, x19, x22 *)
  [`X4`]; (* 1069: umulh x20, x4, x20 *)
  [`CF`; `NF`; `ZF`; `VF`; `X9`]; (* 1070: adc x19, x9, xzr *)
  []; (* 1071: usra v1.2d, v21.2d, #32 *)
  []; (* 1072: adds x22, x15, x7 *)
  [`Q21`; `Q23`]; (* 1073: and v26.16b, v21.16b, v23.16b *)
  []; (* 1074: adcs x16, x12, x15 *)
  [`Q27`]; (* 1075: uaddlp v25.2d, v27.4s *)
  []; (* 1076: adcs x9, x19, x12 *)
  [`Q6`]; (* 1077: umlal v26.2d, v6.2s, v24.2s *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1078: adc x4, x19, xzr *)
  []; (* 1079: adds x16, x16, x7 *)
  [`Q25`]; (* 1080: shl v27.2d, v25.2d, #32 *)
  [`X15`]; (* 1081: adcs x9, x9, x15 *)
  [`X12`]; (* 1082: adcs x4, x4, x12 *)
  [`X20`]; (* 1083: eor x12, x20, x5 *)
  [`CF`; `NF`; `ZF`; `VF`; `X19`]; (* 1084: adc x15, x19, xzr *)
  []; (* 1085: subs x20, x6, x13 *)
  []; (* 1086: cneg x20, x20, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1087: cinv x10, x10, cc *)
  []; (* 1088: cmn x5, #0x1 *)
  []; (* 1089: mul x19, x17, x20 *)
  [`X22`]; (* 1090: adcs x11, x22, x11 *)
  [`X16`]; (* 1091: adcs x12, x16, x12 *)
  []; (* 1092: adcs x9, x9, x5 *)
  [`X20`]; (* 1093: umulh x17, x17, x20 *)
  [`X4`]; (* 1094: adcs x22, x4, x5 *)
  [`CF`; `NF`; `ZF`; `VF`; `X15`]; (* 1095: adc x5, x15, x5 *)
  []; (* 1096: subs x16, x21, x8 *)
  [`X16`]; (* 1097: cneg x20, x16, cc *)
  []; (* 1098: eor x19, x19, x10 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1099: csetm x4, cc *)
  []; (* 1100: subs x16, x6, x23 *)
  []; (* 1101: cneg x16, x16, cc *)
  [`Q22`; `Q24`]; (* 1102: umlal v27.2d, v22.2s, v24.2s *)
  []; (* 1103: mul x15, x20, x16 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1104: cinv x4, x4, cc *)
  []; (* 1105: cmn x10, #0x1 *)
  [`Q26`]; (* 1106: usra v1.2d, v26.2d, #32 *)
  [`X12`]; (* 1107: adcs x19, x12, x19 *)
  []; (* 1108: eor x17, x17, x10 *)
  [`X17`]; (* 1109: adcs x9, x9, x17 *)
  []; (* 1110: adcs x22, x22, x10 *)
  []; (* 1111: lsl x12, x7, #32 *)
  [`X16`]; (* 1112: umulh x20, x20, x16 *)
  [`X15`]; (* 1113: eor x16, x15, x4 *)
  []; (* 1114: ldp x15, x17, [x26, #80] *)
  [`X12`; `X7`]; (* 1115: add x2, x12, x7 *)
  [`CF`; `NF`; `ZF`; `VF`; `X10`; `X5`]; (* 1116: adc x7, x5, x10 *)
  []; (* 1117: ldp x5, x10, [x25, #128] *)
  []; (* 1118: lsr x1, x2, #32 *)
  [`X20`]; (* 1119: eor x12, x20, x4 *)
  []; (* 1120: subs x1, x1, x2 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1121: sbc x20, x2, xzr *)
  []; (* 1122: cmn x4, #0x1 *)
  [`X16`]; (* 1123: adcs x9, x9, x16 *)
  []; (* 1124: extr x1, x20, x1, #32 *)
  []; (* 1125: lsr x20, x20, #32 *)
  [`X12`]; (* 1126: adcs x22, x22, x12 *)
  [`CF`; `NF`; `ZF`; `VF`; `X4`; `X7`]; (* 1127: adc x16, x7, x4 *)
  [`X20`]; (* 1128: adds x12, x20, x2 *)
  []; (* 1129: umulh x7, x24, x14 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1130: adc x4, xzr, xzr *)
  [`X11`]; (* 1131: subs x1, x11, x1 *)
  [`X12`; `X19`]; (* 1132: sbcs x20, x19, x12 *)
  [`X4`; `X9`]; (* 1133: sbcs x12, x9, x4 *)
  []; (* 1134: lsl x9, x1, #32 *)
  [`X9`]; (* 1135: add x1, x9, x1 *)
  [`X22`]; (* 1136: sbcs x9, x22, xzr *)
  []; (* 1137: mul x22, x24, x14 *)
  []; (* 1138: sbcs x16, x16, xzr *)
  []; (* 1139: lsr x4, x1, #32 *)
  [`CF`; `NF`; `ZF`; `VF`; `X2`]; (* 1140: sbc x19, x2, xzr *)
  []; (* 1141: subs x4, x4, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1142: sbc x11, x1, xzr *)
  [`X4`]; (* 1143: extr x2, x11, x4, #32 *)
  [`X11`]; (* 1144: lsr x4, x11, #32 *)
  []; (* 1145: adds x4, x4, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1146: adc x11, xzr, xzr *)
  [`X20`]; (* 1147: subs x2, x20, x2 *)
  [`X12`]; (* 1148: sbcs x4, x12, x4 *)
  [`X11`; `X9`]; (* 1149: sbcs x20, x9, x11 *)
  []; (* 1150: lsl x12, x2, #32 *)
  [`X12`]; (* 1151: add x2, x12, x2 *)
  [`X16`]; (* 1152: sbcs x9, x16, xzr *)
  []; (* 1153: lsr x11, x2, #32 *)
  []; (* 1154: sbcs x19, x19, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1155: sbc x1, x1, xzr *)
  [`X11`]; (* 1156: subs x16, x11, x2 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1157: sbc x12, x2, xzr *)
  []; (* 1158: extr x16, x12, x16, #32 *)
  []; (* 1159: lsr x12, x12, #32 *)
  [`X12`]; (* 1160: adds x11, x12, x2 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1161: adc x12, xzr, xzr *)
  [`X4`]; (* 1162: subs x16, x4, x16 *)
  []; (* 1163: mov x4, v27.d[0] *)
  [`X20`]; (* 1164: sbcs x11, x20, x11 *)
  [`X12`; `X9`]; (* 1165: sbcs x20, x9, x12 *)
  [`X11`; `X16`]; (* 1166: stp x16, x11, [sp, #48] *)
  [`X19`]; (* 1167: sbcs x11, x19, xzr *)
  [`X1`]; (* 1168: sbcs x9, x1, xzr *)
  [`X11`; `X20`]; (* 1169: stp x20, x11, [sp, #64] *)
  []; (* 1170: mov x1, v1.d[0] *)
  [`CF`; `NF`; `ZF`; `VF`; `X2`]; (* 1171: sbc x20, x2, xzr *)
  []; (* 1172: subs x12, x24, x5 *)
  [`Q27`]; (* 1173: mov x11, v27.d[1] *)
  [`X12`]; (* 1174: cneg x16, x12, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1175: csetm x2, cc *)
  []; (* 1176: subs x19, x15, x14 *)
  [`Q1`]; (* 1177: mov x12, v1.d[1] *)
  []; (* 1178: cinv x2, x2, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1179: cneg x19, x19, cc *)
  [`X20`; `X9`]; (* 1180: stp x9, x20, [sp, #80] *)
  []; (* 1181: mul x9, x16, x19 *)
  [`X7`]; (* 1182: adds x4, x7, x4 *)
  [`X1`]; (* 1183: adcs x11, x1, x11 *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 1184: adc x1, x12, xzr *)
  []; (* 1185: adds x20, x4, x22 *)
  [`X16`]; (* 1186: umulh x19, x16, x19 *)
  []; (* 1187: adcs x7, x11, x4 *)
  [`X9`]; (* 1188: eor x16, x9, x2 *)
  []; (* 1189: adcs x9, x1, x11 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1190: adc x12, x1, xzr *)
  []; (* 1191: adds x7, x7, x22 *)
  [`X9`]; (* 1192: adcs x4, x9, x4 *)
  [`X11`; `X12`]; (* 1193: adcs x9, x12, x11 *)
  [`CF`; `NF`; `ZF`; `VF`; `X1`]; (* 1194: adc x12, x1, xzr *)
  []; (* 1195: cmn x2, #0x1 *)
  [`X19`]; (* 1196: eor x1, x19, x2 *)
  [`X16`; `X20`]; (* 1197: adcs x11, x20, x16 *)
  [`X1`; `X7`]; (* 1198: adcs x19, x7, x1 *)
  [`X4`]; (* 1199: adcs x1, x4, x2 *)
  [`X9`]; (* 1200: adcs x20, x9, x2 *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 1201: adc x2, x12, x2 *)
  []; (* 1202: subs x12, x24, x10 *)
  [`X12`]; (* 1203: cneg x16, x12, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1204: csetm x12, cc *)
  []; (* 1205: subs x9, x17, x14 *)
  []; (* 1206: cinv x12, x12, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1207: cneg x9, x9, cc *)
  [`X24`]; (* 1208: subs x3, x24, x3 *)
  []; (* 1209: sbcs x21, x5, x21 *)
  []; (* 1210: mul x24, x16, x9 *)
  [`X8`]; (* 1211: sbcs x4, x10, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1212: ngc x8, xzr *)
  [`X5`]; (* 1213: subs x10, x5, x10 *)
  [`X24`]; (* 1214: eor x5, x24, x12 *)
  []; (* 1215: csetm x7, cc *)
  [`CF`; `NF`; `ZF`; `VF`; `X10`]; (* 1216: cneg x24, x10, cc *)
  []; (* 1217: subs x10, x17, x15 *)
  []; (* 1218: cinv x7, x7, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1219: cneg x10, x10, cc *)
  [`X13`]; (* 1220: subs x14, x13, x14 *)
  [`X23`]; (* 1221: sbcs x15, x23, x15 *)
  [`X21`]; (* 1222: eor x13, x21, x8 *)
  []; (* 1223: mul x23, x24, x10 *)
  [`X6`]; (* 1224: sbcs x17, x6, x17 *)
  [`X3`]; (* 1225: eor x6, x3, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1226: ngc x21, xzr *)
  [`X16`]; (* 1227: umulh x9, x16, x9 *)
  []; (* 1228: cmn x8, #0x1 *)
  [`X23`]; (* 1229: eor x3, x23, x7 *)
  [`X6`]; (* 1230: adcs x23, x6, xzr *)
  []; (* 1231: adcs x13, x13, xzr *)
  [`X4`]; (* 1232: eor x16, x4, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1233: adc x16, x16, xzr *)
  [`X17`]; (* 1234: eor x4, x17, x21 *)
  [`X10`; `X24`]; (* 1235: umulh x17, x24, x10 *)
  []; (* 1236: cmn x21, #0x1 *)
  [`X14`]; (* 1237: eor x24, x14, x21 *)
  [`X15`]; (* 1238: eor x6, x15, x21 *)
  [`X24`]; (* 1239: adcs x15, x24, xzr *)
  [`X6`]; (* 1240: adcs x14, x6, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X4`]; (* 1241: adc x6, x4, xzr *)
  []; (* 1242: cmn x12, #0x1 *)
  [`X9`]; (* 1243: eor x4, x9, x12 *)
  [`X5`]; (* 1244: adcs x19, x19, x5 *)
  []; (* 1245: umulh x5, x23, x15 *)
  [`X4`]; (* 1246: adcs x1, x1, x4 *)
  [`X20`]; (* 1247: adcs x10, x20, x12 *)
  [`X17`]; (* 1248: eor x4, x17, x7 *)
  []; (* 1249: ldp x20, x9, [sp, #48] *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 1250: adc x2, x2, x12 *)
  []; (* 1251: cmn x7, #0x1 *)
  [`X1`; `X3`]; (* 1252: adcs x12, x1, x3 *)
  []; (* 1253: ldp x17, x24, [sp, #64] *)
  []; (* 1254: mul x1, x16, x6 *)
  [`X10`; `X4`]; (* 1255: adcs x3, x10, x4 *)
  [`CF`; `NF`; `ZF`; `VF`; `X7`]; (* 1256: adc x2, x2, x7 *)
  []; (* 1257: ldp x7, x4, [sp, #80] *)
  [`X22`]; (* 1258: adds x20, x22, x20 *)
  []; (* 1259: mul x10, x13, x14 *)
  [`X9`]; (* 1260: adcs x11, x11, x9 *)
  [`X21`; `X8`]; (* 1261: eor x9, x8, x21 *)
  [`X17`; `X19`]; (* 1262: adcs x21, x19, x17 *)
  [`X11`; `X20`]; (* 1263: stp x20, x11, [sp, #48] *)
  [`X24`]; (* 1264: adcs x12, x12, x24 *)
  []; (* 1265: mul x8, x23, x15 *)
  [`X7`]; (* 1266: adcs x3, x3, x7 *)
  [`X12`; `X21`]; (* 1267: stp x21, x12, [sp, #64] *)
  [`X2`; `X4`]; (* 1268: adcs x12, x2, x4 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1269: adc x19, xzr, xzr *)
  []; (* 1270: subs x21, x23, x16 *)
  []; (* 1271: umulh x2, x16, x6 *)
  [`X12`; `X3`]; (* 1272: stp x3, x12, [sp, #80] *)
  [`X21`]; (* 1273: cneg x3, x21, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1274: csetm x24, cc *)
  []; (* 1275: umulh x11, x13, x14 *)
  [`X16`]; (* 1276: subs x21, x13, x16 *)
  []; (* 1277: eor x7, x8, x9 *)
  [`X21`]; (* 1278: cneg x17, x21, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1279: csetm x16, cc *)
  []; (* 1280: subs x21, x6, x15 *)
  [`X21`]; (* 1281: cneg x22, x21, cc *)
  [`CF`; `NF`; `ZF`; `VF`; `X24`]; (* 1282: cinv x21, x24, cc *)
  [`X13`; `X23`]; (* 1283: subs x20, x23, x13 *)
  []; (* 1284: umulh x12, x3, x22 *)
  [`X20`]; (* 1285: cneg x23, x20, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1286: csetm x24, cc *)
  [`X15`]; (* 1287: subs x20, x14, x15 *)
  []; (* 1288: cinv x24, x24, cc *)
  [`X3`]; (* 1289: mul x22, x3, x22 *)
  [`CF`; `NF`; `ZF`; `VF`; `X20`]; (* 1290: cneg x3, x20, cc *)
  [`X14`; `X6`]; (* 1291: subs x13, x6, x14 *)
  [`X13`]; (* 1292: cneg x20, x13, cc *)
  [`CF`; `NF`; `ZF`; `VF`; `X16`]; (* 1293: cinv x15, x16, cc *)
  [`X10`; `X5`]; (* 1294: adds x13, x5, x10 *)
  []; (* 1295: mul x4, x23, x3 *)
  [`X1`]; (* 1296: adcs x11, x11, x1 *)
  [`CF`; `NF`; `ZF`; `VF`; `X2`]; (* 1297: adc x14, x2, xzr *)
  []; (* 1298: adds x5, x13, x8 *)
  []; (* 1299: adcs x16, x11, x13 *)
  [`X3`]; (* 1300: umulh x23, x23, x3 *)
  []; (* 1301: adcs x3, x14, x11 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1302: adc x1, x14, xzr *)
  [`X16`; `X8`]; (* 1303: adds x10, x16, x8 *)
  [`X13`; `X3`]; (* 1304: adcs x6, x3, x13 *)
  [`X1`; `X11`]; (* 1305: adcs x8, x1, x11 *)
  []; (* 1306: umulh x13, x17, x20 *)
  [`X4`]; (* 1307: eor x1, x4, x24 *)
  [`CF`; `NF`; `ZF`; `VF`; `X14`]; (* 1308: adc x4, x14, xzr *)
  []; (* 1309: cmn x24, #0x1 *)
  [`X5`]; (* 1310: adcs x1, x5, x1 *)
  [`X23`]; (* 1311: eor x16, x23, x24 *)
  [`X1`]; (* 1312: eor x11, x1, x9 *)
  [`X10`; `X16`]; (* 1313: adcs x23, x10, x16 *)
  [`X22`]; (* 1314: eor x2, x22, x21 *)
  [`X6`]; (* 1315: adcs x3, x6, x24 *)
  [`X17`; `X20`]; (* 1316: mul x14, x17, x20 *)
  [`X13`]; (* 1317: eor x17, x13, x15 *)
  [`X8`]; (* 1318: adcs x13, x8, x24 *)
  [`CF`; `NF`; `ZF`; `VF`; `X24`; `X4`]; (* 1319: adc x8, x4, x24 *)
  []; (* 1320: cmn x21, #0x1 *)
  [`X2`; `X23`]; (* 1321: adcs x6, x23, x2 *)
  []; (* 1322: mov x16, #0xfffffffffffffffe *)
  [`X12`]; (* 1323: eor x20, x12, x21 *)
  [`X3`]; (* 1324: adcs x20, x3, x20 *)
  [`X14`]; (* 1325: eor x23, x14, x15 *)
  [`X13`]; (* 1326: adcs x2, x13, x21 *)
  [`CF`; `NF`; `ZF`; `VF`; `X21`]; (* 1327: adc x8, x8, x21 *)
  []; (* 1328: cmn x15, #0x1 *)
  []; (* 1329: ldp x5, x4, [sp, #48] *)
  []; (* 1330: ldp x21, x12, [sp, #64] *)
  [`X20`; `X23`]; (* 1331: adcs x22, x20, x23 *)
  [`X22`]; (* 1332: eor x23, x22, x9 *)
  [`X2`]; (* 1333: adcs x17, x2, x17 *)
  [`CF`; `NF`; `ZF`; `VF`; `X15`; `X8`]; (* 1334: adc x22, x8, x15 *)
  []; (* 1335: cmn x9, #0x1 *)
  [`X7`]; (* 1336: adcs x15, x7, x5 *)
  []; (* 1337: ldp x10, x14, [sp, #80] *)
  [`X6`]; (* 1338: eor x1, x6, x9 *)
  []; (* 1339: lsl x2, x15, #32 *)
  [`X11`]; (* 1340: adcs x8, x11, x4 *)
  [`X1`]; (* 1341: adcs x13, x1, x21 *)
  [`X22`]; (* 1342: eor x1, x22, x9 *)
  [`X23`]; (* 1343: adcs x24, x23, x12 *)
  [`X17`]; (* 1344: eor x11, x17, x9 *)
  [`X11`]; (* 1345: adcs x23, x11, x10 *)
  [`X1`]; (* 1346: adcs x7, x1, x14 *)
  []; (* 1347: adcs x17, x9, x19 *)
  []; (* 1348: adcs x20, x9, xzr *)
  [`X15`; `X2`]; (* 1349: add x1, x2, x15 *)
  []; (* 1350: lsr x3, x1, #32 *)
  []; (* 1351: adcs x11, x9, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1352: adc x9, x9, xzr *)
  []; (* 1353: subs x3, x3, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1354: sbc x6, x1, xzr *)
  [`X5`]; (* 1355: adds x24, x24, x5 *)
  [`X23`]; (* 1356: adcs x4, x23, x4 *)
  []; (* 1357: extr x3, x6, x3, #32 *)
  []; (* 1358: lsr x6, x6, #32 *)
  [`X7`]; (* 1359: adcs x21, x7, x21 *)
  [`X12`; `X17`]; (* 1360: adcs x15, x17, x12 *)
  [`X10`; `X20`]; (* 1361: adcs x7, x20, x10 *)
  [`X11`; `X14`]; (* 1362: adcs x20, x11, x14 *)
  []; (* 1363: mov x14, #0xffffffff *)
  [`CF`; `NF`; `ZF`; `VF`; `X19`; `X9`]; (* 1364: adc x22, x9, x19 *)
  [`X6`]; (* 1365: adds x12, x6, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1366: adc x10, xzr, xzr *)
  [`X8`]; (* 1367: subs x3, x8, x3 *)
  [`X13`]; (* 1368: sbcs x12, x13, x12 *)
  []; (* 1369: lsl x9, x3, #32 *)
  [`X9`]; (* 1370: add x3, x9, x3 *)
  [`X24`]; (* 1371: sbcs x10, x24, x10 *)
  [`X4`]; (* 1372: sbcs x24, x4, xzr *)
  []; (* 1373: lsr x9, x3, #32 *)
  []; (* 1374: sbcs x21, x21, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1375: sbc x1, x1, xzr *)
  []; (* 1376: subs x9, x9, x3 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1377: sbc x13, x3, xzr *)
  []; (* 1378: extr x9, x13, x9, #32 *)
  []; (* 1379: lsr x13, x13, #32 *)
  []; (* 1380: adds x13, x13, x3 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1381: adc x6, xzr, xzr *)
  [`X9`]; (* 1382: subs x12, x12, x9 *)
  [`X10`; `X13`]; (* 1383: sbcs x17, x10, x13 *)
  []; (* 1384: lsl x2, x12, #32 *)
  [`X24`; `X6`]; (* 1385: sbcs x10, x24, x6 *)
  [`X12`; `X2`]; (* 1386: add x9, x2, x12 *)
  [`X21`]; (* 1387: sbcs x6, x21, xzr *)
  []; (* 1388: lsr x5, x9, #32 *)
  [`X1`]; (* 1389: sbcs x21, x1, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X3`]; (* 1390: sbc x13, x3, xzr *)
  [`X5`]; (* 1391: subs x8, x5, x9 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1392: sbc x19, x9, xzr *)
  []; (* 1393: lsr x12, x19, #32 *)
  [`X19`; `X8`]; (* 1394: extr x3, x19, x8, #32 *)
  [`X12`]; (* 1395: adds x8, x12, x9 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1396: adc x1, xzr, xzr *)
  [`X17`; `X3`]; (* 1397: subs x2, x17, x3 *)
  [`X10`; `X8`]; (* 1398: sbcs x12, x10, x8 *)
  [`X1`; `X6`]; (* 1399: sbcs x5, x6, x1 *)
  [`X21`]; (* 1400: sbcs x3, x21, xzr *)
  [`X13`]; (* 1401: sbcs x19, x13, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X9`]; (* 1402: sbc x24, x9, xzr *)
  [`X15`; `X3`]; (* 1403: adds x23, x15, x3 *)
  [`X19`; `X7`]; (* 1404: adcs x8, x7, x19 *)
  [`X20`; `X24`]; (* 1405: adcs x11, x20, x24 *)
  [`CF`; `NF`; `ZF`; `VF`; `X22`]; (* 1406: adc x9, x22, xzr *)
  [`X9`]; (* 1407: add x24, x9, #0x1 *)
  []; (* 1408: lsl x7, x24, #32 *)
  []; (* 1409: subs x21, x24, x7 *)
  [`CF`; `NF`; `ZF`; `VF`; `X7`]; (* 1410: sbc x10, x7, xzr *)
  [`X2`; `X21`]; (* 1411: adds x6, x2, x21 *)
  [`X10`; `X12`]; (* 1412: adcs x7, x12, x10 *)
  [`X5`]; (* 1413: adcs x24, x5, x24 *)
  [`X23`]; (* 1414: adcs x13, x23, xzr *)
  []; (* 1415: adcs x8, x8, xzr *)
  [`X11`]; (* 1416: adcs x15, x11, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1417: csetm x23, cc *)
  [`X16`]; (* 1418: and x11, x16, x23 *)
  [`X14`]; (* 1419: and x20, x14, x23 *)
  [`X6`]; (* 1420: adds x22, x6, x20 *)
  [`X20`]; (* 1421: eor x3, x20, x23 *)
  [`X3`; `X7`]; (* 1422: adcs x5, x7, x3 *)
  [`X11`; `X24`]; (* 1423: adcs x14, x24, x11 *)
  [`X22`; `X5`]; (* 1424: stp x22, x5, [sp, #48] *)
  [`X13`]; (* 1425: adcs x5, x13, x23 *)
  [`X8`]; (* 1426: adcs x21, x8, x23 *)
  [`X14`; `X5`]; (* 1427: stp x14, x5, [sp, #64] *)
  [`CF`; `NF`; `ZF`; `VF`; `X15`; `X23`]; (* 1428: adc x12, x15, x23 *)
  [`X12`; `X21`]; (* 1429: stp x21, x12, [sp, #80] *)
  []; (* 1430: mov x1, sp *)
  []; (* 1431: ldr q3, [x1] *)
  []; (* 1432: ldr q25, [x26, #0] *)
  []; (* 1433: ldp x13, x23, [x26, #0] *)
  []; (* 1434: ldp x3, x21, [x1] *)
  []; (* 1435: rev64 v23.4s, v25.4s *)
  [`Q25`]; (* 1436: uzp1 v17.4s, v25.4s, v3.4s *)
  []; (* 1437: umulh x15, x3, x13 *)
  [`Q23`]; (* 1438: mul v6.4s, v23.4s, v3.4s *)
  []; (* 1439: uzp1 v3.4s, v3.4s, v3.4s *)
  []; (* 1440: ldr q27, [x26, #32] *)
  []; (* 1441: ldp x8, x24, [x1, #16] *)
  []; (* 1442: subs x6, x3, x21 *)
  []; (* 1443: ldr q0, [x1, #32] *)
  []; (* 1444: movi v23.2d, #0xffffffff *)
  []; (* 1445: csetm x10, cc *)
  []; (* 1446: umulh x19, x21, x23 *)
  []; (* 1447: rev64 v4.4s, v27.4s *)
  []; (* 1448: uzp2 v25.4s, v27.4s, v27.4s *)
  [`CF`; `NF`; `ZF`; `VF`; `X6`]; (* 1449: cneg x4, x6, cc *)
  []; (* 1450: subs x7, x23, x13 *)
  []; (* 1451: xtn v22.2s, v0.2d *)
  [`Q27`]; (* 1452: xtn v24.2s, v27.2d *)
  [`X7`]; (* 1453: cneg x20, x7, cc *)
  []; (* 1454: ldp x6, x14, [x26, #16] *)
  [`Q4`]; (* 1455: mul v27.4s, v4.4s, v0.4s *)
  [`Q6`]; (* 1456: uaddlp v20.2d, v6.4s *)
  [`CF`; `NF`; `ZF`; `VF`; `X10`]; (* 1457: cinv x5, x10, cc *)
  []; (* 1458: mul x16, x4, x20 *)
  [`Q0`]; (* 1459: uzp2 v6.4s, v0.4s, v0.4s *)
  []; (* 1460: umull v21.2d, v22.2s, v25.2s *)
  [`Q20`]; (* 1461: shl v0.2d, v20.2d, #32 *)
  [`Q17`; `Q3`]; (* 1462: umlal v0.2d, v3.2s, v17.2s *)
  []; (* 1463: mul x22, x8, x6 *)
  [`Q25`]; (* 1464: umull v1.2d, v6.2s, v25.2s *)
  []; (* 1465: subs x12, x3, x8 *)
  []; (* 1466: umull v20.2d, v22.2s, v24.2s *)
  [`X12`]; (* 1467: cneg x17, x12, cc *)
  []; (* 1468: umulh x9, x8, x6 *)
  []; (* 1469: mov x12, v0.d[1] *)
  [`X16`]; (* 1470: eor x11, x16, x5 *)
  [`Q0`]; (* 1471: mov x7, v0.d[0] *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1472: csetm x10, cc *)
  [`Q20`]; (* 1473: usra v21.2d, v20.2d, #32 *)
  [`X12`]; (* 1474: adds x15, x15, x12 *)
  [`X19`; `X22`]; (* 1475: adcs x12, x19, x22 *)
  [`X4`]; (* 1476: umulh x20, x4, x20 *)
  [`CF`; `NF`; `ZF`; `VF`; `X9`]; (* 1477: adc x19, x9, xzr *)
  []; (* 1478: usra v1.2d, v21.2d, #32 *)
  []; (* 1479: adds x22, x15, x7 *)
  [`Q21`; `Q23`]; (* 1480: and v26.16b, v21.16b, v23.16b *)
  []; (* 1481: adcs x16, x12, x15 *)
  [`Q27`]; (* 1482: uaddlp v25.2d, v27.4s *)
  []; (* 1483: adcs x9, x19, x12 *)
  [`Q6`]; (* 1484: umlal v26.2d, v6.2s, v24.2s *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1485: adc x4, x19, xzr *)
  []; (* 1486: adds x16, x16, x7 *)
  [`Q25`]; (* 1487: shl v27.2d, v25.2d, #32 *)
  [`X15`]; (* 1488: adcs x9, x9, x15 *)
  [`X12`]; (* 1489: adcs x4, x4, x12 *)
  [`X20`]; (* 1490: eor x12, x20, x5 *)
  [`CF`; `NF`; `ZF`; `VF`; `X19`]; (* 1491: adc x15, x19, xzr *)
  []; (* 1492: subs x20, x6, x13 *)
  []; (* 1493: cneg x20, x20, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1494: cinv x10, x10, cc *)
  []; (* 1495: cmn x5, #0x1 *)
  []; (* 1496: mul x19, x17, x20 *)
  [`X22`]; (* 1497: adcs x11, x22, x11 *)
  [`X16`]; (* 1498: adcs x12, x16, x12 *)
  []; (* 1499: adcs x9, x9, x5 *)
  [`X20`]; (* 1500: umulh x17, x17, x20 *)
  [`X4`]; (* 1501: adcs x22, x4, x5 *)
  [`CF`; `NF`; `ZF`; `VF`; `X15`]; (* 1502: adc x5, x15, x5 *)
  []; (* 1503: subs x16, x21, x8 *)
  [`X16`]; (* 1504: cneg x20, x16, cc *)
  []; (* 1505: eor x19, x19, x10 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1506: csetm x4, cc *)
  []; (* 1507: subs x16, x6, x23 *)
  []; (* 1508: cneg x16, x16, cc *)
  [`Q22`; `Q24`]; (* 1509: umlal v27.2d, v22.2s, v24.2s *)
  []; (* 1510: mul x15, x20, x16 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1511: cinv x4, x4, cc *)
  []; (* 1512: cmn x10, #0x1 *)
  [`Q26`]; (* 1513: usra v1.2d, v26.2d, #32 *)
  [`X12`]; (* 1514: adcs x19, x12, x19 *)
  []; (* 1515: eor x17, x17, x10 *)
  [`X17`]; (* 1516: adcs x9, x9, x17 *)
  []; (* 1517: adcs x22, x22, x10 *)
  []; (* 1518: lsl x12, x7, #32 *)
  [`X16`]; (* 1519: umulh x20, x20, x16 *)
  [`X15`]; (* 1520: eor x16, x15, x4 *)
  []; (* 1521: ldp x15, x17, [x26, #32] *)
  [`X12`; `X7`]; (* 1522: add x2, x12, x7 *)
  [`CF`; `NF`; `ZF`; `VF`; `X10`; `X5`]; (* 1523: adc x7, x5, x10 *)
  [`X1`]; (* 1524: ldp x5, x10, [x1, #32] *)
  []; (* 1525: lsr x1, x2, #32 *)
  [`X20`]; (* 1526: eor x12, x20, x4 *)
  []; (* 1527: subs x1, x1, x2 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1528: sbc x20, x2, xzr *)
  []; (* 1529: cmn x4, #0x1 *)
  [`X16`]; (* 1530: adcs x9, x9, x16 *)
  []; (* 1531: extr x1, x20, x1, #32 *)
  []; (* 1532: lsr x20, x20, #32 *)
  [`X12`]; (* 1533: adcs x22, x22, x12 *)
  [`CF`; `NF`; `ZF`; `VF`; `X4`; `X7`]; (* 1534: adc x16, x7, x4 *)
  [`X20`]; (* 1535: adds x12, x20, x2 *)
  []; (* 1536: umulh x7, x24, x14 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1537: adc x4, xzr, xzr *)
  [`X11`]; (* 1538: subs x1, x11, x1 *)
  [`X12`; `X19`]; (* 1539: sbcs x20, x19, x12 *)
  [`X4`; `X9`]; (* 1540: sbcs x12, x9, x4 *)
  []; (* 1541: lsl x9, x1, #32 *)
  [`X9`]; (* 1542: add x1, x9, x1 *)
  [`X22`]; (* 1543: sbcs x9, x22, xzr *)
  []; (* 1544: mul x22, x24, x14 *)
  []; (* 1545: sbcs x16, x16, xzr *)
  []; (* 1546: lsr x4, x1, #32 *)
  [`CF`; `NF`; `ZF`; `VF`; `X2`]; (* 1547: sbc x19, x2, xzr *)
  []; (* 1548: subs x4, x4, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1549: sbc x11, x1, xzr *)
  [`X4`]; (* 1550: extr x2, x11, x4, #32 *)
  [`X11`]; (* 1551: lsr x4, x11, #32 *)
  []; (* 1552: adds x4, x4, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1553: adc x11, xzr, xzr *)
  [`X20`]; (* 1554: subs x2, x20, x2 *)
  [`X12`]; (* 1555: sbcs x4, x12, x4 *)
  [`X11`; `X9`]; (* 1556: sbcs x20, x9, x11 *)
  []; (* 1557: lsl x12, x2, #32 *)
  [`X12`]; (* 1558: add x2, x12, x2 *)
  [`X16`]; (* 1559: sbcs x9, x16, xzr *)
  []; (* 1560: lsr x11, x2, #32 *)
  []; (* 1561: sbcs x19, x19, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1562: sbc x1, x1, xzr *)
  [`X11`]; (* 1563: subs x16, x11, x2 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1564: sbc x12, x2, xzr *)
  []; (* 1565: extr x16, x12, x16, #32 *)
  []; (* 1566: lsr x12, x12, #32 *)
  [`X12`]; (* 1567: adds x11, x12, x2 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1568: adc x12, xzr, xzr *)
  [`X4`]; (* 1569: subs x16, x4, x16 *)
  []; (* 1570: mov x4, v27.d[0] *)
  [`X20`]; (* 1571: sbcs x11, x20, x11 *)
  [`X12`; `X9`]; (* 1572: sbcs x20, x9, x12 *)
  [`X11`; `X16`]; (* 1573: stp x16, x11, [sp, #96] *)
  [`X19`]; (* 1574: sbcs x11, x19, xzr *)
  [`X1`]; (* 1575: sbcs x9, x1, xzr *)
  [`X11`; `X20`]; (* 1576: stp x20, x11, [sp, #112] *)
  []; (* 1577: mov x1, v1.d[0] *)
  [`CF`; `NF`; `ZF`; `VF`; `X2`]; (* 1578: sbc x20, x2, xzr *)
  []; (* 1579: subs x12, x24, x5 *)
  [`Q27`]; (* 1580: mov x11, v27.d[1] *)
  [`X12`]; (* 1581: cneg x16, x12, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1582: csetm x2, cc *)
  []; (* 1583: subs x19, x15, x14 *)
  [`Q1`]; (* 1584: mov x12, v1.d[1] *)
  []; (* 1585: cinv x2, x2, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1586: cneg x19, x19, cc *)
  [`X20`; `X9`]; (* 1587: stp x9, x20, [sp, #128] *)
  []; (* 1588: mul x9, x16, x19 *)
  [`X7`]; (* 1589: adds x4, x7, x4 *)
  [`X1`]; (* 1590: adcs x11, x1, x11 *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 1591: adc x1, x12, xzr *)
  []; (* 1592: adds x20, x4, x22 *)
  [`X16`]; (* 1593: umulh x19, x16, x19 *)
  []; (* 1594: adcs x7, x11, x4 *)
  [`X9`]; (* 1595: eor x16, x9, x2 *)
  []; (* 1596: adcs x9, x1, x11 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1597: adc x12, x1, xzr *)
  []; (* 1598: adds x7, x7, x22 *)
  [`X9`]; (* 1599: adcs x4, x9, x4 *)
  [`X11`; `X12`]; (* 1600: adcs x9, x12, x11 *)
  [`CF`; `NF`; `ZF`; `VF`; `X1`]; (* 1601: adc x12, x1, xzr *)
  []; (* 1602: cmn x2, #0x1 *)
  [`X19`]; (* 1603: eor x1, x19, x2 *)
  [`X16`; `X20`]; (* 1604: adcs x11, x20, x16 *)
  [`X1`; `X7`]; (* 1605: adcs x19, x7, x1 *)
  [`X4`]; (* 1606: adcs x1, x4, x2 *)
  [`X9`]; (* 1607: adcs x20, x9, x2 *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 1608: adc x2, x12, x2 *)
  []; (* 1609: subs x12, x24, x10 *)
  [`X12`]; (* 1610: cneg x16, x12, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1611: csetm x12, cc *)
  []; (* 1612: subs x9, x17, x14 *)
  []; (* 1613: cinv x12, x12, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1614: cneg x9, x9, cc *)
  [`X24`]; (* 1615: subs x3, x24, x3 *)
  []; (* 1616: sbcs x21, x5, x21 *)
  []; (* 1617: mul x24, x16, x9 *)
  [`X8`]; (* 1618: sbcs x4, x10, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1619: ngc x8, xzr *)
  [`X5`]; (* 1620: subs x10, x5, x10 *)
  [`X24`]; (* 1621: eor x5, x24, x12 *)
  []; (* 1622: csetm x7, cc *)
  [`CF`; `NF`; `ZF`; `VF`; `X10`]; (* 1623: cneg x24, x10, cc *)
  []; (* 1624: subs x10, x17, x15 *)
  []; (* 1625: cinv x7, x7, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1626: cneg x10, x10, cc *)
  [`X13`]; (* 1627: subs x14, x13, x14 *)
  [`X23`]; (* 1628: sbcs x15, x23, x15 *)
  [`X21`]; (* 1629: eor x13, x21, x8 *)
  []; (* 1630: mul x23, x24, x10 *)
  [`X6`]; (* 1631: sbcs x17, x6, x17 *)
  [`X3`]; (* 1632: eor x6, x3, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1633: ngc x21, xzr *)
  [`X16`]; (* 1634: umulh x9, x16, x9 *)
  []; (* 1635: cmn x8, #0x1 *)
  [`X23`]; (* 1636: eor x3, x23, x7 *)
  [`X6`]; (* 1637: adcs x23, x6, xzr *)
  []; (* 1638: adcs x13, x13, xzr *)
  [`X4`]; (* 1639: eor x16, x4, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1640: adc x16, x16, xzr *)
  [`X17`]; (* 1641: eor x4, x17, x21 *)
  [`X10`; `X24`]; (* 1642: umulh x17, x24, x10 *)
  []; (* 1643: cmn x21, #0x1 *)
  [`X14`]; (* 1644: eor x24, x14, x21 *)
  [`X15`]; (* 1645: eor x6, x15, x21 *)
  [`X24`]; (* 1646: adcs x15, x24, xzr *)
  [`X6`]; (* 1647: adcs x14, x6, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X4`]; (* 1648: adc x6, x4, xzr *)
  []; (* 1649: cmn x12, #0x1 *)
  [`X9`]; (* 1650: eor x4, x9, x12 *)
  [`X5`]; (* 1651: adcs x19, x19, x5 *)
  []; (* 1652: umulh x5, x23, x15 *)
  [`X4`]; (* 1653: adcs x1, x1, x4 *)
  [`X20`]; (* 1654: adcs x10, x20, x12 *)
  [`X17`]; (* 1655: eor x4, x17, x7 *)
  []; (* 1656: ldp x20, x9, [sp, #96] *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 1657: adc x2, x2, x12 *)
  []; (* 1658: cmn x7, #0x1 *)
  [`X1`; `X3`]; (* 1659: adcs x12, x1, x3 *)
  []; (* 1660: ldp x17, x24, [sp, #112] *)
  []; (* 1661: mul x1, x16, x6 *)
  [`X10`; `X4`]; (* 1662: adcs x3, x10, x4 *)
  [`CF`; `NF`; `ZF`; `VF`; `X7`]; (* 1663: adc x2, x2, x7 *)
  []; (* 1664: ldp x7, x4, [sp, #128] *)
  [`X22`]; (* 1665: adds x20, x22, x20 *)
  []; (* 1666: mul x10, x13, x14 *)
  [`X9`]; (* 1667: adcs x11, x11, x9 *)
  [`X21`; `X8`]; (* 1668: eor x9, x8, x21 *)
  [`X17`; `X19`]; (* 1669: adcs x21, x19, x17 *)
  [`X11`; `X20`]; (* 1670: stp x20, x11, [sp, #96] *)
  [`X24`]; (* 1671: adcs x12, x12, x24 *)
  []; (* 1672: mul x8, x23, x15 *)
  [`X7`]; (* 1673: adcs x3, x3, x7 *)
  [`X12`; `X21`]; (* 1674: stp x21, x12, [sp, #112] *)
  [`X2`; `X4`]; (* 1675: adcs x12, x2, x4 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1676: adc x19, xzr, xzr *)
  []; (* 1677: subs x21, x23, x16 *)
  []; (* 1678: umulh x2, x16, x6 *)
  [`X12`; `X3`]; (* 1679: stp x3, x12, [sp, #128] *)
  [`X21`]; (* 1680: cneg x3, x21, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1681: csetm x24, cc *)
  []; (* 1682: umulh x11, x13, x14 *)
  [`X16`]; (* 1683: subs x21, x13, x16 *)
  []; (* 1684: eor x7, x8, x9 *)
  [`X21`]; (* 1685: cneg x17, x21, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1686: csetm x16, cc *)
  []; (* 1687: subs x21, x6, x15 *)
  [`X21`]; (* 1688: cneg x22, x21, cc *)
  [`CF`; `NF`; `ZF`; `VF`; `X24`]; (* 1689: cinv x21, x24, cc *)
  [`X13`; `X23`]; (* 1690: subs x20, x23, x13 *)
  []; (* 1691: umulh x12, x3, x22 *)
  [`X20`]; (* 1692: cneg x23, x20, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1693: csetm x24, cc *)
  [`X15`]; (* 1694: subs x20, x14, x15 *)
  []; (* 1695: cinv x24, x24, cc *)
  [`X3`]; (* 1696: mul x22, x3, x22 *)
  [`CF`; `NF`; `ZF`; `VF`; `X20`]; (* 1697: cneg x3, x20, cc *)
  [`X14`; `X6`]; (* 1698: subs x13, x6, x14 *)
  [`X13`]; (* 1699: cneg x20, x13, cc *)
  [`CF`; `NF`; `ZF`; `VF`; `X16`]; (* 1700: cinv x15, x16, cc *)
  [`X10`; `X5`]; (* 1701: adds x13, x5, x10 *)
  []; (* 1702: mul x4, x23, x3 *)
  [`X1`]; (* 1703: adcs x11, x11, x1 *)
  [`CF`; `NF`; `ZF`; `VF`; `X2`]; (* 1704: adc x14, x2, xzr *)
  []; (* 1705: adds x5, x13, x8 *)
  []; (* 1706: adcs x16, x11, x13 *)
  [`X3`]; (* 1707: umulh x23, x23, x3 *)
  []; (* 1708: adcs x3, x14, x11 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1709: adc x1, x14, xzr *)
  [`X16`; `X8`]; (* 1710: adds x10, x16, x8 *)
  [`X13`; `X3`]; (* 1711: adcs x6, x3, x13 *)
  [`X1`; `X11`]; (* 1712: adcs x8, x1, x11 *)
  []; (* 1713: umulh x13, x17, x20 *)
  [`X4`]; (* 1714: eor x1, x4, x24 *)
  [`CF`; `NF`; `ZF`; `VF`; `X14`]; (* 1715: adc x4, x14, xzr *)
  []; (* 1716: cmn x24, #0x1 *)
  [`X5`]; (* 1717: adcs x1, x5, x1 *)
  [`X23`]; (* 1718: eor x16, x23, x24 *)
  [`X1`]; (* 1719: eor x11, x1, x9 *)
  [`X10`; `X16`]; (* 1720: adcs x23, x10, x16 *)
  [`X22`]; (* 1721: eor x2, x22, x21 *)
  [`X6`]; (* 1722: adcs x3, x6, x24 *)
  [`X17`; `X20`]; (* 1723: mul x14, x17, x20 *)
  [`X13`]; (* 1724: eor x17, x13, x15 *)
  [`X8`]; (* 1725: adcs x13, x8, x24 *)
  [`CF`; `NF`; `ZF`; `VF`; `X24`; `X4`]; (* 1726: adc x8, x4, x24 *)
  []; (* 1727: cmn x21, #0x1 *)
  [`X2`; `X23`]; (* 1728: adcs x6, x23, x2 *)
  []; (* 1729: mov x16, #0xfffffffffffffffe *)
  [`X12`]; (* 1730: eor x20, x12, x21 *)
  [`X3`]; (* 1731: adcs x20, x3, x20 *)
  [`X14`]; (* 1732: eor x23, x14, x15 *)
  [`X13`]; (* 1733: adcs x2, x13, x21 *)
  [`CF`; `NF`; `ZF`; `VF`; `X21`]; (* 1734: adc x8, x8, x21 *)
  []; (* 1735: cmn x15, #0x1 *)
  []; (* 1736: ldp x5, x4, [sp, #96] *)
  []; (* 1737: ldp x21, x12, [sp, #112] *)
  [`X20`; `X23`]; (* 1738: adcs x22, x20, x23 *)
  [`X22`]; (* 1739: eor x23, x22, x9 *)
  [`X2`]; (* 1740: adcs x17, x2, x17 *)
  [`CF`; `NF`; `ZF`; `VF`; `X15`; `X8`]; (* 1741: adc x22, x8, x15 *)
  []; (* 1742: cmn x9, #0x1 *)
  [`X7`]; (* 1743: adcs x15, x7, x5 *)
  []; (* 1744: ldp x10, x14, [sp, #128] *)
  [`X6`]; (* 1745: eor x1, x6, x9 *)
  []; (* 1746: lsl x2, x15, #32 *)
  [`X11`]; (* 1747: adcs x8, x11, x4 *)
  [`X1`]; (* 1748: adcs x13, x1, x21 *)
  [`X22`]; (* 1749: eor x1, x22, x9 *)
  [`X23`]; (* 1750: adcs x24, x23, x12 *)
  [`X17`]; (* 1751: eor x11, x17, x9 *)
  [`X11`]; (* 1752: adcs x23, x11, x10 *)
  [`X1`]; (* 1753: adcs x7, x1, x14 *)
  []; (* 1754: adcs x17, x9, x19 *)
  []; (* 1755: adcs x20, x9, xzr *)
  [`X15`; `X2`]; (* 1756: add x1, x2, x15 *)
  []; (* 1757: lsr x3, x1, #32 *)
  []; (* 1758: adcs x11, x9, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1759: adc x9, x9, xzr *)
  []; (* 1760: subs x3, x3, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1761: sbc x6, x1, xzr *)
  [`X5`]; (* 1762: adds x24, x24, x5 *)
  [`X23`]; (* 1763: adcs x4, x23, x4 *)
  []; (* 1764: extr x3, x6, x3, #32 *)
  []; (* 1765: lsr x6, x6, #32 *)
  [`X7`]; (* 1766: adcs x21, x7, x21 *)
  [`X12`; `X17`]; (* 1767: adcs x15, x17, x12 *)
  [`X10`; `X20`]; (* 1768: adcs x7, x20, x10 *)
  [`X11`; `X14`]; (* 1769: adcs x20, x11, x14 *)
  []; (* 1770: mov x14, #0xffffffff *)
  [`CF`; `NF`; `ZF`; `VF`; `X19`; `X9`]; (* 1771: adc x22, x9, x19 *)
  [`X6`]; (* 1772: adds x12, x6, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1773: adc x10, xzr, xzr *)
  [`X8`]; (* 1774: subs x3, x8, x3 *)
  [`X13`]; (* 1775: sbcs x12, x13, x12 *)
  []; (* 1776: lsl x9, x3, #32 *)
  [`X9`]; (* 1777: add x3, x9, x3 *)
  [`X24`]; (* 1778: sbcs x10, x24, x10 *)
  [`X4`]; (* 1779: sbcs x24, x4, xzr *)
  []; (* 1780: lsr x9, x3, #32 *)
  []; (* 1781: sbcs x21, x21, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1782: sbc x1, x1, xzr *)
  []; (* 1783: subs x9, x9, x3 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1784: sbc x13, x3, xzr *)
  []; (* 1785: extr x9, x13, x9, #32 *)
  []; (* 1786: lsr x13, x13, #32 *)
  []; (* 1787: adds x13, x13, x3 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1788: adc x6, xzr, xzr *)
  [`X9`]; (* 1789: subs x12, x12, x9 *)
  [`X10`; `X13`]; (* 1790: sbcs x17, x10, x13 *)
  []; (* 1791: lsl x2, x12, #32 *)
  [`X24`; `X6`]; (* 1792: sbcs x10, x24, x6 *)
  [`X12`; `X2`]; (* 1793: add x9, x2, x12 *)
  [`X21`]; (* 1794: sbcs x6, x21, xzr *)
  []; (* 1795: lsr x5, x9, #32 *)
  [`X1`]; (* 1796: sbcs x21, x1, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X3`]; (* 1797: sbc x13, x3, xzr *)
  [`X5`]; (* 1798: subs x8, x5, x9 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1799: sbc x19, x9, xzr *)
  []; (* 1800: lsr x12, x19, #32 *)
  [`X19`; `X8`]; (* 1801: extr x3, x19, x8, #32 *)
  [`X12`]; (* 1802: adds x8, x12, x9 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1803: adc x1, xzr, xzr *)
  [`X17`; `X3`]; (* 1804: subs x2, x17, x3 *)
  [`X10`; `X8`]; (* 1805: sbcs x12, x10, x8 *)
  [`X1`; `X6`]; (* 1806: sbcs x5, x6, x1 *)
  [`X21`]; (* 1807: sbcs x3, x21, xzr *)
  [`X13`]; (* 1808: sbcs x19, x13, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X9`]; (* 1809: sbc x24, x9, xzr *)
  [`X15`; `X3`]; (* 1810: adds x23, x15, x3 *)
  [`X19`; `X7`]; (* 1811: adcs x8, x7, x19 *)
  [`X20`; `X24`]; (* 1812: adcs x11, x20, x24 *)
  [`CF`; `NF`; `ZF`; `VF`; `X22`]; (* 1813: adc x9, x22, xzr *)
  [`X9`]; (* 1814: add x24, x9, #0x1 *)
  []; (* 1815: lsl x7, x24, #32 *)
  []; (* 1816: subs x21, x24, x7 *)
  [`CF`; `NF`; `ZF`; `VF`; `X7`]; (* 1817: sbc x10, x7, xzr *)
  [`X2`; `X21`]; (* 1818: adds x6, x2, x21 *)
  [`X10`; `X12`]; (* 1819: adcs x7, x12, x10 *)
  [`X5`]; (* 1820: adcs x24, x5, x24 *)
  [`X23`]; (* 1821: adcs x13, x23, xzr *)
  []; (* 1822: adcs x8, x8, xzr *)
  [`X11`]; (* 1823: adcs x15, x11, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1824: csetm x23, cc *)
  [`X16`]; (* 1825: and x11, x16, x23 *)
  [`X14`]; (* 1826: and x20, x14, x23 *)
  [`X6`]; (* 1827: adds x22, x6, x20 *)
  [`X20`]; (* 1828: eor x3, x20, x23 *)
  [`X3`; `X7`]; (* 1829: adcs x5, x7, x3 *)
  [`X11`; `X24`]; (* 1830: adcs x14, x24, x11 *)
  [`X22`; `X5`]; (* 1831: stp x22, x5, [sp, #96] *)
  [`X13`]; (* 1832: adcs x5, x13, x23 *)
  [`X8`]; (* 1833: adcs x21, x8, x23 *)
  [`X14`; `X5`]; (* 1834: stp x14, x5, [sp, #112] *)
  [`CF`; `NF`; `ZF`; `VF`; `X15`; `X23`]; (* 1835: adc x12, x15, x23 *)
  [`X12`; `X21`]; (* 1836: stp x21, x12, [sp, #128] *)
  []; (* 1837: ldr q3, [sp, #240] *)
  []; (* 1838: ldr q25, [x25, #0] *)
  []; (* 1839: ldp x13, x23, [x25, #0] *)
  []; (* 1840: ldp x3, x21, [sp, #240] *)
  []; (* 1841: rev64 v23.4s, v25.4s *)
  [`Q25`]; (* 1842: uzp1 v17.4s, v25.4s, v3.4s *)
  []; (* 1843: umulh x15, x3, x13 *)
  [`Q23`]; (* 1844: mul v6.4s, v23.4s, v3.4s *)
  []; (* 1845: uzp1 v3.4s, v3.4s, v3.4s *)
  []; (* 1846: ldr q27, [x25, #32] *)
  []; (* 1847: ldp x8, x24, [sp, #256] *)
  []; (* 1848: subs x6, x3, x21 *)
  []; (* 1849: ldr q0, [sp, #272] *)
  []; (* 1850: movi v23.2d, #0xffffffff *)
  []; (* 1851: csetm x10, cc *)
  []; (* 1852: umulh x19, x21, x23 *)
  []; (* 1853: rev64 v4.4s, v27.4s *)
  []; (* 1854: uzp2 v25.4s, v27.4s, v27.4s *)
  [`CF`; `NF`; `ZF`; `VF`; `X6`]; (* 1855: cneg x4, x6, cc *)
  []; (* 1856: subs x7, x23, x13 *)
  []; (* 1857: xtn v22.2s, v0.2d *)
  [`Q27`]; (* 1858: xtn v24.2s, v27.2d *)
  [`X7`]; (* 1859: cneg x20, x7, cc *)
  []; (* 1860: ldp x6, x14, [x25, #16] *)
  [`Q4`]; (* 1861: mul v27.4s, v4.4s, v0.4s *)
  [`Q6`]; (* 1862: uaddlp v20.2d, v6.4s *)
  [`CF`; `NF`; `ZF`; `VF`; `X10`]; (* 1863: cinv x5, x10, cc *)
  []; (* 1864: mul x16, x4, x20 *)
  [`Q0`]; (* 1865: uzp2 v6.4s, v0.4s, v0.4s *)
  []; (* 1866: umull v21.2d, v22.2s, v25.2s *)
  [`Q20`]; (* 1867: shl v0.2d, v20.2d, #32 *)
  [`Q17`; `Q3`]; (* 1868: umlal v0.2d, v3.2s, v17.2s *)
  []; (* 1869: mul x22, x8, x6 *)
  [`Q25`]; (* 1870: umull v1.2d, v6.2s, v25.2s *)
  []; (* 1871: subs x12, x3, x8 *)
  []; (* 1872: umull v20.2d, v22.2s, v24.2s *)
  [`X12`]; (* 1873: cneg x17, x12, cc *)
  []; (* 1874: umulh x9, x8, x6 *)
  []; (* 1875: mov x12, v0.d[1] *)
  [`X16`]; (* 1876: eor x11, x16, x5 *)
  [`Q0`]; (* 1877: mov x7, v0.d[0] *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1878: csetm x10, cc *)
  [`Q20`]; (* 1879: usra v21.2d, v20.2d, #32 *)
  [`X12`]; (* 1880: adds x15, x15, x12 *)
  [`X19`; `X22`]; (* 1881: adcs x12, x19, x22 *)
  [`X4`]; (* 1882: umulh x20, x4, x20 *)
  [`CF`; `NF`; `ZF`; `VF`; `X9`]; (* 1883: adc x19, x9, xzr *)
  []; (* 1884: usra v1.2d, v21.2d, #32 *)
  []; (* 1885: adds x22, x15, x7 *)
  [`Q21`; `Q23`]; (* 1886: and v26.16b, v21.16b, v23.16b *)
  []; (* 1887: adcs x16, x12, x15 *)
  [`Q27`]; (* 1888: uaddlp v25.2d, v27.4s *)
  []; (* 1889: adcs x9, x19, x12 *)
  [`Q6`]; (* 1890: umlal v26.2d, v6.2s, v24.2s *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1891: adc x4, x19, xzr *)
  []; (* 1892: adds x16, x16, x7 *)
  [`Q25`]; (* 1893: shl v27.2d, v25.2d, #32 *)
  [`X15`]; (* 1894: adcs x9, x9, x15 *)
  [`X12`]; (* 1895: adcs x4, x4, x12 *)
  [`X20`]; (* 1896: eor x12, x20, x5 *)
  [`CF`; `NF`; `ZF`; `VF`; `X19`]; (* 1897: adc x15, x19, xzr *)
  []; (* 1898: subs x20, x6, x13 *)
  []; (* 1899: cneg x20, x20, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1900: cinv x10, x10, cc *)
  []; (* 1901: cmn x5, #0x1 *)
  []; (* 1902: mul x19, x17, x20 *)
  [`X22`]; (* 1903: adcs x11, x22, x11 *)
  [`X16`]; (* 1904: adcs x12, x16, x12 *)
  []; (* 1905: adcs x9, x9, x5 *)
  [`X20`]; (* 1906: umulh x17, x17, x20 *)
  [`X4`]; (* 1907: adcs x22, x4, x5 *)
  [`CF`; `NF`; `ZF`; `VF`; `X15`]; (* 1908: adc x5, x15, x5 *)
  []; (* 1909: subs x16, x21, x8 *)
  [`X16`]; (* 1910: cneg x20, x16, cc *)
  []; (* 1911: eor x19, x19, x10 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1912: csetm x4, cc *)
  []; (* 1913: subs x16, x6, x23 *)
  []; (* 1914: cneg x16, x16, cc *)
  [`Q22`; `Q24`]; (* 1915: umlal v27.2d, v22.2s, v24.2s *)
  []; (* 1916: mul x15, x20, x16 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1917: cinv x4, x4, cc *)
  []; (* 1918: cmn x10, #0x1 *)
  [`Q26`]; (* 1919: usra v1.2d, v26.2d, #32 *)
  [`X12`]; (* 1920: adcs x19, x12, x19 *)
  []; (* 1921: eor x17, x17, x10 *)
  [`X17`]; (* 1922: adcs x9, x9, x17 *)
  []; (* 1923: adcs x22, x22, x10 *)
  []; (* 1924: lsl x12, x7, #32 *)
  [`X16`]; (* 1925: umulh x20, x20, x16 *)
  [`X15`]; (* 1926: eor x16, x15, x4 *)
  []; (* 1927: ldp x15, x17, [x25, #32] *)
  [`X12`; `X7`]; (* 1928: add x2, x12, x7 *)
  [`CF`; `NF`; `ZF`; `VF`; `X10`; `X5`]; (* 1929: adc x7, x5, x10 *)
  []; (* 1930: ldp x5, x10, [sp, #272] *)
  []; (* 1931: lsr x1, x2, #32 *)
  [`X20`]; (* 1932: eor x12, x20, x4 *)
  []; (* 1933: subs x1, x1, x2 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1934: sbc x20, x2, xzr *)
  []; (* 1935: cmn x4, #0x1 *)
  [`X16`]; (* 1936: adcs x9, x9, x16 *)
  []; (* 1937: extr x1, x20, x1, #32 *)
  []; (* 1938: lsr x20, x20, #32 *)
  [`X12`]; (* 1939: adcs x22, x22, x12 *)
  [`CF`; `NF`; `ZF`; `VF`; `X4`; `X7`]; (* 1940: adc x16, x7, x4 *)
  [`X20`]; (* 1941: adds x12, x20, x2 *)
  []; (* 1942: umulh x7, x24, x14 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1943: adc x4, xzr, xzr *)
  [`X11`]; (* 1944: subs x1, x11, x1 *)
  [`X12`; `X19`]; (* 1945: sbcs x20, x19, x12 *)
  [`X4`; `X9`]; (* 1946: sbcs x12, x9, x4 *)
  []; (* 1947: lsl x9, x1, #32 *)
  [`X9`]; (* 1948: add x1, x9, x1 *)
  [`X22`]; (* 1949: sbcs x9, x22, xzr *)
  []; (* 1950: mul x22, x24, x14 *)
  []; (* 1951: sbcs x16, x16, xzr *)
  []; (* 1952: lsr x4, x1, #32 *)
  [`CF`; `NF`; `ZF`; `VF`; `X2`]; (* 1953: sbc x19, x2, xzr *)
  []; (* 1954: subs x4, x4, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1955: sbc x11, x1, xzr *)
  [`X4`]; (* 1956: extr x2, x11, x4, #32 *)
  [`X11`]; (* 1957: lsr x4, x11, #32 *)
  []; (* 1958: adds x4, x4, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1959: adc x11, xzr, xzr *)
  [`X20`]; (* 1960: subs x2, x20, x2 *)
  [`X12`]; (* 1961: sbcs x4, x12, x4 *)
  [`X11`; `X9`]; (* 1962: sbcs x20, x9, x11 *)
  []; (* 1963: lsl x12, x2, #32 *)
  [`X12`]; (* 1964: add x2, x12, x2 *)
  [`X16`]; (* 1965: sbcs x9, x16, xzr *)
  []; (* 1966: lsr x11, x2, #32 *)
  []; (* 1967: sbcs x19, x19, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1968: sbc x1, x1, xzr *)
  [`X11`]; (* 1969: subs x16, x11, x2 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1970: sbc x12, x2, xzr *)
  []; (* 1971: extr x16, x12, x16, #32 *)
  []; (* 1972: lsr x12, x12, #32 *)
  [`X12`]; (* 1973: adds x11, x12, x2 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1974: adc x12, xzr, xzr *)
  [`X4`]; (* 1975: subs x16, x4, x16 *)
  []; (* 1976: mov x4, v27.d[0] *)
  [`X20`]; (* 1977: sbcs x11, x20, x11 *)
  [`X12`; `X9`]; (* 1978: sbcs x20, x9, x12 *)
  [`X11`; `X16`]; (* 1979: stp x16, x11, [sp, #192] *)
  [`X19`]; (* 1980: sbcs x11, x19, xzr *)
  [`X1`]; (* 1981: sbcs x9, x1, xzr *)
  [`X11`; `X20`]; (* 1982: stp x20, x11, [sp, #208] *)
  []; (* 1983: mov x1, v1.d[0] *)
  [`CF`; `NF`; `ZF`; `VF`; `X2`]; (* 1984: sbc x20, x2, xzr *)
  []; (* 1985: subs x12, x24, x5 *)
  [`Q27`]; (* 1986: mov x11, v27.d[1] *)
  [`X12`]; (* 1987: cneg x16, x12, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1988: csetm x2, cc *)
  []; (* 1989: subs x19, x15, x14 *)
  [`Q1`]; (* 1990: mov x12, v1.d[1] *)
  []; (* 1991: cinv x2, x2, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 1992: cneg x19, x19, cc *)
  [`X20`; `X9`]; (* 1993: stp x9, x20, [sp, #224] *)
  []; (* 1994: mul x9, x16, x19 *)
  [`X7`]; (* 1995: adds x4, x7, x4 *)
  [`X1`]; (* 1996: adcs x11, x1, x11 *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 1997: adc x1, x12, xzr *)
  []; (* 1998: adds x20, x4, x22 *)
  [`X16`]; (* 1999: umulh x19, x16, x19 *)
  []; (* 2000: adcs x7, x11, x4 *)
  [`X9`]; (* 2001: eor x16, x9, x2 *)
  []; (* 2002: adcs x9, x1, x11 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2003: adc x12, x1, xzr *)
  []; (* 2004: adds x7, x7, x22 *)
  [`X9`]; (* 2005: adcs x4, x9, x4 *)
  [`X11`; `X12`]; (* 2006: adcs x9, x12, x11 *)
  [`CF`; `NF`; `ZF`; `VF`; `X1`]; (* 2007: adc x12, x1, xzr *)
  []; (* 2008: cmn x2, #0x1 *)
  [`X19`]; (* 2009: eor x1, x19, x2 *)
  [`X16`; `X20`]; (* 2010: adcs x11, x20, x16 *)
  [`X1`; `X7`]; (* 2011: adcs x19, x7, x1 *)
  [`X4`]; (* 2012: adcs x1, x4, x2 *)
  [`X9`]; (* 2013: adcs x20, x9, x2 *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 2014: adc x2, x12, x2 *)
  []; (* 2015: subs x12, x24, x10 *)
  [`X12`]; (* 2016: cneg x16, x12, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2017: csetm x12, cc *)
  []; (* 2018: subs x9, x17, x14 *)
  []; (* 2019: cinv x12, x12, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2020: cneg x9, x9, cc *)
  [`X24`]; (* 2021: subs x3, x24, x3 *)
  []; (* 2022: sbcs x21, x5, x21 *)
  []; (* 2023: mul x24, x16, x9 *)
  [`X8`]; (* 2024: sbcs x4, x10, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2025: ngc x8, xzr *)
  [`X5`]; (* 2026: subs x10, x5, x10 *)
  [`X24`]; (* 2027: eor x5, x24, x12 *)
  []; (* 2028: csetm x7, cc *)
  [`CF`; `NF`; `ZF`; `VF`; `X10`]; (* 2029: cneg x24, x10, cc *)
  []; (* 2030: subs x10, x17, x15 *)
  []; (* 2031: cinv x7, x7, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2032: cneg x10, x10, cc *)
  [`X13`]; (* 2033: subs x14, x13, x14 *)
  [`X23`]; (* 2034: sbcs x15, x23, x15 *)
  [`X21`]; (* 2035: eor x13, x21, x8 *)
  []; (* 2036: mul x23, x24, x10 *)
  [`X6`]; (* 2037: sbcs x17, x6, x17 *)
  [`X3`]; (* 2038: eor x6, x3, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2039: ngc x21, xzr *)
  [`X16`]; (* 2040: umulh x9, x16, x9 *)
  []; (* 2041: cmn x8, #0x1 *)
  [`X23`]; (* 2042: eor x3, x23, x7 *)
  [`X6`]; (* 2043: adcs x23, x6, xzr *)
  []; (* 2044: adcs x13, x13, xzr *)
  [`X4`]; (* 2045: eor x16, x4, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2046: adc x16, x16, xzr *)
  [`X17`]; (* 2047: eor x4, x17, x21 *)
  [`X10`; `X24`]; (* 2048: umulh x17, x24, x10 *)
  []; (* 2049: cmn x21, #0x1 *)
  [`X14`]; (* 2050: eor x24, x14, x21 *)
  [`X15`]; (* 2051: eor x6, x15, x21 *)
  [`X24`]; (* 2052: adcs x15, x24, xzr *)
  [`X6`]; (* 2053: adcs x14, x6, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X4`]; (* 2054: adc x6, x4, xzr *)
  []; (* 2055: cmn x12, #0x1 *)
  [`X9`]; (* 2056: eor x4, x9, x12 *)
  [`X5`]; (* 2057: adcs x19, x19, x5 *)
  []; (* 2058: umulh x5, x23, x15 *)
  [`X4`]; (* 2059: adcs x1, x1, x4 *)
  [`X20`]; (* 2060: adcs x10, x20, x12 *)
  [`X17`]; (* 2061: eor x4, x17, x7 *)
  []; (* 2062: ldp x20, x9, [sp, #192] *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 2063: adc x2, x2, x12 *)
  []; (* 2064: cmn x7, #0x1 *)
  [`X1`; `X3`]; (* 2065: adcs x12, x1, x3 *)
  []; (* 2066: ldp x17, x24, [sp, #208] *)
  []; (* 2067: mul x1, x16, x6 *)
  [`X10`; `X4`]; (* 2068: adcs x3, x10, x4 *)
  [`CF`; `NF`; `ZF`; `VF`; `X7`]; (* 2069: adc x2, x2, x7 *)
  []; (* 2070: ldp x7, x4, [sp, #224] *)
  [`X22`]; (* 2071: adds x20, x22, x20 *)
  []; (* 2072: mul x10, x13, x14 *)
  [`X9`]; (* 2073: adcs x11, x11, x9 *)
  [`X21`; `X8`]; (* 2074: eor x9, x8, x21 *)
  [`X17`; `X19`]; (* 2075: adcs x21, x19, x17 *)
  [`X11`; `X20`]; (* 2076: stp x20, x11, [sp, #192] *)
  [`X24`]; (* 2077: adcs x12, x12, x24 *)
  []; (* 2078: mul x8, x23, x15 *)
  [`X7`]; (* 2079: adcs x3, x3, x7 *)
  [`X12`; `X21`]; (* 2080: stp x21, x12, [sp, #208] *)
  [`X2`; `X4`]; (* 2081: adcs x12, x2, x4 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2082: adc x19, xzr, xzr *)
  []; (* 2083: subs x21, x23, x16 *)
  []; (* 2084: umulh x2, x16, x6 *)
  [`X12`; `X3`]; (* 2085: stp x3, x12, [sp, #224] *)
  [`X21`]; (* 2086: cneg x3, x21, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2087: csetm x24, cc *)
  []; (* 2088: umulh x11, x13, x14 *)
  [`X16`]; (* 2089: subs x21, x13, x16 *)
  []; (* 2090: eor x7, x8, x9 *)
  [`X21`]; (* 2091: cneg x17, x21, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2092: csetm x16, cc *)
  []; (* 2093: subs x21, x6, x15 *)
  [`X21`]; (* 2094: cneg x22, x21, cc *)
  [`CF`; `NF`; `ZF`; `VF`; `X24`]; (* 2095: cinv x21, x24, cc *)
  [`X13`; `X23`]; (* 2096: subs x20, x23, x13 *)
  []; (* 2097: umulh x12, x3, x22 *)
  [`X20`]; (* 2098: cneg x23, x20, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2099: csetm x24, cc *)
  [`X15`]; (* 2100: subs x20, x14, x15 *)
  []; (* 2101: cinv x24, x24, cc *)
  [`X3`]; (* 2102: mul x22, x3, x22 *)
  [`CF`; `NF`; `ZF`; `VF`; `X20`]; (* 2103: cneg x3, x20, cc *)
  [`X14`; `X6`]; (* 2104: subs x13, x6, x14 *)
  [`X13`]; (* 2105: cneg x20, x13, cc *)
  [`CF`; `NF`; `ZF`; `VF`; `X16`]; (* 2106: cinv x15, x16, cc *)
  [`X10`; `X5`]; (* 2107: adds x13, x5, x10 *)
  []; (* 2108: mul x4, x23, x3 *)
  [`X1`]; (* 2109: adcs x11, x11, x1 *)
  [`CF`; `NF`; `ZF`; `VF`; `X2`]; (* 2110: adc x14, x2, xzr *)
  []; (* 2111: adds x5, x13, x8 *)
  []; (* 2112: adcs x16, x11, x13 *)
  [`X3`]; (* 2113: umulh x23, x23, x3 *)
  []; (* 2114: adcs x3, x14, x11 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2115: adc x1, x14, xzr *)
  [`X16`; `X8`]; (* 2116: adds x10, x16, x8 *)
  [`X13`; `X3`]; (* 2117: adcs x6, x3, x13 *)
  [`X1`; `X11`]; (* 2118: adcs x8, x1, x11 *)
  []; (* 2119: umulh x13, x17, x20 *)
  [`X4`]; (* 2120: eor x1, x4, x24 *)
  [`CF`; `NF`; `ZF`; `VF`; `X14`]; (* 2121: adc x4, x14, xzr *)
  []; (* 2122: cmn x24, #0x1 *)
  [`X5`]; (* 2123: adcs x1, x5, x1 *)
  [`X23`]; (* 2124: eor x16, x23, x24 *)
  [`X1`]; (* 2125: eor x11, x1, x9 *)
  [`X10`; `X16`]; (* 2126: adcs x23, x10, x16 *)
  [`X22`]; (* 2127: eor x2, x22, x21 *)
  [`X6`]; (* 2128: adcs x3, x6, x24 *)
  [`X17`; `X20`]; (* 2129: mul x14, x17, x20 *)
  [`X13`]; (* 2130: eor x17, x13, x15 *)
  [`X8`]; (* 2131: adcs x13, x8, x24 *)
  [`CF`; `NF`; `ZF`; `VF`; `X24`; `X4`]; (* 2132: adc x8, x4, x24 *)
  []; (* 2133: cmn x21, #0x1 *)
  [`X2`; `X23`]; (* 2134: adcs x6, x23, x2 *)
  []; (* 2135: mov x16, #0xfffffffffffffffe *)
  [`X12`]; (* 2136: eor x20, x12, x21 *)
  [`X3`]; (* 2137: adcs x20, x3, x20 *)
  [`X14`]; (* 2138: eor x23, x14, x15 *)
  [`X13`]; (* 2139: adcs x2, x13, x21 *)
  [`CF`; `NF`; `ZF`; `VF`; `X21`]; (* 2140: adc x8, x8, x21 *)
  []; (* 2141: cmn x15, #0x1 *)
  []; (* 2142: ldp x5, x4, [sp, #192] *)
  []; (* 2143: ldp x21, x12, [sp, #208] *)
  [`X20`; `X23`]; (* 2144: adcs x22, x20, x23 *)
  [`X22`]; (* 2145: eor x23, x22, x9 *)
  [`X2`]; (* 2146: adcs x17, x2, x17 *)
  [`CF`; `NF`; `ZF`; `VF`; `X15`; `X8`]; (* 2147: adc x22, x8, x15 *)
  []; (* 2148: cmn x9, #0x1 *)
  [`X7`]; (* 2149: adcs x15, x7, x5 *)
  []; (* 2150: ldp x10, x14, [sp, #224] *)
  [`X6`]; (* 2151: eor x1, x6, x9 *)
  []; (* 2152: lsl x2, x15, #32 *)
  [`X11`]; (* 2153: adcs x8, x11, x4 *)
  [`X1`]; (* 2154: adcs x13, x1, x21 *)
  [`X22`]; (* 2155: eor x1, x22, x9 *)
  [`X23`]; (* 2156: adcs x24, x23, x12 *)
  [`X17`]; (* 2157: eor x11, x17, x9 *)
  [`X11`]; (* 2158: adcs x23, x11, x10 *)
  [`X1`]; (* 2159: adcs x7, x1, x14 *)
  []; (* 2160: adcs x17, x9, x19 *)
  []; (* 2161: adcs x20, x9, xzr *)
  [`X15`; `X2`]; (* 2162: add x1, x2, x15 *)
  []; (* 2163: lsr x3, x1, #32 *)
  []; (* 2164: adcs x11, x9, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2165: adc x9, x9, xzr *)
  []; (* 2166: subs x3, x3, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2167: sbc x6, x1, xzr *)
  [`X5`]; (* 2168: adds x24, x24, x5 *)
  [`X23`]; (* 2169: adcs x4, x23, x4 *)
  []; (* 2170: extr x3, x6, x3, #32 *)
  []; (* 2171: lsr x6, x6, #32 *)
  [`X7`]; (* 2172: adcs x21, x7, x21 *)
  [`X12`; `X17`]; (* 2173: adcs x15, x17, x12 *)
  [`X10`; `X20`]; (* 2174: adcs x7, x20, x10 *)
  [`X11`; `X14`]; (* 2175: adcs x20, x11, x14 *)
  []; (* 2176: mov x14, #0xffffffff *)
  [`CF`; `NF`; `ZF`; `VF`; `X19`; `X9`]; (* 2177: adc x22, x9, x19 *)
  [`X6`]; (* 2178: adds x12, x6, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2179: adc x10, xzr, xzr *)
  [`X8`]; (* 2180: subs x3, x8, x3 *)
  [`X13`]; (* 2181: sbcs x12, x13, x12 *)
  []; (* 2182: lsl x9, x3, #32 *)
  [`X9`]; (* 2183: add x3, x9, x3 *)
  [`X24`]; (* 2184: sbcs x10, x24, x10 *)
  [`X4`]; (* 2185: sbcs x24, x4, xzr *)
  []; (* 2186: lsr x9, x3, #32 *)
  []; (* 2187: sbcs x21, x21, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2188: sbc x1, x1, xzr *)
  []; (* 2189: subs x9, x9, x3 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2190: sbc x13, x3, xzr *)
  []; (* 2191: extr x9, x13, x9, #32 *)
  []; (* 2192: lsr x13, x13, #32 *)
  []; (* 2193: adds x13, x13, x3 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2194: adc x6, xzr, xzr *)
  [`X9`]; (* 2195: subs x12, x12, x9 *)
  [`X10`; `X13`]; (* 2196: sbcs x17, x10, x13 *)
  []; (* 2197: lsl x2, x12, #32 *)
  [`X24`; `X6`]; (* 2198: sbcs x10, x24, x6 *)
  [`X12`; `X2`]; (* 2199: add x9, x2, x12 *)
  [`X21`]; (* 2200: sbcs x6, x21, xzr *)
  []; (* 2201: lsr x5, x9, #32 *)
  [`X1`]; (* 2202: sbcs x21, x1, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X3`]; (* 2203: sbc x13, x3, xzr *)
  [`X5`]; (* 2204: subs x8, x5, x9 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2205: sbc x19, x9, xzr *)
  []; (* 2206: lsr x12, x19, #32 *)
  [`X19`; `X8`]; (* 2207: extr x3, x19, x8, #32 *)
  [`X12`]; (* 2208: adds x8, x12, x9 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2209: adc x1, xzr, xzr *)
  [`X17`; `X3`]; (* 2210: subs x2, x17, x3 *)
  [`X10`; `X8`]; (* 2211: sbcs x12, x10, x8 *)
  [`X1`; `X6`]; (* 2212: sbcs x5, x6, x1 *)
  [`X21`]; (* 2213: sbcs x3, x21, xzr *)
  [`X13`]; (* 2214: sbcs x19, x13, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X9`]; (* 2215: sbc x24, x9, xzr *)
  [`X15`; `X3`]; (* 2216: adds x23, x15, x3 *)
  [`X19`; `X7`]; (* 2217: adcs x8, x7, x19 *)
  [`X20`; `X24`]; (* 2218: adcs x11, x20, x24 *)
  [`CF`; `NF`; `ZF`; `VF`; `X22`]; (* 2219: adc x9, x22, xzr *)
  [`X9`]; (* 2220: add x24, x9, #0x1 *)
  []; (* 2221: lsl x7, x24, #32 *)
  []; (* 2222: subs x21, x24, x7 *)
  [`CF`; `NF`; `ZF`; `VF`; `X7`]; (* 2223: sbc x10, x7, xzr *)
  [`X2`; `X21`]; (* 2224: adds x6, x2, x21 *)
  [`X10`; `X12`]; (* 2225: adcs x7, x12, x10 *)
  [`X5`]; (* 2226: adcs x24, x5, x24 *)
  [`X23`]; (* 2227: adcs x13, x23, xzr *)
  []; (* 2228: adcs x8, x8, xzr *)
  [`X11`]; (* 2229: adcs x15, x11, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2230: csetm x23, cc *)
  [`X16`]; (* 2231: and x11, x16, x23 *)
  [`X14`]; (* 2232: and x20, x14, x23 *)
  [`X6`]; (* 2233: adds x22, x6, x20 *)
  [`X20`]; (* 2234: eor x3, x20, x23 *)
  [`X3`; `X7`]; (* 2235: adcs x5, x7, x3 *)
  [`X11`; `X24`]; (* 2236: adcs x14, x24, x11 *)
  [`X22`; `X5`]; (* 2237: stp x22, x5, [sp, #192] *)
  [`X13`]; (* 2238: adcs x5, x13, x23 *)
  [`X8`]; (* 2239: adcs x21, x8, x23 *)
  [`X14`; `X5`]; (* 2240: stp x14, x5, [sp, #208] *)
  [`CF`; `NF`; `ZF`; `VF`; `X15`; `X23`]; (* 2241: adc x12, x15, x23 *)
  [`X12`; `X21`]; (* 2242: stp x21, x12, [sp, #224] *)
  []; (* 2243: mov x1, sp *)
  []; (* 2244: ldr q3, [x1] *)
  []; (* 2245: ldr q25, [sp, #48] *)
  []; (* 2246: ldp x13, x23, [sp, #48] *)
  []; (* 2247: ldp x3, x21, [x1] *)
  []; (* 2248: rev64 v23.4s, v25.4s *)
  [`Q25`]; (* 2249: uzp1 v17.4s, v25.4s, v3.4s *)
  []; (* 2250: umulh x15, x3, x13 *)
  [`Q23`]; (* 2251: mul v6.4s, v23.4s, v3.4s *)
  []; (* 2252: uzp1 v3.4s, v3.4s, v3.4s *)
  []; (* 2253: ldr q27, [sp, #80] *)
  []; (* 2254: ldp x8, x24, [x1, #16] *)
  []; (* 2255: subs x6, x3, x21 *)
  []; (* 2256: ldr q0, [x1, #32] *)
  []; (* 2257: movi v23.2d, #0xffffffff *)
  []; (* 2258: csetm x10, cc *)
  []; (* 2259: umulh x19, x21, x23 *)
  []; (* 2260: rev64 v4.4s, v27.4s *)
  []; (* 2261: uzp2 v25.4s, v27.4s, v27.4s *)
  [`CF`; `NF`; `ZF`; `VF`; `X6`]; (* 2262: cneg x4, x6, cc *)
  []; (* 2263: subs x7, x23, x13 *)
  []; (* 2264: xtn v22.2s, v0.2d *)
  [`Q27`]; (* 2265: xtn v24.2s, v27.2d *)
  [`X7`]; (* 2266: cneg x20, x7, cc *)
  []; (* 2267: ldp x6, x14, [sp, #64] *)
  [`Q4`]; (* 2268: mul v27.4s, v4.4s, v0.4s *)
  [`Q6`]; (* 2269: uaddlp v20.2d, v6.4s *)
  [`CF`; `NF`; `ZF`; `VF`; `X10`]; (* 2270: cinv x5, x10, cc *)
  []; (* 2271: mul x16, x4, x20 *)
  [`Q0`]; (* 2272: uzp2 v6.4s, v0.4s, v0.4s *)
  []; (* 2273: umull v21.2d, v22.2s, v25.2s *)
  [`Q20`]; (* 2274: shl v0.2d, v20.2d, #32 *)
  [`Q17`; `Q3`]; (* 2275: umlal v0.2d, v3.2s, v17.2s *)
  []; (* 2276: mul x22, x8, x6 *)
  [`Q25`]; (* 2277: umull v1.2d, v6.2s, v25.2s *)
  []; (* 2278: subs x12, x3, x8 *)
  []; (* 2279: umull v20.2d, v22.2s, v24.2s *)
  [`X12`]; (* 2280: cneg x17, x12, cc *)
  []; (* 2281: umulh x9, x8, x6 *)
  []; (* 2282: mov x12, v0.d[1] *)
  [`X16`]; (* 2283: eor x11, x16, x5 *)
  [`Q0`]; (* 2284: mov x7, v0.d[0] *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2285: csetm x10, cc *)
  [`Q20`]; (* 2286: usra v21.2d, v20.2d, #32 *)
  [`X12`]; (* 2287: adds x15, x15, x12 *)
  [`X19`; `X22`]; (* 2288: adcs x12, x19, x22 *)
  [`X4`]; (* 2289: umulh x20, x4, x20 *)
  [`CF`; `NF`; `ZF`; `VF`; `X9`]; (* 2290: adc x19, x9, xzr *)
  []; (* 2291: usra v1.2d, v21.2d, #32 *)
  []; (* 2292: adds x22, x15, x7 *)
  [`Q21`; `Q23`]; (* 2293: and v26.16b, v21.16b, v23.16b *)
  []; (* 2294: adcs x16, x12, x15 *)
  [`Q27`]; (* 2295: uaddlp v25.2d, v27.4s *)
  []; (* 2296: adcs x9, x19, x12 *)
  [`Q6`]; (* 2297: umlal v26.2d, v6.2s, v24.2s *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2298: adc x4, x19, xzr *)
  []; (* 2299: adds x16, x16, x7 *)
  [`Q25`]; (* 2300: shl v27.2d, v25.2d, #32 *)
  [`X15`]; (* 2301: adcs x9, x9, x15 *)
  [`X12`]; (* 2302: adcs x4, x4, x12 *)
  [`X20`]; (* 2303: eor x12, x20, x5 *)
  [`CF`; `NF`; `ZF`; `VF`; `X19`]; (* 2304: adc x15, x19, xzr *)
  []; (* 2305: subs x20, x6, x13 *)
  []; (* 2306: cneg x20, x20, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2307: cinv x10, x10, cc *)
  []; (* 2308: cmn x5, #0x1 *)
  []; (* 2309: mul x19, x17, x20 *)
  [`X22`]; (* 2310: adcs x11, x22, x11 *)
  [`X16`]; (* 2311: adcs x12, x16, x12 *)
  []; (* 2312: adcs x9, x9, x5 *)
  [`X20`]; (* 2313: umulh x17, x17, x20 *)
  [`X4`]; (* 2314: adcs x22, x4, x5 *)
  [`CF`; `NF`; `ZF`; `VF`; `X15`]; (* 2315: adc x5, x15, x5 *)
  []; (* 2316: subs x16, x21, x8 *)
  [`X16`]; (* 2317: cneg x20, x16, cc *)
  []; (* 2318: eor x19, x19, x10 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2319: csetm x4, cc *)
  []; (* 2320: subs x16, x6, x23 *)
  []; (* 2321: cneg x16, x16, cc *)
  [`Q22`; `Q24`]; (* 2322: umlal v27.2d, v22.2s, v24.2s *)
  []; (* 2323: mul x15, x20, x16 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2324: cinv x4, x4, cc *)
  []; (* 2325: cmn x10, #0x1 *)
  [`Q26`]; (* 2326: usra v1.2d, v26.2d, #32 *)
  [`X12`]; (* 2327: adcs x19, x12, x19 *)
  []; (* 2328: eor x17, x17, x10 *)
  [`X17`]; (* 2329: adcs x9, x9, x17 *)
  []; (* 2330: adcs x22, x22, x10 *)
  []; (* 2331: lsl x12, x7, #32 *)
  [`X16`]; (* 2332: umulh x20, x20, x16 *)
  [`X15`]; (* 2333: eor x16, x15, x4 *)
  []; (* 2334: ldp x15, x17, [sp, #80] *)
  [`X12`; `X7`]; (* 2335: add x2, x12, x7 *)
  [`CF`; `NF`; `ZF`; `VF`; `X10`; `X5`]; (* 2336: adc x7, x5, x10 *)
  [`X1`]; (* 2337: ldp x5, x10, [x1, #32] *)
  []; (* 2338: lsr x1, x2, #32 *)
  [`X20`]; (* 2339: eor x12, x20, x4 *)
  []; (* 2340: subs x1, x1, x2 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2341: sbc x20, x2, xzr *)
  []; (* 2342: cmn x4, #0x1 *)
  [`X16`]; (* 2343: adcs x9, x9, x16 *)
  []; (* 2344: extr x1, x20, x1, #32 *)
  []; (* 2345: lsr x20, x20, #32 *)
  [`X12`]; (* 2346: adcs x22, x22, x12 *)
  [`CF`; `NF`; `ZF`; `VF`; `X4`; `X7`]; (* 2347: adc x16, x7, x4 *)
  [`X20`]; (* 2348: adds x12, x20, x2 *)
  []; (* 2349: umulh x7, x24, x14 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2350: adc x4, xzr, xzr *)
  [`X11`]; (* 2351: subs x1, x11, x1 *)
  [`X12`; `X19`]; (* 2352: sbcs x20, x19, x12 *)
  [`X4`; `X9`]; (* 2353: sbcs x12, x9, x4 *)
  []; (* 2354: lsl x9, x1, #32 *)
  [`X9`]; (* 2355: add x1, x9, x1 *)
  [`X22`]; (* 2356: sbcs x9, x22, xzr *)
  []; (* 2357: mul x22, x24, x14 *)
  []; (* 2358: sbcs x16, x16, xzr *)
  []; (* 2359: lsr x4, x1, #32 *)
  [`CF`; `NF`; `ZF`; `VF`; `X2`]; (* 2360: sbc x19, x2, xzr *)
  []; (* 2361: subs x4, x4, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2362: sbc x11, x1, xzr *)
  [`X4`]; (* 2363: extr x2, x11, x4, #32 *)
  [`X11`]; (* 2364: lsr x4, x11, #32 *)
  []; (* 2365: adds x4, x4, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2366: adc x11, xzr, xzr *)
  [`X20`]; (* 2367: subs x2, x20, x2 *)
  [`X12`]; (* 2368: sbcs x4, x12, x4 *)
  [`X11`; `X9`]; (* 2369: sbcs x20, x9, x11 *)
  []; (* 2370: lsl x12, x2, #32 *)
  [`X12`]; (* 2371: add x2, x12, x2 *)
  [`X16`]; (* 2372: sbcs x9, x16, xzr *)
  []; (* 2373: lsr x11, x2, #32 *)
  []; (* 2374: sbcs x19, x19, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2375: sbc x1, x1, xzr *)
  [`X11`]; (* 2376: subs x16, x11, x2 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2377: sbc x12, x2, xzr *)
  []; (* 2378: extr x16, x12, x16, #32 *)
  []; (* 2379: lsr x12, x12, #32 *)
  [`X12`]; (* 2380: adds x11, x12, x2 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2381: adc x12, xzr, xzr *)
  [`X4`]; (* 2382: subs x16, x4, x16 *)
  []; (* 2383: mov x4, v27.d[0] *)
  [`X20`]; (* 2384: sbcs x11, x20, x11 *)
  [`X12`; `X9`]; (* 2385: sbcs x20, x9, x12 *)
  [`X11`; `X16`]; (* 2386: stp x16, x11, [sp, #48] *)
  [`X19`]; (* 2387: sbcs x11, x19, xzr *)
  [`X1`]; (* 2388: sbcs x9, x1, xzr *)
  [`X11`; `X20`]; (* 2389: stp x20, x11, [sp, #64] *)
  []; (* 2390: mov x1, v1.d[0] *)
  [`CF`; `NF`; `ZF`; `VF`; `X2`]; (* 2391: sbc x20, x2, xzr *)
  []; (* 2392: subs x12, x24, x5 *)
  [`Q27`]; (* 2393: mov x11, v27.d[1] *)
  [`X12`]; (* 2394: cneg x16, x12, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2395: csetm x2, cc *)
  []; (* 2396: subs x19, x15, x14 *)
  [`Q1`]; (* 2397: mov x12, v1.d[1] *)
  []; (* 2398: cinv x2, x2, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2399: cneg x19, x19, cc *)
  [`X20`; `X9`]; (* 2400: stp x9, x20, [sp, #80] *)
  []; (* 2401: mul x9, x16, x19 *)
  [`X7`]; (* 2402: adds x4, x7, x4 *)
  [`X1`]; (* 2403: adcs x11, x1, x11 *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 2404: adc x1, x12, xzr *)
  []; (* 2405: adds x20, x4, x22 *)
  [`X16`]; (* 2406: umulh x19, x16, x19 *)
  []; (* 2407: adcs x7, x11, x4 *)
  [`X9`]; (* 2408: eor x16, x9, x2 *)
  []; (* 2409: adcs x9, x1, x11 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2410: adc x12, x1, xzr *)
  []; (* 2411: adds x7, x7, x22 *)
  [`X9`]; (* 2412: adcs x4, x9, x4 *)
  [`X11`; `X12`]; (* 2413: adcs x9, x12, x11 *)
  [`CF`; `NF`; `ZF`; `VF`; `X1`]; (* 2414: adc x12, x1, xzr *)
  []; (* 2415: cmn x2, #0x1 *)
  [`X19`]; (* 2416: eor x1, x19, x2 *)
  [`X16`; `X20`]; (* 2417: adcs x11, x20, x16 *)
  [`X1`; `X7`]; (* 2418: adcs x19, x7, x1 *)
  [`X4`]; (* 2419: adcs x1, x4, x2 *)
  [`X9`]; (* 2420: adcs x20, x9, x2 *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 2421: adc x2, x12, x2 *)
  []; (* 2422: subs x12, x24, x10 *)
  [`X12`]; (* 2423: cneg x16, x12, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2424: csetm x12, cc *)
  []; (* 2425: subs x9, x17, x14 *)
  []; (* 2426: cinv x12, x12, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2427: cneg x9, x9, cc *)
  [`X24`]; (* 2428: subs x3, x24, x3 *)
  []; (* 2429: sbcs x21, x5, x21 *)
  []; (* 2430: mul x24, x16, x9 *)
  [`X8`]; (* 2431: sbcs x4, x10, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2432: ngc x8, xzr *)
  [`X5`]; (* 2433: subs x10, x5, x10 *)
  [`X24`]; (* 2434: eor x5, x24, x12 *)
  []; (* 2435: csetm x7, cc *)
  [`CF`; `NF`; `ZF`; `VF`; `X10`]; (* 2436: cneg x24, x10, cc *)
  []; (* 2437: subs x10, x17, x15 *)
  []; (* 2438: cinv x7, x7, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2439: cneg x10, x10, cc *)
  [`X13`]; (* 2440: subs x14, x13, x14 *)
  [`X23`]; (* 2441: sbcs x15, x23, x15 *)
  [`X21`]; (* 2442: eor x13, x21, x8 *)
  []; (* 2443: mul x23, x24, x10 *)
  [`X6`]; (* 2444: sbcs x17, x6, x17 *)
  [`X3`]; (* 2445: eor x6, x3, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2446: ngc x21, xzr *)
  [`X16`]; (* 2447: umulh x9, x16, x9 *)
  []; (* 2448: cmn x8, #0x1 *)
  [`X23`]; (* 2449: eor x3, x23, x7 *)
  [`X6`]; (* 2450: adcs x23, x6, xzr *)
  []; (* 2451: adcs x13, x13, xzr *)
  [`X4`]; (* 2452: eor x16, x4, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2453: adc x16, x16, xzr *)
  [`X17`]; (* 2454: eor x4, x17, x21 *)
  [`X10`; `X24`]; (* 2455: umulh x17, x24, x10 *)
  []; (* 2456: cmn x21, #0x1 *)
  [`X14`]; (* 2457: eor x24, x14, x21 *)
  [`X15`]; (* 2458: eor x6, x15, x21 *)
  [`X24`]; (* 2459: adcs x15, x24, xzr *)
  [`X6`]; (* 2460: adcs x14, x6, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X4`]; (* 2461: adc x6, x4, xzr *)
  []; (* 2462: cmn x12, #0x1 *)
  [`X9`]; (* 2463: eor x4, x9, x12 *)
  [`X5`]; (* 2464: adcs x19, x19, x5 *)
  []; (* 2465: umulh x5, x23, x15 *)
  [`X4`]; (* 2466: adcs x1, x1, x4 *)
  [`X20`]; (* 2467: adcs x10, x20, x12 *)
  [`X17`]; (* 2468: eor x4, x17, x7 *)
  []; (* 2469: ldp x20, x9, [sp, #48] *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 2470: adc x2, x2, x12 *)
  []; (* 2471: cmn x7, #0x1 *)
  [`X1`; `X3`]; (* 2472: adcs x12, x1, x3 *)
  []; (* 2473: ldp x17, x24, [sp, #64] *)
  []; (* 2474: mul x1, x16, x6 *)
  [`X10`; `X4`]; (* 2475: adcs x3, x10, x4 *)
  [`CF`; `NF`; `ZF`; `VF`; `X7`]; (* 2476: adc x2, x2, x7 *)
  []; (* 2477: ldp x7, x4, [sp, #80] *)
  [`X22`]; (* 2478: adds x20, x22, x20 *)
  []; (* 2479: mul x10, x13, x14 *)
  [`X9`]; (* 2480: adcs x11, x11, x9 *)
  [`X21`; `X8`]; (* 2481: eor x9, x8, x21 *)
  [`X17`; `X19`]; (* 2482: adcs x21, x19, x17 *)
  [`X11`; `X20`]; (* 2483: stp x20, x11, [sp, #48] *)
  [`X24`]; (* 2484: adcs x12, x12, x24 *)
  []; (* 2485: mul x8, x23, x15 *)
  [`X7`]; (* 2486: adcs x3, x3, x7 *)
  [`X12`; `X21`]; (* 2487: stp x21, x12, [sp, #64] *)
  [`X2`; `X4`]; (* 2488: adcs x12, x2, x4 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2489: adc x19, xzr, xzr *)
  []; (* 2490: subs x21, x23, x16 *)
  []; (* 2491: umulh x2, x16, x6 *)
  [`X12`; `X3`]; (* 2492: stp x3, x12, [sp, #80] *)
  [`X21`]; (* 2493: cneg x3, x21, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2494: csetm x24, cc *)
  []; (* 2495: umulh x11, x13, x14 *)
  [`X16`]; (* 2496: subs x21, x13, x16 *)
  []; (* 2497: eor x7, x8, x9 *)
  [`X21`]; (* 2498: cneg x17, x21, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2499: csetm x16, cc *)
  []; (* 2500: subs x21, x6, x15 *)
  [`X21`]; (* 2501: cneg x22, x21, cc *)
  [`CF`; `NF`; `ZF`; `VF`; `X24`]; (* 2502: cinv x21, x24, cc *)
  [`X13`; `X23`]; (* 2503: subs x20, x23, x13 *)
  []; (* 2504: umulh x12, x3, x22 *)
  [`X20`]; (* 2505: cneg x23, x20, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2506: csetm x24, cc *)
  [`X15`]; (* 2507: subs x20, x14, x15 *)
  []; (* 2508: cinv x24, x24, cc *)
  [`X3`]; (* 2509: mul x22, x3, x22 *)
  [`CF`; `NF`; `ZF`; `VF`; `X20`]; (* 2510: cneg x3, x20, cc *)
  [`X14`; `X6`]; (* 2511: subs x13, x6, x14 *)
  [`X13`]; (* 2512: cneg x20, x13, cc *)
  [`CF`; `NF`; `ZF`; `VF`; `X16`]; (* 2513: cinv x15, x16, cc *)
  [`X10`; `X5`]; (* 2514: adds x13, x5, x10 *)
  []; (* 2515: mul x4, x23, x3 *)
  [`X1`]; (* 2516: adcs x11, x11, x1 *)
  [`CF`; `NF`; `ZF`; `VF`; `X2`]; (* 2517: adc x14, x2, xzr *)
  []; (* 2518: adds x5, x13, x8 *)
  []; (* 2519: adcs x16, x11, x13 *)
  [`X3`]; (* 2520: umulh x23, x23, x3 *)
  []; (* 2521: adcs x3, x14, x11 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2522: adc x1, x14, xzr *)
  [`X16`; `X8`]; (* 2523: adds x10, x16, x8 *)
  [`X13`; `X3`]; (* 2524: adcs x6, x3, x13 *)
  [`X1`; `X11`]; (* 2525: adcs x8, x1, x11 *)
  []; (* 2526: umulh x13, x17, x20 *)
  [`X4`]; (* 2527: eor x1, x4, x24 *)
  [`CF`; `NF`; `ZF`; `VF`; `X14`]; (* 2528: adc x4, x14, xzr *)
  []; (* 2529: cmn x24, #0x1 *)
  [`X5`]; (* 2530: adcs x1, x5, x1 *)
  [`X23`]; (* 2531: eor x16, x23, x24 *)
  [`X1`]; (* 2532: eor x11, x1, x9 *)
  [`X10`; `X16`]; (* 2533: adcs x23, x10, x16 *)
  [`X22`]; (* 2534: eor x2, x22, x21 *)
  [`X6`]; (* 2535: adcs x3, x6, x24 *)
  [`X17`; `X20`]; (* 2536: mul x14, x17, x20 *)
  [`X13`]; (* 2537: eor x17, x13, x15 *)
  [`X8`]; (* 2538: adcs x13, x8, x24 *)
  [`CF`; `NF`; `ZF`; `VF`; `X24`; `X4`]; (* 2539: adc x8, x4, x24 *)
  []; (* 2540: cmn x21, #0x1 *)
  [`X2`; `X23`]; (* 2541: adcs x6, x23, x2 *)
  []; (* 2542: mov x16, #0xfffffffffffffffe *)
  [`X12`]; (* 2543: eor x20, x12, x21 *)
  [`X3`]; (* 2544: adcs x20, x3, x20 *)
  [`X14`]; (* 2545: eor x23, x14, x15 *)
  [`X13`]; (* 2546: adcs x2, x13, x21 *)
  [`CF`; `NF`; `ZF`; `VF`; `X21`]; (* 2547: adc x8, x8, x21 *)
  []; (* 2548: cmn x15, #0x1 *)
  []; (* 2549: ldp x5, x4, [sp, #48] *)
  []; (* 2550: ldp x21, x12, [sp, #64] *)
  [`X20`; `X23`]; (* 2551: adcs x22, x20, x23 *)
  [`X22`]; (* 2552: eor x23, x22, x9 *)
  [`X2`]; (* 2553: adcs x17, x2, x17 *)
  [`CF`; `NF`; `ZF`; `VF`; `X15`; `X8`]; (* 2554: adc x22, x8, x15 *)
  []; (* 2555: cmn x9, #0x1 *)
  [`X7`]; (* 2556: adcs x15, x7, x5 *)
  []; (* 2557: ldp x10, x14, [sp, #80] *)
  [`X6`]; (* 2558: eor x1, x6, x9 *)
  []; (* 2559: lsl x2, x15, #32 *)
  [`X11`]; (* 2560: adcs x8, x11, x4 *)
  [`X1`]; (* 2561: adcs x13, x1, x21 *)
  [`X22`]; (* 2562: eor x1, x22, x9 *)
  [`X23`]; (* 2563: adcs x24, x23, x12 *)
  [`X17`]; (* 2564: eor x11, x17, x9 *)
  [`X11`]; (* 2565: adcs x23, x11, x10 *)
  [`X1`]; (* 2566: adcs x7, x1, x14 *)
  []; (* 2567: adcs x17, x9, x19 *)
  []; (* 2568: adcs x20, x9, xzr *)
  [`X15`; `X2`]; (* 2569: add x1, x2, x15 *)
  []; (* 2570: lsr x3, x1, #32 *)
  []; (* 2571: adcs x11, x9, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2572: adc x9, x9, xzr *)
  []; (* 2573: subs x3, x3, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2574: sbc x6, x1, xzr *)
  [`X5`]; (* 2575: adds x24, x24, x5 *)
  [`X23`]; (* 2576: adcs x4, x23, x4 *)
  []; (* 2577: extr x3, x6, x3, #32 *)
  []; (* 2578: lsr x6, x6, #32 *)
  [`X7`]; (* 2579: adcs x21, x7, x21 *)
  [`X12`; `X17`]; (* 2580: adcs x15, x17, x12 *)
  [`X10`; `X20`]; (* 2581: adcs x7, x20, x10 *)
  [`X11`; `X14`]; (* 2582: adcs x20, x11, x14 *)
  []; (* 2583: mov x14, #0xffffffff *)
  [`CF`; `NF`; `ZF`; `VF`; `X19`; `X9`]; (* 2584: adc x22, x9, x19 *)
  [`X6`]; (* 2585: adds x12, x6, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2586: adc x10, xzr, xzr *)
  [`X8`]; (* 2587: subs x3, x8, x3 *)
  [`X13`]; (* 2588: sbcs x12, x13, x12 *)
  []; (* 2589: lsl x9, x3, #32 *)
  [`X9`]; (* 2590: add x3, x9, x3 *)
  [`X24`]; (* 2591: sbcs x10, x24, x10 *)
  [`X4`]; (* 2592: sbcs x24, x4, xzr *)
  []; (* 2593: lsr x9, x3, #32 *)
  []; (* 2594: sbcs x21, x21, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2595: sbc x1, x1, xzr *)
  []; (* 2596: subs x9, x9, x3 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2597: sbc x13, x3, xzr *)
  []; (* 2598: extr x9, x13, x9, #32 *)
  []; (* 2599: lsr x13, x13, #32 *)
  []; (* 2600: adds x13, x13, x3 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2601: adc x6, xzr, xzr *)
  [`X9`]; (* 2602: subs x12, x12, x9 *)
  [`X10`; `X13`]; (* 2603: sbcs x17, x10, x13 *)
  []; (* 2604: lsl x2, x12, #32 *)
  [`X24`; `X6`]; (* 2605: sbcs x10, x24, x6 *)
  [`X12`; `X2`]; (* 2606: add x9, x2, x12 *)
  [`X21`]; (* 2607: sbcs x6, x21, xzr *)
  []; (* 2608: lsr x5, x9, #32 *)
  [`X1`]; (* 2609: sbcs x21, x1, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X3`]; (* 2610: sbc x13, x3, xzr *)
  [`X5`]; (* 2611: subs x8, x5, x9 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2612: sbc x19, x9, xzr *)
  []; (* 2613: lsr x12, x19, #32 *)
  [`X19`; `X8`]; (* 2614: extr x3, x19, x8, #32 *)
  [`X12`]; (* 2615: adds x8, x12, x9 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2616: adc x1, xzr, xzr *)
  [`X17`; `X3`]; (* 2617: subs x2, x17, x3 *)
  [`X10`; `X8`]; (* 2618: sbcs x12, x10, x8 *)
  [`X1`; `X6`]; (* 2619: sbcs x5, x6, x1 *)
  [`X21`]; (* 2620: sbcs x3, x21, xzr *)
  [`X13`]; (* 2621: sbcs x19, x13, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X9`]; (* 2622: sbc x24, x9, xzr *)
  [`X15`; `X3`]; (* 2623: adds x23, x15, x3 *)
  [`X19`; `X7`]; (* 2624: adcs x8, x7, x19 *)
  [`X20`; `X24`]; (* 2625: adcs x11, x20, x24 *)
  [`CF`; `NF`; `ZF`; `VF`; `X22`]; (* 2626: adc x9, x22, xzr *)
  [`X9`]; (* 2627: add x24, x9, #0x1 *)
  []; (* 2628: lsl x7, x24, #32 *)
  []; (* 2629: subs x21, x24, x7 *)
  [`CF`; `NF`; `ZF`; `VF`; `X7`]; (* 2630: sbc x10, x7, xzr *)
  [`X2`; `X21`]; (* 2631: adds x6, x2, x21 *)
  [`X10`; `X12`]; (* 2632: adcs x7, x12, x10 *)
  [`X5`]; (* 2633: adcs x24, x5, x24 *)
  [`X23`]; (* 2634: adcs x13, x23, xzr *)
  []; (* 2635: adcs x8, x8, xzr *)
  [`X11`]; (* 2636: adcs x15, x11, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2637: csetm x23, cc *)
  [`X16`]; (* 2638: and x11, x16, x23 *)
  [`X14`]; (* 2639: and x20, x14, x23 *)
  [`X6`]; (* 2640: adds x22, x6, x20 *)
  [`X20`]; (* 2641: eor x3, x20, x23 *)
  [`X3`; `X7`]; (* 2642: adcs x5, x7, x3 *)
  [`X11`; `X24`]; (* 2643: adcs x14, x24, x11 *)
  [`X22`; `X5`]; (* 2644: stp x22, x5, [sp, #48] *)
  [`X13`]; (* 2645: adcs x5, x13, x23 *)
  [`X8`]; (* 2646: adcs x21, x8, x23 *)
  [`X14`; `X5`]; (* 2647: stp x14, x5, [sp, #64] *)
  [`CF`; `NF`; `ZF`; `VF`; `X15`; `X23`]; (* 2648: adc x12, x15, x23 *)
  [`X12`; `X21`]; (* 2649: stp x21, x12, [sp, #80] *)
  []; (* 2650: ldr q3, [sp, #240] *)
  []; (* 2651: ldr q25, [sp, #288] *)
  []; (* 2652: ldp x13, x23, [sp, #288] *)
  []; (* 2653: ldp x3, x21, [sp, #240] *)
  []; (* 2654: rev64 v23.4s, v25.4s *)
  [`Q25`]; (* 2655: uzp1 v17.4s, v25.4s, v3.4s *)
  []; (* 2656: umulh x15, x3, x13 *)
  [`Q23`]; (* 2657: mul v6.4s, v23.4s, v3.4s *)
  []; (* 2658: uzp1 v3.4s, v3.4s, v3.4s *)
  []; (* 2659: ldr q27, [sp, #320] *)
  []; (* 2660: ldp x8, x24, [sp, #256] *)
  []; (* 2661: subs x6, x3, x21 *)
  []; (* 2662: ldr q0, [sp, #272] *)
  []; (* 2663: movi v23.2d, #0xffffffff *)
  []; (* 2664: csetm x10, cc *)
  []; (* 2665: umulh x19, x21, x23 *)
  []; (* 2666: rev64 v4.4s, v27.4s *)
  []; (* 2667: uzp2 v25.4s, v27.4s, v27.4s *)
  [`CF`; `NF`; `ZF`; `VF`; `X6`]; (* 2668: cneg x4, x6, cc *)
  []; (* 2669: subs x7, x23, x13 *)
  []; (* 2670: xtn v22.2s, v0.2d *)
  [`Q27`]; (* 2671: xtn v24.2s, v27.2d *)
  [`X7`]; (* 2672: cneg x20, x7, cc *)
  []; (* 2673: ldp x6, x14, [sp, #304] *)
  [`Q4`]; (* 2674: mul v27.4s, v4.4s, v0.4s *)
  [`Q6`]; (* 2675: uaddlp v20.2d, v6.4s *)
  [`CF`; `NF`; `ZF`; `VF`; `X10`]; (* 2676: cinv x5, x10, cc *)
  []; (* 2677: mul x16, x4, x20 *)
  [`Q0`]; (* 2678: uzp2 v6.4s, v0.4s, v0.4s *)
  []; (* 2679: umull v21.2d, v22.2s, v25.2s *)
  [`Q20`]; (* 2680: shl v0.2d, v20.2d, #32 *)
  [`Q17`; `Q3`]; (* 2681: umlal v0.2d, v3.2s, v17.2s *)
  []; (* 2682: mul x22, x8, x6 *)
  [`Q25`]; (* 2683: umull v1.2d, v6.2s, v25.2s *)
  []; (* 2684: subs x12, x3, x8 *)
  []; (* 2685: umull v20.2d, v22.2s, v24.2s *)
  [`X12`]; (* 2686: cneg x17, x12, cc *)
  []; (* 2687: umulh x9, x8, x6 *)
  []; (* 2688: mov x12, v0.d[1] *)
  [`X16`]; (* 2689: eor x11, x16, x5 *)
  [`Q0`]; (* 2690: mov x7, v0.d[0] *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2691: csetm x10, cc *)
  [`Q20`]; (* 2692: usra v21.2d, v20.2d, #32 *)
  [`X12`]; (* 2693: adds x15, x15, x12 *)
  [`X19`; `X22`]; (* 2694: adcs x12, x19, x22 *)
  [`X4`]; (* 2695: umulh x20, x4, x20 *)
  [`CF`; `NF`; `ZF`; `VF`; `X9`]; (* 2696: adc x19, x9, xzr *)
  []; (* 2697: usra v1.2d, v21.2d, #32 *)
  []; (* 2698: adds x22, x15, x7 *)
  [`Q21`; `Q23`]; (* 2699: and v26.16b, v21.16b, v23.16b *)
  []; (* 2700: adcs x16, x12, x15 *)
  [`Q27`]; (* 2701: uaddlp v25.2d, v27.4s *)
  []; (* 2702: adcs x9, x19, x12 *)
  [`Q6`]; (* 2703: umlal v26.2d, v6.2s, v24.2s *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2704: adc x4, x19, xzr *)
  []; (* 2705: adds x16, x16, x7 *)
  [`Q25`]; (* 2706: shl v27.2d, v25.2d, #32 *)
  [`X15`]; (* 2707: adcs x9, x9, x15 *)
  [`X12`]; (* 2708: adcs x4, x4, x12 *)
  [`X20`]; (* 2709: eor x12, x20, x5 *)
  [`CF`; `NF`; `ZF`; `VF`; `X19`]; (* 2710: adc x15, x19, xzr *)
  []; (* 2711: subs x20, x6, x13 *)
  []; (* 2712: cneg x20, x20, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2713: cinv x10, x10, cc *)
  []; (* 2714: cmn x5, #0x1 *)
  []; (* 2715: mul x19, x17, x20 *)
  [`X22`]; (* 2716: adcs x11, x22, x11 *)
  [`X16`]; (* 2717: adcs x12, x16, x12 *)
  []; (* 2718: adcs x9, x9, x5 *)
  [`X20`]; (* 2719: umulh x17, x17, x20 *)
  [`X4`]; (* 2720: adcs x22, x4, x5 *)
  [`CF`; `NF`; `ZF`; `VF`; `X15`]; (* 2721: adc x5, x15, x5 *)
  []; (* 2722: subs x16, x21, x8 *)
  [`X16`]; (* 2723: cneg x20, x16, cc *)
  []; (* 2724: eor x19, x19, x10 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2725: csetm x4, cc *)
  []; (* 2726: subs x16, x6, x23 *)
  []; (* 2727: cneg x16, x16, cc *)
  [`Q22`; `Q24`]; (* 2728: umlal v27.2d, v22.2s, v24.2s *)
  []; (* 2729: mul x15, x20, x16 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2730: cinv x4, x4, cc *)
  []; (* 2731: cmn x10, #0x1 *)
  [`Q26`]; (* 2732: usra v1.2d, v26.2d, #32 *)
  [`X12`]; (* 2733: adcs x19, x12, x19 *)
  []; (* 2734: eor x17, x17, x10 *)
  [`X17`]; (* 2735: adcs x9, x9, x17 *)
  []; (* 2736: adcs x22, x22, x10 *)
  []; (* 2737: lsl x12, x7, #32 *)
  [`X16`]; (* 2738: umulh x20, x20, x16 *)
  [`X15`]; (* 2739: eor x16, x15, x4 *)
  []; (* 2740: ldp x15, x17, [sp, #320] *)
  [`X12`; `X7`]; (* 2741: add x2, x12, x7 *)
  [`CF`; `NF`; `ZF`; `VF`; `X10`; `X5`]; (* 2742: adc x7, x5, x10 *)
  []; (* 2743: ldp x5, x10, [sp, #272] *)
  []; (* 2744: lsr x1, x2, #32 *)
  [`X20`]; (* 2745: eor x12, x20, x4 *)
  []; (* 2746: subs x1, x1, x2 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2747: sbc x20, x2, xzr *)
  []; (* 2748: cmn x4, #0x1 *)
  [`X16`]; (* 2749: adcs x9, x9, x16 *)
  []; (* 2750: extr x1, x20, x1, #32 *)
  []; (* 2751: lsr x20, x20, #32 *)
  [`X12`]; (* 2752: adcs x22, x22, x12 *)
  [`CF`; `NF`; `ZF`; `VF`; `X4`; `X7`]; (* 2753: adc x16, x7, x4 *)
  [`X20`]; (* 2754: adds x12, x20, x2 *)
  []; (* 2755: umulh x7, x24, x14 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2756: adc x4, xzr, xzr *)
  [`X11`]; (* 2757: subs x1, x11, x1 *)
  [`X12`; `X19`]; (* 2758: sbcs x20, x19, x12 *)
  [`X4`; `X9`]; (* 2759: sbcs x12, x9, x4 *)
  []; (* 2760: lsl x9, x1, #32 *)
  [`X9`]; (* 2761: add x1, x9, x1 *)
  [`X22`]; (* 2762: sbcs x9, x22, xzr *)
  []; (* 2763: mul x22, x24, x14 *)
  []; (* 2764: sbcs x16, x16, xzr *)
  []; (* 2765: lsr x4, x1, #32 *)
  [`CF`; `NF`; `ZF`; `VF`; `X2`]; (* 2766: sbc x19, x2, xzr *)
  []; (* 2767: subs x4, x4, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2768: sbc x11, x1, xzr *)
  [`X4`]; (* 2769: extr x2, x11, x4, #32 *)
  [`X11`]; (* 2770: lsr x4, x11, #32 *)
  []; (* 2771: adds x4, x4, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2772: adc x11, xzr, xzr *)
  [`X20`]; (* 2773: subs x2, x20, x2 *)
  [`X12`]; (* 2774: sbcs x4, x12, x4 *)
  [`X11`; `X9`]; (* 2775: sbcs x20, x9, x11 *)
  []; (* 2776: lsl x12, x2, #32 *)
  [`X12`]; (* 2777: add x2, x12, x2 *)
  [`X16`]; (* 2778: sbcs x9, x16, xzr *)
  []; (* 2779: lsr x11, x2, #32 *)
  []; (* 2780: sbcs x19, x19, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2781: sbc x1, x1, xzr *)
  [`X11`]; (* 2782: subs x16, x11, x2 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2783: sbc x12, x2, xzr *)
  []; (* 2784: extr x16, x12, x16, #32 *)
  []; (* 2785: lsr x12, x12, #32 *)
  [`X12`]; (* 2786: adds x11, x12, x2 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2787: adc x12, xzr, xzr *)
  [`X4`]; (* 2788: subs x16, x4, x16 *)
  []; (* 2789: mov x4, v27.d[0] *)
  [`X20`]; (* 2790: sbcs x11, x20, x11 *)
  [`X12`; `X9`]; (* 2791: sbcs x20, x9, x12 *)
  [`X11`; `X16`]; (* 2792: stp x16, x11, [sp, #288] *)
  [`X19`]; (* 2793: sbcs x11, x19, xzr *)
  [`X1`]; (* 2794: sbcs x9, x1, xzr *)
  [`X11`; `X20`]; (* 2795: stp x20, x11, [sp, #304] *)
  []; (* 2796: mov x1, v1.d[0] *)
  [`CF`; `NF`; `ZF`; `VF`; `X2`]; (* 2797: sbc x20, x2, xzr *)
  []; (* 2798: subs x12, x24, x5 *)
  [`Q27`]; (* 2799: mov x11, v27.d[1] *)
  [`X12`]; (* 2800: cneg x16, x12, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2801: csetm x2, cc *)
  []; (* 2802: subs x19, x15, x14 *)
  [`Q1`]; (* 2803: mov x12, v1.d[1] *)
  []; (* 2804: cinv x2, x2, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2805: cneg x19, x19, cc *)
  [`X20`; `X9`]; (* 2806: stp x9, x20, [sp, #320] *)
  []; (* 2807: mul x9, x16, x19 *)
  [`X7`]; (* 2808: adds x4, x7, x4 *)
  [`X1`]; (* 2809: adcs x11, x1, x11 *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 2810: adc x1, x12, xzr *)
  []; (* 2811: adds x20, x4, x22 *)
  [`X16`]; (* 2812: umulh x19, x16, x19 *)
  []; (* 2813: adcs x7, x11, x4 *)
  [`X9`]; (* 2814: eor x16, x9, x2 *)
  []; (* 2815: adcs x9, x1, x11 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2816: adc x12, x1, xzr *)
  []; (* 2817: adds x7, x7, x22 *)
  [`X9`]; (* 2818: adcs x4, x9, x4 *)
  [`X11`; `X12`]; (* 2819: adcs x9, x12, x11 *)
  [`CF`; `NF`; `ZF`; `VF`; `X1`]; (* 2820: adc x12, x1, xzr *)
  []; (* 2821: cmn x2, #0x1 *)
  [`X19`]; (* 2822: eor x1, x19, x2 *)
  [`X16`; `X20`]; (* 2823: adcs x11, x20, x16 *)
  [`X1`; `X7`]; (* 2824: adcs x19, x7, x1 *)
  [`X4`]; (* 2825: adcs x1, x4, x2 *)
  [`X9`]; (* 2826: adcs x20, x9, x2 *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 2827: adc x2, x12, x2 *)
  []; (* 2828: subs x12, x24, x10 *)
  [`X12`]; (* 2829: cneg x16, x12, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2830: csetm x12, cc *)
  []; (* 2831: subs x9, x17, x14 *)
  []; (* 2832: cinv x12, x12, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2833: cneg x9, x9, cc *)
  [`X24`]; (* 2834: subs x3, x24, x3 *)
  []; (* 2835: sbcs x21, x5, x21 *)
  []; (* 2836: mul x24, x16, x9 *)
  [`X8`]; (* 2837: sbcs x4, x10, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2838: ngc x8, xzr *)
  [`X5`]; (* 2839: subs x10, x5, x10 *)
  [`X24`]; (* 2840: eor x5, x24, x12 *)
  []; (* 2841: csetm x7, cc *)
  [`CF`; `NF`; `ZF`; `VF`; `X10`]; (* 2842: cneg x24, x10, cc *)
  []; (* 2843: subs x10, x17, x15 *)
  []; (* 2844: cinv x7, x7, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2845: cneg x10, x10, cc *)
  [`X13`]; (* 2846: subs x14, x13, x14 *)
  [`X23`]; (* 2847: sbcs x15, x23, x15 *)
  [`X21`]; (* 2848: eor x13, x21, x8 *)
  []; (* 2849: mul x23, x24, x10 *)
  [`X6`]; (* 2850: sbcs x17, x6, x17 *)
  [`X3`]; (* 2851: eor x6, x3, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2852: ngc x21, xzr *)
  [`X16`]; (* 2853: umulh x9, x16, x9 *)
  []; (* 2854: cmn x8, #0x1 *)
  [`X23`]; (* 2855: eor x3, x23, x7 *)
  [`X6`]; (* 2856: adcs x23, x6, xzr *)
  []; (* 2857: adcs x13, x13, xzr *)
  [`X4`]; (* 2858: eor x16, x4, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2859: adc x16, x16, xzr *)
  [`X17`]; (* 2860: eor x4, x17, x21 *)
  [`X10`; `X24`]; (* 2861: umulh x17, x24, x10 *)
  []; (* 2862: cmn x21, #0x1 *)
  [`X14`]; (* 2863: eor x24, x14, x21 *)
  [`X15`]; (* 2864: eor x6, x15, x21 *)
  [`X24`]; (* 2865: adcs x15, x24, xzr *)
  [`X6`]; (* 2866: adcs x14, x6, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X4`]; (* 2867: adc x6, x4, xzr *)
  []; (* 2868: cmn x12, #0x1 *)
  [`X9`]; (* 2869: eor x4, x9, x12 *)
  [`X5`]; (* 2870: adcs x19, x19, x5 *)
  []; (* 2871: umulh x5, x23, x15 *)
  [`X4`]; (* 2872: adcs x1, x1, x4 *)
  [`X20`]; (* 2873: adcs x10, x20, x12 *)
  [`X17`]; (* 2874: eor x4, x17, x7 *)
  []; (* 2875: ldp x20, x9, [sp, #288] *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 2876: adc x2, x2, x12 *)
  []; (* 2877: cmn x7, #0x1 *)
  [`X1`; `X3`]; (* 2878: adcs x12, x1, x3 *)
  []; (* 2879: ldp x17, x24, [sp, #304] *)
  []; (* 2880: mul x1, x16, x6 *)
  [`X10`; `X4`]; (* 2881: adcs x3, x10, x4 *)
  [`CF`; `NF`; `ZF`; `VF`; `X7`]; (* 2882: adc x2, x2, x7 *)
  []; (* 2883: ldp x7, x4, [sp, #320] *)
  [`X22`]; (* 2884: adds x20, x22, x20 *)
  []; (* 2885: mul x10, x13, x14 *)
  [`X9`]; (* 2886: adcs x11, x11, x9 *)
  [`X21`; `X8`]; (* 2887: eor x9, x8, x21 *)
  [`X17`; `X19`]; (* 2888: adcs x21, x19, x17 *)
  [`X11`; `X20`]; (* 2889: stp x20, x11, [sp, #288] *)
  [`X24`]; (* 2890: adcs x12, x12, x24 *)
  []; (* 2891: mul x8, x23, x15 *)
  [`X7`]; (* 2892: adcs x3, x3, x7 *)
  [`X12`; `X21`]; (* 2893: stp x21, x12, [sp, #304] *)
  [`X2`; `X4`]; (* 2894: adcs x12, x2, x4 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2895: adc x19, xzr, xzr *)
  []; (* 2896: subs x21, x23, x16 *)
  []; (* 2897: umulh x2, x16, x6 *)
  [`X12`; `X3`]; (* 2898: stp x3, x12, [sp, #320] *)
  [`X21`]; (* 2899: cneg x3, x21, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2900: csetm x24, cc *)
  []; (* 2901: umulh x11, x13, x14 *)
  [`X16`]; (* 2902: subs x21, x13, x16 *)
  []; (* 2903: eor x7, x8, x9 *)
  [`X21`]; (* 2904: cneg x17, x21, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2905: csetm x16, cc *)
  []; (* 2906: subs x21, x6, x15 *)
  [`X21`]; (* 2907: cneg x22, x21, cc *)
  [`CF`; `NF`; `ZF`; `VF`; `X24`]; (* 2908: cinv x21, x24, cc *)
  [`X13`; `X23`]; (* 2909: subs x20, x23, x13 *)
  []; (* 2910: umulh x12, x3, x22 *)
  [`X20`]; (* 2911: cneg x23, x20, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2912: csetm x24, cc *)
  [`X15`]; (* 2913: subs x20, x14, x15 *)
  []; (* 2914: cinv x24, x24, cc *)
  [`X3`]; (* 2915: mul x22, x3, x22 *)
  [`CF`; `NF`; `ZF`; `VF`; `X20`]; (* 2916: cneg x3, x20, cc *)
  [`X14`; `X6`]; (* 2917: subs x13, x6, x14 *)
  [`X13`]; (* 2918: cneg x20, x13, cc *)
  [`CF`; `NF`; `ZF`; `VF`; `X16`]; (* 2919: cinv x15, x16, cc *)
  [`X10`; `X5`]; (* 2920: adds x13, x5, x10 *)
  []; (* 2921: mul x4, x23, x3 *)
  [`X1`]; (* 2922: adcs x11, x11, x1 *)
  [`CF`; `NF`; `ZF`; `VF`; `X2`]; (* 2923: adc x14, x2, xzr *)
  []; (* 2924: adds x5, x13, x8 *)
  []; (* 2925: adcs x16, x11, x13 *)
  [`X3`]; (* 2926: umulh x23, x23, x3 *)
  []; (* 2927: adcs x3, x14, x11 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2928: adc x1, x14, xzr *)
  [`X16`; `X8`]; (* 2929: adds x10, x16, x8 *)
  [`X13`; `X3`]; (* 2930: adcs x6, x3, x13 *)
  [`X1`; `X11`]; (* 2931: adcs x8, x1, x11 *)
  []; (* 2932: umulh x13, x17, x20 *)
  [`X4`]; (* 2933: eor x1, x4, x24 *)
  [`CF`; `NF`; `ZF`; `VF`; `X14`]; (* 2934: adc x4, x14, xzr *)
  []; (* 2935: cmn x24, #0x1 *)
  [`X5`]; (* 2936: adcs x1, x5, x1 *)
  [`X23`]; (* 2937: eor x16, x23, x24 *)
  [`X1`]; (* 2938: eor x11, x1, x9 *)
  [`X10`; `X16`]; (* 2939: adcs x23, x10, x16 *)
  [`X22`]; (* 2940: eor x2, x22, x21 *)
  [`X6`]; (* 2941: adcs x3, x6, x24 *)
  [`X17`; `X20`]; (* 2942: mul x14, x17, x20 *)
  [`X13`]; (* 2943: eor x17, x13, x15 *)
  [`X8`]; (* 2944: adcs x13, x8, x24 *)
  [`CF`; `NF`; `ZF`; `VF`; `X24`; `X4`]; (* 2945: adc x8, x4, x24 *)
  []; (* 2946: cmn x21, #0x1 *)
  [`X2`; `X23`]; (* 2947: adcs x6, x23, x2 *)
  []; (* 2948: mov x16, #0xfffffffffffffffe *)
  [`X12`]; (* 2949: eor x20, x12, x21 *)
  [`X3`]; (* 2950: adcs x20, x3, x20 *)
  [`X14`]; (* 2951: eor x23, x14, x15 *)
  [`X13`]; (* 2952: adcs x2, x13, x21 *)
  [`CF`; `NF`; `ZF`; `VF`; `X21`]; (* 2953: adc x8, x8, x21 *)
  []; (* 2954: cmn x15, #0x1 *)
  []; (* 2955: ldp x5, x4, [sp, #288] *)
  []; (* 2956: ldp x21, x12, [sp, #304] *)
  [`X20`; `X23`]; (* 2957: adcs x22, x20, x23 *)
  [`X22`]; (* 2958: eor x23, x22, x9 *)
  [`X2`]; (* 2959: adcs x17, x2, x17 *)
  [`CF`; `NF`; `ZF`; `VF`; `X15`; `X8`]; (* 2960: adc x22, x8, x15 *)
  []; (* 2961: cmn x9, #0x1 *)
  [`X7`]; (* 2962: adcs x15, x7, x5 *)
  []; (* 2963: ldp x10, x14, [sp, #320] *)
  [`X6`]; (* 2964: eor x1, x6, x9 *)
  []; (* 2965: lsl x2, x15, #32 *)
  [`X11`]; (* 2966: adcs x8, x11, x4 *)
  [`X1`]; (* 2967: adcs x13, x1, x21 *)
  [`X22`]; (* 2968: eor x1, x22, x9 *)
  [`X23`]; (* 2969: adcs x24, x23, x12 *)
  [`X17`]; (* 2970: eor x11, x17, x9 *)
  [`X11`]; (* 2971: adcs x23, x11, x10 *)
  [`X1`]; (* 2972: adcs x7, x1, x14 *)
  []; (* 2973: adcs x17, x9, x19 *)
  []; (* 2974: adcs x20, x9, xzr *)
  [`X15`; `X2`]; (* 2975: add x1, x2, x15 *)
  []; (* 2976: lsr x3, x1, #32 *)
  []; (* 2977: adcs x11, x9, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2978: adc x9, x9, xzr *)
  []; (* 2979: subs x3, x3, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2980: sbc x6, x1, xzr *)
  [`X5`]; (* 2981: adds x24, x24, x5 *)
  [`X23`]; (* 2982: adcs x4, x23, x4 *)
  []; (* 2983: extr x3, x6, x3, #32 *)
  []; (* 2984: lsr x6, x6, #32 *)
  [`X7`]; (* 2985: adcs x21, x7, x21 *)
  [`X12`; `X17`]; (* 2986: adcs x15, x17, x12 *)
  [`X10`; `X20`]; (* 2987: adcs x7, x20, x10 *)
  [`X11`; `X14`]; (* 2988: adcs x20, x11, x14 *)
  []; (* 2989: mov x14, #0xffffffff *)
  [`CF`; `NF`; `ZF`; `VF`; `X19`; `X9`]; (* 2990: adc x22, x9, x19 *)
  [`X6`]; (* 2991: adds x12, x6, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 2992: adc x10, xzr, xzr *)
  [`X8`]; (* 2993: subs x3, x8, x3 *)
  [`X13`]; (* 2994: sbcs x12, x13, x12 *)
  []; (* 2995: lsl x9, x3, #32 *)
  [`X9`]; (* 2996: add x3, x9, x3 *)
  [`X24`]; (* 2997: sbcs x10, x24, x10 *)
  [`X4`]; (* 2998: sbcs x24, x4, xzr *)
  []; (* 2999: lsr x9, x3, #32 *)
  []; (* 3000: sbcs x21, x21, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3001: sbc x1, x1, xzr *)
  []; (* 3002: subs x9, x9, x3 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3003: sbc x13, x3, xzr *)
  []; (* 3004: extr x9, x13, x9, #32 *)
  []; (* 3005: lsr x13, x13, #32 *)
  []; (* 3006: adds x13, x13, x3 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3007: adc x6, xzr, xzr *)
  [`X9`]; (* 3008: subs x12, x12, x9 *)
  [`X10`; `X13`]; (* 3009: sbcs x17, x10, x13 *)
  []; (* 3010: lsl x2, x12, #32 *)
  [`X24`; `X6`]; (* 3011: sbcs x10, x24, x6 *)
  [`X12`; `X2`]; (* 3012: add x9, x2, x12 *)
  [`X21`]; (* 3013: sbcs x6, x21, xzr *)
  []; (* 3014: lsr x5, x9, #32 *)
  [`X1`]; (* 3015: sbcs x21, x1, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X3`]; (* 3016: sbc x13, x3, xzr *)
  [`X5`]; (* 3017: subs x8, x5, x9 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3018: sbc x19, x9, xzr *)
  []; (* 3019: lsr x12, x19, #32 *)
  [`X19`; `X8`]; (* 3020: extr x3, x19, x8, #32 *)
  [`X12`]; (* 3021: adds x8, x12, x9 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3022: adc x1, xzr, xzr *)
  [`X17`; `X3`]; (* 3023: subs x2, x17, x3 *)
  [`X10`; `X8`]; (* 3024: sbcs x12, x10, x8 *)
  [`X1`; `X6`]; (* 3025: sbcs x5, x6, x1 *)
  [`X21`]; (* 3026: sbcs x3, x21, xzr *)
  [`X13`]; (* 3027: sbcs x19, x13, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X9`]; (* 3028: sbc x24, x9, xzr *)
  [`X15`; `X3`]; (* 3029: adds x23, x15, x3 *)
  [`X19`; `X7`]; (* 3030: adcs x8, x7, x19 *)
  [`X20`; `X24`]; (* 3031: adcs x11, x20, x24 *)
  [`CF`; `NF`; `ZF`; `VF`; `X22`]; (* 3032: adc x9, x22, xzr *)
  [`X9`]; (* 3033: add x24, x9, #0x1 *)
  []; (* 3034: lsl x7, x24, #32 *)
  []; (* 3035: subs x21, x24, x7 *)
  [`CF`; `NF`; `ZF`; `VF`; `X7`]; (* 3036: sbc x10, x7, xzr *)
  [`X2`; `X21`]; (* 3037: adds x6, x2, x21 *)
  [`X10`; `X12`]; (* 3038: adcs x7, x12, x10 *)
  [`X5`]; (* 3039: adcs x24, x5, x24 *)
  [`X23`]; (* 3040: adcs x13, x23, xzr *)
  []; (* 3041: adcs x8, x8, xzr *)
  [`X11`]; (* 3042: adcs x15, x11, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3043: csetm x23, cc *)
  [`X16`]; (* 3044: and x11, x16, x23 *)
  [`X14`]; (* 3045: and x20, x14, x23 *)
  [`X6`]; (* 3046: adds x22, x6, x20 *)
  [`X20`]; (* 3047: eor x3, x20, x23 *)
  [`X3`; `X7`]; (* 3048: adcs x5, x7, x3 *)
  [`X11`; `X24`]; (* 3049: adcs x2, x24, x11 *)
  [`X22`; `X5`]; (* 3050: stp x22, x5, [sp, #288] *)
  [`X13`]; (* 3051: adcs x11, x13, x23 *)
  [`X8`]; (* 3052: adcs x12, x8, x23 *)
  []; (* 3053: stp x2, x11, [sp, #304] *)
  [`CF`; `NF`; `ZF`; `VF`; `X15`; `X23`]; (* 3054: adc x13, x15, x23 *)
  []; (* 3055: stp x12, x13, [sp, #320] *)
  []; (* 3056: ldp x5, x6, [sp, #96] *)
  []; (* 3057: ldp x4, x3, [sp, #192] *)
  [`X4`]; (* 3058: subs x5, x5, x4 *)
  [`X3`]; (* 3059: sbcs x6, x6, x3 *)
  []; (* 3060: ldp x7, x8, [sp, #112] *)
  []; (* 3061: ldp x4, x3, [sp, #208] *)
  [`X4`]; (* 3062: sbcs x7, x7, x4 *)
  [`X3`]; (* 3063: sbcs x8, x8, x3 *)
  []; (* 3064: ldp x9, x10, [sp, #128] *)
  []; (* 3065: ldp x4, x3, [sp, #224] *)
  [`X4`]; (* 3066: sbcs x9, x9, x4 *)
  [`X3`]; (* 3067: sbcs x10, x10, x3 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3068: csetm x3, cc *)
  []; (* 3069: mov x4, #0xffffffff *)
  []; (* 3070: and x4, x4, x3 *)
  []; (* 3071: adds x5, x5, x4 *)
  []; (* 3072: eor x4, x4, x3 *)
  [`X4`]; (* 3073: adcs x6, x6, x4 *)
  []; (* 3074: mov x4, #0xfffffffffffffffe *)
  []; (* 3075: and x4, x4, x3 *)
  [`X4`]; (* 3076: adcs x7, x7, x4 *)
  []; (* 3077: adcs x8, x8, x3 *)
  []; (* 3078: adcs x9, x9, x3 *)
  [`CF`; `NF`; `ZF`; `VF`; `X3`]; (* 3079: adc x10, x10, x3 *)
  [`X5`; `X6`]; (* 3080: stp x5, x6, [sp, #240] *)
  [`X7`; `X8`]; (* 3081: stp x7, x8, [sp, #256] *)
  [`X10`; `X9`]; (* 3082: stp x9, x10, [sp, #272] *)
  []; (* 3083: ldp x5, x6, [sp, #48] *)
  []; (* 3084: ldp x4, x3, [sp, #288] *)
  [`X4`]; (* 3085: subs x5, x5, x4 *)
  [`X3`]; (* 3086: sbcs x6, x6, x3 *)
  []; (* 3087: ldp x7, x8, [sp, #64] *)
  [`X2`]; (* 3088: sbcs x7, x7, x2 *)
  [`X11`]; (* 3089: sbcs x8, x8, x11 *)
  []; (* 3090: ldp x9, x10, [sp, #80] *)
  [`X12`]; (* 3091: sbcs x9, x9, x12 *)
  [`X13`]; (* 3092: sbcs x10, x10, x13 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3093: csetm x3, cc *)
  []; (* 3094: mov x4, #0xffffffff *)
  []; (* 3095: and x4, x4, x3 *)
  []; (* 3096: adds x5, x5, x4 *)
  []; (* 3097: eor x4, x4, x3 *)
  [`X4`]; (* 3098: adcs x6, x6, x4 *)
  []; (* 3099: mov x4, #0xfffffffffffffffe *)
  []; (* 3100: and x4, x4, x3 *)
  [`X4`]; (* 3101: adcs x7, x7, x4 *)
  []; (* 3102: adcs x8, x8, x3 *)
  []; (* 3103: adcs x9, x9, x3 *)
  [`CF`; `NF`; `ZF`; `VF`; `X3`]; (* 3104: adc x10, x10, x3 *)
  [`X5`; `X6`]; (* 3105: stp x5, x6, [sp, #48] *)
  [`X7`; `X8`]; (* 3106: stp x7, x8, [sp, #64] *)
  [`X10`; `X9`]; (* 3107: stp x9, x10, [sp, #80] *)
  []; (* 3108: ldr q1, [sp, #240] *)
  []; (* 3109: ldp x9, x2, [sp, #240] *)
  []; (* 3110: ldr q0, [sp, #240] *)
  []; (* 3111: ldp x4, x6, [sp, #256] *)
  []; (* 3112: rev64 v21.4s, v1.4s *)
  []; (* 3113: uzp2 v28.4s, v1.4s, v1.4s *)
  []; (* 3114: umulh x7, x9, x2 *)
  [`Q1`]; (* 3115: xtn v17.2s, v1.2d *)
  [`Q21`]; (* 3116: mul v27.4s, v21.4s, v0.4s *)
  []; (* 3117: ldr q20, [sp, #272] *)
  []; (* 3118: xtn v30.2s, v0.2d *)
  []; (* 3119: ldr q1, [sp, #272] *)
  [`Q0`]; (* 3120: uzp2 v31.4s, v0.4s, v0.4s *)
  []; (* 3121: ldp x5, x10, [sp, #272] *)
  []; (* 3122: umulh x8, x9, x4 *)
  [`Q27`]; (* 3123: uaddlp v3.2d, v27.4s *)
  []; (* 3124: umull v16.2d, v30.2s, v17.2s *)
  []; (* 3125: mul x16, x9, x4 *)
  []; (* 3126: umull v27.2d, v30.2s, v28.2s *)
  []; (* 3127: shrn v0.2s, v20.2d, #32 *)
  [`Q20`]; (* 3128: xtn v7.2s, v20.2d *)
  [`Q3`]; (* 3129: shl v20.2d, v3.2d, #32 *)
  [`Q28`]; (* 3130: umull v3.2d, v31.2s, v28.2s *)
  []; (* 3131: mul x3, x2, x4 *)
  [`Q30`]; (* 3132: umlal v20.2d, v30.2s, v17.2s *)
  [`Q0`]; (* 3133: umull v22.2d, v7.2s, v0.2s *)
  [`Q16`]; (* 3134: usra v27.2d, v16.2d, #32 *)
  []; (* 3135: umulh x11, x2, x4 *)
  []; (* 3136: movi v21.2d, #0xffffffff *)
  []; (* 3137: uzp2 v28.4s, v1.4s, v1.4s *)
  [`X16`; `X7`]; (* 3138: adds x15, x16, x7 *)
  [`Q21`]; (* 3139: and v5.16b, v27.16b, v21.16b *)
  [`X8`]; (* 3140: adcs x3, x3, x8 *)
  [`Q27`]; (* 3141: usra v3.2d, v27.2d, #32 *)
  []; (* 3142: dup v29.2d, x6 *)
  [`CF`; `NF`; `ZF`; `VF`; `X11`]; (* 3143: adcs x16, x11, xzr *)
  []; (* 3144: mov x14, v20.d[0] *)
  [`Q17`; `Q31`]; (* 3145: umlal v5.2d, v31.2s, v17.2s *)
  []; (* 3146: mul x8, x9, x2 *)
  [`Q20`]; (* 3147: mov x7, v20.d[1] *)
  [`Q22`]; (* 3148: shl v19.2d, v22.2d, #33 *)
  []; (* 3149: xtn v25.2s, v29.2d *)
  []; (* 3150: rev64 v31.4s, v1.4s *)
  []; (* 3151: lsl x13, x14, #32 *)
  []; (* 3152: uzp2 v6.4s, v29.4s, v29.4s *)
  [`Q7`]; (* 3153: umlal v19.2d, v7.2s, v7.2s *)
  [`Q5`]; (* 3154: usra v3.2d, v5.2d, #32 *)
  [`X8`]; (* 3155: adds x1, x8, x8 *)
  []; (* 3156: umulh x8, x4, x4 *)
  [`X13`; `X14`]; (* 3157: add x12, x13, x14 *)
  [`Q29`; `Q31`]; (* 3158: mul v17.4s, v31.4s, v29.4s *)
  [`Q1`]; (* 3159: xtn v4.2s, v1.2d *)
  [`X15`]; (* 3160: adcs x14, x15, x15 *)
  []; (* 3161: lsr x13, x12, #32 *)
  [`X3`]; (* 3162: adcs x15, x3, x3 *)
  []; (* 3163: umull v31.2d, v25.2s, v28.2s *)
  [`X16`]; (* 3164: adcs x11, x16, x16 *)
  []; (* 3165: umull v21.2d, v25.2s, v4.2s *)
  []; (* 3166: mov x17, v3.d[0] *)
  [`Q28`]; (* 3167: umull v18.2d, v6.2s, v28.2s *)
  [`CF`; `NF`; `ZF`; `VF`; `X8`]; (* 3168: adc x16, x8, xzr *)
  [`Q17`]; (* 3169: uaddlp v16.2d, v17.4s *)
  []; (* 3170: movi v1.2d, #0xffffffff *)
  []; (* 3171: subs x13, x13, x12 *)
  [`Q21`]; (* 3172: usra v31.2d, v21.2d, #32 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3173: sbc x8, x12, xzr *)
  [`X1`]; (* 3174: adds x17, x17, x1 *)
  []; (* 3175: mul x1, x4, x4 *)
  [`Q16`]; (* 3176: shl v28.2d, v16.2d, #32 *)
  [`Q3`]; (* 3177: mov x3, v3.d[1] *)
  [`X7`]; (* 3178: adcs x14, x7, x14 *)
  [`X13`]; (* 3179: extr x7, x8, x13, #32 *)
  [`X15`; `X3`]; (* 3180: adcs x13, x3, x15 *)
  [`Q1`]; (* 3181: and v3.16b, v31.16b, v1.16b *)
  [`X1`]; (* 3182: adcs x11, x1, x11 *)
  [`X8`]; (* 3183: lsr x1, x8, #32 *)
  [`Q6`]; (* 3184: umlal v3.2d, v6.2s, v4.2s *)
  [`Q31`]; (* 3185: usra v18.2d, v31.2d, #32 *)
  [`CF`; `NF`; `ZF`; `VF`; `X16`]; (* 3186: adc x3, x16, xzr *)
  []; (* 3187: adds x1, x1, x12 *)
  [`Q25`; `Q4`]; (* 3188: umlal v28.2d, v25.2s, v4.2s *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3189: adc x16, xzr, xzr *)
  [`X17`; `X7`]; (* 3190: subs x15, x17, x7 *)
  [`X1`; `X14`]; (* 3191: sbcs x7, x14, x1 *)
  []; (* 3192: lsl x1, x15, #32 *)
  [`X13`]; (* 3193: sbcs x16, x13, x16 *)
  [`X1`; `X15`]; (* 3194: add x8, x1, x15 *)
  [`Q3`]; (* 3195: usra v18.2d, v3.2d, #32 *)
  [`X11`]; (* 3196: sbcs x14, x11, xzr *)
  []; (* 3197: lsr x1, x8, #32 *)
  [`X3`]; (* 3198: sbcs x17, x3, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 3199: sbc x11, x12, xzr *)
  [`X1`]; (* 3200: subs x13, x1, x8 *)
  []; (* 3201: umulh x12, x4, x10 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3202: sbc x1, x8, xzr *)
  []; (* 3203: extr x13, x1, x13, #32 *)
  []; (* 3204: lsr x1, x1, #32 *)
  [`X1`]; (* 3205: adds x15, x1, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3206: adc x1, xzr, xzr *)
  [`X13`]; (* 3207: subs x7, x7, x13 *)
  [`X15`; `X16`]; (* 3208: sbcs x13, x16, x15 *)
  []; (* 3209: lsl x3, x7, #32 *)
  []; (* 3210: umulh x16, x2, x5 *)
  [`X1`; `X14`]; (* 3211: sbcs x15, x14, x1 *)
  [`X3`]; (* 3212: add x7, x3, x7 *)
  [`X17`]; (* 3213: sbcs x3, x17, xzr *)
  []; (* 3214: lsr x1, x7, #32 *)
  [`X11`]; (* 3215: sbcs x14, x11, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X8`]; (* 3216: sbc x11, x8, xzr *)
  [`X1`]; (* 3217: subs x8, x1, x7 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3218: sbc x1, x7, xzr *)
  []; (* 3219: extr x8, x1, x8, #32 *)
  []; (* 3220: lsr x1, x1, #32 *)
  []; (* 3221: adds x1, x1, x7 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3222: adc x17, xzr, xzr *)
  [`X8`]; (* 3223: subs x13, x13, x8 *)
  []; (* 3224: umulh x8, x9, x6 *)
  [`X15`]; (* 3225: sbcs x1, x15, x1 *)
  [`X17`; `X3`]; (* 3226: sbcs x15, x3, x17 *)
  [`X14`]; (* 3227: sbcs x3, x14, xzr *)
  []; (* 3228: mul x17, x2, x5 *)
  []; (* 3229: sbcs x11, x11, xzr *)
  [`X1`; `X13`]; (* 3230: stp x13, x1, [sp, #144] *)
  [`CF`; `NF`; `ZF`; `VF`; `X7`]; (* 3231: sbc x14, x7, xzr *)
  []; (* 3232: mul x7, x4, x10 *)
  []; (* 3233: subs x1, x9, x2 *)
  [`X15`; `X3`]; (* 3234: stp x15, x3, [sp, #160] *)
  []; (* 3235: csetm x15, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3236: cneg x1, x1, cc *)
  [`X11`; `X14`]; (* 3237: stp x11, x14, [sp, #176] *)
  []; (* 3238: mul x14, x9, x6 *)
  [`X8`]; (* 3239: adds x17, x8, x17 *)
  [`X16`]; (* 3240: adcs x7, x16, x7 *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 3241: adc x13, x12, xzr *)
  []; (* 3242: subs x12, x5, x6 *)
  [`X12`]; (* 3243: cneg x3, x12, cc *)
  [`CF`; `NF`; `ZF`; `VF`; `X15`]; (* 3244: cinv x16, x15, cc *)
  []; (* 3245: mul x8, x1, x3 *)
  [`X3`]; (* 3246: umulh x1, x1, x3 *)
  [`X8`]; (* 3247: eor x12, x8, x16 *)
  []; (* 3248: adds x11, x17, x14 *)
  []; (* 3249: adcs x3, x7, x17 *)
  []; (* 3250: adcs x15, x13, x7 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3251: adc x8, x13, xzr *)
  []; (* 3252: adds x3, x3, x14 *)
  [`X17`]; (* 3253: adcs x15, x15, x17 *)
  [`X7`; `X8`]; (* 3254: adcs x17, x8, x7 *)
  []; (* 3255: eor x1, x1, x16 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3256: adc x13, x13, xzr *)
  []; (* 3257: subs x9, x9, x4 *)
  []; (* 3258: csetm x8, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3259: cneg x9, x9, cc *)
  [`X2`]; (* 3260: subs x4, x2, x4 *)
  []; (* 3261: cneg x4, x4, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3262: csetm x7, cc *)
  []; (* 3263: subs x2, x10, x6 *)
  []; (* 3264: cinv x8, x8, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3265: cneg x2, x2, cc *)
  []; (* 3266: cmn x16, #0x1 *)
  [`X12`]; (* 3267: adcs x11, x11, x12 *)
  []; (* 3268: mul x12, x9, x2 *)
  [`X1`]; (* 3269: adcs x3, x3, x1 *)
  []; (* 3270: adcs x15, x15, x16 *)
  [`X2`]; (* 3271: umulh x9, x9, x2 *)
  []; (* 3272: adcs x17, x17, x16 *)
  [`CF`; `NF`; `ZF`; `VF`; `X16`]; (* 3273: adc x13, x13, x16 *)
  []; (* 3274: subs x1, x10, x5 *)
  [`X7`]; (* 3275: cinv x2, x7, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3276: cneg x1, x1, cc *)
  []; (* 3277: eor x9, x9, x8 *)
  []; (* 3278: cmn x8, #0x1 *)
  [`X12`]; (* 3279: eor x7, x12, x8 *)
  []; (* 3280: mul x12, x4, x1 *)
  [`X7`]; (* 3281: adcs x3, x3, x7 *)
  [`X15`; `X9`]; (* 3282: adcs x7, x15, x9 *)
  [`X17`]; (* 3283: adcs x15, x17, x8 *)
  []; (* 3284: ldp x9, x17, [sp, #160] *)
  [`X1`]; (* 3285: umulh x4, x4, x1 *)
  [`CF`; `NF`; `ZF`; `VF`; `X13`]; (* 3286: adc x8, x13, x8 *)
  []; (* 3287: cmn x2, #0x1 *)
  [`X12`]; (* 3288: eor x1, x12, x2 *)
  [`X7`]; (* 3289: adcs x1, x7, x1 *)
  []; (* 3290: ldp x7, x16, [sp, #144] *)
  [`X4`]; (* 3291: eor x12, x4, x2 *)
  [`X12`; `X15`]; (* 3292: adcs x4, x15, x12 *)
  []; (* 3293: ldp x15, x12, [sp, #176] *)
  [`CF`; `NF`; `ZF`; `VF`; `X2`]; (* 3294: adc x8, x8, x2 *)
  [`X14`]; (* 3295: adds x13, x14, x14 *)
  []; (* 3296: umulh x14, x5, x10 *)
  [`X11`]; (* 3297: adcs x2, x11, x11 *)
  []; (* 3298: adcs x3, x3, x3 *)
  []; (* 3299: adcs x1, x1, x1 *)
  []; (* 3300: adcs x4, x4, x4 *)
  [`X8`]; (* 3301: adcs x11, x8, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3302: adc x8, xzr, xzr *)
  [`X7`]; (* 3303: adds x13, x13, x7 *)
  [`X16`]; (* 3304: adcs x2, x2, x16 *)
  []; (* 3305: mul x16, x5, x10 *)
  [`X9`]; (* 3306: adcs x3, x3, x9 *)
  [`X17`]; (* 3307: adcs x1, x1, x17 *)
  []; (* 3308: umulh x5, x5, x5 *)
  []; (* 3309: lsl x9, x13, #32 *)
  [`X13`]; (* 3310: add x9, x9, x13 *)
  [`X15`]; (* 3311: adcs x4, x4, x15 *)
  []; (* 3312: mov x13, v28.d[1] *)
  [`X11`; `X12`]; (* 3313: adcs x15, x11, x12 *)
  []; (* 3314: lsr x7, x9, #32 *)
  [`CF`; `NF`; `ZF`; `VF`; `X8`]; (* 3315: adc x11, x8, xzr *)
  []; (* 3316: subs x7, x7, x9 *)
  []; (* 3317: umulh x10, x10, x10 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3318: sbc x17, x9, xzr *)
  []; (* 3319: extr x7, x17, x7, #32 *)
  []; (* 3320: lsr x17, x17, #32 *)
  []; (* 3321: adds x17, x17, x9 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3322: adc x12, xzr, xzr *)
  [`X2`; `X7`]; (* 3323: subs x8, x2, x7 *)
  [`X3`]; (* 3324: sbcs x17, x3, x17 *)
  []; (* 3325: lsl x7, x8, #32 *)
  [`X1`; `X12`]; (* 3326: sbcs x2, x1, x12 *)
  [`X7`; `X8`]; (* 3327: add x3, x7, x8 *)
  [`X4`]; (* 3328: sbcs x12, x4, xzr *)
  []; (* 3329: lsr x1, x3, #32 *)
  [`X15`]; (* 3330: sbcs x7, x15, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X9`]; (* 3331: sbc x15, x9, xzr *)
  []; (* 3332: subs x1, x1, x3 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3333: sbc x4, x3, xzr *)
  []; (* 3334: lsr x9, x4, #32 *)
  [`X1`; `X4`]; (* 3335: extr x8, x4, x1, #32 *)
  []; (* 3336: adds x9, x9, x3 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3337: adc x4, xzr, xzr *)
  [`X17`; `X8`]; (* 3338: subs x1, x17, x8 *)
  []; (* 3339: lsl x17, x1, #32 *)
  [`X2`; `X9`]; (* 3340: sbcs x8, x2, x9 *)
  [`X12`; `X4`]; (* 3341: sbcs x9, x12, x4 *)
  [`X1`]; (* 3342: add x17, x17, x1 *)
  []; (* 3343: mov x1, v18.d[1] *)
  []; (* 3344: lsr x2, x17, #32 *)
  []; (* 3345: sbcs x7, x7, xzr *)
  [`Q18`]; (* 3346: mov x12, v18.d[0] *)
  []; (* 3347: sbcs x15, x15, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3348: sbc x3, x3, xzr *)
  [`X2`]; (* 3349: subs x4, x2, x17 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3350: sbc x2, x17, xzr *)
  [`X13`]; (* 3351: adds x12, x13, x12 *)
  [`X1`]; (* 3352: adcs x16, x16, x1 *)
  []; (* 3353: lsr x13, x2, #32 *)
  [`X2`; `X4`]; (* 3354: extr x1, x2, x4, #32 *)
  [`CF`; `NF`; `ZF`; `VF`; `X14`]; (* 3355: adc x2, x14, xzr *)
  [`X13`]; (* 3356: adds x4, x13, x17 *)
  []; (* 3357: mul x13, x6, x6 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3358: adc x14, xzr, xzr *)
  [`X8`]; (* 3359: subs x1, x8, x1 *)
  [`X9`]; (* 3360: sbcs x4, x9, x4 *)
  [`Q28`]; (* 3361: mov x9, v28.d[0] *)
  [`X14`]; (* 3362: sbcs x7, x7, x14 *)
  [`X15`]; (* 3363: sbcs x8, x15, xzr *)
  []; (* 3364: sbcs x3, x3, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X17`]; (* 3365: sbc x14, x17, xzr *)
  [`X9`]; (* 3366: adds x17, x9, x9 *)
  []; (* 3367: adcs x12, x12, x12 *)
  []; (* 3368: mov x15, v19.d[0] *)
  [`X16`]; (* 3369: adcs x9, x16, x16 *)
  []; (* 3370: umulh x6, x6, x6 *)
  [`X2`]; (* 3371: adcs x16, x2, x2 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3372: adc x2, xzr, xzr *)
  [`X8`]; (* 3373: adds x11, x11, x8 *)
  []; (* 3374: adcs x3, x3, xzr *)
  []; (* 3375: adcs x14, x14, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3376: adcs x8, xzr, xzr *)
  [`X1`]; (* 3377: adds x13, x1, x13 *)
  [`Q19`]; (* 3378: mov x1, v19.d[1] *)
  [`X4`]; (* 3379: adcs x6, x4, x6 *)
  []; (* 3380: mov x4, #0xffffffff *)
  [`X7`]; (* 3381: adcs x15, x7, x15 *)
  [`X11`; `X5`]; (* 3382: adcs x7, x11, x5 *)
  [`X3`]; (* 3383: adcs x1, x3, x1 *)
  [`X10`]; (* 3384: adcs x14, x14, x10 *)
  [`CF`; `NF`; `ZF`; `VF`; `X8`]; (* 3385: adc x11, x8, xzr *)
  [`X17`]; (* 3386: adds x6, x6, x17 *)
  [`X12`; `X15`]; (* 3387: adcs x8, x15, x12 *)
  [`X7`; `X9`]; (* 3388: adcs x3, x7, x9 *)
  [`X1`; `X16`]; (* 3389: adcs x15, x1, x16 *)
  []; (* 3390: mov x16, #0xffffffff00000001 *)
  [`X2`]; (* 3391: adcs x14, x14, x2 *)
  []; (* 3392: mov x2, #0x1 *)
  [`CF`; `NF`; `ZF`; `VF`; `X11`]; (* 3393: adc x17, x11, xzr *)
  []; (* 3394: cmn x13, x16 *)
  []; (* 3395: adcs xzr, x6, x4 *)
  []; (* 3396: adcs xzr, x8, x2 *)
  []; (* 3397: adcs xzr, x3, xzr *)
  []; (* 3398: adcs xzr, x15, xzr *)
  []; (* 3399: adcs xzr, x14, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X17`]; (* 3400: adc x1, x17, xzr *)
  [`X1`]; (* 3401: neg x9, x1 *)
  [`X16`]; (* 3402: and x1, x16, x9 *)
  [`X1`; `X13`]; (* 3403: adds x11, x13, x1 *)
  [`X4`]; (* 3404: and x13, x4, x9 *)
  [`X13`; `X6`]; (* 3405: adcs x5, x6, x13 *)
  [`X2`; `X9`]; (* 3406: and x1, x2, x9 *)
  [`X1`; `X8`]; (* 3407: adcs x7, x8, x1 *)
  [`X11`; `X5`]; (* 3408: stp x11, x5, [sp, #144] *)
  [`X3`]; (* 3409: adcs x11, x3, xzr *)
  [`X15`]; (* 3410: adcs x2, x15, xzr *)
  [`X11`; `X7`]; (* 3411: stp x7, x11, [sp, #160] *)
  [`CF`; `NF`; `ZF`; `VF`; `X14`]; (* 3412: adc x17, x14, xzr *)
  [`X17`; `X2`]; (* 3413: stp x2, x17, [sp, #176] *)
  []; (* 3414: mov x0, sp *)
  []; (* 3415: ldr q1, [sp, #48] *)
  []; (* 3416: ldp x9, x2, [sp, #48] *)
  []; (* 3417: ldr q0, [sp, #48] *)
  []; (* 3418: ldp x4, x6, [sp, #64] *)
  []; (* 3419: rev64 v21.4s, v1.4s *)
  []; (* 3420: uzp2 v28.4s, v1.4s, v1.4s *)
  []; (* 3421: umulh x7, x9, x2 *)
  [`Q1`]; (* 3422: xtn v17.2s, v1.2d *)
  [`Q21`]; (* 3423: mul v27.4s, v21.4s, v0.4s *)
  []; (* 3424: ldr q20, [sp, #80] *)
  []; (* 3425: xtn v30.2s, v0.2d *)
  []; (* 3426: ldr q1, [sp, #80] *)
  [`Q0`]; (* 3427: uzp2 v31.4s, v0.4s, v0.4s *)
  []; (* 3428: ldp x5, x10, [sp, #80] *)
  []; (* 3429: umulh x8, x9, x4 *)
  [`Q27`]; (* 3430: uaddlp v3.2d, v27.4s *)
  []; (* 3431: umull v16.2d, v30.2s, v17.2s *)
  []; (* 3432: mul x16, x9, x4 *)
  []; (* 3433: umull v27.2d, v30.2s, v28.2s *)
  []; (* 3434: shrn v0.2s, v20.2d, #32 *)
  [`Q20`]; (* 3435: xtn v7.2s, v20.2d *)
  [`Q3`]; (* 3436: shl v20.2d, v3.2d, #32 *)
  [`Q28`]; (* 3437: umull v3.2d, v31.2s, v28.2s *)
  []; (* 3438: mul x3, x2, x4 *)
  [`Q30`]; (* 3439: umlal v20.2d, v30.2s, v17.2s *)
  [`Q0`]; (* 3440: umull v22.2d, v7.2s, v0.2s *)
  [`Q16`]; (* 3441: usra v27.2d, v16.2d, #32 *)
  []; (* 3442: umulh x11, x2, x4 *)
  []; (* 3443: movi v21.2d, #0xffffffff *)
  []; (* 3444: uzp2 v28.4s, v1.4s, v1.4s *)
  [`X16`; `X7`]; (* 3445: adds x15, x16, x7 *)
  [`Q21`]; (* 3446: and v5.16b, v27.16b, v21.16b *)
  [`X8`]; (* 3447: adcs x3, x3, x8 *)
  [`Q27`]; (* 3448: usra v3.2d, v27.2d, #32 *)
  []; (* 3449: dup v29.2d, x6 *)
  [`CF`; `NF`; `ZF`; `VF`; `X11`]; (* 3450: adcs x16, x11, xzr *)
  []; (* 3451: mov x14, v20.d[0] *)
  [`Q17`; `Q31`]; (* 3452: umlal v5.2d, v31.2s, v17.2s *)
  []; (* 3453: mul x8, x9, x2 *)
  [`Q20`]; (* 3454: mov x7, v20.d[1] *)
  [`Q22`]; (* 3455: shl v19.2d, v22.2d, #33 *)
  []; (* 3456: xtn v25.2s, v29.2d *)
  []; (* 3457: rev64 v31.4s, v1.4s *)
  []; (* 3458: lsl x13, x14, #32 *)
  []; (* 3459: uzp2 v6.4s, v29.4s, v29.4s *)
  [`Q7`]; (* 3460: umlal v19.2d, v7.2s, v7.2s *)
  [`Q5`]; (* 3461: usra v3.2d, v5.2d, #32 *)
  [`X8`]; (* 3462: adds x1, x8, x8 *)
  []; (* 3463: umulh x8, x4, x4 *)
  [`X13`; `X14`]; (* 3464: add x12, x13, x14 *)
  [`Q29`; `Q31`]; (* 3465: mul v17.4s, v31.4s, v29.4s *)
  [`Q1`]; (* 3466: xtn v4.2s, v1.2d *)
  [`X15`]; (* 3467: adcs x14, x15, x15 *)
  []; (* 3468: lsr x13, x12, #32 *)
  [`X3`]; (* 3469: adcs x15, x3, x3 *)
  []; (* 3470: umull v31.2d, v25.2s, v28.2s *)
  [`X16`]; (* 3471: adcs x11, x16, x16 *)
  []; (* 3472: umull v21.2d, v25.2s, v4.2s *)
  []; (* 3473: mov x17, v3.d[0] *)
  [`Q28`]; (* 3474: umull v18.2d, v6.2s, v28.2s *)
  [`CF`; `NF`; `ZF`; `VF`; `X8`]; (* 3475: adc x16, x8, xzr *)
  [`Q17`]; (* 3476: uaddlp v16.2d, v17.4s *)
  []; (* 3477: movi v1.2d, #0xffffffff *)
  []; (* 3478: subs x13, x13, x12 *)
  [`Q21`]; (* 3479: usra v31.2d, v21.2d, #32 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3480: sbc x8, x12, xzr *)
  [`X1`]; (* 3481: adds x17, x17, x1 *)
  []; (* 3482: mul x1, x4, x4 *)
  [`Q16`]; (* 3483: shl v28.2d, v16.2d, #32 *)
  [`Q3`]; (* 3484: mov x3, v3.d[1] *)
  [`X7`]; (* 3485: adcs x14, x7, x14 *)
  [`X13`]; (* 3486: extr x7, x8, x13, #32 *)
  [`X15`; `X3`]; (* 3487: adcs x13, x3, x15 *)
  [`Q1`]; (* 3488: and v3.16b, v31.16b, v1.16b *)
  [`X1`]; (* 3489: adcs x11, x1, x11 *)
  [`X8`]; (* 3490: lsr x1, x8, #32 *)
  [`Q6`]; (* 3491: umlal v3.2d, v6.2s, v4.2s *)
  [`Q31`]; (* 3492: usra v18.2d, v31.2d, #32 *)
  [`CF`; `NF`; `ZF`; `VF`; `X16`]; (* 3493: adc x3, x16, xzr *)
  []; (* 3494: adds x1, x1, x12 *)
  [`Q25`; `Q4`]; (* 3495: umlal v28.2d, v25.2s, v4.2s *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3496: adc x16, xzr, xzr *)
  [`X17`; `X7`]; (* 3497: subs x15, x17, x7 *)
  [`X1`; `X14`]; (* 3498: sbcs x7, x14, x1 *)
  []; (* 3499: lsl x1, x15, #32 *)
  [`X13`]; (* 3500: sbcs x16, x13, x16 *)
  [`X1`; `X15`]; (* 3501: add x8, x1, x15 *)
  [`Q3`]; (* 3502: usra v18.2d, v3.2d, #32 *)
  [`X11`]; (* 3503: sbcs x14, x11, xzr *)
  []; (* 3504: lsr x1, x8, #32 *)
  [`X3`]; (* 3505: sbcs x17, x3, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 3506: sbc x11, x12, xzr *)
  [`X1`]; (* 3507: subs x13, x1, x8 *)
  []; (* 3508: umulh x12, x4, x10 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3509: sbc x1, x8, xzr *)
  []; (* 3510: extr x13, x1, x13, #32 *)
  []; (* 3511: lsr x1, x1, #32 *)
  [`X1`]; (* 3512: adds x15, x1, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3513: adc x1, xzr, xzr *)
  [`X13`]; (* 3514: subs x7, x7, x13 *)
  [`X15`; `X16`]; (* 3515: sbcs x13, x16, x15 *)
  []; (* 3516: lsl x3, x7, #32 *)
  []; (* 3517: umulh x16, x2, x5 *)
  [`X1`; `X14`]; (* 3518: sbcs x15, x14, x1 *)
  [`X3`]; (* 3519: add x7, x3, x7 *)
  [`X17`]; (* 3520: sbcs x3, x17, xzr *)
  []; (* 3521: lsr x1, x7, #32 *)
  [`X11`]; (* 3522: sbcs x14, x11, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X8`]; (* 3523: sbc x11, x8, xzr *)
  [`X1`]; (* 3524: subs x8, x1, x7 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3525: sbc x1, x7, xzr *)
  []; (* 3526: extr x8, x1, x8, #32 *)
  []; (* 3527: lsr x1, x1, #32 *)
  []; (* 3528: adds x1, x1, x7 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3529: adc x17, xzr, xzr *)
  [`X8`]; (* 3530: subs x13, x13, x8 *)
  []; (* 3531: umulh x8, x9, x6 *)
  [`X15`]; (* 3532: sbcs x1, x15, x1 *)
  [`X17`; `X3`]; (* 3533: sbcs x15, x3, x17 *)
  [`X14`]; (* 3534: sbcs x3, x14, xzr *)
  []; (* 3535: mul x17, x2, x5 *)
  []; (* 3536: sbcs x11, x11, xzr *)
  [`X1`; `X13`]; (* 3537: stp x13, x1, [x0] *)
  [`CF`; `NF`; `ZF`; `VF`; `X7`]; (* 3538: sbc x14, x7, xzr *)
  []; (* 3539: mul x7, x4, x10 *)
  []; (* 3540: subs x1, x9, x2 *)
  [`X15`; `X3`]; (* 3541: stp x15, x3, [x0, #16] *)
  []; (* 3542: csetm x15, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3543: cneg x1, x1, cc *)
  [`X11`; `X14`]; (* 3544: stp x11, x14, [x0, #32] *)
  []; (* 3545: mul x14, x9, x6 *)
  [`X8`]; (* 3546: adds x17, x8, x17 *)
  [`X16`]; (* 3547: adcs x7, x16, x7 *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 3548: adc x13, x12, xzr *)
  []; (* 3549: subs x12, x5, x6 *)
  [`X12`]; (* 3550: cneg x3, x12, cc *)
  [`CF`; `NF`; `ZF`; `VF`; `X15`]; (* 3551: cinv x16, x15, cc *)
  []; (* 3552: mul x8, x1, x3 *)
  [`X3`]; (* 3553: umulh x1, x1, x3 *)
  [`X8`]; (* 3554: eor x12, x8, x16 *)
  []; (* 3555: adds x11, x17, x14 *)
  []; (* 3556: adcs x3, x7, x17 *)
  []; (* 3557: adcs x15, x13, x7 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3558: adc x8, x13, xzr *)
  []; (* 3559: adds x3, x3, x14 *)
  [`X17`]; (* 3560: adcs x15, x15, x17 *)
  [`X7`; `X8`]; (* 3561: adcs x17, x8, x7 *)
  []; (* 3562: eor x1, x1, x16 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3563: adc x13, x13, xzr *)
  []; (* 3564: subs x9, x9, x4 *)
  []; (* 3565: csetm x8, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3566: cneg x9, x9, cc *)
  [`X2`]; (* 3567: subs x4, x2, x4 *)
  []; (* 3568: cneg x4, x4, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3569: csetm x7, cc *)
  []; (* 3570: subs x2, x10, x6 *)
  []; (* 3571: cinv x8, x8, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3572: cneg x2, x2, cc *)
  []; (* 3573: cmn x16, #0x1 *)
  [`X12`]; (* 3574: adcs x11, x11, x12 *)
  []; (* 3575: mul x12, x9, x2 *)
  [`X1`]; (* 3576: adcs x3, x3, x1 *)
  []; (* 3577: adcs x15, x15, x16 *)
  [`X2`]; (* 3578: umulh x9, x9, x2 *)
  []; (* 3579: adcs x17, x17, x16 *)
  [`CF`; `NF`; `ZF`; `VF`; `X16`]; (* 3580: adc x13, x13, x16 *)
  []; (* 3581: subs x1, x10, x5 *)
  [`X7`]; (* 3582: cinv x2, x7, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3583: cneg x1, x1, cc *)
  []; (* 3584: eor x9, x9, x8 *)
  []; (* 3585: cmn x8, #0x1 *)
  [`X12`]; (* 3586: eor x7, x12, x8 *)
  []; (* 3587: mul x12, x4, x1 *)
  [`X7`]; (* 3588: adcs x3, x3, x7 *)
  [`X15`; `X9`]; (* 3589: adcs x7, x15, x9 *)
  [`X17`]; (* 3590: adcs x15, x17, x8 *)
  []; (* 3591: ldp x9, x17, [x0, #16] *)
  [`X1`]; (* 3592: umulh x4, x4, x1 *)
  [`CF`; `NF`; `ZF`; `VF`; `X13`]; (* 3593: adc x8, x13, x8 *)
  []; (* 3594: cmn x2, #0x1 *)
  [`X12`]; (* 3595: eor x1, x12, x2 *)
  [`X7`]; (* 3596: adcs x1, x7, x1 *)
  []; (* 3597: ldp x7, x16, [x0] *)
  [`X4`]; (* 3598: eor x12, x4, x2 *)
  [`X12`; `X15`]; (* 3599: adcs x4, x15, x12 *)
  []; (* 3600: ldp x15, x12, [x0, #32] *)
  [`CF`; `NF`; `ZF`; `VF`; `X2`]; (* 3601: adc x8, x8, x2 *)
  [`X14`]; (* 3602: adds x13, x14, x14 *)
  []; (* 3603: umulh x14, x5, x10 *)
  [`X11`]; (* 3604: adcs x2, x11, x11 *)
  []; (* 3605: adcs x3, x3, x3 *)
  []; (* 3606: adcs x1, x1, x1 *)
  []; (* 3607: adcs x4, x4, x4 *)
  [`X8`]; (* 3608: adcs x11, x8, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3609: adc x8, xzr, xzr *)
  [`X7`]; (* 3610: adds x13, x13, x7 *)
  [`X16`]; (* 3611: adcs x2, x2, x16 *)
  []; (* 3612: mul x16, x5, x10 *)
  [`X9`]; (* 3613: adcs x3, x3, x9 *)
  [`X17`]; (* 3614: adcs x1, x1, x17 *)
  []; (* 3615: umulh x5, x5, x5 *)
  []; (* 3616: lsl x9, x13, #32 *)
  [`X13`]; (* 3617: add x9, x9, x13 *)
  [`X15`]; (* 3618: adcs x4, x4, x15 *)
  []; (* 3619: mov x13, v28.d[1] *)
  [`X11`; `X12`]; (* 3620: adcs x15, x11, x12 *)
  []; (* 3621: lsr x7, x9, #32 *)
  [`CF`; `NF`; `ZF`; `VF`; `X8`]; (* 3622: adc x11, x8, xzr *)
  []; (* 3623: subs x7, x7, x9 *)
  []; (* 3624: umulh x10, x10, x10 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3625: sbc x17, x9, xzr *)
  []; (* 3626: extr x7, x17, x7, #32 *)
  []; (* 3627: lsr x17, x17, #32 *)
  []; (* 3628: adds x17, x17, x9 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3629: adc x12, xzr, xzr *)
  [`X2`; `X7`]; (* 3630: subs x8, x2, x7 *)
  [`X3`]; (* 3631: sbcs x17, x3, x17 *)
  []; (* 3632: lsl x7, x8, #32 *)
  [`X1`; `X12`]; (* 3633: sbcs x2, x1, x12 *)
  [`X7`; `X8`]; (* 3634: add x3, x7, x8 *)
  [`X4`]; (* 3635: sbcs x12, x4, xzr *)
  []; (* 3636: lsr x1, x3, #32 *)
  [`X15`]; (* 3637: sbcs x7, x15, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X9`]; (* 3638: sbc x15, x9, xzr *)
  []; (* 3639: subs x1, x1, x3 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3640: sbc x4, x3, xzr *)
  []; (* 3641: lsr x9, x4, #32 *)
  [`X1`; `X4`]; (* 3642: extr x8, x4, x1, #32 *)
  []; (* 3643: adds x9, x9, x3 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3644: adc x4, xzr, xzr *)
  [`X17`; `X8`]; (* 3645: subs x1, x17, x8 *)
  []; (* 3646: lsl x17, x1, #32 *)
  [`X2`; `X9`]; (* 3647: sbcs x8, x2, x9 *)
  [`X12`; `X4`]; (* 3648: sbcs x9, x12, x4 *)
  [`X1`]; (* 3649: add x17, x17, x1 *)
  []; (* 3650: mov x1, v18.d[1] *)
  []; (* 3651: lsr x2, x17, #32 *)
  []; (* 3652: sbcs x7, x7, xzr *)
  [`Q18`]; (* 3653: mov x12, v18.d[0] *)
  []; (* 3654: sbcs x15, x15, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3655: sbc x3, x3, xzr *)
  [`X2`]; (* 3656: subs x4, x2, x17 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3657: sbc x2, x17, xzr *)
  [`X13`]; (* 3658: adds x12, x13, x12 *)
  [`X1`]; (* 3659: adcs x16, x16, x1 *)
  []; (* 3660: lsr x13, x2, #32 *)
  [`X2`; `X4`]; (* 3661: extr x1, x2, x4, #32 *)
  [`CF`; `NF`; `ZF`; `VF`; `X14`]; (* 3662: adc x2, x14, xzr *)
  [`X13`]; (* 3663: adds x4, x13, x17 *)
  []; (* 3664: mul x13, x6, x6 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3665: adc x14, xzr, xzr *)
  [`X8`]; (* 3666: subs x1, x8, x1 *)
  [`X9`]; (* 3667: sbcs x4, x9, x4 *)
  [`Q28`]; (* 3668: mov x9, v28.d[0] *)
  [`X14`]; (* 3669: sbcs x7, x7, x14 *)
  [`X15`]; (* 3670: sbcs x8, x15, xzr *)
  []; (* 3671: sbcs x3, x3, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X17`]; (* 3672: sbc x14, x17, xzr *)
  [`X9`]; (* 3673: adds x17, x9, x9 *)
  []; (* 3674: adcs x12, x12, x12 *)
  []; (* 3675: mov x15, v19.d[0] *)
  [`X16`]; (* 3676: adcs x9, x16, x16 *)
  []; (* 3677: umulh x6, x6, x6 *)
  [`X2`]; (* 3678: adcs x16, x2, x2 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3679: adc x2, xzr, xzr *)
  [`X8`]; (* 3680: adds x11, x11, x8 *)
  []; (* 3681: adcs x3, x3, xzr *)
  []; (* 3682: adcs x14, x14, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3683: adcs x8, xzr, xzr *)
  [`X1`]; (* 3684: adds x13, x1, x13 *)
  [`Q19`]; (* 3685: mov x1, v19.d[1] *)
  [`X4`]; (* 3686: adcs x6, x4, x6 *)
  []; (* 3687: mov x4, #0xffffffff *)
  [`X7`]; (* 3688: adcs x15, x7, x15 *)
  [`X11`; `X5`]; (* 3689: adcs x7, x11, x5 *)
  [`X3`]; (* 3690: adcs x1, x3, x1 *)
  [`X10`]; (* 3691: adcs x14, x14, x10 *)
  [`CF`; `NF`; `ZF`; `VF`; `X8`]; (* 3692: adc x11, x8, xzr *)
  [`X17`]; (* 3693: adds x6, x6, x17 *)
  [`X12`; `X15`]; (* 3694: adcs x8, x15, x12 *)
  [`X7`; `X9`]; (* 3695: adcs x3, x7, x9 *)
  [`X1`; `X16`]; (* 3696: adcs x15, x1, x16 *)
  []; (* 3697: mov x16, #0xffffffff00000001 *)
  [`X2`]; (* 3698: adcs x14, x14, x2 *)
  []; (* 3699: mov x2, #0x1 *)
  [`CF`; `NF`; `ZF`; `VF`; `X11`]; (* 3700: adc x17, x11, xzr *)
  []; (* 3701: cmn x13, x16 *)
  []; (* 3702: adcs xzr, x6, x4 *)
  []; (* 3703: adcs xzr, x8, x2 *)
  []; (* 3704: adcs xzr, x3, xzr *)
  []; (* 3705: adcs xzr, x15, xzr *)
  []; (* 3706: adcs xzr, x14, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X17`]; (* 3707: adc x1, x17, xzr *)
  [`X1`]; (* 3708: neg x9, x1 *)
  [`X16`]; (* 3709: and x1, x16, x9 *)
  [`X1`; `X13`]; (* 3710: adds x11, x13, x1 *)
  [`X4`]; (* 3711: and x13, x4, x9 *)
  [`X13`; `X6`]; (* 3712: adcs x5, x6, x13 *)
  [`X2`; `X9`]; (* 3713: and x1, x2, x9 *)
  [`X1`; `X8`]; (* 3714: adcs x7, x8, x1 *)
  [`X11`; `X5`]; (* 3715: stp x11, x5, [x0] *)
  [`X3`]; (* 3716: adcs x11, x3, xzr *)
  [`X15`]; (* 3717: adcs x2, x15, xzr *)
  [`X11`; `X7`]; (* 3718: stp x7, x11, [x0, #16] *)
  [`CF`; `NF`; `ZF`; `VF`; `X14`]; (* 3719: adc x17, x14, xzr *)
  [`X0`; `X17`; `X2`]; (* 3720: stp x2, x17, [x0, #32] *)
  []; (* 3721: ldr q3, [sp, #144] *)
  []; (* 3722: ldr q25, [sp, #192] *)
  []; (* 3723: ldp x13, x23, [sp, #192] *)
  []; (* 3724: ldp x3, x21, [sp, #144] *)
  []; (* 3725: rev64 v23.4s, v25.4s *)
  [`Q25`]; (* 3726: uzp1 v17.4s, v25.4s, v3.4s *)
  []; (* 3727: umulh x15, x3, x13 *)
  [`Q23`]; (* 3728: mul v6.4s, v23.4s, v3.4s *)
  []; (* 3729: uzp1 v3.4s, v3.4s, v3.4s *)
  []; (* 3730: ldr q27, [sp, #224] *)
  []; (* 3731: ldp x8, x24, [sp, #160] *)
  []; (* 3732: subs x6, x3, x21 *)
  []; (* 3733: ldr q0, [sp, #176] *)
  []; (* 3734: movi v23.2d, #0xffffffff *)
  []; (* 3735: csetm x10, cc *)
  []; (* 3736: umulh x19, x21, x23 *)
  []; (* 3737: rev64 v4.4s, v27.4s *)
  []; (* 3738: uzp2 v25.4s, v27.4s, v27.4s *)
  [`CF`; `NF`; `ZF`; `VF`; `X6`]; (* 3739: cneg x4, x6, cc *)
  []; (* 3740: subs x7, x23, x13 *)
  []; (* 3741: xtn v22.2s, v0.2d *)
  [`Q27`]; (* 3742: xtn v24.2s, v27.2d *)
  [`X7`]; (* 3743: cneg x20, x7, cc *)
  []; (* 3744: ldp x6, x14, [sp, #208] *)
  [`Q4`]; (* 3745: mul v27.4s, v4.4s, v0.4s *)
  [`Q6`]; (* 3746: uaddlp v20.2d, v6.4s *)
  [`CF`; `NF`; `ZF`; `VF`; `X10`]; (* 3747: cinv x5, x10, cc *)
  []; (* 3748: mul x16, x4, x20 *)
  [`Q0`]; (* 3749: uzp2 v6.4s, v0.4s, v0.4s *)
  []; (* 3750: umull v21.2d, v22.2s, v25.2s *)
  [`Q20`]; (* 3751: shl v0.2d, v20.2d, #32 *)
  [`Q17`; `Q3`]; (* 3752: umlal v0.2d, v3.2s, v17.2s *)
  []; (* 3753: mul x22, x8, x6 *)
  [`Q25`]; (* 3754: umull v1.2d, v6.2s, v25.2s *)
  []; (* 3755: subs x12, x3, x8 *)
  []; (* 3756: umull v20.2d, v22.2s, v24.2s *)
  [`X12`]; (* 3757: cneg x17, x12, cc *)
  []; (* 3758: umulh x9, x8, x6 *)
  []; (* 3759: mov x12, v0.d[1] *)
  [`X16`]; (* 3760: eor x11, x16, x5 *)
  [`Q0`]; (* 3761: mov x7, v0.d[0] *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3762: csetm x10, cc *)
  [`Q20`]; (* 3763: usra v21.2d, v20.2d, #32 *)
  [`X12`]; (* 3764: adds x15, x15, x12 *)
  [`X19`; `X22`]; (* 3765: adcs x12, x19, x22 *)
  [`X4`]; (* 3766: umulh x20, x4, x20 *)
  [`CF`; `NF`; `ZF`; `VF`; `X9`]; (* 3767: adc x19, x9, xzr *)
  []; (* 3768: usra v1.2d, v21.2d, #32 *)
  []; (* 3769: adds x22, x15, x7 *)
  [`Q21`; `Q23`]; (* 3770: and v26.16b, v21.16b, v23.16b *)
  []; (* 3771: adcs x16, x12, x15 *)
  [`Q27`]; (* 3772: uaddlp v25.2d, v27.4s *)
  []; (* 3773: adcs x9, x19, x12 *)
  [`Q6`]; (* 3774: umlal v26.2d, v6.2s, v24.2s *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3775: adc x4, x19, xzr *)
  []; (* 3776: adds x16, x16, x7 *)
  [`Q25`]; (* 3777: shl v27.2d, v25.2d, #32 *)
  [`X15`]; (* 3778: adcs x9, x9, x15 *)
  [`X12`]; (* 3779: adcs x4, x4, x12 *)
  [`X20`]; (* 3780: eor x12, x20, x5 *)
  [`CF`; `NF`; `ZF`; `VF`; `X19`]; (* 3781: adc x15, x19, xzr *)
  []; (* 3782: subs x20, x6, x13 *)
  []; (* 3783: cneg x20, x20, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3784: cinv x10, x10, cc *)
  []; (* 3785: cmn x5, #0x1 *)
  []; (* 3786: mul x19, x17, x20 *)
  [`X22`]; (* 3787: adcs x11, x22, x11 *)
  [`X16`]; (* 3788: adcs x12, x16, x12 *)
  []; (* 3789: adcs x9, x9, x5 *)
  [`X20`]; (* 3790: umulh x17, x17, x20 *)
  [`X4`]; (* 3791: adcs x22, x4, x5 *)
  [`CF`; `NF`; `ZF`; `VF`; `X15`]; (* 3792: adc x5, x15, x5 *)
  []; (* 3793: subs x16, x21, x8 *)
  [`X16`]; (* 3794: cneg x20, x16, cc *)
  []; (* 3795: eor x19, x19, x10 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3796: csetm x4, cc *)
  []; (* 3797: subs x16, x6, x23 *)
  []; (* 3798: cneg x16, x16, cc *)
  [`Q22`; `Q24`]; (* 3799: umlal v27.2d, v22.2s, v24.2s *)
  []; (* 3800: mul x15, x20, x16 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3801: cinv x4, x4, cc *)
  []; (* 3802: cmn x10, #0x1 *)
  [`Q26`]; (* 3803: usra v1.2d, v26.2d, #32 *)
  [`X12`]; (* 3804: adcs x19, x12, x19 *)
  []; (* 3805: eor x17, x17, x10 *)
  [`X17`]; (* 3806: adcs x9, x9, x17 *)
  []; (* 3807: adcs x22, x22, x10 *)
  []; (* 3808: lsl x12, x7, #32 *)
  [`X16`]; (* 3809: umulh x20, x20, x16 *)
  [`X15`]; (* 3810: eor x16, x15, x4 *)
  []; (* 3811: ldp x15, x17, [sp, #224] *)
  [`X12`; `X7`]; (* 3812: add x2, x12, x7 *)
  [`CF`; `NF`; `ZF`; `VF`; `X10`; `X5`]; (* 3813: adc x7, x5, x10 *)
  []; (* 3814: ldp x5, x10, [sp, #176] *)
  []; (* 3815: lsr x1, x2, #32 *)
  [`X20`]; (* 3816: eor x12, x20, x4 *)
  []; (* 3817: subs x1, x1, x2 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3818: sbc x20, x2, xzr *)
  []; (* 3819: cmn x4, #0x1 *)
  [`X16`]; (* 3820: adcs x9, x9, x16 *)
  []; (* 3821: extr x1, x20, x1, #32 *)
  []; (* 3822: lsr x20, x20, #32 *)
  [`X12`]; (* 3823: adcs x22, x22, x12 *)
  [`CF`; `NF`; `ZF`; `VF`; `X4`; `X7`]; (* 3824: adc x16, x7, x4 *)
  [`X20`]; (* 3825: adds x12, x20, x2 *)
  []; (* 3826: umulh x7, x24, x14 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3827: adc x4, xzr, xzr *)
  [`X11`]; (* 3828: subs x1, x11, x1 *)
  [`X12`; `X19`]; (* 3829: sbcs x20, x19, x12 *)
  [`X4`; `X9`]; (* 3830: sbcs x12, x9, x4 *)
  []; (* 3831: lsl x9, x1, #32 *)
  [`X9`]; (* 3832: add x1, x9, x1 *)
  [`X22`]; (* 3833: sbcs x9, x22, xzr *)
  []; (* 3834: mul x22, x24, x14 *)
  []; (* 3835: sbcs x16, x16, xzr *)
  []; (* 3836: lsr x4, x1, #32 *)
  [`CF`; `NF`; `ZF`; `VF`; `X2`]; (* 3837: sbc x19, x2, xzr *)
  []; (* 3838: subs x4, x4, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3839: sbc x11, x1, xzr *)
  [`X4`]; (* 3840: extr x2, x11, x4, #32 *)
  [`X11`]; (* 3841: lsr x4, x11, #32 *)
  []; (* 3842: adds x4, x4, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3843: adc x11, xzr, xzr *)
  [`X20`]; (* 3844: subs x2, x20, x2 *)
  [`X12`]; (* 3845: sbcs x4, x12, x4 *)
  [`X11`; `X9`]; (* 3846: sbcs x20, x9, x11 *)
  []; (* 3847: lsl x12, x2, #32 *)
  [`X12`]; (* 3848: add x2, x12, x2 *)
  [`X16`]; (* 3849: sbcs x9, x16, xzr *)
  []; (* 3850: lsr x11, x2, #32 *)
  []; (* 3851: sbcs x19, x19, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3852: sbc x1, x1, xzr *)
  [`X11`]; (* 3853: subs x16, x11, x2 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3854: sbc x12, x2, xzr *)
  []; (* 3855: extr x16, x12, x16, #32 *)
  []; (* 3856: lsr x12, x12, #32 *)
  [`X12`]; (* 3857: adds x11, x12, x2 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3858: adc x12, xzr, xzr *)
  [`X4`]; (* 3859: subs x16, x4, x16 *)
  []; (* 3860: mov x4, v27.d[0] *)
  [`X20`]; (* 3861: sbcs x11, x20, x11 *)
  [`X12`; `X9`]; (* 3862: sbcs x20, x9, x12 *)
  [`X11`; `X16`]; (* 3863: stp x16, x11, [sp, #192] *)
  [`X19`]; (* 3864: sbcs x11, x19, xzr *)
  [`X1`]; (* 3865: sbcs x9, x1, xzr *)
  [`X11`; `X20`]; (* 3866: stp x20, x11, [sp, #208] *)
  []; (* 3867: mov x1, v1.d[0] *)
  [`CF`; `NF`; `ZF`; `VF`; `X2`]; (* 3868: sbc x20, x2, xzr *)
  []; (* 3869: subs x12, x24, x5 *)
  [`Q27`]; (* 3870: mov x11, v27.d[1] *)
  [`X12`]; (* 3871: cneg x16, x12, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3872: csetm x2, cc *)
  []; (* 3873: subs x19, x15, x14 *)
  [`Q1`]; (* 3874: mov x12, v1.d[1] *)
  []; (* 3875: cinv x2, x2, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3876: cneg x19, x19, cc *)
  [`X20`; `X9`]; (* 3877: stp x9, x20, [sp, #224] *)
  []; (* 3878: mul x9, x16, x19 *)
  [`X7`]; (* 3879: adds x4, x7, x4 *)
  [`X1`]; (* 3880: adcs x11, x1, x11 *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 3881: adc x1, x12, xzr *)
  []; (* 3882: adds x20, x4, x22 *)
  [`X16`]; (* 3883: umulh x19, x16, x19 *)
  []; (* 3884: adcs x7, x11, x4 *)
  [`X9`]; (* 3885: eor x16, x9, x2 *)
  []; (* 3886: adcs x9, x1, x11 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3887: adc x12, x1, xzr *)
  []; (* 3888: adds x7, x7, x22 *)
  [`X9`]; (* 3889: adcs x4, x9, x4 *)
  [`X11`; `X12`]; (* 3890: adcs x9, x12, x11 *)
  [`CF`; `NF`; `ZF`; `VF`; `X1`]; (* 3891: adc x12, x1, xzr *)
  []; (* 3892: cmn x2, #0x1 *)
  [`X19`]; (* 3893: eor x1, x19, x2 *)
  [`X16`; `X20`]; (* 3894: adcs x11, x20, x16 *)
  [`X1`; `X7`]; (* 3895: adcs x19, x7, x1 *)
  [`X4`]; (* 3896: adcs x1, x4, x2 *)
  [`X9`]; (* 3897: adcs x20, x9, x2 *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 3898: adc x2, x12, x2 *)
  []; (* 3899: subs x12, x24, x10 *)
  [`X12`]; (* 3900: cneg x16, x12, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3901: csetm x12, cc *)
  []; (* 3902: subs x9, x17, x14 *)
  []; (* 3903: cinv x12, x12, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3904: cneg x9, x9, cc *)
  [`X24`]; (* 3905: subs x3, x24, x3 *)
  []; (* 3906: sbcs x21, x5, x21 *)
  []; (* 3907: mul x24, x16, x9 *)
  [`X8`]; (* 3908: sbcs x4, x10, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3909: ngc x8, xzr *)
  [`X5`]; (* 3910: subs x10, x5, x10 *)
  [`X24`]; (* 3911: eor x5, x24, x12 *)
  []; (* 3912: csetm x7, cc *)
  [`CF`; `NF`; `ZF`; `VF`; `X10`]; (* 3913: cneg x24, x10, cc *)
  []; (* 3914: subs x10, x17, x15 *)
  []; (* 3915: cinv x7, x7, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3916: cneg x10, x10, cc *)
  [`X13`]; (* 3917: subs x14, x13, x14 *)
  [`X23`]; (* 3918: sbcs x15, x23, x15 *)
  [`X21`]; (* 3919: eor x13, x21, x8 *)
  []; (* 3920: mul x23, x24, x10 *)
  [`X6`]; (* 3921: sbcs x17, x6, x17 *)
  [`X3`]; (* 3922: eor x6, x3, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3923: ngc x21, xzr *)
  [`X16`]; (* 3924: umulh x9, x16, x9 *)
  []; (* 3925: cmn x8, #0x1 *)
  [`X23`]; (* 3926: eor x3, x23, x7 *)
  [`X6`]; (* 3927: adcs x23, x6, xzr *)
  []; (* 3928: adcs x13, x13, xzr *)
  [`X4`]; (* 3929: eor x16, x4, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3930: adc x16, x16, xzr *)
  [`X17`]; (* 3931: eor x4, x17, x21 *)
  [`X10`; `X24`]; (* 3932: umulh x17, x24, x10 *)
  []; (* 3933: cmn x21, #0x1 *)
  [`X14`]; (* 3934: eor x24, x14, x21 *)
  [`X15`]; (* 3935: eor x6, x15, x21 *)
  [`X24`]; (* 3936: adcs x15, x24, xzr *)
  [`X6`]; (* 3937: adcs x14, x6, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X4`]; (* 3938: adc x6, x4, xzr *)
  []; (* 3939: cmn x12, #0x1 *)
  [`X9`]; (* 3940: eor x4, x9, x12 *)
  [`X5`]; (* 3941: adcs x19, x19, x5 *)
  []; (* 3942: umulh x5, x23, x15 *)
  [`X4`]; (* 3943: adcs x1, x1, x4 *)
  [`X20`]; (* 3944: adcs x10, x20, x12 *)
  [`X17`]; (* 3945: eor x4, x17, x7 *)
  []; (* 3946: ldp x20, x9, [sp, #192] *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 3947: adc x2, x2, x12 *)
  []; (* 3948: cmn x7, #0x1 *)
  [`X1`; `X3`]; (* 3949: adcs x12, x1, x3 *)
  []; (* 3950: ldp x17, x24, [sp, #208] *)
  []; (* 3951: mul x1, x16, x6 *)
  [`X10`; `X4`]; (* 3952: adcs x3, x10, x4 *)
  [`CF`; `NF`; `ZF`; `VF`; `X7`]; (* 3953: adc x2, x2, x7 *)
  []; (* 3954: ldp x7, x4, [sp, #224] *)
  [`X22`]; (* 3955: adds x20, x22, x20 *)
  []; (* 3956: mul x10, x13, x14 *)
  [`X9`]; (* 3957: adcs x11, x11, x9 *)
  [`X21`; `X8`]; (* 3958: eor x9, x8, x21 *)
  [`X17`; `X19`]; (* 3959: adcs x21, x19, x17 *)
  [`X11`; `X20`]; (* 3960: stp x20, x11, [sp, #192] *)
  [`X24`]; (* 3961: adcs x12, x12, x24 *)
  []; (* 3962: mul x8, x23, x15 *)
  [`X7`]; (* 3963: adcs x3, x3, x7 *)
  [`X12`; `X21`]; (* 3964: stp x21, x12, [sp, #208] *)
  [`X2`; `X4`]; (* 3965: adcs x12, x2, x4 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3966: adc x19, xzr, xzr *)
  []; (* 3967: subs x21, x23, x16 *)
  []; (* 3968: umulh x2, x16, x6 *)
  [`X12`; `X3`]; (* 3969: stp x3, x12, [sp, #224] *)
  [`X21`]; (* 3970: cneg x3, x21, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3971: csetm x24, cc *)
  []; (* 3972: umulh x11, x13, x14 *)
  [`X16`]; (* 3973: subs x21, x13, x16 *)
  []; (* 3974: eor x7, x8, x9 *)
  [`X21`]; (* 3975: cneg x17, x21, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3976: csetm x16, cc *)
  []; (* 3977: subs x21, x6, x15 *)
  [`X21`]; (* 3978: cneg x22, x21, cc *)
  [`CF`; `NF`; `ZF`; `VF`; `X24`]; (* 3979: cinv x21, x24, cc *)
  [`X13`; `X23`]; (* 3980: subs x20, x23, x13 *)
  []; (* 3981: umulh x12, x3, x22 *)
  [`X20`]; (* 3982: cneg x23, x20, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3983: csetm x24, cc *)
  [`X15`]; (* 3984: subs x20, x14, x15 *)
  []; (* 3985: cinv x24, x24, cc *)
  [`X3`]; (* 3986: mul x22, x3, x22 *)
  [`CF`; `NF`; `ZF`; `VF`; `X20`]; (* 3987: cneg x3, x20, cc *)
  [`X14`; `X6`]; (* 3988: subs x13, x6, x14 *)
  [`X13`]; (* 3989: cneg x20, x13, cc *)
  [`CF`; `NF`; `ZF`; `VF`; `X16`]; (* 3990: cinv x15, x16, cc *)
  [`X10`; `X5`]; (* 3991: adds x13, x5, x10 *)
  []; (* 3992: mul x4, x23, x3 *)
  [`X1`]; (* 3993: adcs x11, x11, x1 *)
  [`CF`; `NF`; `ZF`; `VF`; `X2`]; (* 3994: adc x14, x2, xzr *)
  []; (* 3995: adds x5, x13, x8 *)
  []; (* 3996: adcs x16, x11, x13 *)
  [`X3`]; (* 3997: umulh x23, x23, x3 *)
  []; (* 3998: adcs x3, x14, x11 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 3999: adc x1, x14, xzr *)
  [`X16`; `X8`]; (* 4000: adds x10, x16, x8 *)
  [`X13`; `X3`]; (* 4001: adcs x6, x3, x13 *)
  [`X1`; `X11`]; (* 4002: adcs x8, x1, x11 *)
  []; (* 4003: umulh x13, x17, x20 *)
  [`X4`]; (* 4004: eor x1, x4, x24 *)
  [`CF`; `NF`; `ZF`; `VF`; `X14`]; (* 4005: adc x4, x14, xzr *)
  []; (* 4006: cmn x24, #0x1 *)
  [`X5`]; (* 4007: adcs x1, x5, x1 *)
  [`X23`]; (* 4008: eor x16, x23, x24 *)
  [`X1`]; (* 4009: eor x11, x1, x9 *)
  [`X10`; `X16`]; (* 4010: adcs x23, x10, x16 *)
  [`X22`]; (* 4011: eor x2, x22, x21 *)
  [`X6`]; (* 4012: adcs x3, x6, x24 *)
  [`X17`; `X20`]; (* 4013: mul x14, x17, x20 *)
  [`X13`]; (* 4014: eor x17, x13, x15 *)
  [`X8`]; (* 4015: adcs x13, x8, x24 *)
  [`CF`; `NF`; `ZF`; `VF`; `X24`; `X4`]; (* 4016: adc x8, x4, x24 *)
  []; (* 4017: cmn x21, #0x1 *)
  [`X2`; `X23`]; (* 4018: adcs x6, x23, x2 *)
  []; (* 4019: mov x16, #0xfffffffffffffffe *)
  [`X12`]; (* 4020: eor x20, x12, x21 *)
  [`X3`]; (* 4021: adcs x20, x3, x20 *)
  [`X14`]; (* 4022: eor x23, x14, x15 *)
  [`X13`]; (* 4023: adcs x2, x13, x21 *)
  [`CF`; `NF`; `ZF`; `VF`; `X21`]; (* 4024: adc x8, x8, x21 *)
  []; (* 4025: cmn x15, #0x1 *)
  []; (* 4026: ldp x5, x4, [sp, #192] *)
  []; (* 4027: ldp x21, x12, [sp, #208] *)
  [`X20`; `X23`]; (* 4028: adcs x22, x20, x23 *)
  [`X22`]; (* 4029: eor x23, x22, x9 *)
  [`X2`]; (* 4030: adcs x17, x2, x17 *)
  [`CF`; `NF`; `ZF`; `VF`; `X15`; `X8`]; (* 4031: adc x22, x8, x15 *)
  []; (* 4032: cmn x9, #0x1 *)
  [`X7`]; (* 4033: adcs x15, x7, x5 *)
  []; (* 4034: ldp x10, x14, [sp, #224] *)
  [`X6`]; (* 4035: eor x1, x6, x9 *)
  []; (* 4036: lsl x2, x15, #32 *)
  [`X11`]; (* 4037: adcs x8, x11, x4 *)
  [`X1`]; (* 4038: adcs x13, x1, x21 *)
  [`X22`]; (* 4039: eor x1, x22, x9 *)
  [`X23`]; (* 4040: adcs x24, x23, x12 *)
  [`X17`]; (* 4041: eor x11, x17, x9 *)
  [`X11`]; (* 4042: adcs x23, x11, x10 *)
  [`X1`]; (* 4043: adcs x7, x1, x14 *)
  []; (* 4044: adcs x17, x9, x19 *)
  []; (* 4045: adcs x20, x9, xzr *)
  [`X15`; `X2`]; (* 4046: add x1, x2, x15 *)
  []; (* 4047: lsr x3, x1, #32 *)
  []; (* 4048: adcs x11, x9, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4049: adc x9, x9, xzr *)
  []; (* 4050: subs x3, x3, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4051: sbc x6, x1, xzr *)
  [`X5`]; (* 4052: adds x24, x24, x5 *)
  [`X23`]; (* 4053: adcs x4, x23, x4 *)
  []; (* 4054: extr x3, x6, x3, #32 *)
  []; (* 4055: lsr x6, x6, #32 *)
  [`X7`]; (* 4056: adcs x21, x7, x21 *)
  [`X12`; `X17`]; (* 4057: adcs x15, x17, x12 *)
  [`X10`; `X20`]; (* 4058: adcs x7, x20, x10 *)
  [`X11`; `X14`]; (* 4059: adcs x20, x11, x14 *)
  []; (* 4060: mov x14, #0xffffffff *)
  [`CF`; `NF`; `ZF`; `VF`; `X19`; `X9`]; (* 4061: adc x22, x9, x19 *)
  [`X6`]; (* 4062: adds x12, x6, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4063: adc x10, xzr, xzr *)
  [`X8`]; (* 4064: subs x3, x8, x3 *)
  [`X13`]; (* 4065: sbcs x12, x13, x12 *)
  []; (* 4066: lsl x9, x3, #32 *)
  [`X9`]; (* 4067: add x3, x9, x3 *)
  [`X24`]; (* 4068: sbcs x10, x24, x10 *)
  [`X4`]; (* 4069: sbcs x24, x4, xzr *)
  []; (* 4070: lsr x9, x3, #32 *)
  []; (* 4071: sbcs x21, x21, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4072: sbc x1, x1, xzr *)
  []; (* 4073: subs x9, x9, x3 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4074: sbc x13, x3, xzr *)
  []; (* 4075: extr x9, x13, x9, #32 *)
  []; (* 4076: lsr x13, x13, #32 *)
  []; (* 4077: adds x13, x13, x3 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4078: adc x6, xzr, xzr *)
  [`X9`]; (* 4079: subs x12, x12, x9 *)
  [`X10`; `X13`]; (* 4080: sbcs x17, x10, x13 *)
  []; (* 4081: lsl x2, x12, #32 *)
  [`X24`; `X6`]; (* 4082: sbcs x10, x24, x6 *)
  [`X12`; `X2`]; (* 4083: add x9, x2, x12 *)
  [`X21`]; (* 4084: sbcs x6, x21, xzr *)
  []; (* 4085: lsr x5, x9, #32 *)
  [`X1`]; (* 4086: sbcs x21, x1, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X3`]; (* 4087: sbc x13, x3, xzr *)
  [`X5`]; (* 4088: subs x8, x5, x9 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4089: sbc x19, x9, xzr *)
  []; (* 4090: lsr x12, x19, #32 *)
  [`X19`; `X8`]; (* 4091: extr x3, x19, x8, #32 *)
  [`X12`]; (* 4092: adds x8, x12, x9 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4093: adc x1, xzr, xzr *)
  [`X17`; `X3`]; (* 4094: subs x2, x17, x3 *)
  [`X10`; `X8`]; (* 4095: sbcs x12, x10, x8 *)
  [`X1`; `X6`]; (* 4096: sbcs x5, x6, x1 *)
  [`X21`]; (* 4097: sbcs x3, x21, xzr *)
  [`X13`]; (* 4098: sbcs x19, x13, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X9`]; (* 4099: sbc x24, x9, xzr *)
  [`X15`; `X3`]; (* 4100: adds x23, x15, x3 *)
  [`X19`; `X7`]; (* 4101: adcs x8, x7, x19 *)
  [`X20`; `X24`]; (* 4102: adcs x11, x20, x24 *)
  [`CF`; `NF`; `ZF`; `VF`; `X22`]; (* 4103: adc x9, x22, xzr *)
  [`X9`]; (* 4104: add x24, x9, #0x1 *)
  []; (* 4105: lsl x7, x24, #32 *)
  []; (* 4106: subs x21, x24, x7 *)
  [`CF`; `NF`; `ZF`; `VF`; `X7`]; (* 4107: sbc x10, x7, xzr *)
  [`X2`; `X21`]; (* 4108: adds x6, x2, x21 *)
  [`X10`; `X12`]; (* 4109: adcs x7, x12, x10 *)
  [`X5`]; (* 4110: adcs x24, x5, x24 *)
  [`X23`]; (* 4111: adcs x13, x23, xzr *)
  []; (* 4112: adcs x8, x8, xzr *)
  [`X11`]; (* 4113: adcs x15, x11, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4114: csetm x23, cc *)
  [`X16`]; (* 4115: and x11, x16, x23 *)
  [`X14`]; (* 4116: and x20, x14, x23 *)
  [`X6`]; (* 4117: adds x22, x6, x20 *)
  [`X20`]; (* 4118: eor x3, x20, x23 *)
  [`X3`; `X7`]; (* 4119: adcs x5, x7, x3 *)
  [`X11`; `X24`]; (* 4120: adcs x14, x24, x11 *)
  [`X22`; `X5`]; (* 4121: stp x22, x5, [sp, #192] *)
  [`X13`]; (* 4122: adcs x5, x13, x23 *)
  [`X8`]; (* 4123: adcs x21, x8, x23 *)
  [`X14`; `X5`]; (* 4124: stp x14, x5, [sp, #208] *)
  [`CF`; `NF`; `ZF`; `VF`; `X15`; `X23`]; (* 4125: adc x12, x15, x23 *)
  [`X12`; `X21`]; (* 4126: stp x21, x12, [sp, #224] *)
  []; (* 4127: ldr q3, [sp, #144] *)
  []; (* 4128: ldr q25, [sp, #96] *)
  []; (* 4129: ldp x13, x23, [sp, #96] *)
  []; (* 4130: ldp x3, x21, [sp, #144] *)
  []; (* 4131: rev64 v23.4s, v25.4s *)
  [`Q25`]; (* 4132: uzp1 v17.4s, v25.4s, v3.4s *)
  []; (* 4133: umulh x15, x3, x13 *)
  [`Q23`]; (* 4134: mul v6.4s, v23.4s, v3.4s *)
  []; (* 4135: uzp1 v3.4s, v3.4s, v3.4s *)
  []; (* 4136: ldr q27, [sp, #128] *)
  []; (* 4137: ldp x8, x24, [sp, #160] *)
  []; (* 4138: subs x6, x3, x21 *)
  []; (* 4139: ldr q0, [sp, #176] *)
  []; (* 4140: movi v23.2d, #0xffffffff *)
  []; (* 4141: csetm x10, cc *)
  []; (* 4142: umulh x19, x21, x23 *)
  []; (* 4143: rev64 v4.4s, v27.4s *)
  []; (* 4144: uzp2 v25.4s, v27.4s, v27.4s *)
  [`CF`; `NF`; `ZF`; `VF`; `X6`]; (* 4145: cneg x4, x6, cc *)
  []; (* 4146: subs x7, x23, x13 *)
  []; (* 4147: xtn v22.2s, v0.2d *)
  [`Q27`]; (* 4148: xtn v24.2s, v27.2d *)
  [`X7`]; (* 4149: cneg x20, x7, cc *)
  []; (* 4150: ldp x6, x14, [sp, #112] *)
  [`Q4`]; (* 4151: mul v27.4s, v4.4s, v0.4s *)
  [`Q6`]; (* 4152: uaddlp v20.2d, v6.4s *)
  [`CF`; `NF`; `ZF`; `VF`; `X10`]; (* 4153: cinv x5, x10, cc *)
  []; (* 4154: mul x16, x4, x20 *)
  [`Q0`]; (* 4155: uzp2 v6.4s, v0.4s, v0.4s *)
  []; (* 4156: umull v21.2d, v22.2s, v25.2s *)
  [`Q20`]; (* 4157: shl v0.2d, v20.2d, #32 *)
  [`Q17`; `Q3`]; (* 4158: umlal v0.2d, v3.2s, v17.2s *)
  []; (* 4159: mul x22, x8, x6 *)
  [`Q25`]; (* 4160: umull v1.2d, v6.2s, v25.2s *)
  []; (* 4161: subs x12, x3, x8 *)
  []; (* 4162: umull v20.2d, v22.2s, v24.2s *)
  [`X12`]; (* 4163: cneg x17, x12, cc *)
  []; (* 4164: umulh x9, x8, x6 *)
  []; (* 4165: mov x12, v0.d[1] *)
  [`X16`]; (* 4166: eor x11, x16, x5 *)
  [`Q0`]; (* 4167: mov x7, v0.d[0] *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4168: csetm x10, cc *)
  [`Q20`]; (* 4169: usra v21.2d, v20.2d, #32 *)
  [`X12`]; (* 4170: adds x15, x15, x12 *)
  [`X19`; `X22`]; (* 4171: adcs x12, x19, x22 *)
  [`X4`]; (* 4172: umulh x20, x4, x20 *)
  [`CF`; `NF`; `ZF`; `VF`; `X9`]; (* 4173: adc x19, x9, xzr *)
  []; (* 4174: usra v1.2d, v21.2d, #32 *)
  []; (* 4175: adds x22, x15, x7 *)
  [`Q21`; `Q23`]; (* 4176: and v26.16b, v21.16b, v23.16b *)
  []; (* 4177: adcs x16, x12, x15 *)
  [`Q27`]; (* 4178: uaddlp v25.2d, v27.4s *)
  []; (* 4179: adcs x9, x19, x12 *)
  [`Q6`]; (* 4180: umlal v26.2d, v6.2s, v24.2s *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4181: adc x4, x19, xzr *)
  []; (* 4182: adds x16, x16, x7 *)
  [`Q25`]; (* 4183: shl v27.2d, v25.2d, #32 *)
  [`X15`]; (* 4184: adcs x9, x9, x15 *)
  [`X12`]; (* 4185: adcs x4, x4, x12 *)
  [`X20`]; (* 4186: eor x12, x20, x5 *)
  [`CF`; `NF`; `ZF`; `VF`; `X19`]; (* 4187: adc x15, x19, xzr *)
  []; (* 4188: subs x20, x6, x13 *)
  []; (* 4189: cneg x20, x20, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4190: cinv x10, x10, cc *)
  []; (* 4191: cmn x5, #0x1 *)
  []; (* 4192: mul x19, x17, x20 *)
  [`X22`]; (* 4193: adcs x11, x22, x11 *)
  [`X16`]; (* 4194: adcs x12, x16, x12 *)
  []; (* 4195: adcs x9, x9, x5 *)
  [`X20`]; (* 4196: umulh x17, x17, x20 *)
  [`X4`]; (* 4197: adcs x22, x4, x5 *)
  [`CF`; `NF`; `ZF`; `VF`; `X15`]; (* 4198: adc x5, x15, x5 *)
  []; (* 4199: subs x16, x21, x8 *)
  [`X16`]; (* 4200: cneg x20, x16, cc *)
  []; (* 4201: eor x19, x19, x10 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4202: csetm x4, cc *)
  []; (* 4203: subs x16, x6, x23 *)
  []; (* 4204: cneg x16, x16, cc *)
  [`Q22`; `Q24`]; (* 4205: umlal v27.2d, v22.2s, v24.2s *)
  []; (* 4206: mul x15, x20, x16 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4207: cinv x4, x4, cc *)
  []; (* 4208: cmn x10, #0x1 *)
  [`Q26`]; (* 4209: usra v1.2d, v26.2d, #32 *)
  [`X12`]; (* 4210: adcs x19, x12, x19 *)
  []; (* 4211: eor x17, x17, x10 *)
  [`X17`]; (* 4212: adcs x9, x9, x17 *)
  []; (* 4213: adcs x22, x22, x10 *)
  []; (* 4214: lsl x12, x7, #32 *)
  [`X16`]; (* 4215: umulh x20, x20, x16 *)
  [`X15`]; (* 4216: eor x16, x15, x4 *)
  []; (* 4217: ldp x15, x17, [sp, #128] *)
  [`X12`; `X7`]; (* 4218: add x2, x12, x7 *)
  [`CF`; `NF`; `ZF`; `VF`; `X10`; `X5`]; (* 4219: adc x7, x5, x10 *)
  []; (* 4220: ldp x5, x10, [sp, #176] *)
  []; (* 4221: lsr x1, x2, #32 *)
  [`X20`]; (* 4222: eor x12, x20, x4 *)
  []; (* 4223: subs x1, x1, x2 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4224: sbc x20, x2, xzr *)
  []; (* 4225: cmn x4, #0x1 *)
  [`X16`]; (* 4226: adcs x9, x9, x16 *)
  []; (* 4227: extr x1, x20, x1, #32 *)
  []; (* 4228: lsr x20, x20, #32 *)
  [`X12`]; (* 4229: adcs x22, x22, x12 *)
  [`CF`; `NF`; `ZF`; `VF`; `X4`; `X7`]; (* 4230: adc x16, x7, x4 *)
  [`X20`]; (* 4231: adds x12, x20, x2 *)
  []; (* 4232: umulh x7, x24, x14 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4233: adc x4, xzr, xzr *)
  [`X11`]; (* 4234: subs x1, x11, x1 *)
  [`X12`; `X19`]; (* 4235: sbcs x20, x19, x12 *)
  [`X4`; `X9`]; (* 4236: sbcs x12, x9, x4 *)
  []; (* 4237: lsl x9, x1, #32 *)
  [`X9`]; (* 4238: add x1, x9, x1 *)
  [`X22`]; (* 4239: sbcs x9, x22, xzr *)
  []; (* 4240: mul x22, x24, x14 *)
  []; (* 4241: sbcs x16, x16, xzr *)
  []; (* 4242: lsr x4, x1, #32 *)
  [`CF`; `NF`; `ZF`; `VF`; `X2`]; (* 4243: sbc x19, x2, xzr *)
  []; (* 4244: subs x4, x4, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4245: sbc x11, x1, xzr *)
  [`X4`]; (* 4246: extr x2, x11, x4, #32 *)
  [`X11`]; (* 4247: lsr x4, x11, #32 *)
  []; (* 4248: adds x4, x4, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4249: adc x11, xzr, xzr *)
  [`X20`]; (* 4250: subs x2, x20, x2 *)
  [`X12`]; (* 4251: sbcs x4, x12, x4 *)
  [`X11`; `X9`]; (* 4252: sbcs x20, x9, x11 *)
  []; (* 4253: lsl x12, x2, #32 *)
  [`X12`]; (* 4254: add x2, x12, x2 *)
  [`X16`]; (* 4255: sbcs x9, x16, xzr *)
  []; (* 4256: lsr x11, x2, #32 *)
  []; (* 4257: sbcs x19, x19, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4258: sbc x1, x1, xzr *)
  [`X11`]; (* 4259: subs x16, x11, x2 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4260: sbc x12, x2, xzr *)
  []; (* 4261: extr x16, x12, x16, #32 *)
  []; (* 4262: lsr x12, x12, #32 *)
  [`X12`]; (* 4263: adds x11, x12, x2 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4264: adc x12, xzr, xzr *)
  [`X4`]; (* 4265: subs x16, x4, x16 *)
  []; (* 4266: mov x4, v27.d[0] *)
  [`X20`]; (* 4267: sbcs x11, x20, x11 *)
  [`X12`; `X9`]; (* 4268: sbcs x20, x9, x12 *)
  [`X11`; `X16`]; (* 4269: stp x16, x11, [sp, #96] *)
  [`X19`]; (* 4270: sbcs x11, x19, xzr *)
  [`X1`]; (* 4271: sbcs x9, x1, xzr *)
  [`X11`; `X20`]; (* 4272: stp x20, x11, [sp, #112] *)
  []; (* 4273: mov x1, v1.d[0] *)
  [`CF`; `NF`; `ZF`; `VF`; `X2`]; (* 4274: sbc x20, x2, xzr *)
  []; (* 4275: subs x12, x24, x5 *)
  [`Q27`]; (* 4276: mov x11, v27.d[1] *)
  [`X12`]; (* 4277: cneg x16, x12, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4278: csetm x2, cc *)
  []; (* 4279: subs x19, x15, x14 *)
  [`Q1`]; (* 4280: mov x12, v1.d[1] *)
  []; (* 4281: cinv x2, x2, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4282: cneg x19, x19, cc *)
  [`X20`; `X9`]; (* 4283: stp x9, x20, [sp, #128] *)
  []; (* 4284: mul x9, x16, x19 *)
  [`X7`]; (* 4285: adds x4, x7, x4 *)
  [`X1`]; (* 4286: adcs x11, x1, x11 *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 4287: adc x1, x12, xzr *)
  []; (* 4288: adds x20, x4, x22 *)
  [`X16`]; (* 4289: umulh x19, x16, x19 *)
  []; (* 4290: adcs x7, x11, x4 *)
  [`X9`]; (* 4291: eor x16, x9, x2 *)
  []; (* 4292: adcs x9, x1, x11 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4293: adc x12, x1, xzr *)
  []; (* 4294: adds x7, x7, x22 *)
  [`X9`]; (* 4295: adcs x4, x9, x4 *)
  [`X11`; `X12`]; (* 4296: adcs x9, x12, x11 *)
  [`CF`; `NF`; `ZF`; `VF`; `X1`]; (* 4297: adc x12, x1, xzr *)
  []; (* 4298: cmn x2, #0x1 *)
  [`X19`]; (* 4299: eor x1, x19, x2 *)
  [`X16`; `X20`]; (* 4300: adcs x11, x20, x16 *)
  [`X1`; `X7`]; (* 4301: adcs x19, x7, x1 *)
  [`X4`]; (* 4302: adcs x1, x4, x2 *)
  [`X9`]; (* 4303: adcs x20, x9, x2 *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 4304: adc x2, x12, x2 *)
  []; (* 4305: subs x12, x24, x10 *)
  [`X12`]; (* 4306: cneg x16, x12, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4307: csetm x12, cc *)
  []; (* 4308: subs x9, x17, x14 *)
  []; (* 4309: cinv x12, x12, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4310: cneg x9, x9, cc *)
  [`X24`]; (* 4311: subs x3, x24, x3 *)
  []; (* 4312: sbcs x21, x5, x21 *)
  []; (* 4313: mul x24, x16, x9 *)
  [`X8`]; (* 4314: sbcs x4, x10, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4315: ngc x8, xzr *)
  [`X5`]; (* 4316: subs x10, x5, x10 *)
  [`X24`]; (* 4317: eor x5, x24, x12 *)
  []; (* 4318: csetm x7, cc *)
  [`CF`; `NF`; `ZF`; `VF`; `X10`]; (* 4319: cneg x24, x10, cc *)
  []; (* 4320: subs x10, x17, x15 *)
  []; (* 4321: cinv x7, x7, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4322: cneg x10, x10, cc *)
  [`X13`]; (* 4323: subs x14, x13, x14 *)
  [`X23`]; (* 4324: sbcs x15, x23, x15 *)
  [`X21`]; (* 4325: eor x13, x21, x8 *)
  []; (* 4326: mul x23, x24, x10 *)
  [`X6`]; (* 4327: sbcs x17, x6, x17 *)
  [`X3`]; (* 4328: eor x6, x3, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4329: ngc x21, xzr *)
  [`X16`]; (* 4330: umulh x9, x16, x9 *)
  []; (* 4331: cmn x8, #0x1 *)
  [`X23`]; (* 4332: eor x3, x23, x7 *)
  [`X6`]; (* 4333: adcs x23, x6, xzr *)
  []; (* 4334: adcs x13, x13, xzr *)
  [`X4`]; (* 4335: eor x16, x4, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4336: adc x16, x16, xzr *)
  [`X17`]; (* 4337: eor x4, x17, x21 *)
  [`X10`; `X24`]; (* 4338: umulh x17, x24, x10 *)
  []; (* 4339: cmn x21, #0x1 *)
  [`X14`]; (* 4340: eor x24, x14, x21 *)
  [`X15`]; (* 4341: eor x6, x15, x21 *)
  [`X24`]; (* 4342: adcs x15, x24, xzr *)
  [`X6`]; (* 4343: adcs x14, x6, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X4`]; (* 4344: adc x6, x4, xzr *)
  []; (* 4345: cmn x12, #0x1 *)
  [`X9`]; (* 4346: eor x4, x9, x12 *)
  [`X5`]; (* 4347: adcs x19, x19, x5 *)
  []; (* 4348: umulh x5, x23, x15 *)
  [`X4`]; (* 4349: adcs x1, x1, x4 *)
  [`X20`]; (* 4350: adcs x10, x20, x12 *)
  [`X17`]; (* 4351: eor x4, x17, x7 *)
  []; (* 4352: ldp x20, x9, [sp, #96] *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 4353: adc x2, x2, x12 *)
  []; (* 4354: cmn x7, #0x1 *)
  [`X1`; `X3`]; (* 4355: adcs x12, x1, x3 *)
  []; (* 4356: ldp x17, x24, [sp, #112] *)
  []; (* 4357: mul x1, x16, x6 *)
  [`X10`; `X4`]; (* 4358: adcs x3, x10, x4 *)
  [`CF`; `NF`; `ZF`; `VF`; `X7`]; (* 4359: adc x2, x2, x7 *)
  []; (* 4360: ldp x7, x4, [sp, #128] *)
  [`X22`]; (* 4361: adds x20, x22, x20 *)
  []; (* 4362: mul x10, x13, x14 *)
  [`X9`]; (* 4363: adcs x11, x11, x9 *)
  [`X21`; `X8`]; (* 4364: eor x9, x8, x21 *)
  [`X17`; `X19`]; (* 4365: adcs x21, x19, x17 *)
  [`X11`; `X20`]; (* 4366: stp x20, x11, [sp, #96] *)
  [`X24`]; (* 4367: adcs x12, x12, x24 *)
  []; (* 4368: mul x8, x23, x15 *)
  [`X7`]; (* 4369: adcs x3, x3, x7 *)
  [`X12`; `X21`]; (* 4370: stp x21, x12, [sp, #112] *)
  [`X2`; `X4`]; (* 4371: adcs x12, x2, x4 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4372: adc x19, xzr, xzr *)
  []; (* 4373: subs x21, x23, x16 *)
  []; (* 4374: umulh x2, x16, x6 *)
  [`X12`; `X3`]; (* 4375: stp x3, x12, [sp, #128] *)
  [`X21`]; (* 4376: cneg x3, x21, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4377: csetm x24, cc *)
  []; (* 4378: umulh x11, x13, x14 *)
  [`X16`]; (* 4379: subs x21, x13, x16 *)
  []; (* 4380: eor x7, x8, x9 *)
  [`X21`]; (* 4381: cneg x17, x21, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4382: csetm x16, cc *)
  []; (* 4383: subs x21, x6, x15 *)
  [`X21`]; (* 4384: cneg x22, x21, cc *)
  [`CF`; `NF`; `ZF`; `VF`; `X24`]; (* 4385: cinv x21, x24, cc *)
  [`X13`; `X23`]; (* 4386: subs x20, x23, x13 *)
  []; (* 4387: umulh x12, x3, x22 *)
  [`X20`]; (* 4388: cneg x23, x20, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4389: csetm x24, cc *)
  [`X15`]; (* 4390: subs x20, x14, x15 *)
  []; (* 4391: cinv x24, x24, cc *)
  [`X3`]; (* 4392: mul x22, x3, x22 *)
  [`CF`; `NF`; `ZF`; `VF`; `X20`]; (* 4393: cneg x3, x20, cc *)
  [`X14`; `X6`]; (* 4394: subs x13, x6, x14 *)
  [`X13`]; (* 4395: cneg x20, x13, cc *)
  [`CF`; `NF`; `ZF`; `VF`; `X16`]; (* 4396: cinv x15, x16, cc *)
  [`X10`; `X5`]; (* 4397: adds x13, x5, x10 *)
  []; (* 4398: mul x4, x23, x3 *)
  [`X1`]; (* 4399: adcs x11, x11, x1 *)
  [`CF`; `NF`; `ZF`; `VF`; `X2`]; (* 4400: adc x14, x2, xzr *)
  []; (* 4401: adds x5, x13, x8 *)
  []; (* 4402: adcs x16, x11, x13 *)
  [`X3`]; (* 4403: umulh x23, x23, x3 *)
  []; (* 4404: adcs x3, x14, x11 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4405: adc x1, x14, xzr *)
  [`X16`; `X8`]; (* 4406: adds x10, x16, x8 *)
  [`X13`; `X3`]; (* 4407: adcs x6, x3, x13 *)
  [`X1`; `X11`]; (* 4408: adcs x8, x1, x11 *)
  []; (* 4409: umulh x13, x17, x20 *)
  [`X4`]; (* 4410: eor x1, x4, x24 *)
  [`CF`; `NF`; `ZF`; `VF`; `X14`]; (* 4411: adc x4, x14, xzr *)
  []; (* 4412: cmn x24, #0x1 *)
  [`X5`]; (* 4413: adcs x1, x5, x1 *)
  [`X23`]; (* 4414: eor x16, x23, x24 *)
  [`X1`]; (* 4415: eor x11, x1, x9 *)
  [`X10`; `X16`]; (* 4416: adcs x23, x10, x16 *)
  [`X22`]; (* 4417: eor x2, x22, x21 *)
  [`X6`]; (* 4418: adcs x3, x6, x24 *)
  [`X17`; `X20`]; (* 4419: mul x14, x17, x20 *)
  [`X13`]; (* 4420: eor x17, x13, x15 *)
  [`X8`]; (* 4421: adcs x13, x8, x24 *)
  [`CF`; `NF`; `ZF`; `VF`; `X24`; `X4`]; (* 4422: adc x8, x4, x24 *)
  []; (* 4423: cmn x21, #0x1 *)
  [`X2`; `X23`]; (* 4424: adcs x6, x23, x2 *)
  []; (* 4425: mov x16, #0xfffffffffffffffe *)
  [`X12`]; (* 4426: eor x20, x12, x21 *)
  [`X3`]; (* 4427: adcs x20, x3, x20 *)
  [`X14`]; (* 4428: eor x23, x14, x15 *)
  [`X13`]; (* 4429: adcs x2, x13, x21 *)
  [`CF`; `NF`; `ZF`; `VF`; `X21`]; (* 4430: adc x8, x8, x21 *)
  []; (* 4431: cmn x15, #0x1 *)
  []; (* 4432: ldp x5, x4, [sp, #96] *)
  []; (* 4433: ldp x21, x12, [sp, #112] *)
  [`X20`; `X23`]; (* 4434: adcs x22, x20, x23 *)
  [`X22`]; (* 4435: eor x23, x22, x9 *)
  [`X2`]; (* 4436: adcs x17, x2, x17 *)
  [`CF`; `NF`; `ZF`; `VF`; `X15`; `X8`]; (* 4437: adc x22, x8, x15 *)
  []; (* 4438: cmn x9, #0x1 *)
  [`X7`]; (* 4439: adcs x15, x7, x5 *)
  []; (* 4440: ldp x10, x14, [sp, #128] *)
  [`X6`]; (* 4441: eor x1, x6, x9 *)
  []; (* 4442: lsl x2, x15, #32 *)
  [`X11`]; (* 4443: adcs x8, x11, x4 *)
  [`X1`]; (* 4444: adcs x13, x1, x21 *)
  [`X22`]; (* 4445: eor x1, x22, x9 *)
  [`X23`]; (* 4446: adcs x24, x23, x12 *)
  [`X17`]; (* 4447: eor x11, x17, x9 *)
  [`X11`]; (* 4448: adcs x23, x11, x10 *)
  [`X1`]; (* 4449: adcs x7, x1, x14 *)
  []; (* 4450: adcs x17, x9, x19 *)
  []; (* 4451: adcs x20, x9, xzr *)
  [`X15`; `X2`]; (* 4452: add x1, x2, x15 *)
  []; (* 4453: lsr x3, x1, #32 *)
  []; (* 4454: adcs x11, x9, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4455: adc x9, x9, xzr *)
  []; (* 4456: subs x3, x3, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4457: sbc x6, x1, xzr *)
  [`X5`]; (* 4458: adds x24, x24, x5 *)
  [`X23`]; (* 4459: adcs x4, x23, x4 *)
  []; (* 4460: extr x3, x6, x3, #32 *)
  []; (* 4461: lsr x6, x6, #32 *)
  [`X7`]; (* 4462: adcs x21, x7, x21 *)
  [`X12`; `X17`]; (* 4463: adcs x15, x17, x12 *)
  [`X10`; `X20`]; (* 4464: adcs x7, x20, x10 *)
  [`X11`; `X14`]; (* 4465: adcs x20, x11, x14 *)
  []; (* 4466: mov x14, #0xffffffff *)
  [`CF`; `NF`; `ZF`; `VF`; `X19`; `X9`]; (* 4467: adc x22, x9, x19 *)
  [`X6`]; (* 4468: adds x12, x6, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4469: adc x10, xzr, xzr *)
  [`X8`]; (* 4470: subs x3, x8, x3 *)
  [`X13`]; (* 4471: sbcs x12, x13, x12 *)
  []; (* 4472: lsl x9, x3, #32 *)
  [`X9`]; (* 4473: add x3, x9, x3 *)
  [`X24`]; (* 4474: sbcs x10, x24, x10 *)
  [`X4`]; (* 4475: sbcs x24, x4, xzr *)
  []; (* 4476: lsr x9, x3, #32 *)
  []; (* 4477: sbcs x21, x21, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4478: sbc x1, x1, xzr *)
  []; (* 4479: subs x9, x9, x3 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4480: sbc x13, x3, xzr *)
  []; (* 4481: extr x9, x13, x9, #32 *)
  []; (* 4482: lsr x13, x13, #32 *)
  []; (* 4483: adds x13, x13, x3 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4484: adc x6, xzr, xzr *)
  [`X9`]; (* 4485: subs x12, x12, x9 *)
  [`X10`; `X13`]; (* 4486: sbcs x17, x10, x13 *)
  []; (* 4487: lsl x2, x12, #32 *)
  [`X24`; `X6`]; (* 4488: sbcs x10, x24, x6 *)
  [`X12`; `X2`]; (* 4489: add x9, x2, x12 *)
  [`X21`]; (* 4490: sbcs x6, x21, xzr *)
  []; (* 4491: lsr x5, x9, #32 *)
  [`X1`]; (* 4492: sbcs x21, x1, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X3`]; (* 4493: sbc x13, x3, xzr *)
  [`X5`]; (* 4494: subs x8, x5, x9 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4495: sbc x19, x9, xzr *)
  []; (* 4496: lsr x12, x19, #32 *)
  [`X19`; `X8`]; (* 4497: extr x3, x19, x8, #32 *)
  [`X12`]; (* 4498: adds x8, x12, x9 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4499: adc x1, xzr, xzr *)
  [`X17`; `X3`]; (* 4500: subs x2, x17, x3 *)
  [`X10`; `X8`]; (* 4501: sbcs x12, x10, x8 *)
  [`X1`; `X6`]; (* 4502: sbcs x5, x6, x1 *)
  [`X21`]; (* 4503: sbcs x3, x21, xzr *)
  [`X13`]; (* 4504: sbcs x19, x13, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X9`]; (* 4505: sbc x24, x9, xzr *)
  [`X15`; `X3`]; (* 4506: adds x23, x15, x3 *)
  [`X19`; `X7`]; (* 4507: adcs x8, x7, x19 *)
  [`X20`; `X24`]; (* 4508: adcs x11, x20, x24 *)
  [`CF`; `NF`; `ZF`; `VF`; `X22`]; (* 4509: adc x9, x22, xzr *)
  [`X9`]; (* 4510: add x24, x9, #0x1 *)
  []; (* 4511: lsl x7, x24, #32 *)
  []; (* 4512: subs x21, x24, x7 *)
  [`CF`; `NF`; `ZF`; `VF`; `X7`]; (* 4513: sbc x10, x7, xzr *)
  [`X2`; `X21`]; (* 4514: adds x6, x2, x21 *)
  [`X10`; `X12`]; (* 4515: adcs x7, x12, x10 *)
  [`X5`]; (* 4516: adcs x24, x5, x24 *)
  [`X23`]; (* 4517: adcs x13, x23, xzr *)
  []; (* 4518: adcs x8, x8, xzr *)
  [`X11`]; (* 4519: adcs x15, x11, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4520: csetm x23, cc *)
  [`X16`]; (* 4521: and x11, x16, x23 *)
  [`X14`]; (* 4522: and x20, x14, x23 *)
  [`X6`]; (* 4523: adds x22, x6, x20 *)
  [`X20`]; (* 4524: eor x3, x20, x23 *)
  [`X3`; `X7`]; (* 4525: adcs x5, x7, x3 *)
  [`X11`; `X24`]; (* 4526: adcs x2, x24, x11 *)
  [`X22`; `X5`]; (* 4527: stp x22, x5, [sp, #96] *)
  [`X13`]; (* 4528: adcs x11, x13, x23 *)
  [`X8`]; (* 4529: adcs x12, x8, x23 *)
  []; (* 4530: stp x2, x11, [sp, #112] *)
  [`CF`; `NF`; `ZF`; `VF`; `X15`; `X23`]; (* 4531: adc x13, x15, x23 *)
  []; (* 4532: stp x12, x13, [sp, #128] *)
  []; (* 4533: mov x0, sp *)
  []; (* 4534: mov x1, sp *)
  []; (* 4535: ldp x5, x6, [x1] *)
  []; (* 4536: ldp x4, x3, [sp, #192] *)
  [`X4`]; (* 4537: subs x5, x5, x4 *)
  [`X3`]; (* 4538: sbcs x6, x6, x3 *)
  []; (* 4539: ldp x7, x8, [x1, #16] *)
  []; (* 4540: ldp x4, x3, [sp, #208] *)
  [`X4`]; (* 4541: sbcs x7, x7, x4 *)
  [`X3`]; (* 4542: sbcs x8, x8, x3 *)
  [`X1`]; (* 4543: ldp x9, x10, [x1, #32] *)
  []; (* 4544: ldp x4, x3, [sp, #224] *)
  [`X4`]; (* 4545: sbcs x9, x9, x4 *)
  [`X3`]; (* 4546: sbcs x10, x10, x3 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4547: csetm x3, cc *)
  []; (* 4548: mov x4, #0xffffffff *)
  []; (* 4549: and x4, x4, x3 *)
  []; (* 4550: adds x5, x5, x4 *)
  []; (* 4551: eor x4, x4, x3 *)
  [`X4`]; (* 4552: adcs x6, x6, x4 *)
  []; (* 4553: mov x4, #0xfffffffffffffffe *)
  []; (* 4554: and x4, x4, x3 *)
  [`X4`]; (* 4555: adcs x7, x7, x4 *)
  []; (* 4556: adcs x8, x8, x3 *)
  []; (* 4557: adcs x9, x9, x3 *)
  [`CF`; `NF`; `ZF`; `VF`; `X3`]; (* 4558: adc x10, x10, x3 *)
  [`X5`; `X6`]; (* 4559: stp x5, x6, [x0] *)
  [`X7`; `X8`]; (* 4560: stp x7, x8, [x0, #16] *)
  [`X0`; `X10`; `X9`]; (* 4561: stp x9, x10, [x0, #32] *)
  []; (* 4562: ldp x5, x6, [sp, #96] *)
  []; (* 4563: ldp x4, x3, [sp, #192] *)
  [`X4`]; (* 4564: subs x5, x5, x4 *)
  [`X3`]; (* 4565: sbcs x6, x6, x3 *)
  []; (* 4566: ldp x4, x3, [sp, #208] *)
  [`X2`; `X4`]; (* 4567: sbcs x7, x2, x4 *)
  [`X11`; `X3`]; (* 4568: sbcs x8, x11, x3 *)
  []; (* 4569: ldp x4, x3, [sp, #224] *)
  [`X12`; `X4`]; (* 4570: sbcs x9, x12, x4 *)
  [`X13`; `X3`]; (* 4571: sbcs x10, x13, x3 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4572: csetm x3, cc *)
  []; (* 4573: mov x4, #0xffffffff *)
  []; (* 4574: and x4, x4, x3 *)
  []; (* 4575: adds x5, x5, x4 *)
  []; (* 4576: eor x4, x4, x3 *)
  [`X4`]; (* 4577: adcs x6, x6, x4 *)
  []; (* 4578: mov x4, #0xfffffffffffffffe *)
  []; (* 4579: and x4, x4, x3 *)
  [`X4`]; (* 4580: adcs x7, x7, x4 *)
  []; (* 4581: adcs x8, x8, x3 *)
  []; (* 4582: adcs x9, x9, x3 *)
  [`CF`; `NF`; `ZF`; `VF`; `X3`]; (* 4583: adc x10, x10, x3 *)
  [`X5`; `X6`]; (* 4584: stp x5, x6, [sp, #144] *)
  [`X7`; `X8`]; (* 4585: stp x7, x8, [sp, #160] *)
  [`X10`; `X9`]; (* 4586: stp x9, x10, [sp, #176] *)
  []; (* 4587: ldr q3, [sp, #240] *)
  []; (* 4588: ldr q25, [x25, #96] *)
  []; (* 4589: ldp x13, x23, [x25, #96] *)
  []; (* 4590: ldp x3, x21, [sp, #240] *)
  []; (* 4591: rev64 v23.4s, v25.4s *)
  [`Q25`]; (* 4592: uzp1 v17.4s, v25.4s, v3.4s *)
  []; (* 4593: umulh x15, x3, x13 *)
  [`Q23`]; (* 4594: mul v6.4s, v23.4s, v3.4s *)
  []; (* 4595: uzp1 v3.4s, v3.4s, v3.4s *)
  []; (* 4596: ldr q27, [x25, #128] *)
  []; (* 4597: ldp x8, x24, [sp, #256] *)
  []; (* 4598: subs x6, x3, x21 *)
  []; (* 4599: ldr q0, [sp, #272] *)
  []; (* 4600: movi v23.2d, #0xffffffff *)
  []; (* 4601: csetm x10, cc *)
  []; (* 4602: umulh x19, x21, x23 *)
  []; (* 4603: rev64 v4.4s, v27.4s *)
  []; (* 4604: uzp2 v25.4s, v27.4s, v27.4s *)
  [`CF`; `NF`; `ZF`; `VF`; `X6`]; (* 4605: cneg x4, x6, cc *)
  []; (* 4606: subs x7, x23, x13 *)
  []; (* 4607: xtn v22.2s, v0.2d *)
  [`Q27`]; (* 4608: xtn v24.2s, v27.2d *)
  [`X7`]; (* 4609: cneg x20, x7, cc *)
  []; (* 4610: ldp x6, x14, [x25, #112] *)
  [`Q4`]; (* 4611: mul v27.4s, v4.4s, v0.4s *)
  [`Q6`]; (* 4612: uaddlp v20.2d, v6.4s *)
  [`CF`; `NF`; `ZF`; `VF`; `X10`]; (* 4613: cinv x5, x10, cc *)
  []; (* 4614: mul x16, x4, x20 *)
  [`Q0`]; (* 4615: uzp2 v6.4s, v0.4s, v0.4s *)
  []; (* 4616: umull v21.2d, v22.2s, v25.2s *)
  [`Q20`]; (* 4617: shl v0.2d, v20.2d, #32 *)
  [`Q17`; `Q3`]; (* 4618: umlal v0.2d, v3.2s, v17.2s *)
  []; (* 4619: mul x22, x8, x6 *)
  [`Q25`]; (* 4620: umull v1.2d, v6.2s, v25.2s *)
  []; (* 4621: subs x12, x3, x8 *)
  []; (* 4622: umull v20.2d, v22.2s, v24.2s *)
  [`X12`]; (* 4623: cneg x17, x12, cc *)
  []; (* 4624: umulh x9, x8, x6 *)
  []; (* 4625: mov x12, v0.d[1] *)
  [`X16`]; (* 4626: eor x11, x16, x5 *)
  [`Q0`]; (* 4627: mov x7, v0.d[0] *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4628: csetm x10, cc *)
  [`Q20`]; (* 4629: usra v21.2d, v20.2d, #32 *)
  [`X12`]; (* 4630: adds x15, x15, x12 *)
  [`X19`; `X22`]; (* 4631: adcs x12, x19, x22 *)
  [`X4`]; (* 4632: umulh x20, x4, x20 *)
  [`CF`; `NF`; `ZF`; `VF`; `X9`]; (* 4633: adc x19, x9, xzr *)
  []; (* 4634: usra v1.2d, v21.2d, #32 *)
  []; (* 4635: adds x22, x15, x7 *)
  [`Q21`; `Q23`]; (* 4636: and v26.16b, v21.16b, v23.16b *)
  []; (* 4637: adcs x16, x12, x15 *)
  [`Q27`]; (* 4638: uaddlp v25.2d, v27.4s *)
  []; (* 4639: adcs x9, x19, x12 *)
  [`Q6`]; (* 4640: umlal v26.2d, v6.2s, v24.2s *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4641: adc x4, x19, xzr *)
  []; (* 4642: adds x16, x16, x7 *)
  [`Q25`]; (* 4643: shl v27.2d, v25.2d, #32 *)
  [`X15`]; (* 4644: adcs x9, x9, x15 *)
  [`X12`]; (* 4645: adcs x4, x4, x12 *)
  [`X20`]; (* 4646: eor x12, x20, x5 *)
  [`CF`; `NF`; `ZF`; `VF`; `X19`]; (* 4647: adc x15, x19, xzr *)
  []; (* 4648: subs x20, x6, x13 *)
  []; (* 4649: cneg x20, x20, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4650: cinv x10, x10, cc *)
  []; (* 4651: cmn x5, #0x1 *)
  []; (* 4652: mul x19, x17, x20 *)
  [`X22`]; (* 4653: adcs x11, x22, x11 *)
  [`X16`]; (* 4654: adcs x12, x16, x12 *)
  []; (* 4655: adcs x9, x9, x5 *)
  [`X20`]; (* 4656: umulh x17, x17, x20 *)
  [`X4`]; (* 4657: adcs x22, x4, x5 *)
  [`CF`; `NF`; `ZF`; `VF`; `X15`]; (* 4658: adc x5, x15, x5 *)
  []; (* 4659: subs x16, x21, x8 *)
  [`X16`]; (* 4660: cneg x20, x16, cc *)
  []; (* 4661: eor x19, x19, x10 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4662: csetm x4, cc *)
  []; (* 4663: subs x16, x6, x23 *)
  []; (* 4664: cneg x16, x16, cc *)
  [`Q22`; `Q24`]; (* 4665: umlal v27.2d, v22.2s, v24.2s *)
  []; (* 4666: mul x15, x20, x16 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4667: cinv x4, x4, cc *)
  []; (* 4668: cmn x10, #0x1 *)
  [`Q26`]; (* 4669: usra v1.2d, v26.2d, #32 *)
  [`X12`]; (* 4670: adcs x19, x12, x19 *)
  []; (* 4671: eor x17, x17, x10 *)
  [`X17`]; (* 4672: adcs x9, x9, x17 *)
  []; (* 4673: adcs x22, x22, x10 *)
  []; (* 4674: lsl x12, x7, #32 *)
  [`X16`]; (* 4675: umulh x20, x20, x16 *)
  [`X15`]; (* 4676: eor x16, x15, x4 *)
  []; (* 4677: ldp x15, x17, [x25, #128] *)
  [`X12`; `X7`]; (* 4678: add x2, x12, x7 *)
  [`CF`; `NF`; `ZF`; `VF`; `X10`; `X5`]; (* 4679: adc x7, x5, x10 *)
  []; (* 4680: ldp x5, x10, [sp, #272] *)
  []; (* 4681: lsr x1, x2, #32 *)
  [`X20`]; (* 4682: eor x12, x20, x4 *)
  []; (* 4683: subs x1, x1, x2 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4684: sbc x20, x2, xzr *)
  []; (* 4685: cmn x4, #0x1 *)
  [`X16`]; (* 4686: adcs x9, x9, x16 *)
  []; (* 4687: extr x1, x20, x1, #32 *)
  []; (* 4688: lsr x20, x20, #32 *)
  [`X12`]; (* 4689: adcs x22, x22, x12 *)
  [`CF`; `NF`; `ZF`; `VF`; `X4`; `X7`]; (* 4690: adc x16, x7, x4 *)
  [`X20`]; (* 4691: adds x12, x20, x2 *)
  []; (* 4692: umulh x7, x24, x14 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4693: adc x4, xzr, xzr *)
  [`X11`]; (* 4694: subs x1, x11, x1 *)
  [`X12`; `X19`]; (* 4695: sbcs x20, x19, x12 *)
  [`X4`; `X9`]; (* 4696: sbcs x12, x9, x4 *)
  []; (* 4697: lsl x9, x1, #32 *)
  [`X9`]; (* 4698: add x1, x9, x1 *)
  [`X22`]; (* 4699: sbcs x9, x22, xzr *)
  []; (* 4700: mul x22, x24, x14 *)
  []; (* 4701: sbcs x16, x16, xzr *)
  []; (* 4702: lsr x4, x1, #32 *)
  [`CF`; `NF`; `ZF`; `VF`; `X2`]; (* 4703: sbc x19, x2, xzr *)
  []; (* 4704: subs x4, x4, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4705: sbc x11, x1, xzr *)
  [`X4`]; (* 4706: extr x2, x11, x4, #32 *)
  [`X11`]; (* 4707: lsr x4, x11, #32 *)
  []; (* 4708: adds x4, x4, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4709: adc x11, xzr, xzr *)
  [`X20`]; (* 4710: subs x2, x20, x2 *)
  [`X12`]; (* 4711: sbcs x4, x12, x4 *)
  [`X11`; `X9`]; (* 4712: sbcs x20, x9, x11 *)
  []; (* 4713: lsl x12, x2, #32 *)
  [`X12`]; (* 4714: add x2, x12, x2 *)
  [`X16`]; (* 4715: sbcs x9, x16, xzr *)
  []; (* 4716: lsr x11, x2, #32 *)
  []; (* 4717: sbcs x19, x19, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4718: sbc x1, x1, xzr *)
  [`X11`]; (* 4719: subs x16, x11, x2 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4720: sbc x12, x2, xzr *)
  []; (* 4721: extr x16, x12, x16, #32 *)
  []; (* 4722: lsr x12, x12, #32 *)
  [`X12`]; (* 4723: adds x11, x12, x2 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4724: adc x12, xzr, xzr *)
  [`X4`]; (* 4725: subs x16, x4, x16 *)
  []; (* 4726: mov x4, v27.d[0] *)
  [`X20`]; (* 4727: sbcs x11, x20, x11 *)
  [`X12`; `X9`]; (* 4728: sbcs x20, x9, x12 *)
  [`X11`; `X16`]; (* 4729: stp x16, x11, [sp, #240] *)
  [`X19`]; (* 4730: sbcs x11, x19, xzr *)
  [`X1`]; (* 4731: sbcs x9, x1, xzr *)
  [`X11`; `X20`]; (* 4732: stp x20, x11, [sp, #256] *)
  []; (* 4733: mov x1, v1.d[0] *)
  [`CF`; `NF`; `ZF`; `VF`; `X2`]; (* 4734: sbc x20, x2, xzr *)
  []; (* 4735: subs x12, x24, x5 *)
  [`Q27`]; (* 4736: mov x11, v27.d[1] *)
  [`X12`]; (* 4737: cneg x16, x12, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4738: csetm x2, cc *)
  []; (* 4739: subs x19, x15, x14 *)
  [`Q1`]; (* 4740: mov x12, v1.d[1] *)
  []; (* 4741: cinv x2, x2, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4742: cneg x19, x19, cc *)
  [`X20`; `X9`]; (* 4743: stp x9, x20, [sp, #272] *)
  []; (* 4744: mul x9, x16, x19 *)
  [`X7`]; (* 4745: adds x4, x7, x4 *)
  [`X1`]; (* 4746: adcs x11, x1, x11 *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 4747: adc x1, x12, xzr *)
  []; (* 4748: adds x20, x4, x22 *)
  [`X16`]; (* 4749: umulh x19, x16, x19 *)
  []; (* 4750: adcs x7, x11, x4 *)
  [`X9`]; (* 4751: eor x16, x9, x2 *)
  []; (* 4752: adcs x9, x1, x11 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4753: adc x12, x1, xzr *)
  []; (* 4754: adds x7, x7, x22 *)
  [`X9`]; (* 4755: adcs x4, x9, x4 *)
  [`X11`; `X12`]; (* 4756: adcs x9, x12, x11 *)
  [`CF`; `NF`; `ZF`; `VF`; `X1`]; (* 4757: adc x12, x1, xzr *)
  []; (* 4758: cmn x2, #0x1 *)
  [`X19`]; (* 4759: eor x1, x19, x2 *)
  [`X16`; `X20`]; (* 4760: adcs x11, x20, x16 *)
  [`X1`; `X7`]; (* 4761: adcs x19, x7, x1 *)
  [`X4`]; (* 4762: adcs x1, x4, x2 *)
  [`X9`]; (* 4763: adcs x20, x9, x2 *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 4764: adc x2, x12, x2 *)
  []; (* 4765: subs x12, x24, x10 *)
  [`X12`]; (* 4766: cneg x16, x12, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4767: csetm x12, cc *)
  []; (* 4768: subs x9, x17, x14 *)
  []; (* 4769: cinv x12, x12, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4770: cneg x9, x9, cc *)
  [`X24`]; (* 4771: subs x3, x24, x3 *)
  []; (* 4772: sbcs x21, x5, x21 *)
  []; (* 4773: mul x24, x16, x9 *)
  [`X8`]; (* 4774: sbcs x4, x10, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4775: ngc x8, xzr *)
  [`X5`]; (* 4776: subs x10, x5, x10 *)
  [`X24`]; (* 4777: eor x5, x24, x12 *)
  []; (* 4778: csetm x7, cc *)
  [`CF`; `NF`; `ZF`; `VF`; `X10`]; (* 4779: cneg x24, x10, cc *)
  []; (* 4780: subs x10, x17, x15 *)
  []; (* 4781: cinv x7, x7, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4782: cneg x10, x10, cc *)
  [`X13`]; (* 4783: subs x14, x13, x14 *)
  [`X23`]; (* 4784: sbcs x15, x23, x15 *)
  [`X21`]; (* 4785: eor x13, x21, x8 *)
  []; (* 4786: mul x23, x24, x10 *)
  [`X6`]; (* 4787: sbcs x17, x6, x17 *)
  [`X3`]; (* 4788: eor x6, x3, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4789: ngc x21, xzr *)
  [`X16`]; (* 4790: umulh x9, x16, x9 *)
  []; (* 4791: cmn x8, #0x1 *)
  [`X23`]; (* 4792: eor x3, x23, x7 *)
  [`X6`]; (* 4793: adcs x23, x6, xzr *)
  []; (* 4794: adcs x13, x13, xzr *)
  [`X4`]; (* 4795: eor x16, x4, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4796: adc x16, x16, xzr *)
  [`X17`]; (* 4797: eor x4, x17, x21 *)
  [`X10`; `X24`]; (* 4798: umulh x17, x24, x10 *)
  []; (* 4799: cmn x21, #0x1 *)
  [`X14`]; (* 4800: eor x24, x14, x21 *)
  [`X15`]; (* 4801: eor x6, x15, x21 *)
  [`X24`]; (* 4802: adcs x15, x24, xzr *)
  [`X6`]; (* 4803: adcs x14, x6, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X4`]; (* 4804: adc x6, x4, xzr *)
  []; (* 4805: cmn x12, #0x1 *)
  [`X9`]; (* 4806: eor x4, x9, x12 *)
  [`X5`]; (* 4807: adcs x19, x19, x5 *)
  []; (* 4808: umulh x5, x23, x15 *)
  [`X4`]; (* 4809: adcs x1, x1, x4 *)
  [`X20`]; (* 4810: adcs x10, x20, x12 *)
  [`X17`]; (* 4811: eor x4, x17, x7 *)
  []; (* 4812: ldp x20, x9, [sp, #240] *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 4813: adc x2, x2, x12 *)
  []; (* 4814: cmn x7, #0x1 *)
  [`X1`; `X3`]; (* 4815: adcs x12, x1, x3 *)
  []; (* 4816: ldp x17, x24, [sp, #256] *)
  []; (* 4817: mul x1, x16, x6 *)
  [`X10`; `X4`]; (* 4818: adcs x3, x10, x4 *)
  [`CF`; `NF`; `ZF`; `VF`; `X7`]; (* 4819: adc x2, x2, x7 *)
  []; (* 4820: ldp x7, x4, [sp, #272] *)
  [`X22`]; (* 4821: adds x20, x22, x20 *)
  []; (* 4822: mul x10, x13, x14 *)
  [`X9`]; (* 4823: adcs x11, x11, x9 *)
  [`X21`; `X8`]; (* 4824: eor x9, x8, x21 *)
  [`X17`; `X19`]; (* 4825: adcs x21, x19, x17 *)
  [`X11`; `X20`]; (* 4826: stp x20, x11, [sp, #240] *)
  [`X24`]; (* 4827: adcs x12, x12, x24 *)
  []; (* 4828: mul x8, x23, x15 *)
  [`X7`]; (* 4829: adcs x3, x3, x7 *)
  [`X12`; `X21`]; (* 4830: stp x21, x12, [sp, #256] *)
  [`X2`; `X4`]; (* 4831: adcs x12, x2, x4 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4832: adc x19, xzr, xzr *)
  []; (* 4833: subs x21, x23, x16 *)
  []; (* 4834: umulh x2, x16, x6 *)
  [`X12`; `X3`]; (* 4835: stp x3, x12, [sp, #272] *)
  [`X21`]; (* 4836: cneg x3, x21, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4837: csetm x24, cc *)
  []; (* 4838: umulh x11, x13, x14 *)
  [`X16`]; (* 4839: subs x21, x13, x16 *)
  []; (* 4840: eor x7, x8, x9 *)
  [`X21`]; (* 4841: cneg x17, x21, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4842: csetm x16, cc *)
  []; (* 4843: subs x21, x6, x15 *)
  [`X21`]; (* 4844: cneg x22, x21, cc *)
  [`CF`; `NF`; `ZF`; `VF`; `X24`]; (* 4845: cinv x21, x24, cc *)
  [`X13`; `X23`]; (* 4846: subs x20, x23, x13 *)
  []; (* 4847: umulh x12, x3, x22 *)
  [`X20`]; (* 4848: cneg x23, x20, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4849: csetm x24, cc *)
  [`X15`]; (* 4850: subs x20, x14, x15 *)
  []; (* 4851: cinv x24, x24, cc *)
  [`X3`]; (* 4852: mul x22, x3, x22 *)
  [`CF`; `NF`; `ZF`; `VF`; `X20`]; (* 4853: cneg x3, x20, cc *)
  [`X14`; `X6`]; (* 4854: subs x13, x6, x14 *)
  [`X13`]; (* 4855: cneg x20, x13, cc *)
  [`CF`; `NF`; `ZF`; `VF`; `X16`]; (* 4856: cinv x15, x16, cc *)
  [`X10`; `X5`]; (* 4857: adds x13, x5, x10 *)
  []; (* 4858: mul x4, x23, x3 *)
  [`X1`]; (* 4859: adcs x11, x11, x1 *)
  [`CF`; `NF`; `ZF`; `VF`; `X2`]; (* 4860: adc x14, x2, xzr *)
  []; (* 4861: adds x5, x13, x8 *)
  []; (* 4862: adcs x16, x11, x13 *)
  [`X3`]; (* 4863: umulh x23, x23, x3 *)
  []; (* 4864: adcs x3, x14, x11 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4865: adc x1, x14, xzr *)
  [`X16`; `X8`]; (* 4866: adds x10, x16, x8 *)
  [`X13`; `X3`]; (* 4867: adcs x6, x3, x13 *)
  [`X1`; `X11`]; (* 4868: adcs x8, x1, x11 *)
  []; (* 4869: umulh x13, x17, x20 *)
  [`X4`]; (* 4870: eor x1, x4, x24 *)
  [`CF`; `NF`; `ZF`; `VF`; `X14`]; (* 4871: adc x4, x14, xzr *)
  []; (* 4872: cmn x24, #0x1 *)
  [`X5`]; (* 4873: adcs x1, x5, x1 *)
  [`X23`]; (* 4874: eor x16, x23, x24 *)
  [`X1`]; (* 4875: eor x11, x1, x9 *)
  [`X10`; `X16`]; (* 4876: adcs x23, x10, x16 *)
  [`X22`]; (* 4877: eor x2, x22, x21 *)
  [`X6`]; (* 4878: adcs x3, x6, x24 *)
  [`X17`; `X20`]; (* 4879: mul x14, x17, x20 *)
  [`X13`]; (* 4880: eor x17, x13, x15 *)
  [`X8`]; (* 4881: adcs x13, x8, x24 *)
  [`CF`; `NF`; `ZF`; `VF`; `X24`; `X4`]; (* 4882: adc x8, x4, x24 *)
  []; (* 4883: cmn x21, #0x1 *)
  [`X2`; `X23`]; (* 4884: adcs x6, x23, x2 *)
  []; (* 4885: mov x16, #0xfffffffffffffffe *)
  [`X12`]; (* 4886: eor x20, x12, x21 *)
  [`X3`]; (* 4887: adcs x20, x3, x20 *)
  [`X14`]; (* 4888: eor x23, x14, x15 *)
  [`X13`]; (* 4889: adcs x2, x13, x21 *)
  [`CF`; `NF`; `ZF`; `VF`; `X21`]; (* 4890: adc x8, x8, x21 *)
  []; (* 4891: cmn x15, #0x1 *)
  []; (* 4892: ldp x5, x4, [sp, #240] *)
  []; (* 4893: ldp x21, x12, [sp, #256] *)
  [`X20`; `X23`]; (* 4894: adcs x22, x20, x23 *)
  [`X22`]; (* 4895: eor x23, x22, x9 *)
  [`X2`]; (* 4896: adcs x17, x2, x17 *)
  [`CF`; `NF`; `ZF`; `VF`; `X15`; `X8`]; (* 4897: adc x22, x8, x15 *)
  []; (* 4898: cmn x9, #0x1 *)
  [`X7`]; (* 4899: adcs x15, x7, x5 *)
  []; (* 4900: ldp x10, x14, [sp, #272] *)
  [`X6`]; (* 4901: eor x1, x6, x9 *)
  []; (* 4902: lsl x2, x15, #32 *)
  [`X11`]; (* 4903: adcs x8, x11, x4 *)
  [`X1`]; (* 4904: adcs x13, x1, x21 *)
  [`X22`]; (* 4905: eor x1, x22, x9 *)
  [`X23`]; (* 4906: adcs x24, x23, x12 *)
  [`X17`]; (* 4907: eor x11, x17, x9 *)
  [`X11`]; (* 4908: adcs x23, x11, x10 *)
  [`X1`]; (* 4909: adcs x7, x1, x14 *)
  []; (* 4910: adcs x17, x9, x19 *)
  []; (* 4911: adcs x20, x9, xzr *)
  [`X15`; `X2`]; (* 4912: add x1, x2, x15 *)
  []; (* 4913: lsr x3, x1, #32 *)
  []; (* 4914: adcs x11, x9, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4915: adc x9, x9, xzr *)
  []; (* 4916: subs x3, x3, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4917: sbc x6, x1, xzr *)
  [`X5`]; (* 4918: adds x24, x24, x5 *)
  [`X23`]; (* 4919: adcs x4, x23, x4 *)
  []; (* 4920: extr x3, x6, x3, #32 *)
  []; (* 4921: lsr x6, x6, #32 *)
  [`X7`]; (* 4922: adcs x21, x7, x21 *)
  [`X12`; `X17`]; (* 4923: adcs x15, x17, x12 *)
  [`X10`; `X20`]; (* 4924: adcs x7, x20, x10 *)
  [`X11`; `X14`]; (* 4925: adcs x20, x11, x14 *)
  []; (* 4926: mov x14, #0xffffffff *)
  [`CF`; `NF`; `ZF`; `VF`; `X19`; `X9`]; (* 4927: adc x22, x9, x19 *)
  [`X6`]; (* 4928: adds x12, x6, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4929: adc x10, xzr, xzr *)
  [`X8`]; (* 4930: subs x3, x8, x3 *)
  [`X13`]; (* 4931: sbcs x12, x13, x12 *)
  []; (* 4932: lsl x9, x3, #32 *)
  [`X9`]; (* 4933: add x3, x9, x3 *)
  [`X24`]; (* 4934: sbcs x10, x24, x10 *)
  [`X4`]; (* 4935: sbcs x24, x4, xzr *)
  []; (* 4936: lsr x9, x3, #32 *)
  []; (* 4937: sbcs x21, x21, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4938: sbc x1, x1, xzr *)
  []; (* 4939: subs x9, x9, x3 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4940: sbc x13, x3, xzr *)
  []; (* 4941: extr x9, x13, x9, #32 *)
  []; (* 4942: lsr x13, x13, #32 *)
  []; (* 4943: adds x13, x13, x3 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4944: adc x6, xzr, xzr *)
  [`X9`]; (* 4945: subs x12, x12, x9 *)
  [`X10`; `X13`]; (* 4946: sbcs x17, x10, x13 *)
  []; (* 4947: lsl x2, x12, #32 *)
  [`X24`; `X6`]; (* 4948: sbcs x10, x24, x6 *)
  [`X12`; `X2`]; (* 4949: add x9, x2, x12 *)
  [`X21`]; (* 4950: sbcs x6, x21, xzr *)
  []; (* 4951: lsr x5, x9, #32 *)
  [`X1`]; (* 4952: sbcs x21, x1, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X3`]; (* 4953: sbc x13, x3, xzr *)
  [`X5`]; (* 4954: subs x8, x5, x9 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4955: sbc x19, x9, xzr *)
  []; (* 4956: lsr x12, x19, #32 *)
  [`X19`; `X8`]; (* 4957: extr x3, x19, x8, #32 *)
  [`X12`]; (* 4958: adds x8, x12, x9 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4959: adc x1, xzr, xzr *)
  [`X17`; `X3`]; (* 4960: subs x2, x17, x3 *)
  [`X10`; `X8`]; (* 4961: sbcs x12, x10, x8 *)
  [`X1`; `X6`]; (* 4962: sbcs x5, x6, x1 *)
  [`X21`]; (* 4963: sbcs x3, x21, xzr *)
  [`X13`]; (* 4964: sbcs x19, x13, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X9`]; (* 4965: sbc x24, x9, xzr *)
  [`X15`; `X3`]; (* 4966: adds x23, x15, x3 *)
  [`X19`; `X7`]; (* 4967: adcs x8, x7, x19 *)
  [`X20`; `X24`]; (* 4968: adcs x11, x20, x24 *)
  [`CF`; `NF`; `ZF`; `VF`; `X22`]; (* 4969: adc x9, x22, xzr *)
  [`X9`]; (* 4970: add x24, x9, #0x1 *)
  []; (* 4971: lsl x7, x24, #32 *)
  []; (* 4972: subs x21, x24, x7 *)
  [`CF`; `NF`; `ZF`; `VF`; `X7`]; (* 4973: sbc x10, x7, xzr *)
  [`X2`; `X21`]; (* 4974: adds x6, x2, x21 *)
  [`X10`; `X12`]; (* 4975: adcs x7, x12, x10 *)
  [`X5`]; (* 4976: adcs x24, x5, x24 *)
  [`X23`]; (* 4977: adcs x13, x23, xzr *)
  []; (* 4978: adcs x8, x8, xzr *)
  [`X11`]; (* 4979: adcs x15, x11, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 4980: csetm x23, cc *)
  [`X16`]; (* 4981: and x11, x16, x23 *)
  [`X14`]; (* 4982: and x20, x14, x23 *)
  [`X6`]; (* 4983: adds x22, x6, x20 *)
  [`X20`]; (* 4984: eor x3, x20, x23 *)
  [`X3`; `X7`]; (* 4985: adcs x5, x7, x3 *)
  [`X11`; `X24`]; (* 4986: adcs x14, x24, x11 *)
  [`X22`; `X5`]; (* 4987: stp x22, x5, [sp, #240] *)
  [`X13`]; (* 4988: adcs x5, x13, x23 *)
  [`X8`]; (* 4989: adcs x21, x8, x23 *)
  [`X14`; `X5`]; (* 4990: stp x14, x5, [sp, #256] *)
  [`CF`; `NF`; `ZF`; `VF`; `X15`; `X23`]; (* 4991: adc x12, x15, x23 *)
  [`X12`; `X21`]; (* 4992: stp x21, x12, [sp, #272] *)
  []; (* 4993: mov x0, sp *)
  []; (* 4994: mov x1, sp *)
  []; (* 4995: ldp x5, x6, [x1] *)
  []; (* 4996: ldp x4, x3, [sp, #96] *)
  [`X4`]; (* 4997: subs x5, x5, x4 *)
  [`X3`]; (* 4998: sbcs x6, x6, x3 *)
  []; (* 4999: ldp x7, x8, [x1, #16] *)
  []; (* 5000: ldp x4, x3, [sp, #112] *)
  [`X4`]; (* 5001: sbcs x7, x7, x4 *)
  [`X3`]; (* 5002: sbcs x8, x8, x3 *)
  [`X1`]; (* 5003: ldp x9, x10, [x1, #32] *)
  []; (* 5004: ldp x4, x3, [sp, #128] *)
  [`X4`]; (* 5005: sbcs x9, x9, x4 *)
  [`X3`]; (* 5006: sbcs x10, x10, x3 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5007: csetm x3, cc *)
  []; (* 5008: mov x4, #0xffffffff *)
  []; (* 5009: and x4, x4, x3 *)
  [`X5`]; (* 5010: adds x2, x5, x4 *)
  []; (* 5011: eor x4, x4, x3 *)
  [`X4`; `X6`]; (* 5012: adcs x11, x6, x4 *)
  []; (* 5013: mov x4, #0xfffffffffffffffe *)
  []; (* 5014: and x4, x4, x3 *)
  [`X7`]; (* 5015: adcs x4, x7, x4 *)
  [`X8`]; (* 5016: adcs x12, x8, x3 *)
  [`X9`]; (* 5017: adcs x13, x9, x3 *)
  [`CF`; `NF`; `ZF`; `VF`; `X10`]; (* 5018: adc x3, x10, x3 *)
  []; (* 5019: stp x2, x11, [x0] *)
  []; (* 5020: stp x4, x12, [x0, #16] *)
  [`X0`]; (* 5021: stp x13, x3, [x0, #32] *)
  []; (* 5022: ldp x5, x6, [sp, #192] *)
  [`X2`]; (* 5023: subs x5, x5, x2 *)
  [`X11`]; (* 5024: sbcs x6, x6, x11 *)
  []; (* 5025: ldp x7, x8, [sp, #208] *)
  [`X4`]; (* 5026: sbcs x7, x7, x4 *)
  [`X12`]; (* 5027: sbcs x8, x8, x12 *)
  []; (* 5028: ldp x9, x10, [sp, #224] *)
  [`X13`]; (* 5029: sbcs x9, x9, x13 *)
  [`X3`]; (* 5030: sbcs x10, x10, x3 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5031: csetm x3, cc *)
  []; (* 5032: mov x4, #0xffffffff *)
  []; (* 5033: and x4, x4, x3 *)
  []; (* 5034: adds x5, x5, x4 *)
  []; (* 5035: eor x4, x4, x3 *)
  [`X4`]; (* 5036: adcs x6, x6, x4 *)
  []; (* 5037: mov x4, #0xfffffffffffffffe *)
  []; (* 5038: and x4, x4, x3 *)
  [`X4`]; (* 5039: adcs x7, x7, x4 *)
  []; (* 5040: adcs x8, x8, x3 *)
  []; (* 5041: adcs x9, x9, x3 *)
  [`CF`; `NF`; `ZF`; `VF`; `X3`]; (* 5042: adc x10, x10, x3 *)
  [`X5`; `X6`]; (* 5043: stp x5, x6, [sp, #192] *)
  [`X7`; `X8`]; (* 5044: stp x7, x8, [sp, #208] *)
  [`X10`; `X9`]; (* 5045: stp x9, x10, [sp, #224] *)
  []; (* 5046: ldr q3, [sp, #144] *)
  []; (* 5047: ldr q25, [sp, #288] *)
  []; (* 5048: ldp x13, x23, [sp, #288] *)
  []; (* 5049: ldp x3, x21, [sp, #144] *)
  []; (* 5050: rev64 v23.4s, v25.4s *)
  [`Q25`]; (* 5051: uzp1 v17.4s, v25.4s, v3.4s *)
  []; (* 5052: umulh x15, x3, x13 *)
  [`Q23`]; (* 5053: mul v6.4s, v23.4s, v3.4s *)
  []; (* 5054: uzp1 v3.4s, v3.4s, v3.4s *)
  []; (* 5055: ldr q27, [sp, #320] *)
  []; (* 5056: ldp x8, x24, [sp, #160] *)
  []; (* 5057: subs x6, x3, x21 *)
  []; (* 5058: ldr q0, [sp, #176] *)
  []; (* 5059: movi v23.2d, #0xffffffff *)
  []; (* 5060: csetm x10, cc *)
  []; (* 5061: umulh x19, x21, x23 *)
  []; (* 5062: rev64 v4.4s, v27.4s *)
  []; (* 5063: uzp2 v25.4s, v27.4s, v27.4s *)
  [`CF`; `NF`; `ZF`; `VF`; `X6`]; (* 5064: cneg x4, x6, cc *)
  []; (* 5065: subs x7, x23, x13 *)
  []; (* 5066: xtn v22.2s, v0.2d *)
  [`Q27`]; (* 5067: xtn v24.2s, v27.2d *)
  [`X7`]; (* 5068: cneg x20, x7, cc *)
  []; (* 5069: ldp x6, x14, [sp, #304] *)
  [`Q4`]; (* 5070: mul v27.4s, v4.4s, v0.4s *)
  [`Q6`]; (* 5071: uaddlp v20.2d, v6.4s *)
  [`CF`; `NF`; `ZF`; `VF`; `X10`]; (* 5072: cinv x5, x10, cc *)
  []; (* 5073: mul x16, x4, x20 *)
  [`Q0`]; (* 5074: uzp2 v6.4s, v0.4s, v0.4s *)
  []; (* 5075: umull v21.2d, v22.2s, v25.2s *)
  [`Q20`]; (* 5076: shl v0.2d, v20.2d, #32 *)
  [`Q17`; `Q3`]; (* 5077: umlal v0.2d, v3.2s, v17.2s *)
  []; (* 5078: mul x22, x8, x6 *)
  [`Q25`]; (* 5079: umull v1.2d, v6.2s, v25.2s *)
  []; (* 5080: subs x12, x3, x8 *)
  []; (* 5081: umull v20.2d, v22.2s, v24.2s *)
  [`X12`]; (* 5082: cneg x17, x12, cc *)
  []; (* 5083: umulh x9, x8, x6 *)
  []; (* 5084: mov x12, v0.d[1] *)
  [`X16`]; (* 5085: eor x11, x16, x5 *)
  [`Q0`]; (* 5086: mov x7, v0.d[0] *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5087: csetm x10, cc *)
  [`Q20`]; (* 5088: usra v21.2d, v20.2d, #32 *)
  [`X12`]; (* 5089: adds x15, x15, x12 *)
  [`X19`; `X22`]; (* 5090: adcs x12, x19, x22 *)
  [`X4`]; (* 5091: umulh x20, x4, x20 *)
  [`CF`; `NF`; `ZF`; `VF`; `X9`]; (* 5092: adc x19, x9, xzr *)
  []; (* 5093: usra v1.2d, v21.2d, #32 *)
  []; (* 5094: adds x22, x15, x7 *)
  [`Q21`; `Q23`]; (* 5095: and v26.16b, v21.16b, v23.16b *)
  []; (* 5096: adcs x16, x12, x15 *)
  [`Q27`]; (* 5097: uaddlp v25.2d, v27.4s *)
  []; (* 5098: adcs x9, x19, x12 *)
  [`Q6`]; (* 5099: umlal v26.2d, v6.2s, v24.2s *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5100: adc x4, x19, xzr *)
  []; (* 5101: adds x16, x16, x7 *)
  [`Q25`]; (* 5102: shl v27.2d, v25.2d, #32 *)
  [`X15`]; (* 5103: adcs x9, x9, x15 *)
  [`X12`]; (* 5104: adcs x4, x4, x12 *)
  [`X20`]; (* 5105: eor x12, x20, x5 *)
  [`CF`; `NF`; `ZF`; `VF`; `X19`]; (* 5106: adc x15, x19, xzr *)
  []; (* 5107: subs x20, x6, x13 *)
  []; (* 5108: cneg x20, x20, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5109: cinv x10, x10, cc *)
  []; (* 5110: cmn x5, #0x1 *)
  []; (* 5111: mul x19, x17, x20 *)
  [`X22`]; (* 5112: adcs x11, x22, x11 *)
  [`X16`]; (* 5113: adcs x12, x16, x12 *)
  []; (* 5114: adcs x9, x9, x5 *)
  [`X20`]; (* 5115: umulh x17, x17, x20 *)
  [`X4`]; (* 5116: adcs x22, x4, x5 *)
  [`CF`; `NF`; `ZF`; `VF`; `X15`]; (* 5117: adc x5, x15, x5 *)
  []; (* 5118: subs x16, x21, x8 *)
  [`X16`]; (* 5119: cneg x20, x16, cc *)
  []; (* 5120: eor x19, x19, x10 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5121: csetm x4, cc *)
  []; (* 5122: subs x16, x6, x23 *)
  []; (* 5123: cneg x16, x16, cc *)
  [`Q22`; `Q24`]; (* 5124: umlal v27.2d, v22.2s, v24.2s *)
  []; (* 5125: mul x15, x20, x16 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5126: cinv x4, x4, cc *)
  []; (* 5127: cmn x10, #0x1 *)
  [`Q26`]; (* 5128: usra v1.2d, v26.2d, #32 *)
  [`X12`]; (* 5129: adcs x19, x12, x19 *)
  []; (* 5130: eor x17, x17, x10 *)
  [`X17`]; (* 5131: adcs x9, x9, x17 *)
  []; (* 5132: adcs x22, x22, x10 *)
  []; (* 5133: lsl x12, x7, #32 *)
  [`X16`]; (* 5134: umulh x20, x20, x16 *)
  [`X15`]; (* 5135: eor x16, x15, x4 *)
  []; (* 5136: ldp x15, x17, [sp, #320] *)
  [`X12`; `X7`]; (* 5137: add x2, x12, x7 *)
  [`CF`; `NF`; `ZF`; `VF`; `X10`; `X5`]; (* 5138: adc x7, x5, x10 *)
  []; (* 5139: ldp x5, x10, [sp, #176] *)
  []; (* 5140: lsr x1, x2, #32 *)
  [`X20`]; (* 5141: eor x12, x20, x4 *)
  []; (* 5142: subs x1, x1, x2 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5143: sbc x20, x2, xzr *)
  []; (* 5144: cmn x4, #0x1 *)
  [`X16`]; (* 5145: adcs x9, x9, x16 *)
  []; (* 5146: extr x1, x20, x1, #32 *)
  []; (* 5147: lsr x20, x20, #32 *)
  [`X12`]; (* 5148: adcs x22, x22, x12 *)
  [`CF`; `NF`; `ZF`; `VF`; `X4`; `X7`]; (* 5149: adc x16, x7, x4 *)
  [`X20`]; (* 5150: adds x12, x20, x2 *)
  []; (* 5151: umulh x7, x24, x14 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5152: adc x4, xzr, xzr *)
  [`X11`]; (* 5153: subs x1, x11, x1 *)
  [`X12`; `X19`]; (* 5154: sbcs x20, x19, x12 *)
  [`X4`; `X9`]; (* 5155: sbcs x12, x9, x4 *)
  []; (* 5156: lsl x9, x1, #32 *)
  [`X9`]; (* 5157: add x1, x9, x1 *)
  [`X22`]; (* 5158: sbcs x9, x22, xzr *)
  []; (* 5159: mul x22, x24, x14 *)
  []; (* 5160: sbcs x16, x16, xzr *)
  []; (* 5161: lsr x4, x1, #32 *)
  [`CF`; `NF`; `ZF`; `VF`; `X2`]; (* 5162: sbc x19, x2, xzr *)
  []; (* 5163: subs x4, x4, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5164: sbc x11, x1, xzr *)
  [`X4`]; (* 5165: extr x2, x11, x4, #32 *)
  [`X11`]; (* 5166: lsr x4, x11, #32 *)
  []; (* 5167: adds x4, x4, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5168: adc x11, xzr, xzr *)
  [`X20`]; (* 5169: subs x2, x20, x2 *)
  [`X12`]; (* 5170: sbcs x4, x12, x4 *)
  [`X11`; `X9`]; (* 5171: sbcs x20, x9, x11 *)
  []; (* 5172: lsl x12, x2, #32 *)
  [`X12`]; (* 5173: add x2, x12, x2 *)
  [`X16`]; (* 5174: sbcs x9, x16, xzr *)
  []; (* 5175: lsr x11, x2, #32 *)
  []; (* 5176: sbcs x19, x19, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5177: sbc x1, x1, xzr *)
  [`X11`]; (* 5178: subs x16, x11, x2 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5179: sbc x12, x2, xzr *)
  []; (* 5180: extr x16, x12, x16, #32 *)
  []; (* 5181: lsr x12, x12, #32 *)
  [`X12`]; (* 5182: adds x11, x12, x2 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5183: adc x12, xzr, xzr *)
  [`X4`]; (* 5184: subs x16, x4, x16 *)
  []; (* 5185: mov x4, v27.d[0] *)
  [`X20`]; (* 5186: sbcs x11, x20, x11 *)
  [`X12`; `X9`]; (* 5187: sbcs x20, x9, x12 *)
  [`X11`; `X16`]; (* 5188: stp x16, x11, [sp, #144] *)
  [`X19`]; (* 5189: sbcs x11, x19, xzr *)
  [`X1`]; (* 5190: sbcs x9, x1, xzr *)
  [`X11`; `X20`]; (* 5191: stp x20, x11, [sp, #160] *)
  []; (* 5192: mov x1, v1.d[0] *)
  [`CF`; `NF`; `ZF`; `VF`; `X2`]; (* 5193: sbc x20, x2, xzr *)
  []; (* 5194: subs x12, x24, x5 *)
  [`Q27`]; (* 5195: mov x11, v27.d[1] *)
  [`X12`]; (* 5196: cneg x16, x12, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5197: csetm x2, cc *)
  []; (* 5198: subs x19, x15, x14 *)
  [`Q1`]; (* 5199: mov x12, v1.d[1] *)
  []; (* 5200: cinv x2, x2, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5201: cneg x19, x19, cc *)
  [`X20`; `X9`]; (* 5202: stp x9, x20, [sp, #176] *)
  []; (* 5203: mul x9, x16, x19 *)
  [`X7`]; (* 5204: adds x4, x7, x4 *)
  [`X1`]; (* 5205: adcs x11, x1, x11 *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 5206: adc x1, x12, xzr *)
  []; (* 5207: adds x20, x4, x22 *)
  [`X16`]; (* 5208: umulh x19, x16, x19 *)
  []; (* 5209: adcs x7, x11, x4 *)
  [`X9`]; (* 5210: eor x16, x9, x2 *)
  []; (* 5211: adcs x9, x1, x11 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5212: adc x12, x1, xzr *)
  []; (* 5213: adds x7, x7, x22 *)
  [`X9`]; (* 5214: adcs x4, x9, x4 *)
  [`X11`; `X12`]; (* 5215: adcs x9, x12, x11 *)
  [`CF`; `NF`; `ZF`; `VF`; `X1`]; (* 5216: adc x12, x1, xzr *)
  []; (* 5217: cmn x2, #0x1 *)
  [`X19`]; (* 5218: eor x1, x19, x2 *)
  [`X16`; `X20`]; (* 5219: adcs x11, x20, x16 *)
  [`X1`; `X7`]; (* 5220: adcs x19, x7, x1 *)
  [`X4`]; (* 5221: adcs x1, x4, x2 *)
  [`X9`]; (* 5222: adcs x20, x9, x2 *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 5223: adc x2, x12, x2 *)
  []; (* 5224: subs x12, x24, x10 *)
  [`X12`]; (* 5225: cneg x16, x12, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5226: csetm x12, cc *)
  []; (* 5227: subs x9, x17, x14 *)
  []; (* 5228: cinv x12, x12, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5229: cneg x9, x9, cc *)
  [`X24`]; (* 5230: subs x3, x24, x3 *)
  []; (* 5231: sbcs x21, x5, x21 *)
  []; (* 5232: mul x24, x16, x9 *)
  [`X8`]; (* 5233: sbcs x4, x10, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5234: ngc x8, xzr *)
  [`X5`]; (* 5235: subs x10, x5, x10 *)
  [`X24`]; (* 5236: eor x5, x24, x12 *)
  []; (* 5237: csetm x7, cc *)
  [`CF`; `NF`; `ZF`; `VF`; `X10`]; (* 5238: cneg x24, x10, cc *)
  []; (* 5239: subs x10, x17, x15 *)
  []; (* 5240: cinv x7, x7, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5241: cneg x10, x10, cc *)
  [`X13`]; (* 5242: subs x14, x13, x14 *)
  [`X23`]; (* 5243: sbcs x15, x23, x15 *)
  [`X21`]; (* 5244: eor x13, x21, x8 *)
  []; (* 5245: mul x23, x24, x10 *)
  [`X6`]; (* 5246: sbcs x17, x6, x17 *)
  [`X3`]; (* 5247: eor x6, x3, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5248: ngc x21, xzr *)
  [`X16`]; (* 5249: umulh x9, x16, x9 *)
  []; (* 5250: cmn x8, #0x1 *)
  [`X23`]; (* 5251: eor x3, x23, x7 *)
  [`X6`]; (* 5252: adcs x23, x6, xzr *)
  []; (* 5253: adcs x13, x13, xzr *)
  [`X4`]; (* 5254: eor x16, x4, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5255: adc x16, x16, xzr *)
  [`X17`]; (* 5256: eor x4, x17, x21 *)
  [`X10`; `X24`]; (* 5257: umulh x17, x24, x10 *)
  []; (* 5258: cmn x21, #0x1 *)
  [`X14`]; (* 5259: eor x24, x14, x21 *)
  [`X15`]; (* 5260: eor x6, x15, x21 *)
  [`X24`]; (* 5261: adcs x15, x24, xzr *)
  [`X6`]; (* 5262: adcs x14, x6, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X4`]; (* 5263: adc x6, x4, xzr *)
  []; (* 5264: cmn x12, #0x1 *)
  [`X9`]; (* 5265: eor x4, x9, x12 *)
  [`X5`]; (* 5266: adcs x19, x19, x5 *)
  []; (* 5267: umulh x5, x23, x15 *)
  [`X4`]; (* 5268: adcs x1, x1, x4 *)
  [`X20`]; (* 5269: adcs x10, x20, x12 *)
  [`X17`]; (* 5270: eor x4, x17, x7 *)
  []; (* 5271: ldp x20, x9, [sp, #144] *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 5272: adc x2, x2, x12 *)
  []; (* 5273: cmn x7, #0x1 *)
  [`X1`; `X3`]; (* 5274: adcs x12, x1, x3 *)
  []; (* 5275: ldp x17, x24, [sp, #160] *)
  []; (* 5276: mul x1, x16, x6 *)
  [`X10`; `X4`]; (* 5277: adcs x3, x10, x4 *)
  [`CF`; `NF`; `ZF`; `VF`; `X7`]; (* 5278: adc x2, x2, x7 *)
  []; (* 5279: ldp x7, x4, [sp, #176] *)
  [`X22`]; (* 5280: adds x20, x22, x20 *)
  []; (* 5281: mul x10, x13, x14 *)
  [`X9`]; (* 5282: adcs x11, x11, x9 *)
  [`X21`; `X8`]; (* 5283: eor x9, x8, x21 *)
  [`X17`; `X19`]; (* 5284: adcs x21, x19, x17 *)
  [`X11`; `X20`]; (* 5285: stp x20, x11, [sp, #144] *)
  [`X24`]; (* 5286: adcs x12, x12, x24 *)
  []; (* 5287: mul x8, x23, x15 *)
  [`X7`]; (* 5288: adcs x3, x3, x7 *)
  [`X12`; `X21`]; (* 5289: stp x21, x12, [sp, #160] *)
  [`X2`; `X4`]; (* 5290: adcs x12, x2, x4 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5291: adc x19, xzr, xzr *)
  []; (* 5292: subs x21, x23, x16 *)
  []; (* 5293: umulh x2, x16, x6 *)
  [`X12`; `X3`]; (* 5294: stp x3, x12, [sp, #176] *)
  [`X21`]; (* 5295: cneg x3, x21, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5296: csetm x24, cc *)
  []; (* 5297: umulh x11, x13, x14 *)
  [`X16`]; (* 5298: subs x21, x13, x16 *)
  []; (* 5299: eor x7, x8, x9 *)
  [`X21`]; (* 5300: cneg x17, x21, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5301: csetm x16, cc *)
  []; (* 5302: subs x21, x6, x15 *)
  [`X21`]; (* 5303: cneg x22, x21, cc *)
  [`CF`; `NF`; `ZF`; `VF`; `X24`]; (* 5304: cinv x21, x24, cc *)
  [`X13`; `X23`]; (* 5305: subs x20, x23, x13 *)
  []; (* 5306: umulh x12, x3, x22 *)
  [`X20`]; (* 5307: cneg x23, x20, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5308: csetm x24, cc *)
  [`X15`]; (* 5309: subs x20, x14, x15 *)
  []; (* 5310: cinv x24, x24, cc *)
  [`X3`]; (* 5311: mul x22, x3, x22 *)
  [`CF`; `NF`; `ZF`; `VF`; `X20`]; (* 5312: cneg x3, x20, cc *)
  [`X14`; `X6`]; (* 5313: subs x13, x6, x14 *)
  [`X13`]; (* 5314: cneg x20, x13, cc *)
  [`CF`; `NF`; `ZF`; `VF`; `X16`]; (* 5315: cinv x15, x16, cc *)
  [`X10`; `X5`]; (* 5316: adds x13, x5, x10 *)
  []; (* 5317: mul x4, x23, x3 *)
  [`X1`]; (* 5318: adcs x11, x11, x1 *)
  [`CF`; `NF`; `ZF`; `VF`; `X2`]; (* 5319: adc x14, x2, xzr *)
  []; (* 5320: adds x5, x13, x8 *)
  []; (* 5321: adcs x16, x11, x13 *)
  [`X3`]; (* 5322: umulh x23, x23, x3 *)
  []; (* 5323: adcs x3, x14, x11 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5324: adc x1, x14, xzr *)
  [`X16`; `X8`]; (* 5325: adds x10, x16, x8 *)
  [`X13`; `X3`]; (* 5326: adcs x6, x3, x13 *)
  [`X1`; `X11`]; (* 5327: adcs x8, x1, x11 *)
  []; (* 5328: umulh x13, x17, x20 *)
  [`X4`]; (* 5329: eor x1, x4, x24 *)
  [`CF`; `NF`; `ZF`; `VF`; `X14`]; (* 5330: adc x4, x14, xzr *)
  []; (* 5331: cmn x24, #0x1 *)
  [`X5`]; (* 5332: adcs x1, x5, x1 *)
  [`X23`]; (* 5333: eor x16, x23, x24 *)
  [`X1`]; (* 5334: eor x11, x1, x9 *)
  [`X10`; `X16`]; (* 5335: adcs x23, x10, x16 *)
  [`X22`]; (* 5336: eor x2, x22, x21 *)
  [`X6`]; (* 5337: adcs x3, x6, x24 *)
  [`X17`; `X20`]; (* 5338: mul x14, x17, x20 *)
  [`X13`]; (* 5339: eor x17, x13, x15 *)
  [`X8`]; (* 5340: adcs x13, x8, x24 *)
  [`CF`; `NF`; `ZF`; `VF`; `X24`; `X4`]; (* 5341: adc x8, x4, x24 *)
  []; (* 5342: cmn x21, #0x1 *)
  [`X2`; `X23`]; (* 5343: adcs x6, x23, x2 *)
  []; (* 5344: mov x16, #0xfffffffffffffffe *)
  [`X12`]; (* 5345: eor x20, x12, x21 *)
  [`X3`]; (* 5346: adcs x20, x3, x20 *)
  [`X14`]; (* 5347: eor x23, x14, x15 *)
  [`X13`]; (* 5348: adcs x2, x13, x21 *)
  [`CF`; `NF`; `ZF`; `VF`; `X21`]; (* 5349: adc x8, x8, x21 *)
  []; (* 5350: cmn x15, #0x1 *)
  []; (* 5351: ldp x5, x4, [sp, #144] *)
  []; (* 5352: ldp x21, x12, [sp, #160] *)
  [`X20`; `X23`]; (* 5353: adcs x22, x20, x23 *)
  [`X22`]; (* 5354: eor x23, x22, x9 *)
  [`X2`]; (* 5355: adcs x17, x2, x17 *)
  [`CF`; `NF`; `ZF`; `VF`; `X15`; `X8`]; (* 5356: adc x22, x8, x15 *)
  []; (* 5357: cmn x9, #0x1 *)
  [`X7`]; (* 5358: adcs x15, x7, x5 *)
  []; (* 5359: ldp x10, x14, [sp, #176] *)
  [`X6`]; (* 5360: eor x1, x6, x9 *)
  []; (* 5361: lsl x2, x15, #32 *)
  [`X11`]; (* 5362: adcs x8, x11, x4 *)
  [`X1`]; (* 5363: adcs x13, x1, x21 *)
  [`X22`]; (* 5364: eor x1, x22, x9 *)
  [`X23`]; (* 5365: adcs x24, x23, x12 *)
  [`X17`]; (* 5366: eor x11, x17, x9 *)
  [`X11`]; (* 5367: adcs x23, x11, x10 *)
  [`X1`]; (* 5368: adcs x7, x1, x14 *)
  []; (* 5369: adcs x17, x9, x19 *)
  []; (* 5370: adcs x20, x9, xzr *)
  [`X15`; `X2`]; (* 5371: add x1, x2, x15 *)
  []; (* 5372: lsr x3, x1, #32 *)
  []; (* 5373: adcs x11, x9, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5374: adc x9, x9, xzr *)
  []; (* 5375: subs x3, x3, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5376: sbc x6, x1, xzr *)
  [`X5`]; (* 5377: adds x24, x24, x5 *)
  [`X23`]; (* 5378: adcs x4, x23, x4 *)
  []; (* 5379: extr x3, x6, x3, #32 *)
  []; (* 5380: lsr x6, x6, #32 *)
  [`X7`]; (* 5381: adcs x21, x7, x21 *)
  [`X12`; `X17`]; (* 5382: adcs x15, x17, x12 *)
  [`X10`; `X20`]; (* 5383: adcs x7, x20, x10 *)
  [`X11`; `X14`]; (* 5384: adcs x20, x11, x14 *)
  []; (* 5385: mov x14, #0xffffffff *)
  [`CF`; `NF`; `ZF`; `VF`; `X19`; `X9`]; (* 5386: adc x22, x9, x19 *)
  [`X6`]; (* 5387: adds x12, x6, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5388: adc x10, xzr, xzr *)
  [`X8`]; (* 5389: subs x3, x8, x3 *)
  [`X13`]; (* 5390: sbcs x12, x13, x12 *)
  []; (* 5391: lsl x9, x3, #32 *)
  [`X9`]; (* 5392: add x3, x9, x3 *)
  [`X24`]; (* 5393: sbcs x10, x24, x10 *)
  [`X4`]; (* 5394: sbcs x24, x4, xzr *)
  []; (* 5395: lsr x9, x3, #32 *)
  []; (* 5396: sbcs x21, x21, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5397: sbc x1, x1, xzr *)
  []; (* 5398: subs x9, x9, x3 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5399: sbc x13, x3, xzr *)
  []; (* 5400: extr x9, x13, x9, #32 *)
  []; (* 5401: lsr x13, x13, #32 *)
  []; (* 5402: adds x13, x13, x3 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5403: adc x6, xzr, xzr *)
  [`X9`]; (* 5404: subs x12, x12, x9 *)
  [`X10`; `X13`]; (* 5405: sbcs x17, x10, x13 *)
  []; (* 5406: lsl x2, x12, #32 *)
  [`X24`; `X6`]; (* 5407: sbcs x10, x24, x6 *)
  [`X12`; `X2`]; (* 5408: add x9, x2, x12 *)
  [`X21`]; (* 5409: sbcs x6, x21, xzr *)
  []; (* 5410: lsr x5, x9, #32 *)
  [`X1`]; (* 5411: sbcs x21, x1, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X3`]; (* 5412: sbc x13, x3, xzr *)
  [`X5`]; (* 5413: subs x8, x5, x9 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5414: sbc x19, x9, xzr *)
  []; (* 5415: lsr x12, x19, #32 *)
  [`X19`; `X8`]; (* 5416: extr x3, x19, x8, #32 *)
  [`X12`]; (* 5417: adds x8, x12, x9 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5418: adc x1, xzr, xzr *)
  [`X17`; `X3`]; (* 5419: subs x2, x17, x3 *)
  [`X10`; `X8`]; (* 5420: sbcs x12, x10, x8 *)
  [`X1`; `X6`]; (* 5421: sbcs x5, x6, x1 *)
  [`X21`]; (* 5422: sbcs x3, x21, xzr *)
  [`X13`]; (* 5423: sbcs x19, x13, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X9`]; (* 5424: sbc x24, x9, xzr *)
  [`X15`; `X3`]; (* 5425: adds x23, x15, x3 *)
  [`X19`; `X7`]; (* 5426: adcs x8, x7, x19 *)
  [`X20`; `X24`]; (* 5427: adcs x11, x20, x24 *)
  [`CF`; `NF`; `ZF`; `VF`; `X22`]; (* 5428: adc x9, x22, xzr *)
  [`X9`]; (* 5429: add x24, x9, #0x1 *)
  []; (* 5430: lsl x7, x24, #32 *)
  []; (* 5431: subs x21, x24, x7 *)
  [`CF`; `NF`; `ZF`; `VF`; `X7`]; (* 5432: sbc x10, x7, xzr *)
  [`X2`; `X21`]; (* 5433: adds x6, x2, x21 *)
  [`X10`; `X12`]; (* 5434: adcs x7, x12, x10 *)
  [`X5`]; (* 5435: adcs x24, x5, x24 *)
  [`X23`]; (* 5436: adcs x13, x23, xzr *)
  []; (* 5437: adcs x8, x8, xzr *)
  [`X11`]; (* 5438: adcs x15, x11, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5439: csetm x23, cc *)
  [`X16`]; (* 5440: and x11, x16, x23 *)
  [`X14`]; (* 5441: and x20, x14, x23 *)
  [`X6`]; (* 5442: adds x22, x6, x20 *)
  [`X20`]; (* 5443: eor x3, x20, x23 *)
  [`X3`; `X7`]; (* 5444: adcs x5, x7, x3 *)
  [`X11`; `X24`]; (* 5445: adcs x14, x24, x11 *)
  [`X22`; `X5`]; (* 5446: stp x22, x5, [sp, #144] *)
  [`X13`]; (* 5447: adcs x5, x13, x23 *)
  [`X8`]; (* 5448: adcs x21, x8, x23 *)
  [`X14`; `X5`]; (* 5449: stp x14, x5, [sp, #160] *)
  [`CF`; `NF`; `ZF`; `VF`; `X15`; `X23`]; (* 5450: adc x12, x15, x23 *)
  [`X12`; `X21`]; (* 5451: stp x21, x12, [sp, #176] *)
  []; (* 5452: ldr q3, [sp, #240] *)
  []; (* 5453: ldr q25, [x26, #96] *)
  []; (* 5454: ldp x13, x23, [x26, #96] *)
  []; (* 5455: ldp x3, x21, [sp, #240] *)
  []; (* 5456: rev64 v23.4s, v25.4s *)
  [`Q25`]; (* 5457: uzp1 v17.4s, v25.4s, v3.4s *)
  []; (* 5458: umulh x15, x3, x13 *)
  [`Q23`]; (* 5459: mul v6.4s, v23.4s, v3.4s *)
  []; (* 5460: uzp1 v3.4s, v3.4s, v3.4s *)
  []; (* 5461: ldr q27, [x26, #128] *)
  []; (* 5462: ldp x8, x24, [sp, #256] *)
  []; (* 5463: subs x6, x3, x21 *)
  []; (* 5464: ldr q0, [sp, #272] *)
  []; (* 5465: movi v23.2d, #0xffffffff *)
  []; (* 5466: csetm x10, cc *)
  []; (* 5467: umulh x19, x21, x23 *)
  []; (* 5468: rev64 v4.4s, v27.4s *)
  []; (* 5469: uzp2 v25.4s, v27.4s, v27.4s *)
  [`CF`; `NF`; `ZF`; `VF`; `X6`]; (* 5470: cneg x4, x6, cc *)
  []; (* 5471: subs x7, x23, x13 *)
  []; (* 5472: xtn v22.2s, v0.2d *)
  [`Q27`]; (* 5473: xtn v24.2s, v27.2d *)
  [`X7`]; (* 5474: cneg x20, x7, cc *)
  []; (* 5475: ldp x6, x14, [x26, #112] *)
  [`Q4`]; (* 5476: mul v27.4s, v4.4s, v0.4s *)
  [`Q6`]; (* 5477: uaddlp v20.2d, v6.4s *)
  [`CF`; `NF`; `ZF`; `VF`; `X10`]; (* 5478: cinv x5, x10, cc *)
  []; (* 5479: mul x16, x4, x20 *)
  [`Q0`]; (* 5480: uzp2 v6.4s, v0.4s, v0.4s *)
  []; (* 5481: umull v21.2d, v22.2s, v25.2s *)
  [`Q20`]; (* 5482: shl v0.2d, v20.2d, #32 *)
  [`Q17`; `Q3`]; (* 5483: umlal v0.2d, v3.2s, v17.2s *)
  []; (* 5484: mul x22, x8, x6 *)
  [`Q25`]; (* 5485: umull v1.2d, v6.2s, v25.2s *)
  []; (* 5486: subs x12, x3, x8 *)
  []; (* 5487: umull v20.2d, v22.2s, v24.2s *)
  [`X12`]; (* 5488: cneg x17, x12, cc *)
  []; (* 5489: umulh x9, x8, x6 *)
  []; (* 5490: mov x12, v0.d[1] *)
  [`X16`]; (* 5491: eor x11, x16, x5 *)
  [`Q0`]; (* 5492: mov x7, v0.d[0] *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5493: csetm x10, cc *)
  [`Q20`]; (* 5494: usra v21.2d, v20.2d, #32 *)
  [`X12`]; (* 5495: adds x15, x15, x12 *)
  [`X19`; `X22`]; (* 5496: adcs x12, x19, x22 *)
  [`X4`]; (* 5497: umulh x20, x4, x20 *)
  [`CF`; `NF`; `ZF`; `VF`; `X9`]; (* 5498: adc x19, x9, xzr *)
  []; (* 5499: usra v1.2d, v21.2d, #32 *)
  []; (* 5500: adds x22, x15, x7 *)
  [`Q21`; `Q23`]; (* 5501: and v26.16b, v21.16b, v23.16b *)
  []; (* 5502: adcs x16, x12, x15 *)
  [`Q27`]; (* 5503: uaddlp v25.2d, v27.4s *)
  []; (* 5504: adcs x9, x19, x12 *)
  [`Q6`]; (* 5505: umlal v26.2d, v6.2s, v24.2s *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5506: adc x4, x19, xzr *)
  []; (* 5507: adds x16, x16, x7 *)
  [`Q25`]; (* 5508: shl v27.2d, v25.2d, #32 *)
  [`X15`]; (* 5509: adcs x9, x9, x15 *)
  [`X12`]; (* 5510: adcs x4, x4, x12 *)
  [`X20`]; (* 5511: eor x12, x20, x5 *)
  [`CF`; `NF`; `ZF`; `VF`; `X19`]; (* 5512: adc x15, x19, xzr *)
  []; (* 5513: subs x20, x6, x13 *)
  []; (* 5514: cneg x20, x20, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5515: cinv x10, x10, cc *)
  []; (* 5516: cmn x5, #0x1 *)
  []; (* 5517: mul x19, x17, x20 *)
  [`X22`]; (* 5518: adcs x11, x22, x11 *)
  [`X16`]; (* 5519: adcs x12, x16, x12 *)
  []; (* 5520: adcs x9, x9, x5 *)
  [`X20`]; (* 5521: umulh x17, x17, x20 *)
  [`X4`]; (* 5522: adcs x22, x4, x5 *)
  [`CF`; `NF`; `ZF`; `VF`; `X15`]; (* 5523: adc x5, x15, x5 *)
  []; (* 5524: subs x16, x21, x8 *)
  [`X16`]; (* 5525: cneg x20, x16, cc *)
  []; (* 5526: eor x19, x19, x10 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5527: csetm x4, cc *)
  []; (* 5528: subs x16, x6, x23 *)
  []; (* 5529: cneg x16, x16, cc *)
  [`Q22`; `Q24`]; (* 5530: umlal v27.2d, v22.2s, v24.2s *)
  []; (* 5531: mul x15, x20, x16 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5532: cinv x4, x4, cc *)
  []; (* 5533: cmn x10, #0x1 *)
  [`Q26`]; (* 5534: usra v1.2d, v26.2d, #32 *)
  [`X12`]; (* 5535: adcs x19, x12, x19 *)
  []; (* 5536: eor x17, x17, x10 *)
  [`X17`]; (* 5537: adcs x9, x9, x17 *)
  []; (* 5538: adcs x22, x22, x10 *)
  []; (* 5539: lsl x12, x7, #32 *)
  [`X16`]; (* 5540: umulh x20, x20, x16 *)
  [`X15`]; (* 5541: eor x16, x15, x4 *)
  []; (* 5542: ldp x15, x17, [x26, #128] *)
  [`X12`; `X7`]; (* 5543: add x2, x12, x7 *)
  [`CF`; `NF`; `ZF`; `VF`; `X10`; `X5`]; (* 5544: adc x7, x5, x10 *)
  []; (* 5545: ldp x5, x10, [sp, #272] *)
  []; (* 5546: lsr x1, x2, #32 *)
  [`X20`]; (* 5547: eor x12, x20, x4 *)
  []; (* 5548: subs x1, x1, x2 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5549: sbc x20, x2, xzr *)
  []; (* 5550: cmn x4, #0x1 *)
  [`X16`]; (* 5551: adcs x9, x9, x16 *)
  []; (* 5552: extr x1, x20, x1, #32 *)
  []; (* 5553: lsr x20, x20, #32 *)
  [`X12`]; (* 5554: adcs x22, x22, x12 *)
  [`CF`; `NF`; `ZF`; `VF`; `X4`; `X7`]; (* 5555: adc x16, x7, x4 *)
  [`X20`]; (* 5556: adds x12, x20, x2 *)
  []; (* 5557: umulh x7, x24, x14 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5558: adc x4, xzr, xzr *)
  [`X11`]; (* 5559: subs x1, x11, x1 *)
  [`X12`; `X19`]; (* 5560: sbcs x20, x19, x12 *)
  [`X4`; `X9`]; (* 5561: sbcs x12, x9, x4 *)
  []; (* 5562: lsl x9, x1, #32 *)
  [`X9`]; (* 5563: add x1, x9, x1 *)
  [`X22`]; (* 5564: sbcs x9, x22, xzr *)
  []; (* 5565: mul x22, x24, x14 *)
  []; (* 5566: sbcs x16, x16, xzr *)
  []; (* 5567: lsr x4, x1, #32 *)
  [`CF`; `NF`; `ZF`; `VF`; `X2`]; (* 5568: sbc x19, x2, xzr *)
  []; (* 5569: subs x4, x4, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5570: sbc x11, x1, xzr *)
  [`X4`]; (* 5571: extr x2, x11, x4, #32 *)
  [`X11`]; (* 5572: lsr x4, x11, #32 *)
  []; (* 5573: adds x4, x4, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5574: adc x11, xzr, xzr *)
  [`X20`]; (* 5575: subs x2, x20, x2 *)
  [`X12`]; (* 5576: sbcs x4, x12, x4 *)
  [`X11`; `X9`]; (* 5577: sbcs x20, x9, x11 *)
  []; (* 5578: lsl x12, x2, #32 *)
  [`X12`]; (* 5579: add x2, x12, x2 *)
  [`X16`]; (* 5580: sbcs x9, x16, xzr *)
  []; (* 5581: lsr x11, x2, #32 *)
  []; (* 5582: sbcs x19, x19, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5583: sbc x1, x1, xzr *)
  [`X11`]; (* 5584: subs x16, x11, x2 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5585: sbc x12, x2, xzr *)
  []; (* 5586: extr x16, x12, x16, #32 *)
  []; (* 5587: lsr x12, x12, #32 *)
  [`X12`]; (* 5588: adds x11, x12, x2 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5589: adc x12, xzr, xzr *)
  [`X4`]; (* 5590: subs x16, x4, x16 *)
  []; (* 5591: mov x4, v27.d[0] *)
  [`X20`]; (* 5592: sbcs x11, x20, x11 *)
  [`X12`; `X9`]; (* 5593: sbcs x20, x9, x12 *)
  [`X11`; `X16`]; (* 5594: stp x16, x11, [sp, #240] *)
  [`X19`]; (* 5595: sbcs x11, x19, xzr *)
  [`X1`]; (* 5596: sbcs x9, x1, xzr *)
  [`X11`; `X20`]; (* 5597: stp x20, x11, [sp, #256] *)
  []; (* 5598: mov x1, v1.d[0] *)
  [`CF`; `NF`; `ZF`; `VF`; `X2`]; (* 5599: sbc x20, x2, xzr *)
  []; (* 5600: subs x12, x24, x5 *)
  [`Q27`]; (* 5601: mov x11, v27.d[1] *)
  [`X12`]; (* 5602: cneg x16, x12, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5603: csetm x2, cc *)
  []; (* 5604: subs x19, x15, x14 *)
  [`Q1`]; (* 5605: mov x12, v1.d[1] *)
  []; (* 5606: cinv x2, x2, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5607: cneg x19, x19, cc *)
  [`X20`; `X9`]; (* 5608: stp x9, x20, [sp, #272] *)
  []; (* 5609: mul x9, x16, x19 *)
  [`X7`]; (* 5610: adds x4, x7, x4 *)
  [`X1`]; (* 5611: adcs x11, x1, x11 *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 5612: adc x1, x12, xzr *)
  []; (* 5613: adds x20, x4, x22 *)
  [`X16`]; (* 5614: umulh x19, x16, x19 *)
  []; (* 5615: adcs x7, x11, x4 *)
  [`X9`]; (* 5616: eor x16, x9, x2 *)
  []; (* 5617: adcs x9, x1, x11 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5618: adc x12, x1, xzr *)
  []; (* 5619: adds x7, x7, x22 *)
  [`X9`]; (* 5620: adcs x4, x9, x4 *)
  [`X11`; `X12`]; (* 5621: adcs x9, x12, x11 *)
  [`CF`; `NF`; `ZF`; `VF`; `X1`]; (* 5622: adc x12, x1, xzr *)
  []; (* 5623: cmn x2, #0x1 *)
  [`X19`]; (* 5624: eor x1, x19, x2 *)
  [`X16`; `X20`]; (* 5625: adcs x11, x20, x16 *)
  [`X1`; `X7`]; (* 5626: adcs x19, x7, x1 *)
  [`X4`]; (* 5627: adcs x1, x4, x2 *)
  [`X9`]; (* 5628: adcs x20, x9, x2 *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 5629: adc x2, x12, x2 *)
  []; (* 5630: subs x12, x24, x10 *)
  [`X12`]; (* 5631: cneg x16, x12, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5632: csetm x12, cc *)
  []; (* 5633: subs x9, x17, x14 *)
  []; (* 5634: cinv x12, x12, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5635: cneg x9, x9, cc *)
  [`X24`]; (* 5636: subs x3, x24, x3 *)
  []; (* 5637: sbcs x21, x5, x21 *)
  []; (* 5638: mul x24, x16, x9 *)
  [`X8`]; (* 5639: sbcs x4, x10, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5640: ngc x8, xzr *)
  [`X5`]; (* 5641: subs x10, x5, x10 *)
  [`X24`]; (* 5642: eor x5, x24, x12 *)
  []; (* 5643: csetm x7, cc *)
  [`CF`; `NF`; `ZF`; `VF`; `X10`]; (* 5644: cneg x24, x10, cc *)
  []; (* 5645: subs x10, x17, x15 *)
  []; (* 5646: cinv x7, x7, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5647: cneg x10, x10, cc *)
  [`X13`]; (* 5648: subs x14, x13, x14 *)
  [`X23`]; (* 5649: sbcs x15, x23, x15 *)
  [`X21`]; (* 5650: eor x13, x21, x8 *)
  []; (* 5651: mul x23, x24, x10 *)
  [`X6`]; (* 5652: sbcs x17, x6, x17 *)
  [`X3`]; (* 5653: eor x6, x3, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5654: ngc x21, xzr *)
  [`X16`]; (* 5655: umulh x9, x16, x9 *)
  []; (* 5656: cmn x8, #0x1 *)
  [`X23`]; (* 5657: eor x3, x23, x7 *)
  [`X6`]; (* 5658: adcs x23, x6, xzr *)
  []; (* 5659: adcs x13, x13, xzr *)
  [`X4`]; (* 5660: eor x16, x4, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5661: adc x16, x16, xzr *)
  [`X17`]; (* 5662: eor x4, x17, x21 *)
  [`X10`; `X24`]; (* 5663: umulh x17, x24, x10 *)
  []; (* 5664: cmn x21, #0x1 *)
  [`X14`]; (* 5665: eor x24, x14, x21 *)
  [`X15`]; (* 5666: eor x6, x15, x21 *)
  [`X24`]; (* 5667: adcs x15, x24, xzr *)
  [`X6`]; (* 5668: adcs x14, x6, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X4`]; (* 5669: adc x6, x4, xzr *)
  []; (* 5670: cmn x12, #0x1 *)
  [`X9`]; (* 5671: eor x4, x9, x12 *)
  [`X5`]; (* 5672: adcs x19, x19, x5 *)
  []; (* 5673: umulh x5, x23, x15 *)
  [`X4`]; (* 5674: adcs x1, x1, x4 *)
  [`X20`]; (* 5675: adcs x10, x20, x12 *)
  [`X17`]; (* 5676: eor x4, x17, x7 *)
  []; (* 5677: ldp x20, x9, [sp, #240] *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 5678: adc x2, x2, x12 *)
  []; (* 5679: cmn x7, #0x1 *)
  [`X1`; `X3`]; (* 5680: adcs x12, x1, x3 *)
  []; (* 5681: ldp x17, x24, [sp, #256] *)
  []; (* 5682: mul x1, x16, x6 *)
  [`X10`; `X4`]; (* 5683: adcs x3, x10, x4 *)
  [`CF`; `NF`; `ZF`; `VF`; `X7`]; (* 5684: adc x2, x2, x7 *)
  []; (* 5685: ldp x7, x4, [sp, #272] *)
  [`X22`]; (* 5686: adds x20, x22, x20 *)
  []; (* 5687: mul x10, x13, x14 *)
  [`X9`]; (* 5688: adcs x11, x11, x9 *)
  [`X21`; `X8`]; (* 5689: eor x9, x8, x21 *)
  [`X17`; `X19`]; (* 5690: adcs x21, x19, x17 *)
  [`X11`; `X20`]; (* 5691: stp x20, x11, [sp, #240] *)
  [`X24`]; (* 5692: adcs x12, x12, x24 *)
  []; (* 5693: mul x8, x23, x15 *)
  [`X7`]; (* 5694: adcs x3, x3, x7 *)
  [`X12`; `X21`]; (* 5695: stp x21, x12, [sp, #256] *)
  [`X2`; `X4`]; (* 5696: adcs x12, x2, x4 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5697: adc x19, xzr, xzr *)
  []; (* 5698: subs x21, x23, x16 *)
  []; (* 5699: umulh x2, x16, x6 *)
  [`X12`; `X3`]; (* 5700: stp x3, x12, [sp, #272] *)
  [`X21`]; (* 5701: cneg x3, x21, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5702: csetm x24, cc *)
  []; (* 5703: umulh x11, x13, x14 *)
  [`X16`]; (* 5704: subs x21, x13, x16 *)
  []; (* 5705: eor x7, x8, x9 *)
  [`X21`]; (* 5706: cneg x17, x21, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5707: csetm x16, cc *)
  []; (* 5708: subs x21, x6, x15 *)
  [`X21`]; (* 5709: cneg x22, x21, cc *)
  [`CF`; `NF`; `ZF`; `VF`; `X24`]; (* 5710: cinv x21, x24, cc *)
  [`X13`; `X23`]; (* 5711: subs x20, x23, x13 *)
  []; (* 5712: umulh x12, x3, x22 *)
  [`X20`]; (* 5713: cneg x23, x20, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5714: csetm x24, cc *)
  [`X15`]; (* 5715: subs x20, x14, x15 *)
  []; (* 5716: cinv x24, x24, cc *)
  [`X3`]; (* 5717: mul x22, x3, x22 *)
  [`CF`; `NF`; `ZF`; `VF`; `X20`]; (* 5718: cneg x3, x20, cc *)
  [`X14`; `X6`]; (* 5719: subs x13, x6, x14 *)
  [`X13`]; (* 5720: cneg x20, x13, cc *)
  [`CF`; `NF`; `ZF`; `VF`; `X16`]; (* 5721: cinv x15, x16, cc *)
  [`X10`; `X5`]; (* 5722: adds x13, x5, x10 *)
  []; (* 5723: mul x4, x23, x3 *)
  [`X1`]; (* 5724: adcs x11, x11, x1 *)
  [`CF`; `NF`; `ZF`; `VF`; `X2`]; (* 5725: adc x14, x2, xzr *)
  []; (* 5726: adds x5, x13, x8 *)
  []; (* 5727: adcs x16, x11, x13 *)
  [`X3`]; (* 5728: umulh x23, x23, x3 *)
  []; (* 5729: adcs x3, x14, x11 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5730: adc x1, x14, xzr *)
  [`X16`; `X8`]; (* 5731: adds x10, x16, x8 *)
  [`X13`; `X3`]; (* 5732: adcs x6, x3, x13 *)
  [`X1`; `X11`]; (* 5733: adcs x8, x1, x11 *)
  []; (* 5734: umulh x13, x17, x20 *)
  [`X4`]; (* 5735: eor x1, x4, x24 *)
  [`CF`; `NF`; `ZF`; `VF`; `X14`]; (* 5736: adc x4, x14, xzr *)
  []; (* 5737: cmn x24, #0x1 *)
  [`X5`]; (* 5738: adcs x1, x5, x1 *)
  [`X23`]; (* 5739: eor x16, x23, x24 *)
  [`X1`]; (* 5740: eor x11, x1, x9 *)
  [`X10`; `X16`]; (* 5741: adcs x23, x10, x16 *)
  [`X22`]; (* 5742: eor x2, x22, x21 *)
  [`X6`]; (* 5743: adcs x3, x6, x24 *)
  [`X17`; `X20`]; (* 5744: mul x14, x17, x20 *)
  [`X13`]; (* 5745: eor x17, x13, x15 *)
  [`X8`]; (* 5746: adcs x13, x8, x24 *)
  [`CF`; `NF`; `ZF`; `VF`; `X24`; `X4`]; (* 5747: adc x8, x4, x24 *)
  []; (* 5748: cmn x21, #0x1 *)
  [`X2`; `X23`]; (* 5749: adcs x6, x23, x2 *)
  []; (* 5750: mov x16, #0xfffffffffffffffe *)
  [`X12`]; (* 5751: eor x20, x12, x21 *)
  [`X3`]; (* 5752: adcs x20, x3, x20 *)
  [`X14`]; (* 5753: eor x23, x14, x15 *)
  [`X13`]; (* 5754: adcs x2, x13, x21 *)
  [`CF`; `NF`; `ZF`; `VF`; `X21`]; (* 5755: adc x8, x8, x21 *)
  []; (* 5756: cmn x15, #0x1 *)
  []; (* 5757: ldp x5, x4, [sp, #240] *)
  []; (* 5758: ldp x21, x12, [sp, #256] *)
  [`X20`; `X23`]; (* 5759: adcs x22, x20, x23 *)
  [`X22`]; (* 5760: eor x23, x22, x9 *)
  [`X2`]; (* 5761: adcs x17, x2, x17 *)
  [`CF`; `NF`; `ZF`; `VF`; `X15`; `X8`]; (* 5762: adc x22, x8, x15 *)
  []; (* 5763: cmn x9, #0x1 *)
  [`X7`]; (* 5764: adcs x15, x7, x5 *)
  []; (* 5765: ldp x10, x14, [sp, #272] *)
  [`X6`]; (* 5766: eor x1, x6, x9 *)
  []; (* 5767: lsl x2, x15, #32 *)
  [`X11`]; (* 5768: adcs x8, x11, x4 *)
  [`X1`]; (* 5769: adcs x13, x1, x21 *)
  [`X22`]; (* 5770: eor x1, x22, x9 *)
  [`X23`]; (* 5771: adcs x24, x23, x12 *)
  [`X17`]; (* 5772: eor x11, x17, x9 *)
  [`X11`]; (* 5773: adcs x23, x11, x10 *)
  [`X1`]; (* 5774: adcs x7, x1, x14 *)
  []; (* 5775: adcs x17, x9, x19 *)
  []; (* 5776: adcs x20, x9, xzr *)
  [`X15`; `X2`]; (* 5777: add x1, x2, x15 *)
  []; (* 5778: lsr x3, x1, #32 *)
  []; (* 5779: adcs x11, x9, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5780: adc x9, x9, xzr *)
  []; (* 5781: subs x3, x3, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5782: sbc x6, x1, xzr *)
  [`X5`]; (* 5783: adds x24, x24, x5 *)
  [`X23`]; (* 5784: adcs x4, x23, x4 *)
  []; (* 5785: extr x3, x6, x3, #32 *)
  []; (* 5786: lsr x6, x6, #32 *)
  [`X7`]; (* 5787: adcs x21, x7, x21 *)
  [`X12`; `X17`]; (* 5788: adcs x15, x17, x12 *)
  [`X10`; `X20`]; (* 5789: adcs x7, x20, x10 *)
  [`X11`; `X14`]; (* 5790: adcs x20, x11, x14 *)
  []; (* 5791: mov x14, #0xffffffff *)
  [`CF`; `NF`; `ZF`; `VF`; `X19`; `X9`]; (* 5792: adc x22, x9, x19 *)
  [`X6`]; (* 5793: adds x12, x6, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5794: adc x10, xzr, xzr *)
  [`X8`]; (* 5795: subs x3, x8, x3 *)
  [`X13`]; (* 5796: sbcs x12, x13, x12 *)
  []; (* 5797: lsl x9, x3, #32 *)
  [`X9`]; (* 5798: add x3, x9, x3 *)
  [`X24`]; (* 5799: sbcs x10, x24, x10 *)
  [`X4`]; (* 5800: sbcs x24, x4, xzr *)
  []; (* 5801: lsr x9, x3, #32 *)
  []; (* 5802: sbcs x21, x21, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5803: sbc x1, x1, xzr *)
  []; (* 5804: subs x9, x9, x3 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5805: sbc x13, x3, xzr *)
  []; (* 5806: extr x9, x13, x9, #32 *)
  []; (* 5807: lsr x13, x13, #32 *)
  []; (* 5808: adds x13, x13, x3 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5809: adc x6, xzr, xzr *)
  [`X9`]; (* 5810: subs x12, x12, x9 *)
  [`X10`; `X13`]; (* 5811: sbcs x17, x10, x13 *)
  []; (* 5812: lsl x2, x12, #32 *)
  [`X24`; `X6`]; (* 5813: sbcs x10, x24, x6 *)
  [`X12`; `X2`]; (* 5814: add x9, x2, x12 *)
  [`X21`]; (* 5815: sbcs x6, x21, xzr *)
  []; (* 5816: lsr x5, x9, #32 *)
  [`X1`]; (* 5817: sbcs x21, x1, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X3`]; (* 5818: sbc x13, x3, xzr *)
  [`X5`]; (* 5819: subs x8, x5, x9 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5820: sbc x19, x9, xzr *)
  []; (* 5821: lsr x12, x19, #32 *)
  [`X19`; `X8`]; (* 5822: extr x3, x19, x8, #32 *)
  [`X12`]; (* 5823: adds x8, x12, x9 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5824: adc x1, xzr, xzr *)
  [`X17`; `X3`]; (* 5825: subs x2, x17, x3 *)
  [`X10`; `X8`]; (* 5826: sbcs x12, x10, x8 *)
  [`X1`; `X6`]; (* 5827: sbcs x5, x6, x1 *)
  [`X21`]; (* 5828: sbcs x3, x21, xzr *)
  [`X13`]; (* 5829: sbcs x19, x13, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X9`]; (* 5830: sbc x24, x9, xzr *)
  [`X15`; `X3`]; (* 5831: adds x23, x15, x3 *)
  [`X19`; `X7`]; (* 5832: adcs x8, x7, x19 *)
  [`X20`; `X24`]; (* 5833: adcs x11, x20, x24 *)
  [`CF`; `NF`; `ZF`; `VF`; `X22`]; (* 5834: adc x9, x22, xzr *)
  [`X9`]; (* 5835: add x24, x9, #0x1 *)
  []; (* 5836: lsl x7, x24, #32 *)
  []; (* 5837: subs x21, x24, x7 *)
  [`CF`; `NF`; `ZF`; `VF`; `X7`]; (* 5838: sbc x10, x7, xzr *)
  [`X2`; `X21`]; (* 5839: adds x6, x2, x21 *)
  [`X10`; `X12`]; (* 5840: adcs x7, x12, x10 *)
  [`X5`]; (* 5841: adcs x24, x5, x24 *)
  [`X23`]; (* 5842: adcs x13, x23, xzr *)
  []; (* 5843: adcs x8, x8, xzr *)
  [`X11`]; (* 5844: adcs x15, x11, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5845: csetm x23, cc *)
  [`X16`]; (* 5846: and x11, x16, x23 *)
  [`X14`]; (* 5847: and x20, x14, x23 *)
  [`X6`]; (* 5848: adds x22, x6, x20 *)
  [`X20`]; (* 5849: eor x3, x20, x23 *)
  [`X3`; `X7`]; (* 5850: adcs x5, x7, x3 *)
  [`X11`; `X24`]; (* 5851: adcs x14, x24, x11 *)
  [`X22`; `X5`]; (* 5852: stp x22, x5, [sp, #240] *)
  [`X13`]; (* 5853: adcs x5, x13, x23 *)
  [`X8`]; (* 5854: adcs x21, x8, x23 *)
  [`X14`; `X5`]; (* 5855: stp x14, x5, [sp, #256] *)
  [`CF`; `NF`; `ZF`; `VF`; `X15`; `X23`]; (* 5856: adc x12, x15, x23 *)
  [`X12`; `X21`]; (* 5857: stp x21, x12, [sp, #272] *)
  [`X2`]; (* 5858: ldp x2, x27, [sp, #-48] *)
  []; (* 5859: ldr q3, [sp, #48] *)
  []; (* 5860: ldr q25, [sp, #192] *)
  []; (* 5861: ldp x13, x23, [sp, #192] *)
  []; (* 5862: ldp x3, x21, [sp, #48] *)
  []; (* 5863: rev64 v23.4s, v25.4s *)
  [`Q25`]; (* 5864: uzp1 v17.4s, v25.4s, v3.4s *)
  []; (* 5865: umulh x15, x3, x13 *)
  [`Q23`]; (* 5866: mul v6.4s, v23.4s, v3.4s *)
  []; (* 5867: uzp1 v3.4s, v3.4s, v3.4s *)
  []; (* 5868: ldr q27, [sp, #224] *)
  []; (* 5869: ldp x8, x24, [sp, #64] *)
  []; (* 5870: subs x6, x3, x21 *)
  []; (* 5871: ldr q0, [sp, #80] *)
  []; (* 5872: movi v23.2d, #0xffffffff *)
  []; (* 5873: csetm x10, cc *)
  []; (* 5874: umulh x19, x21, x23 *)
  []; (* 5875: rev64 v4.4s, v27.4s *)
  []; (* 5876: uzp2 v25.4s, v27.4s, v27.4s *)
  [`CF`; `NF`; `ZF`; `VF`; `X6`]; (* 5877: cneg x4, x6, cc *)
  []; (* 5878: subs x7, x23, x13 *)
  []; (* 5879: xtn v22.2s, v0.2d *)
  [`Q27`]; (* 5880: xtn v24.2s, v27.2d *)
  [`X7`]; (* 5881: cneg x20, x7, cc *)
  []; (* 5882: ldp x6, x14, [sp, #208] *)
  [`Q4`]; (* 5883: mul v27.4s, v4.4s, v0.4s *)
  [`Q6`]; (* 5884: uaddlp v20.2d, v6.4s *)
  [`CF`; `NF`; `ZF`; `VF`; `X10`]; (* 5885: cinv x5, x10, cc *)
  []; (* 5886: mul x16, x4, x20 *)
  [`Q0`]; (* 5887: uzp2 v6.4s, v0.4s, v0.4s *)
  []; (* 5888: umull v21.2d, v22.2s, v25.2s *)
  [`Q20`]; (* 5889: shl v0.2d, v20.2d, #32 *)
  [`Q17`; `Q3`]; (* 5890: umlal v0.2d, v3.2s, v17.2s *)
  []; (* 5891: mul x22, x8, x6 *)
  [`Q25`]; (* 5892: umull v1.2d, v6.2s, v25.2s *)
  []; (* 5893: subs x12, x3, x8 *)
  []; (* 5894: umull v20.2d, v22.2s, v24.2s *)
  [`X12`]; (* 5895: cneg x17, x12, cc *)
  []; (* 5896: umulh x9, x8, x6 *)
  []; (* 5897: mov x12, v0.d[1] *)
  [`X16`]; (* 5898: eor x11, x16, x5 *)
  [`Q0`]; (* 5899: mov x7, v0.d[0] *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5900: csetm x10, cc *)
  [`Q20`]; (* 5901: usra v21.2d, v20.2d, #32 *)
  [`X12`]; (* 5902: adds x15, x15, x12 *)
  [`X19`; `X22`]; (* 5903: adcs x12, x19, x22 *)
  [`X4`]; (* 5904: umulh x20, x4, x20 *)
  [`CF`; `NF`; `ZF`; `VF`; `X9`]; (* 5905: adc x19, x9, xzr *)
  []; (* 5906: usra v1.2d, v21.2d, #32 *)
  []; (* 5907: adds x22, x15, x7 *)
  [`Q21`; `Q23`]; (* 5908: and v26.16b, v21.16b, v23.16b *)
  []; (* 5909: adcs x16, x12, x15 *)
  [`Q27`]; (* 5910: uaddlp v25.2d, v27.4s *)
  []; (* 5911: adcs x9, x19, x12 *)
  [`Q6`]; (* 5912: umlal v26.2d, v6.2s, v24.2s *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5913: adc x4, x19, xzr *)
  []; (* 5914: adds x16, x16, x7 *)
  [`Q25`]; (* 5915: shl v27.2d, v25.2d, #32 *)
  [`X15`]; (* 5916: adcs x9, x9, x15 *)
  [`X12`]; (* 5917: adcs x4, x4, x12 *)
  [`X20`]; (* 5918: eor x12, x20, x5 *)
  [`CF`; `NF`; `ZF`; `VF`; `X19`]; (* 5919: adc x15, x19, xzr *)
  []; (* 5920: subs x20, x6, x13 *)
  []; (* 5921: cneg x20, x20, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5922: cinv x10, x10, cc *)
  []; (* 5923: cmn x5, #0x1 *)
  []; (* 5924: mul x19, x17, x20 *)
  [`X22`]; (* 5925: adcs x11, x22, x11 *)
  [`X16`]; (* 5926: adcs x12, x16, x12 *)
  []; (* 5927: adcs x9, x9, x5 *)
  [`X20`]; (* 5928: umulh x17, x17, x20 *)
  [`X4`]; (* 5929: adcs x22, x4, x5 *)
  [`CF`; `NF`; `ZF`; `VF`; `X15`]; (* 5930: adc x5, x15, x5 *)
  []; (* 5931: subs x16, x21, x8 *)
  [`X16`]; (* 5932: cneg x20, x16, cc *)
  []; (* 5933: eor x19, x19, x10 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5934: csetm x4, cc *)
  []; (* 5935: subs x16, x6, x23 *)
  []; (* 5936: cneg x16, x16, cc *)
  [`Q22`; `Q24`]; (* 5937: umlal v27.2d, v22.2s, v24.2s *)
  []; (* 5938: mul x15, x20, x16 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5939: cinv x4, x4, cc *)
  []; (* 5940: cmn x10, #0x1 *)
  [`Q26`]; (* 5941: usra v1.2d, v26.2d, #32 *)
  [`X12`]; (* 5942: adcs x19, x12, x19 *)
  []; (* 5943: eor x17, x17, x10 *)
  [`X17`]; (* 5944: adcs x9, x9, x17 *)
  []; (* 5945: adcs x22, x22, x10 *)
  []; (* 5946: lsl x12, x7, #32 *)
  [`X16`]; (* 5947: umulh x20, x20, x16 *)
  [`X15`]; (* 5948: eor x16, x15, x4 *)
  []; (* 5949: ldp x15, x17, [sp, #224] *)
  [`X12`; `X7`]; (* 5950: add x2, x12, x7 *)
  [`CF`; `NF`; `ZF`; `VF`; `X10`; `X5`]; (* 5951: adc x7, x5, x10 *)
  []; (* 5952: ldp x5, x10, [sp, #80] *)
  []; (* 5953: lsr x1, x2, #32 *)
  [`X20`]; (* 5954: eor x12, x20, x4 *)
  []; (* 5955: subs x1, x1, x2 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5956: sbc x20, x2, xzr *)
  []; (* 5957: cmn x4, #0x1 *)
  [`X16`]; (* 5958: adcs x9, x9, x16 *)
  []; (* 5959: extr x1, x20, x1, #32 *)
  []; (* 5960: lsr x20, x20, #32 *)
  [`X12`]; (* 5961: adcs x22, x22, x12 *)
  [`CF`; `NF`; `ZF`; `VF`; `X4`; `X7`]; (* 5962: adc x16, x7, x4 *)
  [`X20`]; (* 5963: adds x12, x20, x2 *)
  []; (* 5964: umulh x7, x24, x14 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5965: adc x4, xzr, xzr *)
  [`X11`]; (* 5966: subs x1, x11, x1 *)
  [`X12`; `X19`]; (* 5967: sbcs x20, x19, x12 *)
  [`X4`; `X9`]; (* 5968: sbcs x12, x9, x4 *)
  []; (* 5969: lsl x9, x1, #32 *)
  [`X9`]; (* 5970: add x1, x9, x1 *)
  [`X22`]; (* 5971: sbcs x9, x22, xzr *)
  []; (* 5972: mul x22, x24, x14 *)
  []; (* 5973: sbcs x16, x16, xzr *)
  []; (* 5974: lsr x4, x1, #32 *)
  [`CF`; `NF`; `ZF`; `VF`; `X2`]; (* 5975: sbc x19, x2, xzr *)
  []; (* 5976: subs x4, x4, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5977: sbc x11, x1, xzr *)
  [`X4`]; (* 5978: extr x2, x11, x4, #32 *)
  [`X11`]; (* 5979: lsr x4, x11, #32 *)
  []; (* 5980: adds x4, x4, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5981: adc x11, xzr, xzr *)
  [`X20`]; (* 5982: subs x2, x20, x2 *)
  [`X12`]; (* 5983: sbcs x4, x12, x4 *)
  [`X11`; `X9`]; (* 5984: sbcs x20, x9, x11 *)
  []; (* 5985: lsl x12, x2, #32 *)
  [`X12`]; (* 5986: add x2, x12, x2 *)
  [`X16`]; (* 5987: sbcs x9, x16, xzr *)
  []; (* 5988: lsr x11, x2, #32 *)
  []; (* 5989: sbcs x19, x19, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5990: sbc x1, x1, xzr *)
  [`X11`]; (* 5991: subs x16, x11, x2 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5992: sbc x12, x2, xzr *)
  []; (* 5993: extr x16, x12, x16, #32 *)
  []; (* 5994: lsr x12, x12, #32 *)
  [`X12`]; (* 5995: adds x11, x12, x2 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 5996: adc x12, xzr, xzr *)
  [`X4`]; (* 5997: subs x16, x4, x16 *)
  []; (* 5998: mov x4, v27.d[0] *)
  [`X20`]; (* 5999: sbcs x11, x20, x11 *)
  [`X12`; `X9`]; (* 6000: sbcs x20, x9, x12 *)
  [`X11`; `X16`]; (* 6001: stp x16, x11, [sp, #192] *)
  [`X19`]; (* 6002: sbcs x11, x19, xzr *)
  [`X1`]; (* 6003: sbcs x9, x1, xzr *)
  [`X11`; `X20`]; (* 6004: stp x20, x11, [sp, #208] *)
  []; (* 6005: mov x1, v1.d[0] *)
  [`CF`; `NF`; `ZF`; `VF`; `X2`]; (* 6006: sbc x20, x2, xzr *)
  []; (* 6007: subs x12, x24, x5 *)
  [`Q27`]; (* 6008: mov x11, v27.d[1] *)
  [`X12`]; (* 6009: cneg x16, x12, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 6010: csetm x2, cc *)
  []; (* 6011: subs x19, x15, x14 *)
  [`Q1`]; (* 6012: mov x12, v1.d[1] *)
  []; (* 6013: cinv x2, x2, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 6014: cneg x19, x19, cc *)
  [`X20`; `X9`]; (* 6015: stp x9, x20, [sp, #224] *)
  []; (* 6016: mul x9, x16, x19 *)
  [`X7`]; (* 6017: adds x4, x7, x4 *)
  [`X1`]; (* 6018: adcs x11, x1, x11 *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 6019: adc x1, x12, xzr *)
  []; (* 6020: adds x20, x4, x22 *)
  [`X16`]; (* 6021: umulh x19, x16, x19 *)
  []; (* 6022: adcs x7, x11, x4 *)
  [`X9`]; (* 6023: eor x16, x9, x2 *)
  []; (* 6024: adcs x9, x1, x11 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 6025: adc x12, x1, xzr *)
  []; (* 6026: adds x7, x7, x22 *)
  [`X9`]; (* 6027: adcs x4, x9, x4 *)
  [`X11`; `X12`]; (* 6028: adcs x9, x12, x11 *)
  [`CF`; `NF`; `ZF`; `VF`; `X1`]; (* 6029: adc x12, x1, xzr *)
  []; (* 6030: cmn x2, #0x1 *)
  [`X19`]; (* 6031: eor x1, x19, x2 *)
  [`X16`; `X20`]; (* 6032: adcs x11, x20, x16 *)
  [`X1`; `X7`]; (* 6033: adcs x19, x7, x1 *)
  [`X4`]; (* 6034: adcs x1, x4, x2 *)
  [`X9`]; (* 6035: adcs x20, x9, x2 *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 6036: adc x2, x12, x2 *)
  []; (* 6037: subs x12, x24, x10 *)
  [`X12`]; (* 6038: cneg x16, x12, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 6039: csetm x12, cc *)
  []; (* 6040: subs x9, x17, x14 *)
  []; (* 6041: cinv x12, x12, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 6042: cneg x9, x9, cc *)
  [`X24`]; (* 6043: subs x3, x24, x3 *)
  []; (* 6044: sbcs x21, x5, x21 *)
  []; (* 6045: mul x24, x16, x9 *)
  [`X8`]; (* 6046: sbcs x4, x10, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 6047: ngc x8, xzr *)
  [`X5`]; (* 6048: subs x10, x5, x10 *)
  [`X24`]; (* 6049: eor x5, x24, x12 *)
  []; (* 6050: csetm x7, cc *)
  [`CF`; `NF`; `ZF`; `VF`; `X10`]; (* 6051: cneg x24, x10, cc *)
  []; (* 6052: subs x10, x17, x15 *)
  []; (* 6053: cinv x7, x7, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 6054: cneg x10, x10, cc *)
  [`X13`]; (* 6055: subs x14, x13, x14 *)
  [`X23`]; (* 6056: sbcs x15, x23, x15 *)
  [`X21`]; (* 6057: eor x13, x21, x8 *)
  []; (* 6058: mul x23, x24, x10 *)
  [`X6`]; (* 6059: sbcs x17, x6, x17 *)
  [`X3`]; (* 6060: eor x6, x3, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 6061: ngc x21, xzr *)
  [`X16`]; (* 6062: umulh x9, x16, x9 *)
  []; (* 6063: cmn x8, #0x1 *)
  [`X23`]; (* 6064: eor x3, x23, x7 *)
  [`X6`]; (* 6065: adcs x23, x6, xzr *)
  []; (* 6066: adcs x13, x13, xzr *)
  [`X4`]; (* 6067: eor x16, x4, x8 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 6068: adc x16, x16, xzr *)
  [`X17`]; (* 6069: eor x4, x17, x21 *)
  [`X10`; `X24`]; (* 6070: umulh x17, x24, x10 *)
  []; (* 6071: cmn x21, #0x1 *)
  [`X14`]; (* 6072: eor x24, x14, x21 *)
  [`X15`]; (* 6073: eor x6, x15, x21 *)
  [`X24`]; (* 6074: adcs x15, x24, xzr *)
  [`X6`]; (* 6075: adcs x14, x6, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X4`]; (* 6076: adc x6, x4, xzr *)
  []; (* 6077: cmn x12, #0x1 *)
  [`X9`]; (* 6078: eor x4, x9, x12 *)
  [`X5`]; (* 6079: adcs x19, x19, x5 *)
  []; (* 6080: umulh x5, x23, x15 *)
  [`X4`]; (* 6081: adcs x1, x1, x4 *)
  [`X20`]; (* 6082: adcs x10, x20, x12 *)
  [`X17`]; (* 6083: eor x4, x17, x7 *)
  []; (* 6084: ldp x20, x9, [sp, #192] *)
  [`CF`; `NF`; `ZF`; `VF`; `X12`]; (* 6085: adc x2, x2, x12 *)
  []; (* 6086: cmn x7, #0x1 *)
  [`X1`; `X3`]; (* 6087: adcs x12, x1, x3 *)
  []; (* 6088: ldp x17, x24, [sp, #208] *)
  []; (* 6089: mul x1, x16, x6 *)
  [`X10`; `X4`]; (* 6090: adcs x3, x10, x4 *)
  [`CF`; `NF`; `ZF`; `VF`; `X7`]; (* 6091: adc x2, x2, x7 *)
  []; (* 6092: ldp x7, x4, [sp, #224] *)
  [`X22`]; (* 6093: adds x20, x22, x20 *)
  []; (* 6094: mul x10, x13, x14 *)
  [`X9`]; (* 6095: adcs x11, x11, x9 *)
  [`X21`; `X8`]; (* 6096: eor x9, x8, x21 *)
  [`X17`; `X19`]; (* 6097: adcs x21, x19, x17 *)
  [`X11`; `X20`]; (* 6098: stp x20, x11, [sp, #192] *)
  [`X24`]; (* 6099: adcs x12, x12, x24 *)
  []; (* 6100: mul x8, x23, x15 *)
  [`X7`]; (* 6101: adcs x3, x3, x7 *)
  [`X12`; `X21`]; (* 6102: stp x21, x12, [sp, #208] *)
  [`X2`; `X4`]; (* 6103: adcs x12, x2, x4 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 6104: adc x19, xzr, xzr *)
  []; (* 6105: subs x21, x23, x16 *)
  []; (* 6106: umulh x2, x16, x6 *)
  [`X12`; `X3`]; (* 6107: stp x3, x12, [sp, #224] *)
  [`X21`]; (* 6108: cneg x3, x21, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 6109: csetm x24, cc *)
  []; (* 6110: umulh x11, x13, x14 *)
  [`X16`]; (* 6111: subs x21, x13, x16 *)
  []; (* 6112: eor x7, x8, x9 *)
  [`X21`]; (* 6113: cneg x17, x21, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 6114: csetm x16, cc *)
  []; (* 6115: subs x21, x6, x15 *)
  [`X21`]; (* 6116: cneg x22, x21, cc *)
  [`CF`; `NF`; `ZF`; `VF`; `X24`]; (* 6117: cinv x21, x24, cc *)
  [`X13`; `X23`]; (* 6118: subs x20, x23, x13 *)
  []; (* 6119: umulh x12, x3, x22 *)
  [`X20`]; (* 6120: cneg x23, x20, cc *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 6121: csetm x24, cc *)
  [`X15`]; (* 6122: subs x20, x14, x15 *)
  []; (* 6123: cinv x24, x24, cc *)
  [`X3`]; (* 6124: mul x22, x3, x22 *)
  [`CF`; `NF`; `ZF`; `VF`; `X20`]; (* 6125: cneg x3, x20, cc *)
  [`X14`; `X6`]; (* 6126: subs x13, x6, x14 *)
  [`X13`]; (* 6127: cneg x20, x13, cc *)
  [`CF`; `NF`; `ZF`; `VF`; `X16`]; (* 6128: cinv x15, x16, cc *)
  [`X10`; `X5`]; (* 6129: adds x13, x5, x10 *)
  []; (* 6130: mul x4, x23, x3 *)
  [`X1`]; (* 6131: adcs x11, x11, x1 *)
  [`CF`; `NF`; `ZF`; `VF`; `X2`]; (* 6132: adc x14, x2, xzr *)
  []; (* 6133: adds x5, x13, x8 *)
  []; (* 6134: adcs x16, x11, x13 *)
  [`X3`]; (* 6135: umulh x23, x23, x3 *)
  []; (* 6136: adcs x3, x14, x11 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 6137: adc x1, x14, xzr *)
  [`X16`; `X8`]; (* 6138: adds x10, x16, x8 *)
  [`X13`; `X3`]; (* 6139: adcs x6, x3, x13 *)
  [`X1`; `X11`]; (* 6140: adcs x8, x1, x11 *)
  []; (* 6141: umulh x13, x17, x20 *)
  [`X4`]; (* 6142: eor x1, x4, x24 *)
  [`CF`; `NF`; `ZF`; `VF`; `X14`]; (* 6143: adc x4, x14, xzr *)
  []; (* 6144: cmn x24, #0x1 *)
  [`X5`]; (* 6145: adcs x1, x5, x1 *)
  [`X23`]; (* 6146: eor x16, x23, x24 *)
  [`X1`]; (* 6147: eor x11, x1, x9 *)
  [`X10`; `X16`]; (* 6148: adcs x23, x10, x16 *)
  [`X22`]; (* 6149: eor x2, x22, x21 *)
  [`X6`]; (* 6150: adcs x3, x6, x24 *)
  [`X17`; `X20`]; (* 6151: mul x14, x17, x20 *)
  [`X13`]; (* 6152: eor x17, x13, x15 *)
  [`X8`]; (* 6153: adcs x13, x8, x24 *)
  [`CF`; `NF`; `ZF`; `VF`; `X24`; `X4`]; (* 6154: adc x8, x4, x24 *)
  []; (* 6155: cmn x21, #0x1 *)
  [`X2`; `X23`]; (* 6156: adcs x6, x23, x2 *)
  []; (* 6157: mov x16, #0xfffffffffffffffe *)
  [`X12`]; (* 6158: eor x20, x12, x21 *)
  [`X3`]; (* 6159: adcs x20, x3, x20 *)
  [`X14`]; (* 6160: eor x23, x14, x15 *)
  [`X13`]; (* 6161: adcs x2, x13, x21 *)
  [`CF`; `NF`; `ZF`; `VF`; `X21`]; (* 6162: adc x8, x8, x21 *)
  []; (* 6163: cmn x15, #0x1 *)
  []; (* 6164: ldp x5, x4, [sp, #192] *)
  []; (* 6165: ldp x21, x12, [sp, #208] *)
  [`X20`; `X23`]; (* 6166: adcs x22, x20, x23 *)
  [`X22`]; (* 6167: eor x23, x22, x9 *)
  [`X2`]; (* 6168: adcs x17, x2, x17 *)
  [`CF`; `NF`; `ZF`; `VF`; `X15`; `X8`]; (* 6169: adc x22, x8, x15 *)
  []; (* 6170: cmn x9, #0x1 *)
  [`X7`]; (* 6171: adcs x15, x7, x5 *)
  []; (* 6172: ldp x10, x14, [sp, #224] *)
  [`X6`]; (* 6173: eor x1, x6, x9 *)
  []; (* 6174: lsl x2, x15, #32 *)
  [`X11`]; (* 6175: adcs x8, x11, x4 *)
  [`X1`]; (* 6176: adcs x13, x1, x21 *)
  [`X22`]; (* 6177: eor x1, x22, x9 *)
  [`X23`]; (* 6178: adcs x24, x23, x12 *)
  [`X17`]; (* 6179: eor x11, x17, x9 *)
  [`X11`]; (* 6180: adcs x23, x11, x10 *)
  [`X1`]; (* 6181: adcs x7, x1, x14 *)
  []; (* 6182: adcs x17, x9, x19 *)
  []; (* 6183: adcs x20, x9, xzr *)
  [`X15`; `X2`]; (* 6184: add x1, x2, x15 *)
  []; (* 6185: lsr x3, x1, #32 *)
  []; (* 6186: adcs x11, x9, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 6187: adc x9, x9, xzr *)
  []; (* 6188: subs x3, x3, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 6189: sbc x6, x1, xzr *)
  [`X5`]; (* 6190: adds x24, x24, x5 *)
  [`X23`]; (* 6191: adcs x4, x23, x4 *)
  []; (* 6192: extr x3, x6, x3, #32 *)
  []; (* 6193: lsr x6, x6, #32 *)
  [`X7`]; (* 6194: adcs x21, x7, x21 *)
  [`X12`; `X17`]; (* 6195: adcs x15, x17, x12 *)
  [`X10`; `X20`]; (* 6196: adcs x7, x20, x10 *)
  [`X11`; `X14`]; (* 6197: adcs x20, x11, x14 *)
  []; (* 6198: mov x14, #0xffffffff *)
  [`CF`; `NF`; `ZF`; `VF`; `X19`; `X9`]; (* 6199: adc x22, x9, x19 *)
  [`X6`]; (* 6200: adds x12, x6, x1 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 6201: adc x10, xzr, xzr *)
  [`X8`]; (* 6202: subs x3, x8, x3 *)
  [`X13`]; (* 6203: sbcs x12, x13, x12 *)
  []; (* 6204: lsl x9, x3, #32 *)
  [`X9`]; (* 6205: add x3, x9, x3 *)
  [`X24`]; (* 6206: sbcs x10, x24, x10 *)
  [`X4`]; (* 6207: sbcs x24, x4, xzr *)
  []; (* 6208: lsr x9, x3, #32 *)
  []; (* 6209: sbcs x21, x21, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 6210: sbc x1, x1, xzr *)
  []; (* 6211: subs x9, x9, x3 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 6212: sbc x13, x3, xzr *)
  []; (* 6213: extr x9, x13, x9, #32 *)
  []; (* 6214: lsr x13, x13, #32 *)
  []; (* 6215: adds x13, x13, x3 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 6216: adc x6, xzr, xzr *)
  [`X9`]; (* 6217: subs x12, x12, x9 *)
  [`X10`; `X13`]; (* 6218: sbcs x17, x10, x13 *)
  []; (* 6219: lsl x2, x12, #32 *)
  [`X24`; `X6`]; (* 6220: sbcs x10, x24, x6 *)
  [`X12`; `X2`]; (* 6221: add x9, x2, x12 *)
  [`X21`]; (* 6222: sbcs x6, x21, xzr *)
  []; (* 6223: lsr x5, x9, #32 *)
  [`X1`]; (* 6224: sbcs x21, x1, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X3`]; (* 6225: sbc x13, x3, xzr *)
  [`X5`]; (* 6226: subs x8, x5, x9 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 6227: sbc x19, x9, xzr *)
  []; (* 6228: lsr x12, x19, #32 *)
  [`X19`; `X8`]; (* 6229: extr x3, x19, x8, #32 *)
  [`X12`]; (* 6230: adds x8, x12, x9 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 6231: adc x1, xzr, xzr *)
  [`X17`; `X3`]; (* 6232: subs x2, x17, x3 *)
  [`X10`; `X8`]; (* 6233: sbcs x12, x10, x8 *)
  [`X1`; `X6`]; (* 6234: sbcs x5, x6, x1 *)
  [`X21`]; (* 6235: sbcs x3, x21, xzr *)
  [`X13`]; (* 6236: sbcs x19, x13, xzr *)
  [`CF`; `NF`; `ZF`; `VF`; `X9`]; (* 6237: sbc x24, x9, xzr *)
  [`X15`; `X3`]; (* 6238: adds x23, x15, x3 *)
  [`X19`; `X7`]; (* 6239: adcs x8, x7, x19 *)
  [`X20`; `X24`]; (* 6240: adcs x11, x20, x24 *)
  [`CF`; `NF`; `ZF`; `VF`; `X22`]; (* 6241: adc x9, x22, xzr *)
  [`X9`]; (* 6242: add x24, x9, #0x1 *)
  []; (* 6243: lsl x7, x24, #32 *)
  []; (* 6244: subs x21, x24, x7 *)
  [`CF`; `NF`; `ZF`; `VF`; `X7`]; (* 6245: sbc x10, x7, xzr *)
  [`X2`; `X21`]; (* 6246: adds x6, x2, x21 *)
  [`X10`; `X12`]; (* 6247: adcs x7, x12, x10 *)
  [`X5`]; (* 6248: adcs x24, x5, x24 *)
  [`X23`]; (* 6249: adcs x13, x23, xzr *)
  []; (* 6250: adcs x8, x8, xzr *)
  [`X11`]; (* 6251: adcs x15, x11, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 6252: csetm x23, cc *)
  [`X16`]; (* 6253: and x11, x16, x23 *)
  [`X14`]; (* 6254: and x20, x14, x23 *)
  [`X6`]; (* 6255: adds x2, x6, x20 *)
  [`X20`]; (* 6256: eor x3, x20, x23 *)
  [`X3`; `X7`]; (* 6257: adcs x6, x7, x3 *)
  [`X11`; `X24`]; (* 6258: adcs x7, x24, x11 *)
  [`X13`]; (* 6259: adcs x9, x13, x23 *)
  [`X8`]; (* 6260: adcs x10, x8, x23 *)
  [`CF`; `NF`; `ZF`; `VF`; `X15`; `X23`]; (* 6261: adc x11, x15, x23 *)
  []; (* 6262: ldp x4, x3, [sp, #144] *)
  [`X2`; `X4`]; (* 6263: subs x5, x2, x4 *)
  [`X3`]; (* 6264: sbcs x6, x6, x3 *)
  []; (* 6265: ldp x4, x3, [sp, #160] *)
  [`X4`]; (* 6266: sbcs x7, x7, x4 *)
  [`X3`; `X9`]; (* 6267: sbcs x8, x9, x3 *)
  []; (* 6268: ldp x4, x3, [sp, #176] *)
  [`X10`; `X4`]; (* 6269: sbcs x9, x10, x4 *)
  [`X11`; `X3`]; (* 6270: sbcs x10, x11, x3 *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 6271: csetm x3, cc *)
  []; (* 6272: mov x4, #0xffffffff *)
  []; (* 6273: and x4, x4, x3 *)
  [`X5`]; (* 6274: adds x19, x5, x4 *)
  []; (* 6275: eor x4, x4, x3 *)
  [`X4`; `X6`]; (* 6276: adcs x24, x6, x4 *)
  []; (* 6277: mov x4, #0xfffffffffffffffe *)
  []; (* 6278: and x4, x4, x3 *)
  [`X4`]; (* 6279: adcs x7, x7, x4 *)
  []; (* 6280: adcs x8, x8, x3 *)
  []; (* 6281: adcs x9, x9, x3 *)
  [`CF`; `NF`; `ZF`; `VF`; `X3`]; (* 6282: adc x10, x10, x3 *)
  [`X7`; `X8`]; (* 6283: stp x7, x8, [sp, #208] *)
  [`X10`; `X9`]; (* 6284: stp x9, x10, [sp, #224] *)
  []; (* 6285: ldp x0, x1, [x25, #96] *)
  []; (* 6286: ldp x2, x3, [x25, #112] *)
  []; (* 6287: ldp x4, x5, [x25, #128] *)
  []; (* 6288: orr x20, x0, x1 *)
  []; (* 6289: orr x21, x2, x3 *)
  []; (* 6290: orr x22, x4, x5 *)
  [`X21`]; (* 6291: orr x20, x20, x21 *)
  [`X22`]; (* 6292: orr x20, x20, x22 *)
  [`X20`]; (* 6293: cmp x20, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 6294: cset x20, ne *)
  []; (* 6295: ldp x6, x7, [x26, #96] *)
  []; (* 6296: ldp x8, x9, [x26, #112] *)
  []; (* 6297: ldp x10, x11, [x26, #128] *)
  []; (* 6298: orr x21, x6, x7 *)
  []; (* 6299: orr x22, x8, x9 *)
  []; (* 6300: orr x23, x10, x11 *)
  [`X22`]; (* 6301: orr x21, x21, x22 *)
  [`X23`]; (* 6302: orr x21, x21, x23 *)
  [`X21`]; (* 6303: cmp x21, xzr *)
  [`CF`; `NF`; `ZF`; `VF`]; (* 6304: cset x21, ne *)
  [`X20`; `X21`]; (* 6305: cmp x21, x20 *)
  []; (* 6306: ldp x12, x13, [sp, #240] *)
  [`X0`]; (* 6307: csel x12, x0, x12, cc *)
  [`X1`]; (* 6308: csel x13, x1, x13, cc *)
  [`X6`]; (* 6309: csel x12, x6, x12, hi *)
  [`X7`]; (* 6310: csel x13, x7, x13, hi *)
  []; (* 6311: ldp x14, x15, [sp, #256] *)
  [`X2`]; (* 6312: csel x14, x2, x14, cc *)
  [`X3`]; (* 6313: csel x15, x3, x15, cc *)
  [`X8`]; (* 6314: csel x14, x8, x14, hi *)
  [`X9`]; (* 6315: csel x15, x9, x15, hi *)
  []; (* 6316: ldp x16, x17, [sp, #272] *)
  [`X4`]; (* 6317: csel x16, x4, x16, cc *)
  [`X5`]; (* 6318: csel x17, x5, x17, cc *)
  [`X10`]; (* 6319: csel x16, x10, x16, hi *)
  [`X11`]; (* 6320: csel x17, x11, x17, hi *)
  []; (* 6321: ldp x20, x21, [x25] *)
  []; (* 6322: ldp x0, x1, [sp, #0] *)
  [`X20`]; (* 6323: csel x0, x20, x0, cc *)
  [`X21`]; (* 6324: csel x1, x21, x1, cc *)
  []; (* 6325: ldp x20, x21, [x26] *)
  [`X20`]; (* 6326: csel x0, x20, x0, hi *)
  [`X21`]; (* 6327: csel x1, x21, x1, hi *)
  []; (* 6328: ldp x20, x21, [x25, #16] *)
  []; (* 6329: ldp x2, x3, [sp, #16] *)
  [`X20`]; (* 6330: csel x2, x20, x2, cc *)
  [`X21`]; (* 6331: csel x3, x21, x3, cc *)
  []; (* 6332: ldp x20, x21, [x26, #16] *)
  [`X20`]; (* 6333: csel x2, x20, x2, hi *)
  [`X21`]; (* 6334: csel x3, x21, x3, hi *)
  []; (* 6335: ldp x20, x21, [x25, #32] *)
  []; (* 6336: ldp x4, x5, [sp, #32] *)
  [`X20`]; (* 6337: csel x4, x20, x4, cc *)
  [`X21`]; (* 6338: csel x5, x21, x5, cc *)
  []; (* 6339: ldp x20, x21, [x26, #32] *)
  [`X20`]; (* 6340: csel x4, x20, x4, hi *)
  [`X21`]; (* 6341: csel x5, x21, x5, hi *)
  []; (* 6342: ldp x20, x21, [x25, #48] *)
  [`X19`; `X20`]; (* 6343: csel x6, x20, x19, cc *)
  [`X21`; `X24`]; (* 6344: csel x7, x21, x24, cc *)
  []; (* 6345: ldp x20, x21, [x26, #48] *)
  [`X20`]; (* 6346: csel x6, x20, x6, hi *)
  [`X21`]; (* 6347: csel x7, x21, x7, hi *)
  []; (* 6348: ldp x20, x21, [x25, #64] *)
  []; (* 6349: ldp x8, x9, [sp, #208] *)
  [`X20`]; (* 6350: csel x8, x20, x8, cc *)
  [`X21`]; (* 6351: csel x9, x21, x9, cc *)
  []; (* 6352: ldp x20, x21, [x26, #64] *)
  [`X20`]; (* 6353: csel x8, x20, x8, hi *)
  [`X21`]; (* 6354: csel x9, x21, x9, hi *)
  [`X25`]; (* 6355: ldp x20, x21, [x25, #80] *)
  []; (* 6356: ldp x10, x11, [sp, #224] *)
  [`X20`]; (* 6357: csel x10, x20, x10, cc *)
  [`X21`]; (* 6358: csel x11, x21, x11, cc *)
  [`X26`]; (* 6359: ldp x20, x21, [x26, #80] *)
  [`X20`]; (* 6360: csel x10, x20, x10, hi *)
  [`CF`; `NF`; `ZF`; `VF`; `X21`]; (* 6361: csel x11, x21, x11, hi *)
  [`X0`; `X1`]; (* 6362: stp x0, x1, [x27] *)
  [`X2`; `X3`]; (* 6363: stp x2, x3, [x27, #16] *)
  [`X4`; `X5`]; (* 6364: stp x4, x5, [x27, #32] *)
  [`X6`; `X7`]; (* 6365: stp x6, x7, [x27, #48] *)
  [`X8`; `X9`]; (* 6366: stp x8, x9, [x27, #64] *)
  [`X10`; `X11`]; (* 6367: stp x10, x11, [x27, #80] *)
  [`X12`; `X13`]; (* 6368: stp x12, x13, [x27, #96] *)
  [`X14`; `X15`]; (* 6369: stp x14, x15, [x27, #112] *)
  [`X16`; `X17`; `X27`]; (* 6370: stp x16, x17, [x27, #128] *)
|];;
